/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#include <stdio.h>
#include <alloca.h>
#include <math.h>
#include <stream.h>
#include "algebra.h"

#define PRINT_OFFa



Real simplify_angle(Real angle)
{
  Real mult;

  if (angle > 2.0*PI) {
    mult = angle/(2.0*PI);
    mult = floor(mult);
    angle -= mult*2.0*PI;

  } else if (angle < -2.0*PI) {
    mult = angle/(-2.0*PI);
    mult = floor(mult);
    angle += mult*2.0*PI;
  }

  return angle;
}

#define ANGLE_TOLERANCE 0.0001

Real simplify_angle_180(Real angle)
{
  Real mult;

  if (angle > PI) {
    mult = angle/(2.0*PI);
    mult = floor(mult);
    angle -= mult*2.0*PI;
    if (angle > PI)
      angle -= 2.0 * PI;

  } else if (angle < PI) {
    mult = angle/(-2.0*PI);
    mult = floor(mult);
    angle += mult*2.0*PI;
    if (angle < -PI)
      angle += 2.0 * PI;
  }

  if (absv(angle+PI) < ANGLE_TOLERANCE)
    angle = PI;

  return angle;
}


// Want angle between the two vectors c-p1 and c-p2.
// Calculate angle as:
//   a.b = |a||b|cos(theta)
Real angle_btw_pts(Real cx, Real cy, Real p1x, Real p1y,
		   Real p2x, Real p2y)
{
  Real numer = p1x*p2x + p1y*p2y;
  Real denom = sqrt(sqr(p1x)+sqr(p1y)) * sqrt(sqr(p2x)+sqr(p2y));

  if (denom < ANGLE_TOLERANCE)
    return 0.0;
  return acos(numer / denom);
}


// Returns the angle between the two vectors vec1 and vec2.
// Calculate angle as:
//   a.b = |a||b|cos(theta)
Real angle_btw_vec3(Real vec1[3], Real vec2[3])
{
  Real numer = vec_dot(vec1, vec2);
  Real denom = vec_dot(vec1, vec1) * vec_dot(vec2, vec2);

  //  v_print(vec1, 0);
  //  v_print(vec2, 0);
  //  printf("%.4f / %.4f\n", numer, sqrt(denom));

  if (denom < ANGLE_TOLERANCE)
    return 0.0;

  // Delay taking the square-root 
  Real angle = numer / sqrt(denom);
  //  printf("angle: %.4f, ", angle);
  angle = acos(angle);
  //  printf(" %.4f <%.4f>\n", angle, angle*RAD2DEG);
  return angle;
}



void m3_print(int m[3][3], char *string)
{
  int i, j;

#ifndef PRINT_OFF
  if (string)
    cout << string;
  for (i=0; i<3; i++)
    {
      for (j=0; j<3; j++)
	printf("%2d ", m[i][j]);
      cout <<endl;
    }
#endif
} 

void m3_print(Real m[3][3], char *string)
{
#ifndef PRINT_OFF
  int i, j;

  if (string)
    cout << string;
  for (i=0; i<3; i++) {
    for (j=0; j<3; j++)
      printf("%.3f  ", m[i][j]);
    printf("\n");
  }
#endif
} 

void m4_print(Real m[4][4], char *string)
{
  int i, j;

#ifndef PRINT_OFF
  if (string)
    cout << string;
  for (i=0; i<4; i++)
    {
      for (j=0; j<4; j++)
	cout << m[i][j]<<"  ";
      cout <<endl;
    }
#endif
} 

void sp_m_print(Real m[6][6], char *string)
{
  int i, j;

#ifndef PRINT_OFF
  if (string)
    cout << string;
  for (i=0; i<6; i++)
    {
      for (j=0; j<6; j++)
	cout << m[i][j]<<" ";
      cout << endl;
    }
#endif
} 

void sp_v_print(Real v[6], char *string)
{
#ifndef PRINT_OFF
  int i;

  if (string)
    cout << string;
  for (i=0; i<6; i++)
    cout << v[i]<<"  ";
  cout << endl;
#endif
} 

void v_print(Real v[3], char *string)
{
#ifndef PRINT_OFF
  int i;
  Real l = 0.0;

  if (string)
    cout << string;
  for (i=0; i<3; i++)
    cout << v[i]<<"  ";

  for (i=0; i<3; i++)
    l += sqr(v[i]);
  l = sqrt(l);
  cout << "("<<l<<")"<<endl;
#endif
} 


void v4_print(Real v[4], char *string)
{
#ifndef PRINT_OFF
  int i;
  Real l = 0.0;

  if (string)
    cout << string;
  for (i=0; i<4; i++)
    cout << v[i]<<"  ";

  for (i=0; i<4; i++)
    l += sqr(v[i]);
  l = sqrt(l);
  cout << "("<<l<<")" << endl;
#endif
} 


int vec_norm(Real v[3])
{
  Real len;
  len = sqrt(sqr(v[0]) + sqr(v[1]) + sqr(v[2]));
  if (len != 0.0) {
    vs_div(v, len);
    return TRUE;
  }
  return FALSE;
}


Real distance2(Real p1[2], Real p2[2])
{
  return sqrt(sqr(p1[0] - p2[0]) + sqr(p1[1] - p2[1]));
}


void mm3_transpose(Real dst[3][3], Real src[3][3])
{
  for (int i=0; i<3; i++)
    for (int j=0; j<3; j++)
      dst[i][j] = src[j][i];
}


Real svTv6_mult(Real v1[6], Real v2[6])
{
  int i;
  Real val = 0.0;

  for (i=0; i<3; i++)
    val += v1[i+3]*v2[i];
  for (i=3; i<6; i++)
    val += v1[i-3]*v2[i];

  return val;
}

Real svv6_mult(Real v1[6], Real v2[6])
{
  int i;
  Real val = 0.0;

  for (i=0; i<6; i++)
    val += v1[i]*v2[i];

  return val;
}


void mm3_copy(Real mdst[3][3], Real msrc[3][3])
{
  mdst[0][0] = msrc[0][0];
  mdst[0][1] = msrc[0][1];
  mdst[0][2] = msrc[0][2];
  mdst[1][0] = msrc[1][0];
  mdst[1][1] = msrc[1][1];
  mdst[1][2] = msrc[1][2];
  mdst[2][0] = msrc[2][0];
  mdst[2][1] = msrc[2][1];
  mdst[2][2] = msrc[2][2];
}

void mm3_copy(int mdst[3][3], int msrc[3][3])
{
  mdst[0][0] = msrc[0][0];
  mdst[0][1] = msrc[0][1];
  mdst[0][2] = msrc[0][2];
  mdst[1][0] = msrc[1][0];
  mdst[1][1] = msrc[1][1];
  mdst[1][2] = msrc[1][2];
  mdst[2][0] = msrc[2][0];
  mdst[2][1] = msrc[2][1];
  mdst[2][2] = msrc[2][2];
}


void mmm3_mult(int mdst[3][3], int mat1[3][3], 
	   int mat2[3][3])
{
   int i, j, k;
   int sum;

   for (i=0; i<3; i++)
      for (j=0; j<3; j++) {
	sum = 0;
	for (k=0; k<3; k++)
	  sum += mval(mat1, i, k)*mval(mat2, k, j);
	mval(mdst, i, j) = sum;
      }
}

void mmm3_mult(Real mdst[3][3], Real mat1[3][3], 
	   Real mat2[3][3])
{
   int i, j, k;
   Real sum;

   for (i=0; i<3; i++)
      for (j=0; j<3; j++) {
	sum = 0.0;
	for (k=0; k<3; k++)
	  sum += mval(mat1, i, k)*mval(mat2, k, j);
	mval(mdst, i, j) = sum;
      }
}

void mmm_mult(Real mdst[4][4], Real mat1[4][4], 
	   Real mat2[4][4])
{
   int i, j, k;
   Real sum;

   for (i=0; i<4; i++)
      for (j=0; j<4; j++)
      {
	 sum = 0.0;
	 for (k=0; k<4; k++)
	    sum += mval(mat1, i, k)*mval(mat2, k, j);
	 mval(mdst, i, j) = sum;
      }
}

void vvm4_mult(Real dst[3], Real src[3], Real mat[4][4])
{
  int i, j;

  for (i=0; i<3; i++)
    {
      dst[i] = 0.0;
      for (j=0; j<3; j++)
	dst[i] += src[j]*mat[j][i];
    }
}

void vvTm6_mult(Real dst[6], Real src[6], Real mat[6][6])
{
  int i, j;

  for (i=0; i<6; i++)
    {
      dst[i] = 0.0;
      for (j=0; j<3; j++)
	dst[i] += src[j+3]*mat[j][i];
      for (j=3; j<6; j++)
	dst[i] += src[j-3]*mat[j][i];
    }
}

void vvm6_mult(Real dst[6], Real src[6], Real mat[6][6])
{
  int i, j;

  for (i=0; i<6; i++)
    {
      dst[i] = 0.0;
      for (j=0; j<6; j++)
	dst[i] += src[j]*mat[j][i];
    }
}

void vmv_mult(Real dst[3], Real mat[3][3], Real src[3])
{
  int i, j;

  for (i=0; i<3; i++)
    {
      dst[i] = 0.0;
      for (j=0; j<3; j++)
	dst[i] += mat[i][j]*src[j];
    }
}


void vmv4_mult(Real dst[3], Real mat[4][4], Real src[3])
{
  int i, j;

  for (i=0; i<3; i++)
    {
      dst[i] = 0.0;
      for (j=0; j<3; j++)
	dst[i] += mat[i][j]*src[j];
    }
}


void vmv443_mult(Real dst[4], Real mat[4][3], Real src[3])
{
  int i, j;

  for (i=0; i<4; i++)
    {
      dst[i] = 0.0;
      for (j=0; j<3; j++)
	dst[i] += mat[i][j]*src[j];
    }
}


void msm6_mult(Real dst[6][6], Real sc, Real src[6][6])
{
  int i, j;

  for (i=0; i<6; i++)
    for (j=0; j<6; j++)
      dst[i][j] = sc * src[i][j];
}


void ms6_mult(Real m[6][6], Real sc)
{
  int i, j;

  for (i=0; i<6; i++)
    for (j=0; j<6; j++)
      m[i][j] = sc*m[i][j];
}

void ms6_div(Real m[6][6], Real sc)
{
  int i, j;

  for (i=0; i<6; i++)
    for (j=0; j<6; j++)
      m[i][j] = m[i][j]/sc;
}

void mvv6_mult(Real m[6][6], Real v1[6], Real v2[6])
{
  int i, j;
  /* handle transpose "funny"? : no */

  for (i=0; i<6; i++)
    for (j=0; j<6; j++)
      m[i][j] = v1[i]*v2[j];
}

void vmv6_mult(Real dst[6], Real mat[6][6], Real src[6])
{
  int i, j;

  for (i=0; i<6; i++)
    {
      dst[i] = 0.0;
      for (j=0; j<6; j++)
	dst[i] += mat[i][j]*src[j];
    }
}

void mm6_plus(Real d[6][6], Real s[6][6])
{
  int i, j;

  for (i=0; i<6; i++)
    for (j=0; j<6; j++)
      d[i][j]+=s[i][j];
}

void mmm6_subt(Real d[6][6], Real m1[6][6], Real m2[6][6])
{
  int i, j;

  for (i=0; i<6; i++)
    for (j=0; j<6; j++)
      d[i][j]=m1[i][j]-m2[i][j];
}

void mmm6_mult(Real mdst[6][6], Real mat1[6][6], 
	       Real mat2[6][6])
{
   int i, j, k;
   Real sum;

   for (i=0; i<6; i++)
      for (j=0; j<6; j++)
      {
	 sum = 0.0;
	 for (k=0; k<6; k++)
	    sum += mval(mat1, i, k)*mval(mat2, k, j);
	 mval(mdst, i, j) = sum;
      }
}


#define ALMOST_ZERO 0.0000001

Real line_segment_distance(Real seg1_a[3], Real seg1_b[3],
			   Real seg2_a[3], Real seg2_b[3])
{
  Real sc, sd, sn, tc, td, tn;
  Real uu, uv, vv, uw, vw, det;
  Real u[3], v[3], w[3];

  vec_subt(u, seg1_b, seg1_a);
  vec_subt(v, seg2_b, seg2_a);
  vec_subt(w, seg1_a, seg2_a);

  uu = vec_dot(u, u);
  uv = vec_dot(u, v);
  vv = vec_dot(v, v);
  uw = vec_dot(u, w);
  vw = vec_dot(v, w);
  
  det = uu*vv - sqr(uv);
  sd = det;
  td = det;

  if (det < ALMOST_ZERO) {
    sn = 0.0;
    sd = 1.0;
    tn = vw;
    td = vv;
  } else {
    sn = uv*vw - vv*uw;
    tn = uu*vw - uv*uw;
    if (sn < 0.0) {
      sn = 0.0;
      tn = vw;
      td = vv;
    } else if (sn > sd) {
      sn = sd;
      tn = vw + uv;
      td = vv;
    }
  }

  if (tn < 0.0) {
    tn = 0.0;
    if (-uw < 0.0) {
      sn = 0.0;
    } else if (-uw > uu) {
      sn = sd;
    } else {
      sn = -uw;
      sd = uu;
    }
  } else if (tn > td) {
    tn = td;
    if (uv-uw < 0.0) {
      sn = 0.0;
    } else if (uv-uw > uu) {
      sn = sd;
    } else {
      sn = uv-uw;
      sd = uu;
    }
  }

  sc = (absv(sn) < ALMOST_ZERO ? 0.0 : sn / sd);
  tc = (absv(tn) < ALMOST_ZERO ? 0.0 : tn / td);

  vs_mult(u, sc);
  vs_mult(v, tc);
  vec_plus(w, w, u);
  vec_subt(w, w, v);

  return vec_dot(w, w);
}


// This function is adapted from geom.cpp in ODE, which
// is copyright Russell Smith.
// given two line segments A and B with endpoints a1-a2 and b1-b2, return the
// points on A and B that are closest to each other (in cp1 and cp2).
// in the case of parallel lines where there are multiple solutions, a
// solution involving the endpoint of at least one line will be returned.
// this will work correctly for zero length lines, e.g. if a1==a2 and/or
// b1==b2.
//
// the algorithm works by applying the voronoi clipping rule to the features
// of the line segments. the three features of each line segment are the two
// endpoints and the line between them. the voronoi clipping rule states that,
// for feature X on line A and feature Y on line B, the closest points PA and
// PB between X and Y are globally the closest points if PA is in V(Y) and
// PB is in V(X), where V(X) is the voronoi region of X.
void line_segment_closest_points(Real a1[3], Real a2[3],
				 Real b1[3], Real b2[3],
				 Real cp1[3], Real cp2[3])
{
  Real la, lb, k, da1, da2, da3, da4, db1, db2, db3, db4, det;
  Real a1a2[3], b1b2[3], a1b1[3], a1b2[3], a2b1[3], a2b2[3], n[3], tmp[3];

  // check vertex-vertex features

  vec_subt(a1a2, a2, a1);
  vec_subt(b1b2, b2, b1);
  vec_subt(a1b1, b1, a1);

  da1 = vec_dot(a1a2, a1b1);
  db1 = vec_dot(b1b2, a1b1);

  if ((da1 <= 0) && (db1 >= 0)) {
    vec_copy(cp1, a1);
    vec_copy(cp2, b1);
    return;
  }

  vec_subt(a1b2, b2, a1);
  da2 = vec_dot(a1a2, a1b2);
  db2 = vec_dot(b1b2,a1b2);
  if ((da2 <= 0) && (db2 <= 0)) {
    vec_copy(cp1, a1);
    vec_copy(cp2, b2);
    return;
  }

  vec_subt(a2b1, b1, a2);
  da3 = vec_dot(a1a2, a2b1);
  db3 = vec_dot(b1b2, a2b1);
  if ((da3 >= 0) && (db3 >= 0)) {
    vec_copy(cp1, a2);
    vec_copy(cp2, b1);
    return;
  }

  vec_subt(a2b2, b2, a2);
  da4 = vec_dot(a1a2, a2b2);
  db4 = vec_dot(b1b2, a2b2);
  if ((da4 >= 0) && (db4 <= 0)) {
    vec_copy(cp1, a2);
    vec_copy(cp2, b2);
    return;
  }

  // check edge-vertex features.
  // if one or both of the lines has zero length, we will never get to here,
  // so we do not have to worry about the following divisions by zero.

  la = vec_dot(a1a2, a1a2);
  if ((da1 >= 0) && (da3 <= 0.0)) {
    k = da1 / la;
    vsv_mult(tmp, k, a1a2);
    vec_subt(n, a1b1, tmp);

    if (vec_dot(b1b2, n) >= 0.0) {
      vec_plus(cp1, a1, tmp);
      vec_copy(cp2, b1);
      return;
    }
  }

  if ((da2 >= 0) && (da4 <= 0)) {
    k = da2 / la;
    vsv_mult(tmp, k, a1a2);
    vec_subt(n, a1b2, tmp);
    if (vec_dot(b1b2, n) <= 0.0) {
      vec_plus(cp1, a1, tmp);
      vec_copy(cp2, b2);
      return;
    }
  }

  lb = vec_dot(b1b2, b1b2);
  if ((db1 <= 0) && (db2 >= 0.0)) {
    k = -db1 / lb;
    vsv_mult(tmp, k, b1b2);
    vec_neg(a1b1);
    vec_subt(n, a1b1, tmp);
    if (vec_dot(a1a2, n) >= 0.0) {
      vec_copy(cp1,a1);
      vec_plus(cp2, b1, tmp);
      return;
    }
  }

  if ((db3 <= 0.0) && (db4 >= 0.0)) {
    k = -db3 / lb;
    vsv_mult(tmp, k, b1b2);
    vec_neg(a2b1);
    vec_subt(n, a2b1, tmp);
    if (vec_dot(a1a2, n) <= 0.0) {
      vec_copy(cp1,a2);
      vec_plus(cp2, b1, tmp);
      return;
    }
  }

  // it must be edge-edge

  k = vec_dot(a1a2, b1b2);
  det = la*lb - k*k;
  if (det <= 0.0) {
    // this should never happen, but just in case...
    vec_copy(cp1, a1);
    vec_copy(cp2, b1);
    return;
  }

  det = 1.0/det;
  Real alpha = (lb*da1 -  k*db1) * det;
  Real beta  = ( k*da1 - la*db1) * det;

  vs_mult(a1a2, alpha);
  vec_plus(cp1, a1, a1a2);

  vs_mult(b1b2, beta);
  vec_plus(cp2, b1, b1b2);
}


Real lines_distance_3d(Real a1[3], Real a2[3],
		       Real b1[3], Real b2[3])
{
  Real cp1[3], cp2[3], len[3];

  line_segment_closest_points(a1, a2, b1, b2, cp1, cp2);

  vec_subt(len, cp2, cp1);
  return vec_dot(len, len);
}
