/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stream.h>
#include <fstream.h>
#include "general.h"
#include "graphic.h"
#include "lsystem.h"
#include "pop.h"
#include "lsystem.h"
#include "bars2d.h"
#include "algebra.h"

#define ANGLE_TOLERANCE 0.0001
#define DIST_TOLERANCE 0.00001
#define SAME_AS_TOLERANCE 0.1
#define DIST_COLLISION 15.0

#define GROUND_TOLERANCE_MAX 1.0
#define GROUND_TOLERANCE_MIN 0.00001


#define MAX_CYCLE_DEPTH 19
#define MAX_ARM_LENGTH 4
#define OFFSET_STEP 0.25

#define DRAW_STEP 0.01
#define TIME_STEP_2D 0.0002
#define HISTORY_STEP 0.2

#define WAIT_CHECK_FAR 0.004
#define WAIT_CHECK_NEAR 0.0004

Real Ground_tolerance;


// Returns TRUE if the two line segments intersect
// each other.
int lines_intersect_2d(Real x1, Real y1, Real x2, Real y2,
		       Real x3, Real y3, Real x4, Real y4)
{
  Real denom, unk_a, unk_b;

  denom = (y4-y3)*(x2-x1) - (x4-x3)*(y2-y1);
  if (absv(denom) < 0.00001)
    // Lines are parallel - assume they don't intersect,
    // (or, if they do, they'll be caught when joints move.
    return FALSE;

  unk_a = (x4-x3)*(y1-y3) - (y4-y3)*(x1-x3);
  unk_a /= denom;

  if ((unk_a > 1.0) || (unk_a < 0.0)) {
    return FALSE;
  }

  unk_b = (x2-x1)*(y1-y3) - (y2-y1)*(x1-x3);
  unk_b /= denom;
  if ((unk_b > 1.0) || (unk_b < 0.0)) {
    return FALSE;
  }

  return TRUE;
}


// Returns TRUE if the two line segments intersect
// each other.
Real lines_distance_2d(Real x1, Real y1, Real x2, Real y2,
		       Real x3, Real y3, Real x4, Real y4)
{
  Real denom, unk_a, unk_b;
  Real loc_a[2], loc_b[2];

  denom = (y4-y3)*(x2-x1) - (x4-x3)*(y2-y1);
  if (absv(denom) < 0.00001) {
    // Lines are parallel - assume they don't intersect,
    // (or, if they do, they'll be caught when joints move.
    return 100.0;
  }

  unk_a = (x4-x3)*(y1-y3) - (y4-y3)*(x1-x3);
  unk_a /= denom;

  if (unk_a >= 1.0) {
    loc_a[0] = x2;
    loc_a[1] = y2;
  } else if (unk_a <= 0.0) {
    loc_a[0] = x1;
    loc_a[1] = y1;
  } else {
    loc_a[0] = x1 + unk_a *(x2-x1);
    loc_a[1] = y1 + unk_a *(y2-y1);
  }

  unk_b = (x2-x1)*(y1-y3) - (y2-y1)*(x1-x3);
  unk_b /= denom;
  if (unk_b >= 1.0) {
    loc_b[0] = x4;
    loc_b[1] = y4;
  } else if (unk_b <= 0.0) {
    loc_b[0] = x3;
    loc_b[1] = y3;
  } else {
    loc_b[0] = x3 + unk_b *(x4-x3);
    loc_b[1] = y3 + unk_b *(y4-y3);
  }

  if (Print[23])
    printf("a: %.3f:%.3f,  b: %.3f:%.3f\n", loc_a[0], loc_a[1],
	   loc_b[0], loc_b[1]);


  return sqr(loc_a[0]-loc_b[0]) + sqr(loc_a[1]-loc_b[1]);
}




BarStruct :: BarStruct(void)
{
  ground_pt = (int*)malloc(MAX_BARS2*sizeof(int));
  ground_pt_prev = (int*)malloc(MAX_BARS2*sizeof(int));
}


void BarStruct :: initialize(void)
{
  ref_pt = 0;
  Bar_stack_ptr = 0;
  Num_bars = 1;
  Num_joints = 0;
  ref_rotation = 0.0;

  Ground_tolerance = random_dbl()*GROUND_TOLERANCE_MAX + GROUND_TOLERANCE_MIN;
  //  Ground_tolerance = GROUND_TOLERANCE_MAX;

  Time_step_2d = TIME_STEP_2D;
  Time = 0.0;
  Time_history_prev = 0.0;

  Move[0] = 0.0;
  Move[1] = 0.0;

  Bar[0].type = BAR_TYPE_FIXED;
  Bar[0].moves = FALSE;
  Bar[0].move_spd = 0.0;
  Bar[0].move_offset = 0.0;
  Bar[0].angle_a = 0.0;
  Bar[0].length = 0;
  Bar[0].end_x = 0.0;
  Bar[0].end_y = 0.0;
  Bar[0].abs_angle = 0.0;
  Bar[0].parent = 0;
  Bar[0].same_as = -1;
  Bar[0].joint_index = 0;
  Bar[0].num_children = 0;
  Bar[0].num_from = 0;
  Bar[0].weight = 1.0;
}


void BarStruct :: set_translate(void)
{
  if (COM_X - draw_loc[0] > DRAW_TOLERANCE) {
    draw_loc[0] = COM_X - DRAW_TOLERANCE;
  } else if (draw_loc[0] - COM_X > DRAW_TOLERANCE) {
    draw_loc[0] = COM_X + DRAW_TOLERANCE;
  }

  win1->translate(-draw_loc[0], -com_y_avg, 0.0);
}



void BarStruct :: draw_history2(int h)
{
  int i;
  //  Real z = 0.0;
  Real z = -Real(h*50+50);
  Real width = Real(HISTORY_LENGTH - h);


  win1->SetLineWidth(width*5.0/Real(HISTORY_LENGTH+1));
  win1->startLine();

  win1->SetColor(col_grey3);
  for (i=0; i<Num_bars; i++) {
    win1->addVertex(Bar[Bar[i].parent].x_history[h], 0.01,
		    -Bar[Bar[i].parent].y_history[h] + z);
    win1->addVertex(Bar[i].x_history[h], 0.01, -Bar[i].y_history[h] + z);
  }

  // Draw creature.
  //  win1->SetColor(col_green2);
  win1->SetColor(col_blue);
  for (i=0; i<Num_bars; i++) {
    win1->addVertex(Bar[Bar[i].parent].x_history[h], 
		    Bar[Bar[i].parent].y_history[h], z);
    win1->addVertex(Bar[i].x_history[h], Bar[i].y_history[h], z);
  }

  for (i=0; i<Num_bars; i++) {
    if ((Bar[i].type != BAR_TYPE_FIXED) &&
	(Bar[i].num_children > 0)) {
      win1->starLocation(Bar[i].x_history[h], Bar[i].y_history[h], z);
    }
  }
  win1->stopLine();
}


void BarStruct :: draw_struct2(void)
{
  int i, ctr;
  Real x, z;


  set_translate();

  //  win1->SetColor(col_surf);
  win1->SetColor(col_surf3);
  //  win1->drawPlaneY(550.0*3.0, -1.0, 35.0*3.0);
  win1->drawPlaneY(25000.0, -0.1, 25000.0);

  win1->SetLineWidth(1.0);
  win1->startLine();

  // Stripes on road.
  ctr = 0;
  for (x=-10200.0; x<=10000.0; x+=100.0) {
    ctr++;
    if ((ctr%5) == 1)
      win1->SetColor(col_stripe1);
    else if ((ctr%5) == 2)
      win1->SetColor(col_stripe2);
    win1->addVertex(x, 0.0, -10000.0);
    win1->addVertex(x, 0.0, 10000.0);
  }
  ctr = 0;
  for (z=-10321.0; z<=10000.0; z+=100.0) {
    ctr++;
    if ((ctr%5) == 1)
      win1->SetColor(col_stripe1);
    else if ((ctr%5) == 2)
      win1->SetColor(col_stripe2);
    win1->addVertex(-10000, 0.0, z);
    win1->addVertex(10000, 0.0, z);
  }
  win1->stopLine();


  if (Option[11] || Option[12]) {
    for (i=0; i<HISTORY_LENGTH; i++) {
      draw_history2(i);
    }
  }

  win1->SetLineWidth(4.0);
  win1->startLine();

  // Draw shadow.
  win1->SetColor(col_grey3);
  for (i=0; i<Num_bars; i++) {
    win1->addVertex(Bar[Bar[i].parent].end_x, 0.01,
		    -Bar[Bar[i].parent].end_y);
    win1->addVertex(Bar[i].end_x, 0.01, -Bar[i].end_y);
  }

  // Draw creature.
  //  win1->SetColor(col_green2);
  win1->SetColor(col_blue);
  for (i=0; i<Num_bars; i++) {
    win1->addVertex(Bar[Bar[i].parent].end_x, Bar[Bar[i].parent].end_y);
    win1->addVertex(Bar[i].end_x, Bar[i].end_y);
  }

  for (i=0; i<Num_bars; i++) {
    if ((Bar[i].type != BAR_TYPE_FIXED) &&
	(Bar[i].num_children > 0)) {
      win1->starLocation(Bar[i].end_x, Bar[i].end_y);
    }
  }
  win1->stopLine();
}


void BarStruct :: draw_history(int h)
{
  int i;
  Real z = -Real(h*30 + 30);
  //  Real z = 0.0;
  Real width = Real(HISTORY_LENGTH - h);
  

  //  win1->SetLineWidth(2.0);
  win1->SetLineWidth(width*5.0/Real(HISTORY_LENGTH+1));
  win1->startLine();

  // Draw creature.
  win1->SetColor(col_green2);
  win1->starLocation(Bar[0].x_history[h], Bar[0].y_history[h], z);

  for (i=0; i<Num_bars; i++) {
    if (!Bar[i].moves) {
      win1->addVertex(Bar[Bar[i].parent].x_history[h], Bar[Bar[i].parent].y_history[h], z);
      win1->addVertex(Bar[i].x_history[h], Bar[i].y_history[h], z);
    }

    if (Bar[i].same_as >= 0)
      continue;

    if ((Bar[i].num_children==0) || 
	(Bar[i].type == BAR_TYPE_FIXED)) {
      win1->starLocation(Bar[i].x_history[h], Bar[i].y_history[h], z);
    }
  }

  win1->SetColor(col_blue);
  for (i=0; i<Num_bars; i++) {
    if (Bar[i].moves) {
      win1->addVertex(Bar[Bar[i].parent].x_history[h], Bar[Bar[i].parent].y_history[h], z);
      win1->addVertex(Bar[i].x_history[h], Bar[i].y_history[h], z);
    }

    if (Bar[i].same_as >= 0)
      continue;

    if ((Bar[i].num_children > 0) &&
	(Bar[i].type != BAR_TYPE_FIXED)) {
      win1->starLocation(Bar[i].x_history[h], Bar[i].y_history[h], z);
    }
  }

  win1->stopLine();
}


void BarStruct :: draw_struct(void)
{
  int i, ctr;
  Real x, z;


  set_translate();

  if (Option[8] != 10) {
    //  win1->SetColor(col_surf);
    if (!Option[8] || Option[3])
      win1->SetColor(col_surf2);
    else
      win1->SetColor(col_surf3);

    //  win1->drawPlaneY(550.0*3.0, -1.0, 35.0*3.0);
    win1->drawPlaneY(25000.0, -1.0, 25000.0);

    if (Option[8] != 3)
      win1->SetLineWidth(1.0);
    else
      win1->SetLineWidth(2.0);
    win1->startLine();

    // Stripes on road.
    ctr = 0;
    for (x=-10200.0; x<=10000.0; x+=100.0) {
      ctr++;
      if ((ctr%5) == 1)
	win1->SetColor(col_stripe1);
      else if ((ctr%5) == 2)
	win1->SetColor(col_stripe2);
      win1->addVertex(x, 0.0, -10000.0);
      win1->addVertex(x, 0.0, 10000.0);
    }
    ctr = 0;
    for (z=-10321.0; z<=10000.0; z+=100.0) {
      ctr++;
      if ((ctr%5) == 1)
	win1->SetColor(col_stripe1);
      else if ((ctr%5) == 2)
	win1->SetColor(col_stripe2);
      win1->addVertex(-10000, 0.0, z);
      win1->addVertex(10000, 0.0, z);
    }
    win1->stopLine();
  }

  if (Option[11] || Option[12]) {
    for (i=0; i<HISTORY_LENGTH; i++) {
      draw_history(i);
    }
  }

  if (Option[8] != 10) {
    win1->SetLineWidth(5.0);
  } else {
    win1->SetLineWidth(8.0);
  }
  win1->startLine();

  // Draw shadow.
  if (!Option[8])
    win1->SetColor(col_grey4);
  else if (Option[8] == 10) {
    win1->SetColor(col_grey10);
  } else
    win1->SetColor(col_grey6);
  for (i=0; i<Num_bars; i++) {
    win1->addVertex(Bar[Bar[i].parent].end_x, 0.01,
		    -Bar[Bar[i].parent].end_y);
    win1->addVertex(Bar[i].end_x, 0.01, -Bar[i].end_y);
  }


  // Draw center of mass.
  //  win1->SetColor(col_black);
  //  win1->starLocation(COM_X, COM_Y);

  // Draw creature.
  //  win1->SetColor(col_black);
  //  win1->SetColor(col_green2);

  /*
  if (Option[8] == 0)
    win1->SetColor(col_black);
  else {
    if (Option[14] == 1) {
      win1->SetColor(col_blue3);
    } else {
      win1->SetColor(col_blue);
    }
  }
  */

  // Ground points.
  if ((Option[8] == 0) || (Option[8] == 10)) {
    win1->SetColor(col_black);

    for (i=0; i<Num_bars; i++) {
      if (Bar[i].same_as >= 0)
	continue;
      
      if (Bar[i].ground_pt) {
	win1->starLocation(Bar[i].end_x, Bar[i].end_y);
      }
    }

  } else {
    //  win1->SetColor(col_green2);
    //  win1->SetColor(col_blue);
    // Fixed bars:
    if (Option[14] == 1) {
      win1->SetColor(col_blue3);
    } else {
      win1->SetColor(col_blue);
    }
    if (!Bar[0].ground_pt) {
      win1->starLocation(Bar[0].end_x, Bar[0].end_y);
    }
  }

  for (i=0; i<Num_bars; i++) {
    if (!Bar[i].moves) {
      win1->addVertex(Bar[Bar[i].parent].end_x, Bar[Bar[i].parent].end_y);
      win1->addVertex(Bar[i].end_x, Bar[i].end_y);
    }

    if (Bar[i].same_as >= 0)
      continue;

    if ((Bar[i].num_children==0) || 
	(Bar[i].type == BAR_TYPE_FIXED)) {
      if (!Bar[i].ground_pt) { 
	win1->starLocation(Bar[i].end_x, Bar[i].end_y);
      }
    }
  }

  // Moving bars:
  //  win1->SetColor(col_blue);
  if (Option[8] == 10) {
    win1->SetColor(col_black);

  } else if (Option[8] != 4) {
    win1->SetColor(col_green6); // col_green4
  } else {
    if (Option[14] == 1) {
      win1->SetColor(col_blue3);
    } else {
      win1->SetColor(col_blue);
    }
  }
  for (i=0; i<Num_bars; i++) {
    if (Bar[i].moves) {
      win1->addVertex(Bar[Bar[i].parent].end_x, Bar[Bar[i].parent].end_y);
      win1->addVertex(Bar[i].end_x, Bar[i].end_y);
    }

    if (Bar[i].same_as >= 0)
      continue;

    if ((Bar[i].num_children > 0) &&
	(Bar[i].type != BAR_TYPE_FIXED)) {
      if (!Bar[i].ground_pt) { 
	win1->starLocation(Bar[i].end_x, Bar[i].end_y);
      }
    }
  }
  win1->stopLine();

  /*
  win1->SetLineWidth(6.0);
  if (Option[8] != 4)
    win1->SetColor(col_green5);
  else {
    if (Option[14] == 1) {
      win1->SetColor(col_blue3);
    } else {
      win1->SetColor(col_blue);
    }
  }
  win1->startLine();
  for (i=0; i<Num_bars; i++) {
    if (Bar[i].same_as >= 0)
      continue;
    
    if (Bar[i].moves)
      if (!Bar[i].ground_pt) {
	win1->starLocationBig(Bar[i].end_x, Bar[i].end_y);
      }
  }
  win1->stopLine();
  */
  /*
  win1->SetColor(col_black);
  for (i=0; i<Num_bars; i++) {
    win1->addVertex(Bar[Bar[i].parent].end_x, Bar[Bar[i].parent].end_y);
    win1->addVertex(Bar[i].end_x, Bar[i].end_y);
    win1->starLocation(Bar[i].end_x, Bar[i].end_y);
  }
  */

}



void BarStruct :: draw_bars(int arg)
{
  if (win1 == 0) {
    printf("win1 not defined\n");
    return;
  }

  if (Draw_arg[2]) {
    Time += Time_step_2d;
    calculate_positions(Time);
    settle_structure();

  } else if (Draw_arg[1]) {
    Real prev_time = Time;
    while (Time < prev_time + DRAW_STEP) {
      Time += Time_step_2d;
      calculate_positions(Time);
      settle_structure();
    }
  }

  if (Option[8] != 2)
    draw_struct();
  else
    draw_struct2();

#ifndef OPTIMIZE
  // Save history.
  if (Option[11] || Draw_arg[3]) {
    if ((Time >= Time_history_prev + HISTORY_STEP) || Draw_arg[3]) {
      Time_history_prev = Time;
      for (int i=0; i<Num_bars; i++) {
	for (int j=HISTORY_LENGTH-1; j>0; j--) {
	  Bar[i].x_history[j] = Bar[i].x_history[j-1];
	  Bar[i].y_history[j] = Bar[i].y_history[j-1];
	}
	Bar[i].x_history[0] = Bar[i].end_x;
	Bar[i].y_history[0] = Bar[i].end_y;
      }
    }
  }
#endif
}


void BarStruct :: add_connection_noise(void)
{
  int i, j, index;

  for (i=0; i<Num_bars; i++) {
    for (j=0; j<Bar[i].num_children; j++) {
      if (Bar[i].in_cycle[j])
	continue;

      index = Bar[i].child[j];
      Bar[index].angle_a = random_dbl()*0.15 - 0.075 + Bar[index].true_angle_a;
      if (Bar[index].moves)
	Bar[index].angle_b = random_dbl()*0.15 - 0.075 + Bar[index].true_angle_b;
    }
  }
}


// Test an arm for length.
int BarStruct :: arm_length(int node, int depth)
{
  int length, max_length = depth;

#ifndef OPTIMIZE
  //  if (Print[25])
  //    printf(" %d : %d\n", node, depth);
#endif
  int i, child, parent;

  Bar[node].updated = depth;

  //  printf("%d (%d)\n", node, depth);

  // Test same_as.
  if (Bar[node].same_as != -1) {
    child = Bar[node].same_as;
    if (Bar[child].updated == 0) {
      length = arm_length(Bar[node].same_as, depth);
      if (length > max_length)
	max_length = length;
    }
  }

  // Test parent.
  parent = Bar[node].parent;
  if (Bar[parent].updated == 0) {
    length = arm_length(parent, depth+1);
    if (length > max_length)
      max_length = length;
  }
  
  // Test children.
  for (i=0; i<Bar[node].num_children; i++) {
    child = Bar[node].child[i];
    if (Bar[child].updated == 0) {
      length = arm_length(child, depth+1);
      if (length > max_length)
	max_length = length;
    }
  }

  // Test from list.
  for (i=0; i<Bar[node].num_from; i++) {
    child = Bar[node].from_list[i];
    if (Bar[child].updated == 0) {
      length = arm_length(child, depth+1);
      if (length > max_length)
	max_length = length;
    }
  }

  Bar[node].updated = 0;
  return max_length;
}


// Returns the length of the longest arm.
int BarStruct :: max_arm_length(void)
{
  int i, j;
  int index_j;
  int prev_i, prev_j;
  int max_length = 1;
  int length1, length2;

  // Initialize data structures.
  for (i=0; i<Num_bars; i++) {
    Bar[i].updated = 0; // distance.
    for (j=0; j<Bar[i].num_children; j++) {
      if (Bar[i].in_cycle[j]) {
	Bar[i].updated = 1;
	Bar[Bar[i].child[j]].updated = 1;
      }
    }

    if (Bar[i].same_as != -1) {
      if (Bar[Bar[i].same_as].updated == 1)
	Bar[i].updated = 1;
      else if (Bar[i].updated == 1)
	Bar[Bar[i].same_as].updated = 1;
    }
  }

#ifndef OPTIMIZE
  if (Print[25]) {
    for (i=0; i<Num_bars; i++) {
      printf("%d(%d): ", i,Bar[i].updated);
      if (Bar[i].same_as == -1) {
	for (j=0; j<Bar[i].num_children; j++) {
	  printf("%d(%d)  ", Bar[i].child[j], Bar[i].in_cycle[j]);
	}
	if (Bar[i].num_from > 0) {
	  printf(":  ");
	  for (j=0; j<Bar[i].num_from; j++) {
	    printf("%d  ", Bar[i].from_list[j]);
	  }
	}
	
      } else {
	printf(" = %d", Bar[i].same_as);
      }
      printf("\n");
    }
  }
#endif



  // Test nodes.
  for (i=0; i<Num_bars; i++) {
    // Test all children.
    for (j=0; j<Bar[i].num_children; j++) {
      if (Bar[i].in_cycle[j]) {
	continue; // Don't need to test length of edge in cycle.
      }

      index_j = Bar[i].child[j];

      prev_i = Bar[i].updated;
      prev_j = Bar[index_j].updated;

      Bar[i].updated = 1;
      length1 = arm_length(index_j, 2);
      Bar[index_j].updated = 1;
      length2 = arm_length(i, 2);

      Bar[i].updated = prev_i;
      Bar[index_j].updated = prev_j;

#ifndef OPTIMIZE
      if (Print[25]) {
	printf("Arm length: %d - %d(%d):  %d, %d\n", i, index_j, j,
	       length1, length2);
      }
#endif

      if (length1 > length2) {
	if (length2 > max_length)
	  max_length = length2;
      } else {
	if (length1 > max_length)
	  max_length = length1;
      }
    }
  }


#ifndef OPTIMIZE
  if (Print[25]) {
    for (i=0; i<Num_bars; i++) {
      printf("%d(%d): ", i,Bar[i].updated);
      if (Bar[i].same_as == -1) {
	for (j=0; j<Bar[i].num_children; j++) {
	  printf("%d(%d)  ", Bar[i].child[j], Bar[i].in_cycle[j]);
	}
	if (Bar[i].num_from > 0) {
	  printf(":  ");
	  for (j=0; j<Bar[i].num_from; j++) {
	    printf("%d  ", Bar[i].from_list[j]);
	  }
	}
	
      } else {
	printf(" = %d", Bar[i].same_as);
      }
      printf("\n");
    }
  }
#endif


#ifndef OPTIMIZE
  if (Print[25]) {
    printf("Max Arm length: %d\n", max_length);
  }
#endif

  return max_length;
}


// Test to see if arms (non-cycle edges) are too long.
int BarStruct :: test_arm_lengths(void)
{
  int i, j;
  int index_j;
  int length1, length2;

  // Initialize data structures.
  for (i=0; i<Num_bars; i++) {
    Bar[i].updated = 0; // distance.
  }


  // Test nodes.
  for (i=0; i<Num_bars; i++) {
    // Test all children.
    for (j=0; j<Bar[i].num_children; j++) {
      if (Bar[i].in_cycle[j]) {
	continue; // Don't need to test length of edge in cycle.
      }

      index_j = Bar[i].child[j];

      Bar[i].updated = 1;
      length1 = arm_length(index_j, 2);
      Bar[index_j].updated = 1;
      length2 = arm_length(i, 2);
      Bar[index_j].updated = 0;

#ifndef OPTIMIZE
      if (Print[25]) {
	printf("Arm length: %d - %d(%d):  %d, %d\n", i, index_j, j,
	       length1, length2);
      }
#endif

      if ((length1 > MAX_ARM_LENGTH) && (length2 > MAX_ARM_LENGTH)) {
#ifndef OPTIMIZE
	if (Print[9]) {
	  printf("Arm too long: %d - %d(%d):  %d,%d\n", i, index_j, j,
		 length1, length2);
	}
#endif
	return FALSE;
      }
    }
  }

  return TRUE;
}



int BarStruct :: test_children(int node, int root, int depth)
{
  if (depth > MAX_CYCLE_DEPTH)
    return FALSE;

  int i, child, parent;

  if (Bar[node].same_as != -1) {
    if (Bar[node].same_as == root) {
      return TRUE;
    }

    if (Bar[Bar[node].same_as].updated == 0) {
      Bar[node].updated = depth + 1;
      Bar[Bar[node].same_as].updated = depth + 1;
      int ret_val = test_children(Bar[node].same_as, root, depth);
      Bar[node].updated = 0;
      Bar[Bar[node].same_as].updated = 0;
      return ret_val;
    }
  }

  // Test parent.
  parent = Bar[node].parent;
  if (depth > 2) {
    if (parent == root) {
      // Cycle!
      return TRUE;
    }
  }
  if (Bar[parent].updated == 0) {
    Bar[parent].updated = depth + 1;
    if (test_children(parent, root, depth+1)) {
      // Cycle!
      Bar[parent].updated = 0;
      return TRUE;
    }
    Bar[parent].updated = 0;
  }

  // Test children.
  for (i=0; i<Bar[node].num_children; i++) {
    child = Bar[node].child[i];
    if (child == root) {
      Bar[node].in_cycle[i] = TRUE;
      return TRUE;
    }

    if (Bar[child].updated != 0)
      continue;

    Bar[child].updated = depth + 1;
    if (test_children(child, root, depth+1)) {
      // cycle!
      Bar[child].updated = 0;
      Bar[node].in_cycle[i] = TRUE;
      return TRUE;
    }
    Bar[child].updated = 0;
  }

  // Test from list.
  for (i=0; i<Bar[node].num_from; i++) {
    child = Bar[node].from_list[i];
    if (depth > 2) {
      if (child == root) {
	return TRUE;
      }
    }

    if (Bar[child].updated != 0)
      continue;

    Bar[child].updated = depth + 1;
    if (test_children(child, root, depth+1)) {
      // cycle!
      Bar[child].updated = 0;
      return TRUE;
    }
    Bar[child].updated = 0;
  }
    
  return FALSE;
}



void BarStruct :: label_cycles(void)
{
  int i, j;
  int index_j;

  // Initialize data structures.
  for (i=0; i<Num_bars; i++) {
    Bar[i].updated = 0; // distance.
    for (j=0; j<Bar[i].num_children; j++) {
      Bar[i].in_cycle[j] = FALSE;
    }
  }

  // Find cycles.
  for (i=0; i<Num_bars; i++) {
    Bar[i].updated = 1; // distance.

    // Test all children.
    for (j=0; j<Bar[i].num_children; j++) {
      if (Bar[i].in_cycle[j]) {
	// If already in cycle, don't need to check.
	continue;
      }

      index_j = Bar[i].child[j];
      Bar[index_j].updated = 2;
      if (test_children(index_j, i, 2)) {
	Bar[i].in_cycle[j] = TRUE;
      }
      Bar[index_j].updated = 0;
    }
    Bar[i].updated = 0;
  }

#ifndef OPTIMIZE
  if (Print[25]) {
    for (i=0; i<Num_bars; i++) {
      printf("%d: ", i);
      if (Bar[i].same_as == -1) {
	for (j=0; j<Bar[i].num_children; j++) {
	  printf("%d(%d)  ", Bar[i].child[j], Bar[i].in_cycle[j]);
	}
	if (Bar[i].num_from > 0) {
	  printf(":  ");
	  for (j=0; j<Bar[i].num_from; j++) {
	    printf("%d  ", Bar[i].from_list[j]);
	  }
	}
	
      } else {
	printf(" = %d", Bar[i].same_as);
      }
      printf("\n");
    }
  }
#endif
}


void BarStruct :: initialize_intersection(void)
{
  int i, j;
  int pi, pj, jt_i, jt_j;
  int near;

  last_check_near = 0.0;
  last_check_far = 0.0;

  for (i=1; i<Num_bars-1; i++) {
    pi = Bar[i].parent;
    jt_i = Bar[i].joint_index;

    near = 0;
    for (j=i+1; j<Num_bars; j++) {
      jt_j = Bar[j].joint_index;
      if (jt_i == jt_j) {
	// Part of same joint structure.
	bar_check[i][j] = 0;
	continue;
      }

      pj = Bar[j].parent;      
      if ((pj == i) || (pj == pi)) {
	// Bars are connected.
	bar_check[i][j] = 0;
	continue;
      }

      // Test for "same_as" connected.
      if ((Bar[j].same_as == i) ||
	  (Bar[j].same_as == pi) ||
	  ((Bar[j].same_as == Bar[i].same_as) &&
	   (Bar[j].same_as != -1))) {
	// Same as connection.
	bar_check[i][j] = 0;
	continue;
      }


      near++;
      bar_check[i][j] = 2;
    }
    num_near[i] = near;
  }

#ifndef OPTIMIZE
  if (Print[25]) {
    for (i=1; i<Num_bars; i++) {
      printf("%2d: #%2d : ", i, num_near[i]);
      for (j=i+1; j<Num_bars; j++)
	printf("%d", bar_check[i][j]);
      printf("\n");
    }
  }
#endif
}


// No graphics: (small = 2 time steps; far = 20 time steps)
// 17s with checking.
// 15s iterating - no checking
// 9s with empty function call.

// Graphics: (small = time_step; far = every 10 time steps)
// 25s with checking
// 20s w/o checking (but performing n^2 looping iterations)
// 10s with empty function call.
int BarStruct :: is_intersection(void)
{
  int i, j;
  int pi, pj;
  int jt_i;
  int check_val, near;
  Real ci_x, ci_y, cj_x, cj_y;
  Real dist, length, dist_apart;


  if (Time - last_check_far >= WAIT_CHECK_FAR) {
    last_check_far = Time;
    last_check_near = Time;
    check_val = 1;

    // Test to to make sure "same_as" nodes have same locatino.
    for (i=1; i<Num_bars; i++) {
      if (Bar[i].same_as != -1) {
	j = Bar[i].same_as;
	if (absv(Bar[i].end_x - Bar[j].end_x) > SAME_AS_TOLERANCE) {
	  if (Print[9]) {
	    printf("Bars %d(%.4f:%.4f):%d(%.4f:%.4f) (%lf:%lf) moved apart.\n",
		   i, Bar[i].end_x, Bar[i].end_y,
		   j, Bar[j].end_x, Bar[j].end_y,
		   absv(Bar[i].end_x - Bar[j].end_x),
		   absv(Bar[i].end_y - Bar[j].end_y));
	  }
	  return TRUE;
	}
	if (absv(Bar[i].end_y - Bar[j].end_y) > SAME_AS_TOLERANCE) {
	  if (Print[9]) {
	    printf("Bars %d:%d moved apart.\n", i, j);
	  }
	  return TRUE;
	}
      }
    }

  } else if (Time - last_check_near >= WAIT_CHECK_NEAR) {
    last_check_near = Time;
    check_val = 2;
  } else {
    // No need to test.
    //    printf("No test: %.5f, %.5f, %.5f\n", Time, last_check_near, last_check_far);
    return FALSE;
  }

  for (i=1; i<Num_bars; i++) {
    pi = Bar[i].parent;
    Bar[i].center_x = (Bar[i].end_x + Bar[pi].end_x)/2.0;
    Bar[i].center_y = (Bar[i].end_y + Bar[pi].end_y)/2.0;
  }
    
  for (i=1; i<Num_bars; i++) {
    pi = Bar[i].parent;
    jt_i = Bar[i].joint_index;
    ci_x = Bar[i].center_x;
    ci_y = Bar[i].center_y;

    if (check_val == 2) {
      if (num_near[i] == 0)
	continue;
    }

    near = 0;
    for (j=i+1; j<Num_bars; j++) {
      if (bar_check[i][j] < check_val) {
	// Don't need to test yet.
	continue;
      }

      cj_x = Bar[j].center_x;
      cj_y = Bar[j].center_y;

      dist = sqr(ci_x - cj_x) + sqr(ci_y-cj_y);
      length = sqr((Bar[i].length+Bar[j].length)/2.0);
      if (dist < 4.0*length) {
	near++;
	bar_check[i][j] = 2;
      } else
	bar_check[i][j] = 1;

      /*
      if (Time < 4.0*TIME_STEP_2D) {
	printf("dist: %.2f, length: %.2f  -> %d\n", dist, length, bar_check[i][j]);
      }
      */
      
      if (dist > length)
	// Too far apart.
	continue;


      pj = Bar[j].parent;
      dist_apart = lines_distance_2d(Bar[pi].end_x, Bar[pi].end_y,
				     Bar[i].end_x, Bar[i].end_y,
				     Bar[pj].end_x, Bar[pj].end_y,
				     Bar[j].end_x, Bar[j].end_y);
      /*
      if (lines_intersect_2d(Bar[pi].end_x, Bar[pi].end_y,
			     Bar[i].end_x, Bar[i].end_y,
			     Bar[pj].end_x, Bar[pj].end_y,
			     Bar[j].end_x, Bar[j].end_y)) {
      */

      if (dist_apart < DIST_COLLISION) {
#ifndef OPTIMIZE
	if (Print[10]) {
	  printf("Intersection between %d and %d\n", i, j);
	  printf("%.2f:%.2f : %.2f:%.2f  -  %.2f:%.2f : %.2f:%.2f\n",
		 Bar[pi].end_x, Bar[pi].end_y,
		 Bar[i].end_x, Bar[i].end_y,
		 Bar[pj].end_x, Bar[pj].end_y,
		 Bar[j].end_x, Bar[j].end_y);
	}
#endif
	return TRUE;
      }
      if (dist_apart < 4.0*length) {
	near++;
	bar_check[i][j] = 2;
      } else
	bar_check[i][j] = 1;
    }
    num_near[i] = near;
  }


#ifndef OPTIMIZE
  if (Print[19]) {
    for (i=1; i<Num_bars; i++) {
      for (j=i+1; j<Num_bars; j++)
	printf("%d", bar_check[i][j]);
      printf("\n");
    }
  }
#endif

  return FALSE;
}


// 72 bars, 7 joints:
// no intersection test: 10s
// is_intersection: 60s.
// is intersection old 1: 80s.
int BarStruct :: is_intersection_old2(void)
{
  int i, j;
  int pi, pj;
  int jt_i, jt_j;
  Real ci_x, ci_y, cj_x, cj_y;
#ifndef OPTIMIZE
  //  int ctr0 = 0, ctr1 = 0, ctr2 = 0, ctr3 = 0;
#endif

  for (i=1; i<Num_bars; i++) {
    pi = Bar[i].parent;
    Bar[i].center_x = (Bar[i].end_x + Bar[pi].end_x)/2.0;
    Bar[i].center_y = (Bar[i].end_y + Bar[pi].end_y)/2.0;
  }


  for (i=1; i<Num_bars; i++) {
    pi = Bar[i].parent;
    jt_i = Bar[i].joint_index;
    ci_x = Bar[i].center_x;
    ci_y = Bar[i].center_y;

    for (j=i+1; j<Num_bars; j++) {
      //      ctr0++;

      jt_j = Bar[j].joint_index;
      if (jt_i == jt_j)
	// Part of same joint structure.
	continue;

      //      ctr1++;

      pj = Bar[j].parent;      
      if (pj == i) {
	// Bars are connected.
	continue;
      }

      //      ctr2++;

      cj_x = Bar[j].center_x;
      cj_y = Bar[j].center_y;

      if ((sqr(ci_x - cj_x) + sqr(ci_y-cj_y)) >
	  sqr((Bar[i].length+Bar[j].length)/2.0))
	// Too far apart.
	continue;

#ifndef OPTIMIZE
      //      ctr3++;
#endif

      if (lines_intersect_2d(Bar[pi].end_x, Bar[pi].end_y,
			     Bar[i].end_x, Bar[i].end_y,
			     Bar[pj].end_x, Bar[pj].end_y,
			     Bar[j].end_x, Bar[j].end_y)) {
#ifndef OPTIMIZE
	if (Print[10]) {
	  printf("Intersection between %d and %d\n", i, j);
	  printf("%.2f:%.2f : %.2f:%.2f  -  %.2f:%.2f : %.2f:%.2f\n",
		 Bar[pi].end_x, Bar[pi].end_y,
		 Bar[i].end_x, Bar[i].end_y,
		 Bar[pj].end_x, Bar[pj].end_y,
		 Bar[j].end_x, Bar[j].end_y);
	}
#endif
	return TRUE;
      }
    }
  }

#ifndef OPTIMIZE
  //  printf("ctr0: %d;  ctr1: %d;  ctr2: %d;  ctr3: %d\n", ctr0, ctr1, ctr2, ctr3);
#endif
  return FALSE;
}



int BarStruct :: is_intersection_old1(void)
{
  int i, j;
  int pi, pj;

  for (i=1; i<Num_bars; i++) {
    pi = Bar[i].parent;
    for (j=i+1; j<Num_bars; j++) {
      pj = Bar[j].parent;
      if ((pj == i) || (pi == j)) {
	// Bars are connected.
	continue;
      }

      if (lines_intersect_2d(Bar[pi].end_x, Bar[pi].end_y,
			     Bar[i].end_x, Bar[i].end_y,
			     Bar[pj].end_x, Bar[pj].end_y,
			     Bar[j].end_x, Bar[j].end_y)) {
#ifndef OPTIMIZE
	if (Print[10]) {
	  printf("Intersection between %d and %d\n", i, j);
	  printf("%.2f:%.2f : %.2f:%.2f  -  %.2f:%.2f : %.2f:%.2f\n",
		 Bar[pi].end_x, Bar[pi].end_y,
		 Bar[i].end_x, Bar[i].end_y,
		 Bar[pj].end_x, Bar[pj].end_y,
		 Bar[j].end_x, Bar[j].end_y);
	}
#endif
	return TRUE;
      }
    }
  }
  return FALSE;
}


void BarStruct :: calculate_up(Real time)
{
  int child=ref_pt, parent;
  Real angle, theta;

#ifndef OPTIMIZE
  if (Print[20]) {
    printf("Calculate up\n");
    printf("\nRef point: %d, %.4f:%.4f (%.4f)\n", ref_pt,
	   Bar[ref_pt].end_x, Bar[ref_pt].end_y, Bar[ref_pt].abs_angle);
  }
#endif

  while (child != 0) {
    parent = Bar[child].parent;

#ifdef OLD
    if (Bar[parent].moves) {
      theta = time * Bar[Bar[parent].parent].move_spd;
      theta -= floor(theta);
      theta *= 2.0;
      if (theta > 1.0)
	theta = 2.0 - theta;

      Bar[parent].abs_angle = Bar[child].abs_angle - Bar[parent].angle_a*(1.0-theta) + Bar[parent].angle_b*theta;
      
    } else {
      Bar[parent].abs_angle = Bar[child].abs_angle - Bar[parent].angle_a;
    }
#endif

    if (Bar[child].moves) {
      theta = time * Bar[parent].move_spd + Bar[parent].move_offset;
      theta -= floor(theta);
      theta *= 2.0;
      if (theta > 1.0)
	theta = 2.0 - theta;
      angle = Bar[child].angle_a*(1.0-theta) + Bar[child].angle_b*theta;

    } else {
      angle = Bar[child].angle_a;
    }


    Bar[parent].abs_angle = Bar[child].abs_angle - angle;
    angle = Bar[child].abs_angle;
    Bar[parent].end_x = Bar[child].end_x - Bar[child].length*cos(angle);
    Bar[parent].end_y = Bar[child].end_y - Bar[child].length*sin(angle);

#ifndef OPTIMIZE
    if (Print[20]) {
      printf("%.3f:%.3f (%.1f:%.3f; %.3f:%.3f)\n", Bar[parent].end_x,
	     Bar[parent].end_y, Bar[child].length, Bar[child].abs_angle,
	     Bar[child].end_x, Bar[child].end_y);
    }
#endif

    child = parent;
  }
}



void BarStruct :: calculate_positions(Real time)
{
  int i, parent;
  Real angle;
  Real theta;

  if (Num_bars == 0)
    return;

  for (i=0; i<Num_bars; i++) {
    Bar[i].updated = FALSE;
    Bar[i].prev_x = Bar[i].end_x;
    Bar[i].prev_y = Bar[i].end_y;
  }

  // Calculate up from reference point.
  Bar[ref_pt].abs_angle += ref_rotation;
  ref_rotation = 0.0;
  if (ref_pt != 0) {
    calculate_up(time);
  }


  for (i=1; i<Num_bars; i++) {
    if (Bar[i].updated)
      continue;

    parent = Bar[i].parent;

    angle = Bar[parent].abs_angle;
    if (Bar[i].moves) {
      theta = time * Bar[parent].move_spd + Bar[parent].move_offset;
      theta -= floor(theta);
      theta *= 2.0;
      if (theta > 1.0)
	theta = 2.0 - theta;

      angle += Bar[i].angle_a*(1.0-theta) + Bar[i].angle_b*theta;
#ifndef OPTIMIZE
      if (Print[12]) {
	printf("spd: %.1f  theta: %.5f, angle: %.5f (%.4f) (%.4f : %.4f)\n",
	       Bar[parent].move_spd, theta, angle, 
	       Bar[parent].abs_angle, Bar[i].angle_a, Bar[i].angle_b);
      }
#endif
    } else {
      angle += Bar[i].angle_a;
    }

    /*
    while (angle > 2.0*PI)
      angle -= 2.0*PI;
    while (angle < 0.0)
      angle += 2.0*PI;
      */

    Bar[i].end_x = Bar[parent].end_x + Bar[i].length*cos(angle);
    Bar[i].end_y = Bar[parent].end_y + Bar[i].length*sin(angle);
    Bar[i].abs_angle = angle;

    /*
    len = sqrt(sqr(Bar[i].end_x-Bar[parent].end_x) +
	       sqr(Bar[i].end_y-Bar[parent].end_y));
    printf("Bar %d(%d:%d): %.1f:%.1f -> (%.1f:%.1f:%.1f) -> %.1f:%.1f\n",
	   i, parent, Bar[i].num_children, Bar[parent].end_x, Bar[parent].end_y,
	   Bar[i].length, len, Bar[i].abs_angle * 180.0/PI,
	   Bar[i].end_x, Bar[i].end_y);
	   */
  }

  calculate_com();

#ifndef OPTIMIZE
  if (Print[20])
    print_struct();
#endif
}

int BarStruct :: cycle_for_intersections(void)
{
  Real time;
  
  for (time = 1.0; time >= 0.0; time -= 0.025) {
    calculate_positions(time);
    if (is_intersection())
      return TRUE;
  }

  return FALSE;
}


// Calculates the center-of-mass of the structure.
void BarStruct :: calculate_com(void)
{
  int i;

  COM_X = 0.0;
  COM_Y = 0.0;
  COM_Z = 0.0;
  for (i=0; i<Num_bars; i++) {
    COM_X += Bar[i].end_x;
    COM_Y += Bar[i].end_y;
  }

  COM_X /= Real(Num_bars);
  COM_Y /= Real(Num_bars);
}


void BarStruct :: print_struct(void)
{
  printf("CoM: %.6f, %.6f\n", COM_X, COM_Y);
  for (int i=0; i<Num_bars; i++) {
    printf("%3d(%3d):  %.4f %.4f  (%.4f:%d)  ", i, Bar[i].same_as,
	   Bar[i].end_x, Bar[i].end_y, Bar[i].abs_angle, Bar[i].ground_pt);
    printf("t:%d  %.4f - %.4f (%.4f) (%d)\n", Bar[i].type,
	   Bar[i].angle_a, Bar[i].angle_b, Bar[i].move_spd,
	   Bar[i].num_children);
  }
}


// Moves the entire structure by x:y.
void BarStruct :: translate_struct(Real x, Real y)
{
  int i;

#ifndef OPTIMIZE
  if (Print[12]) {
    printf("Translate: %.4f, %.4f  - ", x, y);
    printf("CoM: %.6f, %.6f\n", COM_X, COM_Y);
  }
#endif

  COM_X += x;
  COM_Y += y;

  for (i=0; i<Num_bars; i++) {
    Bar[i].end_x += x;
    Bar[i].end_y += y;
  }
}



// Rotates the structure by angle.
void BarStruct :: rotate_struct(Real angle)
{
  int i;
  Real x, y;
  Real cos_ang, sin_ang;

#ifndef OPTIMIZE
  if (Print[12]) {
    printf("Rotate: %.4f             - ", angle);
    printf("CoM: %.6f, %.6f\n", COM_X, COM_Y);
  }
#endif

  cos_ang = cos(angle);
  sin_ang = sin(angle);

  x = cos_ang*COM_X - sin_ang * COM_Y;
  y = sin_ang*COM_X + cos_ang * COM_Y;
  COM_X = x;
  COM_Y = y;

  //  Bar[0].abs_angle += angle;
  ref_rotation += angle;
  for (i=0; i<Num_bars; i++) {
    //  Bar[i].abs_angle += angle;
    x = cos_ang*Bar[i].end_x - sin_ang * Bar[i].end_y;
    y = sin_ang*Bar[i].end_x + cos_ang * Bar[i].end_y;
    Bar[i].end_x = x;
    Bar[i].end_y = y;  
  }
}


int BarStruct :: settle_right(int right_min_index)
{
  int i, bal_index=-1;
  Real slope, slope2;
  Real angle;
  Real old_x, old_y;

  slope = 9999999.0;
  ref_pt = right_min_index;

  // Find left balance point.
  for (i=0; i<Num_bars; i++) {
    if (i == right_min_index)
      continue;

    if (Bar[i].end_x > Bar[right_min_index].end_x) {
      slope2 = (Bar[i].end_y-Bar[right_min_index].end_y) /
	(Bar[i].end_x-Bar[right_min_index].end_x);
      if (slope2 < slope) {
	bal_index = i;
	slope = slope2;
      }
    }
  }

#ifndef OPTIMIZE
  if (Print[12]) {
    printf("Settle right:\n");
    printf("  pivot: %2d  %.4f %.4f\n", right_min_index,
	   Bar[right_min_index].end_x,
	   Bar[right_min_index].end_y);
    printf("  other: %2d  %.4f %.4f\n", bal_index,
	   Bar[bal_index].end_x,
	   Bar[bal_index].end_y);
  }
#endif

  // Rotate clockwise and re-settle().
  angle = atan(slope);
  old_x = Bar[right_min_index].end_x;
  old_y = Bar[right_min_index].end_y;  

  /*
  printf("r pivot before: %2d  %.4f %.4f\n", right_min_index,
	 Bar[right_min_index].end_x,
	 Bar[right_min_index].end_y);
  */

  rotate_struct(-angle);
  translate_struct(old_x - Bar[right_min_index].end_x,
		   old_y - Bar[right_min_index].end_y);

  /*
  printf("  pivot after: %2d  %.4f %.4f\n\n", right_min_index,
	 Bar[right_min_index].end_x,
	 Bar[right_min_index].end_y);
  */

#ifndef OPTIMIZE
  if (Print[12])
    print_struct();
#endif

  return TRUE;
}



int BarStruct :: settle_left(int left_min_index)
{
  int i, bal_index=-1;
  Real slope, slope2;
  Real angle;
  Real old_x, old_y;

  slope = -9999999.0;
  ref_pt = left_min_index;


  // Find left balance point.
  for (i=0; i<Num_bars; i++) {
    if (i == left_min_index)
      continue;

    if (Bar[i].end_x < Bar[left_min_index].end_x) {
      slope2 = (Bar[i].end_y-Bar[left_min_index].end_y) /
	(Bar[i].end_x-Bar[left_min_index].end_x);
      if (slope2 > slope) {
	bal_index = i;
	slope = slope2;
      }
    }
  }

#ifndef OPTIMIZE
  if (Print[12]) {
    printf("Settle left:\n");
    printf("  pivot: %2d  %.4f %.4f\n", left_min_index,
	   Bar[left_min_index].end_x,
	   Bar[left_min_index].end_y);
    printf("  other: %2d  %.4f %.4f\n", bal_index,
	   Bar[bal_index].end_x,
	   Bar[bal_index].end_y);
  }
#endif

  // Rotate clockwise, re-position and re-settle().
  angle = atan(slope);
  old_x = Bar[left_min_index].end_x;
  old_y = Bar[left_min_index].end_y;  
  rotate_struct(-angle);
  translate_struct(old_x - Bar[left_min_index].end_x,
		   old_y - Bar[left_min_index].end_y);

#ifndef OPTIMIZE
  if (Print[12])
    print_struct();
#endif


  return TRUE;
}


int BarStruct :: settle_direction(int *pivot_pt)
{
  int i;
  int left_min_index=-1, right_min_index=-1;
  Real min_y;

  // Find ground level.
  min_y = Bar[0].end_y;
  for (i=1; i<Num_bars; i++) {
    if (Bar[i].end_y < min_y)
      min_y = Bar[i].end_y;
  }


  // Find left-most and right-most supports.
  for (i=0; i<Num_bars; i++) {
    if (Bar[i].end_y-min_y < DIST_TOLERANCE) {
      left_min_index = i;
      right_min_index = i;
      break;
    }
  }

  for (; i<Num_bars; i++) {
    if (Bar[i].end_y-min_y < DIST_TOLERANCE) {
      if (Bar[i].end_x < Bar[left_min_index].end_x)
	left_min_index = i;
      else if (Bar[i].end_x > Bar[right_min_index].end_x)
	right_min_index = i;
    }
  }

#ifndef OPTIMIZE
  if (Print[12]) {
    printf("%.4f Left/right: %.4f:%.4f ; %.4f:%.4f\n",
	   min_y,
	   Bar[left_min_index].end_x,
	   Bar[left_min_index].end_y,
	   Bar[right_min_index].end_x,
	   Bar[right_min_index].end_y);
  }
#endif

  if (left_min_index == right_min_index) {
    if (absv(Bar[left_min_index].end_x - COM_X) < DIST_TOLERANCE) {
#ifndef OPTIMIZE
      if (Print[12]) {
	printf("Balanced: %.6f  <  %.6f  <  %.6f  (%.6f)\n",
	       Bar[left_min_index].end_x, COM_X, Bar[right_min_index].end_x,
	       absv(Bar[left_min_index].end_x - COM_X));
      }
#endif
      // Balanced on a point...
      *pivot_pt = left_min_index;
      return 0;
    }
  }

  
  if (COM_X < Bar[left_min_index].end_x) {
#ifndef OPTIMIZE
    if (Print[12])
      printf("%.6f < %.6f,  %.6f  (settle left)\n", COM_X, Bar[left_min_index].end_x,
	     Bar[right_min_index].end_x);
#endif
    *pivot_pt = left_min_index;
    return -1;

  } else if (COM_X > Bar[right_min_index].end_x) {
#ifndef OPTIMIZE
    if (Print[12])
      printf("%.6f, %.6f  <  %.6f (settle right)\n", Bar[left_min_index].end_x,
	     Bar[right_min_index].end_x, COM_X);
#endif
    *pivot_pt = right_min_index;
    return 1;
  }

#ifndef OPTIMIZE
  if (Print[12])
    printf("%.6f  <  %.6f  <  %.6f (balanced)\n", Bar[left_min_index].end_x,
	   COM_X, Bar[right_min_index].end_x);
#endif

  // Balanced.
  *pivot_pt = left_min_index;
  return 0;
}




// Distributes mass of each vertex to the nearest ground point.
void BarStruct :: distribute_mass(void)
{
  int i;
  int ctr, index, index2, stack_ptr, dist;
  int node_stack[Num_bars];
  int num_visited[Num_bars];
  int from_list[Num_bars][3];
  Real support_weight[Num_bars];

  if (num_grnd_pts == 0) {
    return;
  }

  // Test to see if re-distribution is necessary.
  if (num_grnd_pts == num_grnd_pts_prev) {
    int recalc = FALSE;

    for (i=0; i<num_grnd_pts; i++) {
      if (ground_pt[i] != ground_pt_prev[i]) {
	recalc = TRUE;
	//	printf("Ground point %d different: %d - %d.\n", i,
	//	       ground_pt[i], ground_pt_prev[i]);
	break;
      }
    }
    if (!recalc) {
      // Don't need to re-distribute mass.
      return;
    }
  }
  /*
 else {
    printf("Num ground points different: %d (%d)\n",
	   num_grnd_pts, num_grnd_pts_prev);
  }
  */

  // Put ground points on node list.
  stack_ptr = 0;
  for (i=0; i<Num_bars; i++) {
    if ((Bar[i].ground_pt) && (Bar[i].ground_pt_prev)) {
      Bar[i].updated = 1;
      node_stack[stack_ptr] = i;
      stack_ptr++;

    } else {
      Bar[i].updated = 0;
    }
  }

  /*
  printf("Ground points: %d (%d).\n", stack_ptr, num_grnd_pts);
  for (i=0; i<stack_ptr; i++) {
    printf("%d: %.5f:%.7f:%.5f (%d)\n", ground_pt[i],
	   Bar[node_stack[i]].end_loc[0],
	   Bar[node_stack[i]].end_loc[1],
	   Bar[node_stack[i]].end_loc[2],
	   Bar[node_stack[i]].ground_pt);
  }
  */

  // Initialize variables.
  for (i=0; i<Num_bars; i++) {
    num_visited[i] = 0;
    support_weight[i] = Bar[i].weight;
  }


  // Do a breadth first search to find distance from ground points.
  ctr = 0;
  while (ctr < Num_bars) {
    index = node_stack[ctr];
    ctr++;
    dist = Bar[index].updated + 1;

    // Test parent.
    if (index != 0) {
      index2 = Bar[index].parent;
      if (Bar[index2].updated == 0) {
	Bar[index2].updated = dist;

	from_list[index2][0] = index;
	num_visited[index2] = 1;

	node_stack[stack_ptr] = index2;
	stack_ptr++;

      } else if (Bar[index2].updated == dist) {
	if (num_visited[index2] < 3) {
	  from_list[index2][num_visited[index2]] = index;
	  num_visited[index2]++;
	}
      }
    }

    // Test children.
    for (i=0; i<Bar[index].num_children; i++) {
      index2 = Bar[index].child[i];
      if (Bar[index2].updated == 0) {
	Bar[index2].updated = dist;

	from_list[index2][0] = index;
	num_visited[index2] = 1;

	node_stack[stack_ptr] = index2;
	stack_ptr++;

      } else if (Bar[index2].updated == dist) {
	if (num_visited[index2] < 3) {
	  from_list[index2][num_visited[index2]] = index;
	  num_visited[index2]++;
	}
      }
    }
  }

  // Go in reverse to distribute the weight.
  ctr = Num_bars-1;
  while (ctr >= num_grnd_pts) {
    index = node_stack[ctr];
    ctr--;

    // Distribute weight among its "children".
    if (num_visited[index] == 1) {
      support_weight[from_list[index][0]] += support_weight[index];

    } else if (num_visited[index] == 2) {
      support_weight[from_list[index][0]] += support_weight[index]*0.5;
      support_weight[from_list[index][1]] += support_weight[index]*0.5;

    } else if (num_visited[index] == 3) {
      support_weight[from_list[index][0]] += support_weight[index]*0.333333;
      support_weight[from_list[index][1]] += support_weight[index]*0.333333;
      support_weight[from_list[index][2]] += support_weight[index]*0.333333;
    }
  }

  for (i=0; i<num_grnd_pts; i++) {
    Bar[ground_pt[i]].weight_supported = support_weight[ground_pt[i]];
  //    Bar[node_stack[i]].weight_supported = support_weight[node_stack[i]] / total_weight;
  }

  // Print tree.
  /*
  for (i=0; i<Num_bars; i++) {
    printf("%d: w:%.4f:%.4f; d:%d,  #:%d: ", i, Bar[i].weight,
	   support_weight[i], Bar[i].updated, num_visited[i]);
    for (int j=0; j<num_visited[i]; j++) {
      printf("%d ", from_list[i][j]);
    }
    printf("\n");
  }
  */
  /*
  for (i=0; i<num_grnd_pts; i++) {
    printf("%d: %.4f (%.4f).\n", node_stack[i],
    	   Bar[node_stack[i]].weight_supported, total_weight);
  }
  printf("\n");
  */
}



// Translates and rotates the structure until it
// is in a stable state, resting on the surface.
int BarStruct :: settle_struct(void)
{
  int i;
  int max_tries = Num_bars;
  int settle_attempts = 0;
  int pivot_pt, direction, index;
  int right_min_index, left_min_index;
  Real min_y, move_x, val, val_sum;

  // Notes:
  // Should keep track of all bars resting on ground
  // so as to deterine their influence on object movement.


  left_min_index = 0;
  right_min_index = 0;
  min_y = Bar[0].end_y;
  if (max_tries < 10)
    max_tries = 10;

  while (1) {
    direction = settle_direction(&pivot_pt);
    if (direction == 0)
      break;

    if (settle_attempts++ > max_tries) {
      printf("Settle struct failed.  To many attempts: %d.\n", settle_attempts);
      return FALSE;
    }
    
    if (direction < 0) {
      // Rotate counter clockwise;
      settle_left(pivot_pt);

    } else {
      // Rotate clockwise;
      settle_right(pivot_pt);
    }
  }


  // Settled!
  // Find minimum y.
  min_y = Bar[0].end_y;
  for (i=1; i<Num_bars; i++) {
    if (Bar[i].end_y < min_y)
      min_y = Bar[i].end_y;
  }

  // Determine the ground points.
  for (i=0; i<Num_bars; i++) {
    if (Bar[i].end_y - min_y < Ground_tolerance)
      Bar[i].ground_pt = TRUE;
    else
      Bar[i].ground_pt = FALSE;
  }

  // Determine sequential ground points.
  num_grnd_pts = 0;
  for (i=0; i<Num_bars; i++) {
    if (Bar[i].ground_pt && Bar[i].ground_pt_prev) {
      ground_pt[num_grnd_pts] = i;
      num_grnd_pts++;
    }
  }

  // Find average displacement of sequential ground points.`
  move_x = 0.0;
  if (num_grnd_pts > 0) {
    // Distribute mass on ground points.
    distribute_mass();

    for (i=0; i<num_grnd_pts; i++) {
      index = ground_pt[i];
      move_x += (Bar[index].prev_x - Bar[index].end_x)
	* Bar[index].weight_supported;
#ifndef OPTIMIZE
      if (Print[20] || Print[31]) {
	if (move_x != 0.0) {
	  printf("%d: %.2f   %.3f:%.3f -> %.3f:%.3f (%d->%d) -> %.4f \n",
		 i, move_x, Bar[index].prev_x,
		 Bar[index].prev_y, Bar[index].end_x, Bar[index].end_y,
		 Bar[index].ground_pt_prev, Bar[index].ground_pt,
		 move_x);
	}
      }
#endif
      //      drag_distance += absv(move_x);
    }

    //        move_x /= Real(num_grnd_pts);

    // Set ref_pt as one with largest weight on it.
    int max_index = ground_pt[0];
    Real max_weight = Bar[max_index].weight_supported;
    for (i=0; i<num_grnd_pts; i++) {
      if (Bar[ground_pt[i]].weight_supported > max_weight) {
	max_index = ground_pt[i];
	max_weight = Bar[max_index].weight_supported;
      }
    }
    ref_pt = max_index;
  }


  // Swap ground points.
  int *tmp = ground_pt;
  ground_pt = ground_pt_prev;
  ground_pt_prev = tmp;
  num_grnd_pts_prev = num_grnd_pts;

  translate_struct(move_x, -min_y);

  val_sum = 0.0;
  for (i=0; i<num_grnd_pts; i++) {
      index = ground_pt[i];
      val = (Bar[index].prev_x - Bar[index].end_x) * Bar[index].weight_supported;
      drag_distance += val;
      drag_distance_abs2 += absv(val);
      val_sum += val;
  }
  drag_distance_abs1 += absv(val_sum);

  for (i=0; i<Num_bars; i++) {
    Bar[i].ground_pt_prev = Bar[i].ground_pt;
  }


#ifndef OPTIMIZE
  if (Print[20] || Print[21]) {
    if (move_x != 0.0) {
      printf("move x: %.4f (%d) -> %.4f\n", move_x, num_grnd_pts, Move[0]);
    }
  }
#endif


#ifndef OPTIMIZE
  if (Print[12]) {
    printf("Settled: %.1f:%.1f (%d) (%.3f:%.3f)  CoM: %.1f:%.1f.\n",
	   Bar[ref_pt].end_x, Bar[ref_pt].end_y, ref_pt,
	   move_x, min_y,
	   COM_X, COM_Y);
  }
  if (Print[18])
    printf("CoM: %.6f, %.6f (mv: %.2f) (Time: %.4f)\n", COM_X, COM_Y, Move[0], Time);
  if (Print[20])
    print_struct();
#endif

  return TRUE;
}


int BarStruct :: settle_structure(void)
{
  return settle_struct();
}


Real BarStruct :: animate_struct(Real time_length)
{
  Real prev_time = Time;
  Real time, start_time;

  start_time = Time;
  time = Time + time_length;
  for (; Time <= time; Time+=TIME_STEP_2D) {

    /*
    if ((Time > 0.1649) && (Time < 0.1657)) {
      Print[12] = TRUE;
      Print[20] = TRUE;
    } else if ((Time > 0.8817) && (Time < 0.8841)) {
      Print[12] = TRUE;
      Print[20] = TRUE;
    } else if ((Time > 1.8033) && (Time < 1.8041)) {
      Print[12] = TRUE;
      Print[20] = TRUE;
    } else if ((Time > 1.8567) && (Time < 1.8573)) {
      Print[12] = TRUE;
      Print[20] = TRUE;
    } else if ((Time > 1.8817) && (Time < 1.8837)) {
      Print[12] = TRUE;
      Print[20] = TRUE;
    } else {
      Print[12] = FALSE;
      Print[20] = FALSE;
    }
    */

    calculate_positions(Time);
    if (is_intersection()) {
      // Bars collide.
#ifndef OPTIMIZE
      if (Print[9]) {
	printf("Bars collided.\n");
      }
#endif
      return Time-start_time;
    }
    if (!settle_structure()) {
      printf("Animate_struct() - structure failed to settle.\n");
      return -1;
    }

    if (COM_Y < com_min[1])
      com_min[1] = COM_Y;
    else if (COM_Y > com_max[1])
      com_max[1] = COM_Y;

    if (COM_X < com_min[0])
      com_min[0] = COM_X;
    else if (COM_X > com_max[0])
      com_max[0] = COM_X;

    if (win1 != 0) {
      if (time - prev_time > DRAW_STEP) {
	draw_struct();
	prev_time = time;
      }
    }
  }

  return time_length;
}



int BarStruct :: add_bar(Real len, Real angle, Real spd, Real offset,
			 int type, int p)
{
  int i;
  int child;

#ifndef OPTIMIZE
  if (Print[12]) {
    printf("Add bar: len:%.1f, spd:%.2f, ang:%.3f, type:%d, p:%d\n",
	   len, spd, angle, type, p);
  }
#endif

  if ((absv(absv(angle)-PI) < ANGLE_TOLERANCE) && (p>0)) {
    // Go back up parent.
    //    printf("  Moving up to parent (%.5f).\n",
    //	   absv((absv(angle)-PI)));
    return Bar[p].parent;
  }


  if (Num_bars >= MAX_BARS2) {
    if (Print[9])
      printf("Maximum number of bars exceeded.\n");
    return FALSE;
  }

  if (len == 0.0)
    len = 1.0;

  if (Bar[p].num_children >= MAX_CHILDREN) {
    if (Print[9]) {
      printf("Too many children for bar #%d.\n", p);
    }
    return FALSE;
  }

  if (len > 5.0)
    len = 5.0;
  len *= 10.0;

  if (angle > PI) {
    angle = -(2.0*PI - angle);
    if (absv(angle) < ANGLE_TOLERANCE)
      angle = 0.0;
  }

  if (Bar[p].num_children == 0) {
    // First child.
    if ((angle == 0.0) &&
	(Bar[p].type != BAR_TYPE_FIXED)) {
      Bar[p].type = BAR_TYPE_FIXED;
    }
    
  } else {
    // test for conflicts.
    for (i=0; i<Bar[p].num_children; i++) {
      child = Bar[p].child[i];
      if (absv(Bar[child].angle_a-angle) < ANGLE_TOLERANCE) {
	// Traverse up this bar and end.
	//	printf("  Traversing bar#: %d (child: %d).\n", child, i);
	return child;
      }
    }
  }

  if (spd > 5.0)
    spd = 5.0;
  else if (spd < -5.0)
    spd = 5.0;

  child = Num_bars;
  Bar[p].child[Bar[p].num_children] = child;
  Bar[p].num_children++;
  Bar[child].parent = p;
  Bar[child].move_spd = spd;
  Bar[child].move_offset = 0.0;
  Bar[child].num_children = 0;
  Bar[child].length = len;
  Bar[child].type = type;
  Bar[child].weight = 1.0;
  Bar[child].same_as = -1;
  Num_bars++;
 
  if ((Bar[p].type != BAR_TYPE_FIXED) && 
      (Bar[p].num_children == 1) &&
      (angle != 0.0)) {
    Num_joints++;
    Bar[child].moves = TRUE;
    Bar[child].joint_index = p;

    //    printf(" moves: type:%d, angle:%.3f\n", type, angle);

    if (Bar[p].type == BAR_TYPE_CLOCKWISE) {
      Bar[child].angle_a = 0.0;
      Bar[child].angle_b = angle;
    } else {
      Bar[child].angle_a = angle;
      Bar[child].angle_b = 0.0;
    }
      
  } else {
    Bar[child].moves = FALSE;
    Bar[child].joint_index = Bar[p].joint_index;
    Bar[child].angle_a = angle;
    Bar[child].abs_angle = angle;
  }

  if ((Bar[p].num_children == 1) &&
      (!Bar[child].moves))
    Bar[p].type = BAR_TYPE_FIXED;


  if (Print[22] && !Print[12]) {
    printf("%d(%d): ", child, p);
    printf(" len:%.1f, type:%d, moves:%d, spd:%.3f angle(%.2f:%.2f)\n",
	   Bar[child].length, Bar[child].type, Bar[child].moves,
	   Bar[child].move_spd,
	   Bar[child].angle_a*RAD2DEG, Bar[child].angle_b*RAD2DEG);
  }

#ifndef OPTIMIZE
  if (Print[12] || Print[25]) {
    int parent = p;
    printf("P: ");
    printf("%2d(%2d): %.2f:%.2f  len:%2d  type:%d, moves:%d, spd:%.3f, angle(%.3f:%.3f)\n",
	   parent, Bar[parent].same_as, Bar[parent].end_x, Bar[parent].end_y,
	   int(Bar[parent].length), Bar[parent].type, Bar[parent].moves,
	   Bar[parent].move_spd,
	   Bar[parent].angle_a, Bar[parent].angle_b);
    
    printf("C: ");
    printf("%2d(%2d): %.2f:%.2f  len:%2d  type:%d, moves:%d, spd:%.3f angle(%.3f:%.3f)\n",
	   child, Bar[child].same_as, Bar[child].end_x, Bar[child].end_y,
	   int(Bar[child].length),
	   Bar[child].type, Bar[child].moves,
	   Bar[child].move_spd,
	   Bar[child].angle_a, Bar[child].angle_b);
    printf("\n");

    /*
    printf("Parent: ");
    printf("%d: type:%d, moves:%d, spd:%.3f, angle(%.3f:%.3f)\n",
	   p, Bar[p].type, Bar[p].moves,
	   Bar[p].move_spd,
	   Bar[p].angle_a, Bar[p].angle_b);
    
    printf("Child: ");
    printf("%d: type:%d, moves:%d, spd:%.3f angle(%.3f:%.3f)\n",
	   child, Bar[child].type, Bar[child].moves,
	   Bar[child].move_spd,
	   Bar[child].angle_a, Bar[child].angle_b);
    printf("\n");
    */
  }
#endif

  return child;
}


int BarStruct :: add_bar(Real len, Real angle, Real spd,
			 int type, int p)
{
  return add_bar(len, angle, spd, 0.0, type, p);
}



int BarStruct :: add_bar2(Real len, Real *child_angle, Real spd, Real offset,
			 int type, int *p_index)
{
  int i;
  int child, same_as=-1, child_index;
  int parent = *p_index;
  Real end_x, end_y, angle_b=0.0;
  Real angle = *child_angle;

#ifndef OPTIMIZE
  if (Print[12] || Print[26]) {
    printf("Add bar (%d): len:%.1f, spd:%.2f(%.2f), ang:%.3f, type:%d\n",
	   parent, len, spd, offset, angle, type);
  }
#endif

  if (Num_bars >= MAX_BARS2) {
    if (Print[9])
      printf("Maximum number of bars exceeded.\n");
    return FALSE;
  }

  // Adjust length.
  if (len == 0.0)
    len = 1.0;
  else if (len > 5.0)
    len = 5.0;
  len *= 10.0;

  // Adjust speed.
  if (type == BAR_TYPE_FIXED) {
    spd = 0.0;
    offset = 0.0;
  } else {
    if (spd > 5.0)
      spd = 5.0;
    else if (spd < -5.0)
      spd = 5.0;
  }

  angle = simplify_angle_180(angle);

  if (parent > 0) {
    if ((Bar[parent].type != BAR_TYPE_FIXED) && 
	(Bar[parent].num_children == 0)) {
      if (Bar[parent].type == BAR_TYPE_CLOCKWISE) {
	angle_b = angle - PI/3.0;
      } else if (Bar[parent].type == BAR_TYPE_COUNTER) {
	angle_b = angle + PI/3.0;
      } else if (Bar[parent].type == BAR_TYPE_COUNTER2) {
	angle_b = angle + 2.0*PI/3.0;
      } else if (Bar[parent].type == BAR_TYPE_CLOCKWISE2) {
	angle_b = angle - 2.0*PI/3.0;
      } else {
	printf("Invalid joint type.\n");
      }

    } else {
      angle_b = angle;
    }

    //    printf("angle: %.4f, angle_b: %.4f\n", angle, angle_b);

    if ((absv(absv(angle)-PI) < ANGLE_TOLERANCE) ||
	((absv(angle_b)-PI) - ANGLE_TOLERANCE > 0.0)) {
      // Go back up parent.
      //      printf("  Moving up to parent (%.5f:%.5f).\n",
      //	     absv((absv(angle)-PI)),
      //	     absv((absv(angle_b)-PI)));
      *p_index = Bar[parent].parent;
      *child_angle = PI;
      return TRUE;
    }
  }

  if (Bar[parent].num_children >= MAX_CHILDREN) {
    if (Print[9]) {
      printf("Too many children for bar #%d.\n", parent);
    }
    return FALSE;
  }

  if (Bar[parent].num_children != 0) {
    // test for conflicts.
    if (Bar[parent].type != BAR_TYPE_FIXED) {
      child = Bar[parent].child[0];
      if (Bar[child].angle_b < Bar[child].angle_a) {
	if ((angle + ANGLE_TOLERANCE > Bar[child].angle_b) &&
	    (angle - ANGLE_TOLERANCE < Bar[child].angle_a)) {
	  *p_index = child;
	  *child_angle = 0.0;
	  if (Print[12] || Print[26]) {
	    printf(" same as first child -> %d.\n", child);
	  }
	  return TRUE;
	}

      } else {
	if ((angle - ANGLE_TOLERANCE < Bar[child].angle_b) &&
	    (angle + ANGLE_TOLERANCE > Bar[child].angle_a)) {
	  *p_index = child;
	  *child_angle = 0.0;
	  if (Print[12] || Print[26]) {
	    printf(" same as first child -> %d.\n", child);
	  }
	  return TRUE;
	}
      }
    }
  }
  
  for (i=0; i<Bar[parent].num_children; i++) {
    child = Bar[parent].child[i];
    if (absv(Bar[child].angle_a-angle) < ANGLE_TOLERANCE) {
      // Traverse up this bar and end.
      //	printf("  Traversing bar#: %d (child: %d).\n", child, i);
      if (Bar[child].same_as != -1) {
	*p_index = Bar[child].same_as;
	*child_angle = Bar[child].abs_angle - Bar[*p_index].abs_angle;
      } else {
	*p_index = child;
	*child_angle = 0.0;
      }
      if (Print[12] || Print[26]) {
	printf(" same as child -> %d.\n", child);
      }
      return TRUE;
    }
  }
  

  // Start setting values for child.
  child = Num_bars;
  child_index = Bar[parent].num_children;
  Bar[parent].child[child_index] = child;
  Bar[parent].num_children++;
  Bar[child].parent = parent;
  Bar[child].num_children = 0;
  Bar[child].num_from = 0;

  Bar[child].move_spd = spd;
  Bar[child].move_offset = offset;
  Bar[child].length = len;
  Bar[child].type = type;
  if (same_as == -1)
    Bar[child].weight = 1.0;
  else
    Bar[child].weight = 0.0;

  Bar[child].angle_a = angle;
  Bar[child].angle_b = angle_b;
  if ((Bar[parent].type != BAR_TYPE_FIXED) && 
      (Bar[parent].num_children == 1)) {
    Num_joints++;
    Bar[child].moves = TRUE;
    Bar[child].joint_index = parent;
      
  } else {
    Bar[child].moves = FALSE;
    Bar[child].joint_index = Bar[parent].joint_index;
  }


  // Calculate location and see if it matches anyone else.
  Real theta = offset * 2.0;
  if (theta > 2.0)
    theta = 2.0 - theta;
  angle = Bar[child].angle_a*(1.0-theta) + Bar[child].angle_b*theta;
  Bar[child].abs_angle = Bar[parent].abs_angle + angle;
  end_x = Bar[parent].end_x + len*cos(Bar[child].abs_angle);
  end_y = Bar[parent].end_y + len*sin(Bar[child].abs_angle);
  same_as = -1;
  for (i=0; i<Num_bars; i++) {
    if ((absv(Bar[i].end_x - end_x) < DIST_TOLERANCE) &&
	(absv(Bar[i].end_y - end_y) < DIST_TOLERANCE)) {
      same_as = i;
      break;
    }
  }
  Bar[child].same_as = same_as;
  Bar[child].end_x = end_x;
  Bar[child].end_y = end_y;

  if (same_as != -1) {
    // Connect same_as to parent.
    int index = Bar[same_as].num_from;
    Bar[same_as].from_list[index] = parent;
    Bar[same_as].num_from += 1;
    if (Bar[same_as].num_from >= MAX_CHILDREN) {
      if (Print[9]) {
	printf("Too many from nodes for %d.\n", same_as);
      }
      return FALSE;
    }
  }
  Num_bars++;

#ifndef OPTIMIZE
  if (Print[22] && !Print[12]) {
    printf("%d(%d): ", child, parent);
    printf(" len:%.1f, type:%d, moves:%d, spd:%.3f angle(%.2f:%.2f)\n",
	   Bar[child].length, Bar[child].type, Bar[child].moves,
	   Bar[child].move_spd,
	   Bar[child].angle_a*RAD2DEG, Bar[child].angle_b*RAD2DEG);
  }    

  if (Print[12] || Print[25]) {
    printf("P: ");
    printf("%2d(%2d): %.2f:%.2f  type:%d, moves:%d, spd:%.3f(%.2f), angle(%.3f:%.3f)\n",
	   parent, Bar[parent].same_as, Bar[parent].end_x, Bar[parent].end_y, Bar[parent].type, Bar[parent].moves,
	   Bar[parent].move_spd, Bar[parent].move_offset,
	   Bar[parent].angle_a, Bar[parent].angle_b);
    
    printf("C: ");
    printf("%2d(%2d): %.2f:%.2f  type:%d, moves:%d, spd:%.3f(%.2f) angle(%.3f:%.3f)\n",
	   child, Bar[child].same_as, Bar[child].end_x, Bar[child].end_y,
	   Bar[child].type, Bar[child].moves,
	   Bar[child].move_spd, Bar[parent].move_offset,
	   Bar[child].angle_a, Bar[child].angle_b);
    printf("\n");
  }
#endif
  
  if (same_as != -1) {
    *p_index = Bar[child].same_as;
    *child_angle = Bar[child].abs_angle - Bar[*p_index].abs_angle;
  } else {
    *p_index = child;
    *child_angle = 0.0;
  }
  
  return TRUE;
}


// Build structure using parameterized absolute coordinates.
// This builds the kinematic model for bending pieces.
int BarStruct :: build_structure60(int length, LString *command, Real *args)
{
  int i, j;
  int bar_index, num_bars_prev;
  int ret_val;
  Real angle, spd, offset;

  angle = 0.0;
  bar_index = 0;
  offset = 0.0;
  waste_length = 0;
  initialize();


  for (i=0; i<length; i++) {
#ifndef OPTIMIZE
    if (Print[13]) {
      printf("%d: %c (%d:%.2f)   ang:%.8f\n", i, command[i].symbol,
	     int(command[i].arg), command[i].arg, angle);
    } else if (Print[37]) {
      if ((command[i].symbol == TERM_PUSH) ||
	  (command[i].symbol == TERM_POP)) {
	printf("%c ", command[i].symbol);
      } else {	
	printf("%c(%d) ", command[i].symbol, int(command[i].arg));
      }
    }

#endif

    if (command[i].symbol == 0) {
      if (Print[0] || Print[9]) 
	printf("Build:: unexpected early termination of command sequence @i=%d(!%d)\n", i, length);
      return FALSE;

    } else if (command[i].symbol == TERM_PENUP) {
      Pen_down = FALSE;

    } else if (command[i].symbol == TERM_PENDOWN) {
      Pen_down = TRUE;

    } else if (command[i].symbol == TERM_PUSH) {
      if (Bar_stack_ptr == BAR_STACK_SIZE) {
	printf("Bar stack exceeded.\n");
	return FALSE;
      }
Bar_stack[Bar_stack_ptr].angle = angle;
      Bar_stack[Bar_stack_ptr].bar_index = bar_index;
      Bar_stack[Bar_stack_ptr].offset = offset;
      Bar_stack_ptr++;

    } else if (command[i].symbol == TERM_POP) {
      if (Bar_stack_ptr == 0) {
	printf("Bar stack over-popped.\n");
	return FALSE;
      }
      Bar_stack_ptr--;
      angle = Bar_stack[Bar_stack_ptr].angle;
      bar_index = Bar_stack[Bar_stack_ptr].bar_index;
      offset = Bar_stack[Bar_stack_ptr].offset;

    } else if (command[i].symbol == TERM_BACKWARD) {
      angle = 0.0;
      if (bar_index != 0)
	bar_index = Bar[bar_index].parent;

    } else if (command[i].symbol == TERM_FORWARD) {
      num_bars_prev = Num_bars;
      ret_val = add_bar2(1.0, &angle, 0.0, 0.0, BAR_TYPE_FIXED,
			&bar_index);
      if (ret_val == FALSE)
	return FALSE;
      if (num_bars_prev == Num_bars)
	waste_length++;
      //      printf("Forward.  new index: %d.\n", bar_index);

    } else if (command[i].symbol == TERM_JOINT_CLOCKWISE) {
      spd = floor(command[i].arg);
      if (spd == 0.0)
	spd = 1.0;
      else if (spd < 0.0)
	spd = -spd;
      num_bars_prev = Num_bars;
      ret_val = add_bar2(1.0, &angle, spd, offset, BAR_TYPE_CLOCKWISE, &bar_index);
      if (ret_val == FALSE)
	return FALSE;
      if (num_bars_prev == Num_bars)
	waste_length++;
      //      printf("Straight.  new index: %d.\n", bar_index);

    } else if (command[i].symbol == TERM_JOINT_COUNTER) {
      spd = floor(command[i].arg);
      if (spd == 0.0)
	spd = 1.0;
      else if (spd < 0.0)
	spd = -spd;

      num_bars_prev = Num_bars;
      if (!add_bar2(1.0, &angle, spd, offset, BAR_TYPE_COUNTER, &bar_index))
	return FALSE;
      if (num_bars_prev == Num_bars)
	waste_length++;
      //      printf("Bent.  new index: %d.\n", bar_index);

    } else if (command[i].symbol == TERM_JOINT_COUNTER2) {
      spd = floor(command[i].arg);
      if (spd == 0.0)
	spd = 1.0;
      else if (spd < 0.0)
	spd = -spd;
      num_bars_prev = Num_bars;
      ret_val = add_bar2(1.0, &angle, spd, offset, BAR_TYPE_CLOCKWISE, &bar_index);
      if (ret_val == FALSE)
	return FALSE;
      if (num_bars_prev == Num_bars)
	waste_length++;

    } else if (command[i].symbol == TERM_JOINT_CLOCKWISE2) {
      spd = floor(command[i].arg);
      if (spd == 0.0)
	spd = 1.0;
      else if (spd < 0.0)
	spd = -spd;
      num_bars_prev = Num_bars;
      ret_val = add_bar2(1.0, &angle, spd, offset, BAR_TYPE_CLOCKWISE, &bar_index);
      if (ret_val == FALSE)
	return FALSE;
      if (num_bars_prev == Num_bars)
	waste_length++;

    } else if (command[i].symbol == TERM_TURN_CLOCKWISE) {
      angle -= floor(command[i].arg) * PI/3.0;
      angle = simplify_angle(angle);
#ifndef OPTIMIZE
      if (Print[13])
	printf("arg:%.4f   angle: %.4lf\n", command[i].arg, angle);
#endif

    } else if (command[i].symbol == TERM_TURN_CNTR_CLOCKWISE) {
      angle += floor(command[i].arg) * PI/3.0;
      angle = simplify_angle(angle);
#ifndef OPTIMIZE
      if (Print[13])
	printf("arg:%.4f   angle: %.4lf\n", command[i].arg, angle);
#endif

    } else if (command[i].symbol == TERM_OFFSET_INCREASE) {
      offset += OFFSET_STEP * floor(command[i].arg);
      if (offset > 1.0) {
	offset -= floor(offset);
      }

    } else if (command[i].symbol == TERM_OFFSET_DECREASE) {
      offset -= OFFSET_STEP * floor(command[i].arg);
      if (offset < 0.0) {
	offset -= floor(offset);
      }

    } else {
#ifndef OPTIMIZE
	if (Print[0])
	  printf("Build:: unknown command: (%c:%d)\n", 
		 command[i].symbol, int(command[i].symbol));
#endif
    }     
  }

  for (i=0; i<Num_bars; i++) {
    if (Bar[i].num_children == 0)
      Bar[i].type = BAR_TYPE_FIXED;

    else if (Bar[i].type != BAR_TYPE_FIXED) {
      if (Bar[i].angle_a == Bar[i].angle_b)
	Bar[i].type = BAR_TYPE_FIXED;
    }
  }

  Num_joints = 0;
  for (i=0; i<Num_bars; i++) {
    if (Bar[i].type != BAR_TYPE_FIXED)
      Num_joints++;
  }


  if (Num_bars <= 1) {
    if (Print[9])
      printf("Not enough bars: %d.\n", Num_bars);
    return FALSE;
  }

  if (Num_joints > Max_joints) {
    if (Print[9])
      printf("Too many joints: %d.\n", Num_joints);
return FALSE;
  }

  total_weight = Real(Num_bars);
  for (i=0; i<Num_bars; i++)
    Bar[i].weight /= total_weight;

  calculate_positions(0.0);
  settle_structure();
  translate_struct(-COM_X, 0.0);
  Move[0] = 0.0;
  Move[1] = 0.0;
  com_y_avg = COM_Y / 2.0;

  for (i=0; i<Num_bars; i++) {
    Bar[i].prev_x = Bar[i].end_x;
    Bar[i].prev_y = Bar[i].prev_y;
  }

  if (Option[11] || Option[12]) {
    for (i=0; i<Num_bars; i++) {
      for (j=0; j<HISTORY_LENGTH; j++) {
	Bar[i].x_history[j] = Bar[i].end_x;
	Bar[i].y_history[j] = Bar[i].end_y;
      }
    }
  }

  initialize_intersection();

#ifndef OPTIMIZE
  if (Print[27])
    print_struct();
#endif

  return TRUE;
}


// Build structure using parameterized absolute coordinates.
// This builds the kinematic model for bending pieces.
int BarStruct :: build_structure_param(int length, 
				       LString *command,
				       Real *args)
{
  int i;
  int bar_index;
  int ret_val;
  Real angle, len, spd;

  angle = 0.0;
  bar_index = 0;
  initialize();

  for (i=0; i<length; i++) {
    if (Print[13]) {
      printf("%d: %c(%d)\n", i, command[i].symbol,
	     int(command[i].arg));
    }

    if (command[i].symbol == 0) {
      if (Print[0] || Print[9]) 
	printf("Build:: unexpected early termination of command sequence @:%d(!%d)\n", i, length);
      return FALSE;

    } else if (command[i].symbol == TERM_PENUP) {
      Pen_down = FALSE;

    } else if (command[i].symbol == TERM_PENDOWN) {
      Pen_down = TRUE;

    } else if (command[i].symbol == TERM_PUSH) {
      if (Bar_stack_ptr == BAR_STACK_SIZE) {
	printf("Bar stack exceeded.\n");
	return FALSE;
      }
      Bar_stack[Bar_stack_ptr].angle = angle;
      Bar_stack[Bar_stack_ptr].bar_index = bar_index;
      Bar_stack_ptr++;

    } else if (command[i].symbol == TERM_POP) {
      if (Bar_stack_ptr == 0) {
	printf("Bar stack over-popped.\n");
	return FALSE;
      }
      Bar_stack_ptr--;
      angle = Bar_stack[Bar_stack_ptr].angle;
      bar_index = Bar_stack[Bar_stack_ptr].bar_index;

    } else if (command[i].symbol == TERM_FORWARD) {
      len = command[i].arg;
      if (len == 0.0)
	continue;

      ret_val = add_bar(len, angle, 0.0, BAR_TYPE_FIXED,
			bar_index);
      if (ret_val == FALSE)
	return FALSE;
      if (ret_val == Bar[bar_index].parent) {
	angle = PI;
      } else {
	angle = 0.0;
      }
      bar_index = ret_val;

    } else if (command[i].symbol == TERM_JOINT_CLOCKWISE) {
      spd = command[i].arg;
      if (spd == 0.0)
	spd = 1.0;
      else if (spd < 0.0)
	spd = -spd;
      ret_val = add_bar(1.0, angle, spd, BAR_TYPE_CLOCKWISE,
			bar_index);
      if (ret_val == FALSE)
	return FALSE;
      if (ret_val == Bar[bar_index].parent) {
	angle = PI;
      } else {
	angle = 0.0;
      }
      bar_index = ret_val;

    } else if (command[i].symbol == TERM_JOINT_COUNTER) {
      spd = command[i].arg;
      if (spd == 0.0)
	spd = 1.0;
      else if (spd < 0.0)
	spd = -spd;

      ret_val = add_bar(1.0, angle, spd, BAR_TYPE_COUNTER,
			bar_index);
      if (ret_val == FALSE)
	return FALSE;
      if (ret_val == Bar[bar_index].parent) {
	angle = PI;
      } else {
	angle = 0.0;
      }
      bar_index = ret_val;

    } else if (command[i].symbol == TERM_TURN_CLOCKWISE) {
      angle -= command[i].arg * PI/4.0;
      angle = simplify_angle(angle);
      //      printf("angle: %.3f\n", angle);

    } else if (command[i].symbol == TERM_TURN_CNTR_CLOCKWISE) {
      angle += command[i].arg * PI/4.0;
      angle = simplify_angle(angle);
      //      printf("angle: %.3f\n", angle);

    } else {
#ifndef OPTIMIZE
	if (Print[0])
	  printf("Build:: unknown command: (%c:%d)\n", 
		 command[i].symbol, int(command[i].symbol));
#endif
    }     
  }

  if (Num_bars <= 1) {
    if (Print[9])
      printf("Not enough bars: %d.\n", Num_bars);
    return FALSE;
  }

  if (Num_joints > Max_joints) {
    if (Print[9])
      printf("Too many joints: %d.\n", Num_joints);
    return FALSE;
  }

  total_weight = Real(Num_bars);
  for (i=0; i<Num_bars; i++)
    Bar[i].weight /= total_weight;

  calculate_positions(0.0);
  settle_structure();
  translate_struct(-COM_X, 0.0);
  Move[0] = 0.0;
  Move[1] = 0.0;

  for (i=0; i<Num_bars; i++) {
    Bar[i].prev_x = Bar[i].end_x;
    Bar[i].prev_y = Bar[i].prev_y;
  }

  initialize_intersection();

  if (Print[9]) {
    printf("Built!\n");
  }

  return TRUE;
}


void BarStruct :: initialize_evaluation(void)
{
  int i;

  for (i=0; i<Num_bars; i++) {
    if (Bar[i].same_as == -1) {
      Bar[i].weight = 0.75 + 0.5*random_dbl();
    } else
      Bar[i].weight = 0.0;

    if ((Bar[i].num_children > 0) &&
	(Bar[i].type != BAR_TYPE_FIXED)) {
      Bar[i].weight += 0.5;
    }
  }

  total_weight = 0.0;
  drag_distance = 0.0;
  drag_distance_abs1 = 0.0;
  drag_distance_abs2 = 0.0;
  for (i=0; i<Num_bars; i++) {
    // Give random weights;
    total_weight += Bar[i].weight;
  }    
  for (i=0; i<Num_bars; i++)
    Bar[i].weight /= total_weight;
}



Real BarStruct :: run_distance(ModelStruct *model)
{
  Real distance, val;

  initialize_evaluation();
  Time = 0.0;
  com_max[0] = COM_X;
  com_max[1] = COM_Y;
  com_min[0] = COM_X;
  com_min[1] = COM_Y;

  val = animate_struct(1.0);
  if (val == -1.0) {
    model->error = TRUE;
    printf("Animate structure error\n");
    return 0.0;

  } else if (val < 1.0) {
#ifndef OPTIMIZE
    if (Print[9])
      printf("Animate structure ended prematurely\n");
#endif
    return 0.0;
  }

  if ((com_max[0] != com_min[0]) ||
      (com_max[1] != com_min[1])) {
    com_max[0] = COM_X;
    com_max[1] = COM_Y;
    com_min[0] = COM_X;
    com_min[1] = COM_Y;

    val = animate_struct(10.0);
    if (val == -1.0) {
      model->error = TRUE;
      printf("Animate structure error\n");
      return 0.0;

    } else if (val < 10.0) {
      //      printf("Animate structure ended prematurely\n");
      return 0.0;
    }

    distance = com_max[0]-com_min[0];

  } else {
    distance = 0.0;
  }

  return distance;
}



// Test for cycles and put noise in non-cycle connections.
int BarStruct :: fitness_distance5(ModelStruct *model)
{
  int i, fail = FALSE;
  Real distance=0.0, max_arm_len = 0.0;


  if (Num_bars < 3) {
    if (Print[9])
      printf("Fitness distance :: Not enough bars: %d.\n", Num_bars);
    fail = TRUE;

  } else if (Num_joints == 0) {
    if (Print[9])
      printf("Fitness distance :: Not enough joints: %d.\n", Num_joints);
    fail = TRUE;
  } else {
    fail = TRUE;
    for (i=1; i<Num_bars; i++) {
      if (Bar[i].num_children > 1) {
	fail = FALSE;
	break;
      }
    }
    if (fail && Print[9]) {
      printf("Fitness distance :: single line - fail.\n");
    }

    label_cycles();
    for (i=0; i<Num_bars; i++) {
      Bar[i].true_angle_a = Bar[i].angle_a;
      Bar[i].true_angle_b = Bar[i].angle_b;
    }
  }

  if (!fail) {
    Real dd1=0.0, dd2=0.0, dd3=0.0;
    Real min_distance = 9999999.0;

    for (i=0; i<Number_trials; i++) {
      add_connection_noise();
      distance = run_distance(model);
      if (distance == 0.0) {
	fail = TRUE;
	min_distance = 0.0;
	break;

      } else if (distance < min_distance) {
	min_distance = distance;
	dd1 = absv(drag_distance);
	dd2 = drag_distance_abs1;
	dd3 = drag_distance_abs2;
      }
    }
    distance = min_distance;
    drag_distance = dd1;
    drag_distance_abs1 = dd2;
    drag_distance_abs2 = dd3;
  }


  if (fail || (distance-drag_distance_abs1 <= 0)) {
    model->fitness[0] = 0.0;
  } else {
    Real fitness, len;

    fitness = distance - drag_distance_abs1;
    drag_distance_abs2 = sqrt(drag_distance_abs1);
    max_arm_len = Real(max_arm_length());

    if (fitness/2.0 < drag_distance_abs1)
      fitness /= 2.0;
    else
      fitness -= drag_distance_abs1;

    //    len = pow(model->lstring_length, 0.75);
    len = waste_length;
    if (len > fitness / 2.0)
      fitness /= 2.0;
    else
      fitness -= Real(len);
    //    fitness = floor(fitness/10.0)*10.0 + 1.0;

    if (max_arm_len > 1.0)
      fitness /= max_arm_len;

    if (fitness >= 5.0)
      fitness = floor(fitness/2.0)*2.0;
    else if (distance >= 1.0)
      fitness = 1.0;
    else
      fitness = 0.0;

    model->fitness[0] = fitness;
    //    model->fitness[0] = floor(distance/10.0)*10.0 + 1;
    //    model->fitness[0] = 1.0 + distance;
    //    model->fitness[0] = 1.0 + absv(Move[0]);
  }


  if (Print[6]) {
    printf("COM %.2f:%.2f ; %.2f:%.2f\n", com_min[0], com_max[0],
	   com_min[1], com_max[1]);
    printf("dist:%.1f(%.2f:%.2f:%.2f)  #bars:%d  #joints:%d  max-arm:%.0f  len:%d  waste:%d\n",
	   distance, drag_distance, drag_distance_abs1, drag_distance_abs2,
	   Num_bars, Num_joints, max_arm_len, model->lstring_length, waste_length);
    printf("Fitness #4: %.1f\n", model->fitness[0]);
  }

  if (model->fitness[0] > 100000.0) {
    printf("Fitness distance :: Error - fitness too high: %.3f\n", model->fitness[0]);
    model->error = TRUE;
  } else
    model->error = FALSE;
  translate_struct(-COM_X, 0.0);
  ref_pt = 0;
  ref_rotation = 0.0;

  return TRUE;
}


// Test for cycles and put noise in non-cycle connections.
int BarStruct :: fitness_distance4(ModelStruct *model)
{
  int i, fail = FALSE;
  Real distance=0.0, max_arm_len = 0.0;


  if (Num_bars < 3) {
    if (Print[9])
      printf("Fitness distance :: Not enough bars: %d.\n", Num_bars);
    fail = TRUE;

  } else if (Num_joints == 0) {
    if (Print[9])
      printf("Fitness distance :: Not enough joints: %d.\n", Num_joints);
    fail = TRUE;
  } else {
    fail = TRUE;
    for (i=1; i<Num_bars; i++) {
      if (Bar[i].num_children > 1) {
	fail = FALSE;
	break;
      }
    }
    if (fail && Print[9]) {
      printf("Fitness distance :: single line - fail.\n");
    }

    label_cycles();
    for (i=0; i<Num_bars; i++) {
      Bar[i].true_angle_a = Bar[i].angle_a;
      Bar[i].true_angle_b = Bar[i].angle_b;
    }
  }

  if (!fail) {
    Real dd1=0.0, dd2=0.0, dd3=0.0;
    Real min_distance = 9999999.0;

    for (i=0; i<Number_trials; i++) {
      add_connection_noise();
      distance = run_distance(model);
      if (distance == 0.0) {
	fail = TRUE;
	min_distance = 0.0;
	break;

      } else if (distance < min_distance) {
	min_distance = distance;
	dd1 = absv(drag_distance);
	dd2 = drag_distance_abs1;
	dd3 = drag_distance_abs2;
      }
    }
    distance = min_distance;
    drag_distance = dd1;
    drag_distance_abs1 = dd2;
    drag_distance_abs2 = dd3;
  }


  if (fail || (distance-drag_distance_abs1 <= 0)) {
    model->fitness[0] = 0.0;
  } else {
    Real fitness, len;

    fitness = distance - drag_distance_abs1;
    drag_distance_abs2 = sqrt(drag_distance_abs2);
    max_arm_len = Real(max_arm_length());

    if (fitness/2.0 < drag_distance_abs2)
      fitness /= 2.0;
    else
      fitness -= drag_distance_abs2;

    //    len = pow(model->lstring_length, 0.75);
    len = waste_length;
    if (len > fitness / 2.0)
      fitness /= 2.0;
    else
      fitness -= Real(len);
    //    fitness = floor(fitness/10.0)*10.0 + 1.0;

    if (max_arm_len > 1.0)
      fitness /= max_arm_len;

    if (fitness >= 5.0)
      fitness = floor(fitness/2.0)*2.0;
    else if (distance >= 1.0)
      fitness = 1.0;
    else
      fitness = 0.0;

    model->fitness[0] = fitness;
    //    model->fitness[0] = floor(distance/10.0)*10.0 + 1;
    //    model->fitness[0] = 1.0 + distance;
    //    model->fitness[0] = 1.0 + absv(Move[0]);
  }


  if (Print[6]) {
    printf("COM %.2f:%.2f ; %.2f:%.2f\n", com_min[0], com_max[0],
	   com_min[1], com_max[1]);
    printf("dist:%.1f(%.2f:%.2f:%.2f)  #bars:%d  #joints:%d  max-arm:%.0f  len:%d  waste:%d\n",
	   distance, drag_distance, drag_distance_abs1, drag_distance_abs2,
	   Num_bars, Num_joints, max_arm_len, model->lstring_length, waste_length);
    printf("Fitness #4: %.1f\n", model->fitness[0]);
  }

  if (model->fitness[0] > 100000.0) {
    printf("Fitness distance :: Error - fitness too high: %.3f\n", model->fitness[0]);
    model->error = TRUE;
  } else
    model->error = FALSE;
  translate_struct(-COM_X, 0.0);
  ref_pt = 0;
  ref_rotation = 0.0;

  return TRUE;
}


// Test for cycles and put noise in non-cycle connections.
int BarStruct :: fitness_distance3(ModelStruct *model)
{
  int i, fail = FALSE;
  Real distance=0.0;


  if (Num_bars < 3) {
    if (Print[9])
      printf("Fitness distance :: Not enough bars: %d.\n", Num_bars);
    fail = TRUE;

  } else if (Num_joints == 0) {
    if (Print[9])
      printf("Fitness distance :: Not enough joints: %d.\n", Num_joints);
    fail = TRUE;
  } else {
    fail = TRUE;
    for (i=1; i<Num_bars; i++) {
      if (Bar[i].num_children > 1) {
	fail = FALSE;
	break;
      }
    }
    if (fail && Print[9]) {
      printf("Fitness distance :: single line - fail.\n");
    }

    label_cycles();
    for (i=0; i<Num_bars; i++) {
      Bar[i].true_angle_a = Bar[i].angle_a;
      Bar[i].true_angle_b = Bar[i].angle_b;
    }
  }

  if (!fail) {
    Real dd1=0.0, dd2=0.0, dd3=0.0;
    Real min_distance = 9999999.0;

    for (i=0; i<Number_trials; i++) {
      add_connection_noise();
      distance = run_distance(model);
      if (distance == 0.0) {
	fail = TRUE;
	min_distance = 0.0;
	break;

      } else if (distance < min_distance) {
	min_distance = distance;
	dd1 = absv(drag_distance);
	dd2 = drag_distance_abs1;
	dd3 = drag_distance_abs2;
      }
    }
    distance = min_distance;
    drag_distance = dd1;
    drag_distance_abs1 = dd2;
    drag_distance_abs2 = dd3;
  }

  if (fail || (distance-drag_distance_abs1 <= 0)) {
    model->fitness[0] = 0.0;
  } else {
    Real fitness, len;
    fitness = distance - drag_distance_abs1;
    drag_distance_abs2 = sqrt(drag_distance_abs2);
    if (fitness/2.0 < drag_distance_abs2)
      fitness /= 2.0;
    else
      fitness -= drag_distance_abs2;

    //    len = pow(model->lstring_length, 0.75);
    len = waste_length;
    if (len > fitness / 2.0)
      fitness /= 2.0;
    else
      fitness -= Real(len);
    //    fitness = floor(fitness/10.0)*10.0 + 1.0;

    if (fitness >= 10.0)
      fitness = floor(fitness/10.0)*10.0;
    else if (distance >= 1.0)
      fitness = 1.0;
    else
      fitness = 0.0;

    model->fitness[0] = fitness;
    //    model->fitness[0] = floor(distance/10.0)*10.0 + 1;
    //    model->fitness[0] = 1.0 + distance;
    //    model->fitness[0] = 1.0 + absv(Move[0]);
  }


  if (Print[6]) {
    printf("COM %.2f:%.2f ; %.2f:%.2f\n", com_min[0], com_max[0],
	   com_min[1], com_max[1]);
    printf("dist:%.1f(%.2f:%.2f:%.2f)  #bars: %d, #joints: %d, len: %d  waste:%d  ",
	   distance, drag_distance, drag_distance_abs1, drag_distance_abs2,
	   Num_bars, Num_joints, model->lstring_length, waste_length);
    printf("Fitness #3: %.1f\n", model->fitness[0]);
  }

  if (model->fitness[0] > 100000.0) {
    printf("Fitness distance :: Error - fitness too high: %.3f\n", model->fitness[0]);
    model->error = TRUE;
  } else
    model->error = FALSE;
  translate_struct(-COM_X, 0.0);
  ref_pt = 0;
  ref_rotation = 0.0;

  return TRUE;
}



// Test for cycles and ensure arms are not too long.
int BarStruct :: fitness_distance2(ModelStruct *model)
{
  int i, fail = FALSE;
  Real distance=0.0;

  if (Num_bars < 3) {
    if (Print[9])
      printf("Fitness distance :: Not enough bars: %d.\n", Num_bars);
    fail = TRUE;

  } else if (Num_joints == 0) {
    if (Print[9])
      printf("Fitness distance :: Not enough joints: %d.\n", Num_joints);
    fail = TRUE;
  } else {
    fail = TRUE;
    for (int i=1; i<Num_bars; i++) {
      if (Bar[i].num_children > 1) {
	fail = FALSE;
	break;
      }
    }
    if (fail && Print[9]) {
      printf("Fitness distance :: single line - fail.\n");
    }

    label_cycles();
    if (!test_arm_lengths()) {
      fail = TRUE;
    }
  }

  /*
  if (win1)
    animate_struct(-1);
    */

  if (!fail) {
    Real dd1=0.0, dd2=0.0, dd3=0.0;
    Real min_distance = 9999999.0;

    for (i=0; i<Number_trials; i++) {
      distance = run_distance(model);
      if (distance == 0.0) {
	fail = TRUE;
	min_distance = 0.0;
	break;

      } else if (distance < min_distance) {
	min_distance = distance;
	dd1 = absv(drag_distance);
	dd2 = drag_distance_abs1;
	dd3 = drag_distance_abs2;
      }
    }
    distance = min_distance;
    drag_distance = dd1;
    drag_distance_abs1 = dd2;
    drag_distance_abs2 = dd3;
  }

  if (fail || (distance-drag_distance_abs1 <= 0)) {
    model->fitness[0] = 0.0;
  } else {
    Real fitness, len;
    fitness = distance - drag_distance_abs1;
    drag_distance_abs2 = sqrt(drag_distance_abs2);
    if (fitness/2.0 < drag_distance_abs2)
      fitness /= 2.0;
    else
      fitness -= drag_distance_abs2;

    //    len = pow(model->lstring_length, 0.75);
    len = waste_length;
    if (len > fitness / 2.0)
      fitness /= 2.0;
    else
      fitness -= Real(len);
    //    fitness = floor(fitness/10.0)*10.0 + 1.0;

    if (fitness >= 10.0)
      fitness = floor(fitness/10.0)*10.0;
    else if (distance >= 1.0)
      fitness = 1.0;
    else
      fitness = 0.0;

    model->fitness[0] = fitness;
    //    model->fitness[0] = floor(distance/10.0)*10.0 + 1;
    //    model->fitness[0] = 1.0 + distance;
    //    model->fitness[0] = 1.0 + absv(Move[0]);
  }


  if (Print[6]) {
    printf("COM %.2f:%.2f ; %.2f:%.2f\n", com_min[0], com_max[0],
	   com_min[1], com_max[1]);
    printf("dist:%.1f(%.2f:%.2f:%.2f)  #bars: %d, #joints: %d, len: %d  waste:%d  ",
	   distance, drag_distance, drag_distance_abs1, drag_distance_abs2,
	   Num_bars, Num_joints, model->lstring_length, waste_length);
    printf("Fitness #2: %.1f\n", model->fitness[0]);
  }

  if (model->fitness[0] > 100000.0) {
    printf("Fitness distance :: Error - fitness too high: %.3f\n", model->fitness[0]);
    model->error = TRUE;
  } else
    model->error = FALSE;
  translate_struct(-COM_X, 0.0);
  ref_pt = 0;
  ref_rotation = 0.0;

  return TRUE;
}


// No testing for cycles or arm lengths.
int BarStruct :: fitness_distance(ModelStruct *model)
{
  int i, fail = FALSE;
  Real distance = 0.0;

  if (Num_bars < 3) {
    if (Print[9])
      printf("Fitness distance :: Not enough bars: %d.\n", Num_bars);
    fail = TRUE;

  } else if (Num_joints == 0) {
    if (Print[9])
      printf("Fitness distance :: Not enough joints: %d.\n", Num_joints);
    fail = TRUE;
  } else {
    fail = TRUE;
    for (i=1; i<Num_bars; i++) {
      if (Bar[i].num_children > 1) {
	fail = FALSE;
	break;
      }
    }
    if (fail && Print[9]) {
      printf("Fitness distance :: single line - fail.\n");
    }
  }

  if (!fail) {
    Real dd1 = 0.0, dd2 = 0.0, dd3 = 0.0;
    Real min_distance = 9999999.0;

    for (i=0; i<Number_trials; i++) {
      distance = run_distance(model);
      if (distance == 0.0) {
	fail = TRUE;
	min_distance = 0.0;
	break;

      } else if (distance < min_distance) {
	min_distance = distance;
	dd1 = absv(drag_distance);
	dd2 = drag_distance_abs1;
	dd3 = drag_distance_abs2;
      }
    }
    distance = min_distance;
    drag_distance = dd1;
    drag_distance_abs1 = dd2;
    drag_distance_abs2 = dd3;
  }

  if (fail || (distance-drag_distance_abs1 <= 0)) {
    model->fitness[0] = 0.0;
  } else {
    Real fitness, len;
    fitness = distance - drag_distance_abs1;
    drag_distance_abs2 = sqrt(drag_distance_abs2);
    if (fitness/2.0 < drag_distance_abs2)
      fitness /= 2.0;
    else
      fitness -= drag_distance_abs2;

    //    len = pow(model->lstring_length, 0.75);
    len = waste_length;
    if (len > fitness / 2.0)
      fitness /= 2.0;
    else
      fitness -= Real(len);
    //    fitness = floor(fitness/10.0)*10.0 + 1.0;

    if (fitness >= 10.0)
      fitness = floor(fitness/10.0)*10.0;
    else if (distance >= 1.0)
      fitness = 1.0;
    else
      fitness = 0.0;

    model->fitness[0] = fitness;
    //    model->fitness[0] = floor(distance/10.0)*10.0 + 1;
    //    model->fitness[0] = 1.0 + distance;
    //    model->fitness[0] = 1.0 + absv(Move[0]);
  }


  if (Print[6]) {
    printf("COM %.2f:%.2f ; %.2f:%.2f\n", com_min[0], com_max[0],
	   com_min[1], com_max[1]);
    printf("dist:%.1f(%.2f:%.2f:%.2f)  #bars: %d, #joints: %d, len: %d  waste:%d  ",
	   distance, drag_distance, drag_distance_abs1, drag_distance_abs2,
	   Num_bars, Num_joints, model->lstring_length, waste_length);
    printf("Fitness #1: %.1f\n", model->fitness[0]);
  }

  if (model->fitness[0] > 100000.0) {
    printf("Fitness distance :: Error - fitness too high: %.3f\n", model->fitness[0]);
    model->error = TRUE;
  } else
    model->error = FALSE;
  translate_struct(-COM_X, 0.0);
  ref_pt = 0;
  ref_rotation = 0.0;

  return TRUE;
}


int BarStruct :: fitness_com_height(ModelStruct *model)
{
  int fail = FALSE;
  Real distance = 0.0;

  if (Num_bars < 3)
    fail = TRUE;
  else if (Num_joints == 0)
    fail = TRUE;
  //  else if (cycle_for_intersections()) {
  //    fail = TRUE;
  //  }

  /*
  if (win1)
    animate_struct(-1);
    */

  if (!fail) {
    initialize_evaluation();
    Time = 0.0;
    com_max[0] = COM_X;
    com_max[1] = COM_Y;
    com_min[0] = COM_X;
    com_min[1] = COM_Y;

    if (!animate_struct(1.0)) {
      model->error = TRUE;
      printf("Animate structure error\n");
      return FALSE;
    }

    if ((com_max[0] != com_min[0]) ||
	(com_max[1] != com_min[1])) {
      if (!animate_struct(10.0)) {
	model->error = TRUE;
	printf("Animate structure error\n");
	return FALSE;
      }

      distance = com_max[1]-com_min[1];
    } else {
      distance = 0.0;
    }
  }

  if (fail) {
    model->fitness[0] = 0.0;
  } else {
    model->fitness[0] = 1.0 + distance;
  }


  if (Print[6]) {
    printf("#bars: %d, #joints: %d.\n", Num_bars, Num_joints);
    printf("Fitness: %.1f\n", model->fitness[0]);
  }

  if (model->fitness[0] > 100000.0)
    model->error = TRUE;
  else
    model->error = FALSE;
  translate_struct(-COM_X, 0.0);

  return TRUE;
}


int BarStruct :: fitness_struct(ModelStruct *model)
{
  if (Fitness_func == 1)
    return fitness_distance(model);
  else if (Fitness_func == 2)
    return fitness_distance2(model);
  else if (Fitness_func == 3)
    return fitness_distance3(model);
  else if (Fitness_func == 4)
    return fitness_distance4(model);
  else if (Fitness_func == 5)
    return fitness_distance5(model);
  else if (Fitness_func == 6)
    return fitness_com_height(model);
  return FALSE;
}


int BarStruct :: fitness_blind_watchmaker(int fv_len, 
					  Real *fitness_vec)
{
  int fail = FALSE;

  if (Num_bars < 3)
    fail = TRUE;
  else if (Num_joints == 0)
    fail = TRUE;
  /*
  else if (cycle_for_intersections()) {
    fail = TRUE;
  }
  */

  if (fail) {
    fitness_vec[0] = 0.0;
  } else {
    fitness_vec[0] = 1.0;
  }

  if (Print[6]) {
    printf("#bars: %d, #joints: %d.\n", Num_bars, Num_joints);
    printf("Fitness: %.1f\n", fitness_vec[0]);
  }

  return (!fail);
}



// Translates and rotates the structure until it
// is in a stable state, resting on the surface.
int BarStruct :: settle_struct_old1(void)
{
  int i;
  int num_grnd_pts;
  int left_min_index, right_min_index;
  int max_tries = Num_bars;
  int settle_attempts = 0;
  Real min_y, move_x;

  // Notes:
  // Should keep track of all bars resting on ground
  // so as to deterine their influence on object movement.

  if (settle_attempts++ > Num_bars+2) {
    printf("Settle struct failed.  To many attempts: %d.\n", settle_attempts);
    while(1);
    return FALSE;
  }

  left_min_index = 0;
  right_min_index = 0;
  min_y = Bar[0].end_y;
  if (max_tries < 10)
    max_tries = 10;

  // Find left-most and right-most supports.
  for (i=1; i<Num_bars; i++) {
    if (absv(Bar[i].end_y-min_y) < DIST_TOLERANCE) {
      if (Bar[i].end_x < Bar[left_min_index].end_x) {
	left_min_index = i;
	if (Bar[i].end_y < Bar[right_min_index].end_y) {
	  Bar[right_min_index].end_y = Bar[i].end_y;
	  min_y = Bar[i].end_y;
	}

      } else if (Bar[i].end_x > Bar[right_min_index].end_x) {
	right_min_index = i;
	if (Bar[i].end_y < Bar[left_min_index].end_y) {
	  Bar[left_min_index].end_y = Bar[i].end_y;
	  min_y = Bar[i].end_y;
	}
      }

    } else if (Bar[i].end_y < min_y) {
      min_y = Bar[i].end_y;
      left_min_index = i;
      right_min_index = i;
    }
  }

  if (left_min_index == right_min_index) {
    if (absv(Bar[left_min_index].end_x - COM_X) < DIST_TOLERANCE) {
#ifndef OPTIMIZE
      if (Print[12]) {
	printf("Balanced: %.6f  <  %.6f  <  %.6f\n", Bar[left_min_index].end_x,
	       COM_X, Bar[right_min_index].end_x);
      }
#endif
      // Balanced on a point...
      return TRUE;
    }
  }

#ifndef OPTIMIZE
  if (Print[12])
    printf("%.6f  <  %.6f  <  %.6f\n", Bar[left_min_index].end_x,
	   COM_X, Bar[right_min_index].end_x);
#endif

  if (COM_X < Bar[left_min_index].end_x) {
    // Rotate counter clockwise;
    return settle_left(left_min_index);

  } else if (COM_X > Bar[right_min_index].end_x) {
    // Rotate clockwise;
    return settle_right(right_min_index);
  }
   
  // Settled!
  // Find average x displacement of ground points.
  move_x = 0.0;
  num_grnd_pts = 0;
  for (i=0; i<Num_bars; i++) {
    if (Bar[i].ground_pt) {
      num_grnd_pts++;
      move_x += Bar[i].prev_x - Bar[i].end_x;
    }
  }
  if (num_grnd_pts > 0)
    move_x /= Real(num_grnd_pts);
  else
    move_x = 0.0;

  // Find minimum y.
  min_y = Bar[0].end_y;
  for (i=1; i<Num_bars; i++) {
    if (Bar[i].end_y < min_y)
      min_y = Bar[i].end_y;
  }

  translate_struct(move_x, -min_y);

  // Determine the ground points.
  for (i=0; i<Num_bars; i++) {
    if (Bar[i].end_y < DIST_TOLERANCE)
      Bar[i].ground_pt = TRUE;
    else
      Bar[i].ground_pt = FALSE;
  }

#ifndef OPTIMIZE
  if (Print[12]) {
    printf("Settled: %.1f:%.1f  (%.3f:%.3f)  CoM: %.1f:%.1f.\n",
	   Bar[1].end_x, Bar[1].end_y,
	   move_x, min_y,
	   COM_X, COM_Y);
  }
  if (Print[18])
    printf("CoM: %.6f, %.6f  (Time: %.4f)\n", COM_X, COM_Y, Time);
#endif


  return TRUE;
}


#ifdef OLD
int BarStruct :: add_bar2(Real len, Real *child_angle, Real spd, Real offset,
			 int type, int *p_index)
{
  int i;
  int child, same_as, child_index;
  int parent = *p_index;
  Real end_x, end_y;
  Real angle = *child_angle;

#ifndef OPTIMIZE
  if (Print[12] || Print[26]) {
    printf("Add bar (%d): len:%.1f, spd:%.2f(%.2f), ang:%.3f, type:%d\n",
	   parent, len, spd, offset, angle, type);
  }
#endif

  if (Num_bars >= MAX_BARS2) {
    if (Print[9])
      printf("Maximum number of bars exceeded.\n");
    return FALSE;
  }


  // Adjust length.
  if (len == 0.0)
    len = 1.0;
  else if (len > 5.0)
    len = 5.0;
  len *= 10.0;

  // Adjust speed.
  if (type == BAR_TYPE_FIXED) {
    spd = 0.0;
    offset = 0.0;
  } else {
    if (spd > 5.0)
      spd = 5.0;
    else if (spd < -5.0)
      spd = -5.0;

    if ((type == BAR_TYPE_COUNTER2) ||
	(type == BAR_TYPE_CLOCKWISE2))
      spd /= 2.0;  // hack to make speeds constant for different angle sizes.
  }

  angle = simplify_angle_180(angle);
  if ((absv(absv(angle)-PI) < ANGLE_TOLERANCE) && (parent>0)) {
    // Go back up parent.
    //    printf("  Moving up to parent (%.5f).\n",
    //	   absv((absv(angle)-PI)));
    *p_index = Bar[parent].parent;
    *child_angle = PI;
    return TRUE;
  }

  if (Bar[parent].num_children >= MAX_CHILDREN) {
    if (Print[9]) {
      printf("Too many children for bar #%d.\n", parent);
    }
    return FALSE;
  }

  if (Bar[parent].num_children != 0) {
    // test for conflicts.
    if (Bar[parent].type != BAR_TYPE_FIXED) {
      if (Bar[parent].child_angle_b[0] < Bar[parent].child_angle_a[0]) {
	if ((angle + ANGLE_TOLERANCE > Bar[parent].child_angle_b[0]) &&
	    (angle - ANGLE_TOLERANCE < Bar[parent].child_angle_a[0])) {
	  // Same angle as first child - traverse that child.
	  *p_index = Bar[parent].child[0];
	  *child_angle = 0.0;
	  if (Print[12] || Print[26]) {
	    printf(" same as first child -> %d.\n", child);
	  }
	  return TRUE;
	}

      } else {
	if ((angle - ANGLE_TOLERANCE < Bar[parent].child_angle_b[0]) &&
	    (angle + ANGLE_TOLERANCE > Bar[parent].child_angle_a[0])) {
	  // Same angle as first child - traverse that child.
	  *p_index = Bar[parent].child[0];
	  *child_angle = 0.0;
	  if (Print[12] || Print[26]) {
	    printf(" same as first child -> %d.\n", child);
	  }
	  return TRUE;
	}
      }

#ifdef OLD
      if (absv(Bar[parent].child_angle_b[0]-angle) < ANGLE_TOLERANCE) {
	// Same angle as first child - traverse that child.
	*p_index = Bar[parent].child[0];
	*child_angle = 0.0;
	if (Print[12] || Print[26]) {
	  printf(" same as first child -> %d.\n", child);
	}
	return TRUE;
      }
#endif
#ifdef OLD
      child = Bar[parent].child[0];
      if (absv(Bar[child].angle_b-angle) < ANGLE_TOLERANCE) {
	// Same angle as first child - traverse that child.
	*p_index = child;
	*child_angle = 0.0;
	if (Print[12] || Print[26]) {
	  printf(" same as first child -> %d.\n", child);
	}
	return TRUE;
      }
#endif
    }

    for (i=0; i<Bar[parent].num_children; i++) {
      if (absv(Bar[parent].child_angle_a[i]-angle) < ANGLE_TOLERANCE) {
	// Traverse up this bar and end.
	//	printf("  Traversing bar#: %d (child: %d).\n", child, i);
	child = Bar[parent].child[i];
	if (Bar[child].same_as != -1) {
	  *p_index = Bar[child].same_as;
	  *child_angle = Bar[child].abs_angle - Bar[*p_index].abs_angle;
	} else {
	  *p_index = child;
	  *child_angle = 0.0;
	}
	if (Print[12] || Print[26]) {
	  printf(" same as child -> %d.\n", child);
	}
	return TRUE;
      }
    }

#ifdef OLD
    for (i=0; i<Bar[parent].num_children; i++) {
      child = Bar[parent].child[i];
      if (absv(Bar[child].angle_a-angle) < ANGLE_TOLERANCE) {
	// Traverse up this bar and end.
	//	printf("  Traversing bar#: %d (child: %d).\n", child, i);
	if (Bar[child].same_as != -1) {
	  *p_index = Bar[child].same_as;
	  *child_angle = Bar[child].abs_angle - Bar[*p_index].abs_angle;
	} else {
	  *p_index = child;
	  *child_angle = 0.0;
	}
	if (Print[12] || Print[26]) {
	  printf(" same as child -> %d.\n", child);
	}
	return TRUE;
      }
    }
#endif
  }


  // Start setting values for child.
  child = Num_bars;
  child_index = Bar[parent].num_children;
  Bar[parent].child[child_index] = child;
  Bar[parent].num_children++;
  Bar[child].parent = parent;
  Bar[child].num_children = 0;

  Bar[child].move_spd = spd;
  Bar[child].move_offset = offset;
  Bar[child].length = len;
  Bar[child].type = type;
  if (same_as == -1)
    Bar[child].weight = 1.0;
  else
    Bar[child].weight = 0.0;

  Bar[child].angle_a = angle;
  if ((Bar[parent].type != BAR_TYPE_FIXED) && 
      (Bar[parent].num_children == 1)) {
    Num_joints++;
    Bar[child].moves = TRUE;
    Bar[child].joint_index = parent;
    if (Bar[parent].type == BAR_TYPE_CLOCKWISE) {
      Bar[child].angle_b = angle - PI/3.0;
    } else {
      Bar[child].angle_b = angle + PI/3.0;
    }
    //    printf(" moves: type:%d, angle:%.3f\n", type, angle);
      
  } else {
    Bar[child].angle_b = angle;
    Bar[child].moves = FALSE;
    Bar[child].joint_index = Bar[parent].joint_index;
  }

  Bar[parent].child_angle_a[child_index] = Bar[child].angle_a;
  Bar[parent].child_angle_b[child_index] = Bar[child].angle_b;


  // Calculate location and see if it matches anyone else.
  Real theta = offset * 2.0;
  if (theta > 2.0)
    theta = 2.0 - theta;
  angle = Bar[child].angle_a*(1.0-theta) + Bar[child].angle_b*theta;
  Bar[child].abs_angle = Bar[parent].abs_angle + angle;
  end_x = Bar[parent].end_x + len*cos(Bar[child].abs_angle);
  end_y = Bar[parent].end_y + len*sin(Bar[child].abs_angle);
  same_as = -1;
  for (i=0; i<Num_bars; i++) {
    if ((absv(Bar[i].end_x - end_x) < DIST_TOLERANCE) &&
	(absv(Bar[i].end_y - end_y) < DIST_TOLERANCE)) {
      same_as = i;
      break;
    }
  }
  Bar[child].same_as = same_as;
  Bar[child].end_x = end_x;
  Bar[child].end_y = end_y;

  if (same_as != -1) {
    // Connect parent to child for existing child.
    Real angle2 = (Bar[parent].abs_angle + angle - PI) - Bar[same_as].abs_angle;
    angle2 = simplify_angle_180(angle2);

    Bar[same_as].child[Bar[same_as].num_children] = parent;
    Bar[same_as].num_children++;

    // Connect parent to same_as.
    int child_index2 = child_index + 1;
    Bar[parent].child[child_index2] = same_as;
    Bar[parent].child_angle_a[child_index2] = Bar[parent].child_angle_a[child_index];
    Bar[parent].child_angle_b[child_index2] = Bar[parent].child_angle_b[child_index];
    Bar[parent].num_children++;
  }
  Num_bars++;

#ifndef OPTIMIZE
  if (Print[22] && !Print[12]) {
    printf("%d(%d): ", child, parent);
    printf(" len:%.1f, type:%d, moves:%d, spd:%.3f angle(%.2f:%.2f)\n",
	   Bar[child].length, Bar[child].type, Bar[child].moves,
	   Bar[child].move_spd,
	   Bar[child].angle_a*RAD2DEG, Bar[child].angle_b*RAD2DEG);
  }    

  if (Print[12] || Print[25]) {
    printf("P: ");
    printf("%2d(%2d): %.2f:%.2f  type:%d, moves:%d, spd:%.3f(%.2f), angle(%.3f:%.3f)\n",
	   parent, Bar[parent].same_as, Bar[parent].end_x, Bar[parent].end_y, Bar[parent].type, Bar[parent].moves,
	   Bar[parent].move_spd, Bar[parent].move_offset,
	   Bar[parent].angle_a, Bar[parent].angle_b);
    
    printf("C: ");
    printf("%2d(%2d): %.2f:%.2f  type:%d, moves:%d, spd:%.3f(%.2f) angle(%.3f:%.3f)\n",
	   child, Bar[child].same_as, Bar[child].end_x, Bar[child].end_y,
	   Bar[child].type, Bar[child].moves,
	   Bar[child].move_spd, Bar[parent].move_offset,
	   Bar[child].angle_a, Bar[child].angle_b);
    printf("\n");
  }
#endif

  if (same_as != -1) {
    *p_index = Bar[child].same_as;
    *child_angle = Bar[child].abs_angle - Bar[*p_index].abs_angle;
  } else {
    *p_index = child;
    *child_angle = 0.0;
  }

  /*
  for (i=0; i<Num_bars; i++) {
    printf("%d: ", i);
    if (Bar[i].same_as == -1) {
      for (int j=0; j<Bar[i].num_children; j++) {
	printf("%d(%d)  ", Bar[i].child[j], Bar[i].in_cycle[j]);
      }
    } else {
      printf(" = %d", Bar[i].same_as);
    }
    printf("\n");
  }
  */

  return TRUE;
}
#endif

#ifdef OLD
int BarStruct :: arm_length(int node, int depth)
{
  int length, max_length = depth;

#ifndef OPTIMIZE
  //  if (Print[25])
  //    printf(" %d : %d\n", node, depth);
#endif
  if (depth > MAX_ARM_LENGTH)
    return depth;

  int i, child, parent;

  Bar[node].updated = depth;

  //  printf("%d (%d)\n", node, depth);

  // Test same_as.
  if (Bar[node].same_as != -1) {
    child = Bar[node].same_as;
    if (Bar[child].updated == 0) {
      length = arm_length(Bar[node].same_as, depth);
      if (length > MAX_ARM_LENGTH) {
	Bar[node].updated = 0;
	return length;
      } else if (length > max_length)
	max_length = length;
    }
  }

  // Test parent.
  parent = Bar[node].parent;
  if (Bar[parent].updated == 0) {
    length = arm_length(parent, depth+1);
    if (length > MAX_ARM_LENGTH) {
      Bar[node].updated = 0;
      return length;
    } else if (length > max_length)
      max_length = length;
  }
  
  // Test children.
  for (i=0; i<Bar[node].num_children; i++) {
    child = Bar[node].child[i];
    if (Bar[child].updated == 0) {
      length = arm_length(child, depth+1);
      if (length > MAX_ARM_LENGTH) {
	Bar[node].updated = 0;
	return length;
      } else if (length > max_length)
	max_length = length;
    }
  }

  // Test from list.
  for (i=0; i<Bar[node].num_from; i++) {
    child = Bar[node].from_list[i];
    if (Bar[child].updated == 0) {
      length = arm_length(child, depth+1);
      if (length > MAX_ARM_LENGTH) {
	Bar[node].updated = 0;
	return length;
      } else if (length > max_length)
	max_length = length;
    }
  }

  Bar[node].updated = 0;
  return max_length;
}
#endif
