/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#ifndef BARS_2D_HEADER_FILE
#define BARS_2D_HEADER_FILE

#include "general.h"

class Graphic;
class Individual;
class Pop;
class Deme;

#define MAX_BARS2 1500
#define BAR_STACK_SIZE 75
#define MAX_CHILDREN 7

#define HISTORY_LENGTH 8


typedef struct tBar
{
  int type;
  int moves;
  Real angle_a; // joint angle in state a (also angle for non-joints).
  Real angle_b; // joint angle in state b.
  Real true_angle_a;
  Real true_angle_b;
  Real length;
  Real end_x;
  Real end_y;
  Real abs_angle; // reference angle (angle wrt world coords).
  Real move_spd;
  Real move_offset;

  Real center_x;
  Real center_y;

  int ground_pt; // True if this point is on the ground.
  int ground_pt_prev; // True if ground point at previous iteration.
  Real weight;
  Real weight_supported;

  int updated;
  Real prev_x;
  Real prev_y;
  Real x_history[HISTORY_LENGTH];
  Real y_history[HISTORY_LENGTH];

  int parent; // index of parent.
  int same_as;
  int joint_index;
  int num_children;  // number of children.
  int child[MAX_CHILDREN];  // indexes of children.
  int in_cycle[MAX_CHILDREN]; // TRUE if edge in cycle.

  int num_from;
  int from_list[MAX_CHILDREN];
} Bar_struct;


typedef struct tBarStack
{
  Real angle;
  Real offset;
  int bar_index;
} Bar_Stack;


int lines_intersect_2d(Real x1, Real y1, Real x2, Real y2,
		       Real x3, Real y3, Real x4, Real y4);



class BarStruct
{
public:
  int ref_pt;
  int Num_bars;
  int Num_joints;
  int waste_length;
  Bar_struct Bar[MAX_BARS2];
  Real total_weight;
  Real COM_X, COM_Y, COM_Z, com_y_avg;
  Real Move[3];
  Real drag_distance;  // Average distance that ground points are dragged (pos and neg cancel).
  Real drag_distance_abs1; // Absolute distance ground points are translated.
  Real drag_distance_abs2;  // Average absolute distance ground points are translated (pos and neg are absoluted).
  Real ref_rotation;

  Real draw_loc[1];
  Real com_max[2], com_min[2];

  Real Time_step_2d;
  Real Time;
  Real Time_history_prev;

  char bar_check[MAX_BARS2][MAX_BARS2];
  int num_near[MAX_BARS2];
  Real last_check_near, last_check_far;

  int num_grnd_pts;
  int num_grnd_pts_prev;
  int *ground_pt;
  int *ground_pt_prev;

  int Bar_stack_ptr;
  Bar_Stack Bar_stack[BAR_STACK_SIZE];

  BarStruct(void);
  void initialize(void);
  void set_translate(void);
  void draw_history2(int h);
  void draw_struct2(void);
  void draw_history(int h);
  void draw_struct(void);
  void draw_bars(int arg);

  void add_connection_noise(void);
  int arm_length(int node, int depth);
  int max_arm_length(void);
  int test_arm_lengths(void);
  int test_children(int node, int root, int depth);
  void label_cycles(void);
  void initialize_intersection(void);
  int is_intersection(void);
  int is_intersection_old2(void);
  int is_intersection_old1(void);
  void calculate_up(Real time);
  void calculate_positions(Real time);
  int cycle_for_intersections(void);

  void calculate_com(void);
  void print_struct(void);
  void translate_struct(Real x, Real y);
  void rotate_struct(Real angle);
  int settle_right(int right_min_index);
  int settle_left(int left_min_index);
  int settle_direction(int *pivot_pt);
  void distribute_mass(void);
  int settle_struct(void);
  int settle_structure(void);
  Real animate_struct(Real delta);

  int add_bar(Real len, Real angle, Real spd, Real offset, int type, int p);
  int add_bar(Real len, Real angle, Real spd, int type, int p);
  int add_bar2(Real len, Real *child_angle, Real spd, Real offset,
	       int type, int *p_index);
  int build_structure60(int length, LString *command, Real *args);
  int build_structure_param(int length, LString *command, Real *args);

  void initialize_evaluation(void);
  Real run_distance(ModelStruct *model);
  int fitness_distance5(ModelStruct *model);
  int fitness_distance4(ModelStruct *model);
  int fitness_distance3(ModelStruct *model);
  int fitness_distance2(ModelStruct *model);
  int fitness_distance(ModelStruct *model);
  int fitness_com_height(ModelStruct *model);
  int fitness_struct(ModelStruct *model);
  int fitness_blind_watchmaker(int fv_len, Real *fitness_vec);

  int settle_struct_old1(void);

};

#endif
