/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#ifndef BARS_3D_HEADER_FILE
#define BARS_3D_HEADER_FILE

#include "general.h"
#include "network.h"

class Graphic;
class Individual;
class Pop;
class Deme;

#define MAX_BARS3 400
#define BAR_STACK_SIZE 75
#define MAX_CHILDREN3D 6
#define EDGE_STACK_SIZE 75


#define ROT_NONE 0
#define ROT_X_AXIS 1
#define ROT_Y_AXIS 2
#define ROT_Z_AXIS 3
#define ROT_REVERSE 4


typedef struct HullPoint
{
  int id;
  int inside;
  Real r;
  Real slope;
  Real slope_COM;
} HullPt;


typedef struct tBar3d
{
  int type;
  int moves;

  Real length;
  Real len_short;  // length / 2.0

  int rot_axis;

  Real angle_a; // joint angle in state a (also angle for non-joints).
  Real angle_b; // joint angle in state b.
  Real true_angle_a;
  Real true_angle_b;
  Real move_spd;
  Real move_offset;

  Real angle; // Current angle - only used for network controllers.
  Real desired_angle; // Only used for network controllers.
  Real sensor_value; // Current sensor value.

  Real end_loc[3];
  Real abs_angle;
  Real center[3];
  Real delta_rot_mat[3][3];
  Real inverse_delta_rot[3][3];
  Real true_delta_rot[3][3];
  Real rotation_matrix[3][3];
  Real abs_heading[3][3]; // absolute heading when built.

  Real prev_loc[3]; // Location at previous phase/time step.
  Real saved_loc[3]; // Used for debugging.

  int ground_pt; // True if this point is on the ground.
  int ground_pt_prev;
  Real weight;
  Real weight_supported;
  int updated;
  int parent; // index of parent.
  int same_as;
  int joint_index;
  int num_children;  // number of children.
  int child[MAX_CHILDREN3D];  // indexes of children.
  int in_cycle[MAX_CHILDREN3D];
  int line_length;
  int in_cycle2; // True if this node is part of any cycle.

  int num_from;
  int from_list[MAX_CHILDREN3D];
} Bar_struct_3d;


typedef struct tBarStack3d
{
  int bar_index;
  int axis;
  Real heading[3][3];
  Real offset;
  LinkIndex current_link;
} Bar_Stack_3d;


int lines_intersect_3d(Real x1, Real y1, Real x2, Real y2,
		       Real x3, Real y3, Real x4, Real y4);


void init_heading_matrix(int mat[3][3]);
void rot_xaxis_pos(int mat[3][3]);
void rot_xaxis_neg(int mat[3][3]);
void rot_yaxis_pos(int mat[3][3]);
void rot_yaxis_neg(int mat[3][3]);
void rot_zaxis_pos(int mat[3][3]);
void rot_zaxis_neg(int mat[3][3]);


class BarStruct3D
{
public:
  int ref_pt; // index of first ground point -> reference point where position calculations start from.
  int Num_bars;
  int Num_joints;
  int Num_sensors;
  Bar_struct_3d Bar[MAX_BARS3];
  Real total_weight;
  Real COM[3];
  Real com_max[3], com_min[3];
  Real draw_loc[3];

  int Is_ball;
  Real Ball_radius;
  Real Ball_loc[3];
  Real Ball_start_loc[3];

  int Joint_index[MAX_BARS3];
  int Sensor_index[MAX_BARS3];

  int waste_length;
  Real max_arm_len;
  Real body_ratio;

  Real drag_distance;  // Average distance that ground points are dragged (pos and neg cancel).
  Real drag_distance_abs1; // Absolute distance ground points are translated.
  Real drag_distance_abs2;  // Average absolute distance ground points are translated (pos and neg are absoluted).

  Real Time_step;
  Real Time;
  Real rotation_matrix[3][3];
  Real abs_orient[3][3];

  char bar_check[MAX_BARS3][MAX_BARS3];
  int num_near[MAX_BARS3];
  Real last_check_near, last_check_far;

  int num_grnd_pts;
  int num_grnd_pts_prev;
  int *ground_pt;
  int *ground_pt_prev;

  int Link_stack_ptr;
  LinkIndex Link_stack[EDGE_STACK_SIZE];

  int Bar_stack_ptr;
  Bar_Stack_3d Bar_stack[BAR_STACK_SIZE];

  Network *network;


  BarStruct3D(void);
  void initialize(void);
  void initialize_network(void);
  void set_translate(void);
  void draw_struct_bw(void);
  void draw_struct_tree(void);
  void draw_struct(void);
  void draw_bars(int arg);

  void add_connection_noise(void);
  void add_connection_noise2(void);
  int arm_length(int node, int depth);
  int max_arm_length(void);
  int max_line_length(void);
  int test_arm_lengths(void);
  int test_children(int node, int root, int depth);
  void label_cycles(void);

  void initialize_intersection(void);
  int is_intersection(void);
  void calculate_up(int parent, Real time);
  void calculate_positions(Real phase);
  void calculate_positions_new(Real phase);
  void calc_bar_position(int i, Real phase);
  void calc_positions(Real phase);
  void calculate_bar_positions(Real phase);
  int cycle_for_intersections(void);

  void calculate_com(void);
  void print_struct(void);
  void translate_struct(Real x, Real y, Real z);
  void rotate_struct_Xaxis(Real angle);
  void rotate_struct_Yaxis(Real angle);
  void rotate_struct_Zaxis(Real angle);
  void rotate_struct(int axis, Real angle);

  Real calc_slope(int p1, int p2);
  Real calc_slope_COM(int p1);
  Real calc_COM_proj(int p1, int p2);
  int COM_above_line(int p1, int p2);

  int build_upper_hull(int num_grnd_pts, int *grnd_pt,
		       HullPt *hull_pt);
  int build_lower_hull(int num_grnd_pts, int *grnd_pt,
		       HullPt *hull_pt);
  int is_balanced_1pt(int num_grnd_pts, int *grnd_pt);
  int is_balanced_2pt(int num_grnd_pts, int *grnd_pt);
  int is_balanced_3pt(int *grnd_pt, HullPt *hull_pt);
  int is_balanced(int num_grnd_pts, int *grnd_pt);
  int rotate_to_balance(int *pivot);

  int same_point(int index1, int index2);
  int settle_right(int right_min_index, int rotate_full);
  int settle_left(int left_min_index, int rotate_full);
  int find_ground_pts(int *grnd_pt);
  void distribute_mass(void);
  int settle_struct(void);
  int settle_structure(void);
  void update_ball_sensors(void);
  void update_ball_position(void);
  Real animate_struct(Real time);

  void update_network(void);

  void make_delta_rot_mat(int index, Real angle);
  void make_joint_rot_mat_network(int index, Real rot_mat[3][3]);
  void make_joint_rot_mat(int index, Real time, Real rot_mat[3][3]);
  int add_bar(Real len, Real heading[3][3], int type, int rot_axis,
	      Real rot_angle_a, Real rot_angle_b, Real spd, Real offset,
	      int *p_index);
  int add_branch(Real len, Real rotation[3][3], int *p_index);
  int build_structure_right(ModelStruct *model);
  int build_structure_tree(ModelStruct *model);
  int build_network(ModelStruct *model);
  int build_structure(ModelStruct *model);

  void initialize_evaluation(void);
  int fitness_com_height(ModelStruct *model);
  Real run_distance(ModelStruct *model);

  void randomize_remote(void);
  Real run_distance_remote(ModelStruct *model);
  Real fitness_distance_value4(Real distance);
  int fitness_distance4(ModelStruct *model);
  Real fitness_distance_value3(Real distance);
  int fitness_distance3(ModelStruct *model);
  Real fitness_distance_value2(Real distance);
  int fitness_distance2(ModelStruct *model);
  Real fitness_distance_value(Real distance);
  int fitness_distance(ModelStruct *model);
  Real fitness_ball_value(Real distance);
  Real fitness_remote_value(Real distance);
  int fitness_ball(ModelStruct *model);
  int fitness_remote(ModelStruct *model);
  int fitness_tree(ModelStruct *model);
  Real fitness_network_value(void);
  int fitness_network(ModelStruct *model);
  int calculate_parity(int num, Real *inputs);
  int fitness_net_parity(ModelStruct *model);
  int calculate_parity2(int num, Real *inputs);
  int fitness_net_parity2(ModelStruct *model);
  int fitness_struct(ModelStruct *model);
  int fitness_blind_watchmaker(int fv_len, Real *fitness_vec);
};

#endif
