/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#ifndef IND_LSYS_HEADER_FILE
#define IND_LSYS_HEADER_FILE

#include <iostream.h>
#include <fstream.h>
#include "general.h"
#include "lsystem.h"


#define CREATE_INIT -1
//define CREATE_REEVALUATED -1
#define CREATE_MUTATE 0
#define CREATE_RECOMBINE 1

#define TAPE_END '.'

#define NUM_PRODUCTIONS 15

#define MAX_PROD_BODIES 2
#define MAX_PROD_VARS 3
#define MAX_CONSTANTS 10
#define MAX_FUNCTIONS 0

#define OP_RECOMB 1
#define OP_MUTATE 2

#define OP_REC_RULE 0
#define OP_REC_PROD 1
#define NUM_RECOMB_OPS 2
#define OP_MUT_CONST 2
#define OP_MUT_CHAR 3
#define OP_MUT_ADD 4
#define OP_MUT_DEL 5
#define OP_MUT_REORDER 6
#define OP_MUT_COND 7
#define OP_MUT_REPLICATE 8
#define NUM_MUTATE_OPS 7

#define OP_REC_CONST 9
#define OP_UNKNOWN 9


extern Real Rec_delta;
extern int Num_arguments;
extern int Num_constants;
extern int Num_functions;
extern int Num_productions;
extern int Max_bodies;
extern int Num_operations;
extern Real Prob_mult_seq;
extern Real Prob_branch_edge;

extern int Max_prod_length;
extern int Prod_length;
extern int Max_string_length;


// Constants:
// 0: number of iterations left.
// 1:


// Functions:
// These take a productions arguments/variables as
// their inputs: f1 = prod_value1 op prod_value2.
// This way the lines specifying a function do not
// need to specify their parameters.
// eg: f1=(v1 * k1), f2=(4+v2), ...
// Note: only productions should be allowed to have
// their own functions.  Terminals and blocks should
// only be allowed to use a single prod_value.


class Prod_Body;

typedef struct History
{
  int index;
  int how_created;
  int operation;
  int num_mutate;
  int num_recombine;
  int op_count[15];
  Real delta;
  Real val1;

  Real fitness[10];
  Real fitness_prev[10];
} history;


typedef struct tConstant
{
  Real value;
  int used;
} Constant;

typedef struct tTape
{
  char type;
  char symbol;
  Real *arg;
} Tape;


// Rules.

typedef struct Production_Head
{
  int num_bodies;
  int used;
  Real *max;
  Real *min;
  Prod_Body **body;
} *Prod_Head;





class Individual
{
public:
  int created;
  int evaluated;
  int version;
  int iterations;
  int length;
  Constant constant[MAX_CONSTANTS];
  Prod_Head *rules;
  //  Prod_Head rules[NUM_PRODUCTIONS];
  LString *lstring;
  ModelStruct model;

  history hist_self;
  history hist_parent1;
  history hist_parent2;


  /*************************/

  Individual(void); // perhaps input data struct
  Individual(Individual *sample);
  void clear(void);
  int randomize_rule(int r);
  void make_random(void);
  void init(void);
  void init(int init_h);
  void init_pheno(int init_h); // For making single production rule & body.
  int repair_rule(Prod_Head ph);
  int repair_rule(int i);
  void repair_init(void);

  int create_mutate(void);
  int better(void);
  int better(Individual *other);
  int replace(Individual *other);
  int fitness_fail(void);
  Real get_fitness(void);
  void get_fitness(Real *fitness_vector);
  void set_fitness(Real val);
  Real get_fitness(int index);
  void set_fitness(Real *val);
  void set_fitness(int index, Real val);
  void add_fitness(Real val);
  void div_fitness(Real val);
  void set_creation(int type);
  int get_creation(void);
  void init_hist(history *hist);
  void init_hist(void);
  void duplicate_hist(history *hist_from, history *hist_to);
  void duplicate(Individual *ind);
  int duplicate_pheno(Individual *ind);
  void op_used(int op_num);
  int system_length(void);

  int duplicate_used_bodies(int r);
  int mutate_rule_duplicate(int r);
  int mutate_rule_new(int r);
  int mutate_rule(int r);
  void clear_rule(int r);
  int random_used_body_p2(int r);
  int random_used_body(int r);
  int random_unused_body(int r);
  int is_rule_used(int r);
  int random_used_rule(void);
  int random_unused_rule(void);
  int random_used_constant(void);
  int mutate_constant(void);
  int encapsulate(int r);
  int mutate(Real alpha);
  int mutate(void);
  int mutate_pheno(void);
  int mutate_init(void);
  int recombine_constants(Individual *parent2);
  int recombine_productions(Individual *parent2);
  int recombine_rule(Individual *parent2);
  int join(Individual *parent2);
  int recombine(Individual *parent2);
  int recombine_pheno(Individual *parent2);

  int process_tape(Tape *tape_src, Tape *tape_dest);
  void reset_used(void);
  int start_tape(Real *var, Tape *tape);
  int tape_to_string(Tape *tape, LString *string);
  int edit_distance(Individual *ind2);
  int phenotype_distance(Individual *ind2);
  int phenotype_distance2(Individual *ind2);
  int phenotype_dist(Individual *ind2);
  void display(void);
  Real evaluate_helper(int init, int iterations, Real *argument);
  Real evaluate(int init);
  Real evaluate(void);

  void print_fitness(void);
  void print_history(history *hist);
  void print_history2(history *hist);
  void fprint_history2(history *hist, FILE *log);
  void print_history_self(void);
  void print_history_self2(void);
  void fprint_history_self2(FILE *log);
  void print_history(void);
  void fprint_history(FILE *log);
  void print_history_full(void);
  void print_results(void);
  void info(void);

  int read_history(history *hist, FILE *file);
  int read(FILE *file);
  void write_history(history *hist, FILE *file);
  void write(FILE *file);
  void write_pheno(FILE *file);
};

void print_lchar(int i, LString *lstring);

#endif
