/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stream.h>
#include <iostream.h>
#include <fstream.h>
#include "general.h"
#include "network2.h"
#include "ind_tree.h"
#include "tree.h"

#define GP_VERSION 1


Individual :: Individual(void)
{
  init_hist(&hist_self);
  init_hist(&hist_parent1);
  init_hist(&hist_parent2);
  created = FALSE;
  evaluated = FALSE;
  num_inputs = 1;
  num_outputs = 1;
}

Individual :: Individual(Individual *sample)
{
  init_hist(&hist_self);
  init_hist(&hist_parent1);
  init_hist(&hist_parent2);
  created = FALSE;
  evaluated = FALSE;
  if (sample != NULL) {
    num_inputs = sample->num_inputs;
    num_outputs = sample->num_outputs;
  } else {
    num_inputs = 1;
    num_outputs = 1;
  }
}

void Individual :: init(void)
{
  created = TRUE;

  if (num_inputs > 0)
    input = (Real *)malloc(num_inputs*sizeof(Real));
  if (num_outputs > 0)
    output = (Real *)malloc(num_outputs*sizeof(Real));

  init_hist(&hist_self);
  init_hist(&hist_parent1);
  init_hist(&hist_parent2);
  tree = tree_new();
  tree->source = this;
}


void Individual :: init_hist(history *hist)
{
  int i;

  hist->index = -1;
  hist->operation = OP_INIT;
  hist->how_created = CREATE_INIT;
  hist->num_mutate = 0;
  hist->num_recombine = 0;
  hist->delta = 0.0;
  hist->op1 = 0;
  hist->op2 = 0;
  hist->op3 = 0;
  hist->fitness = 0.0;
  hist->fitness_prev = 0.0;

  for (i=0; i<7; i++)
    hist->op_count[i] = 0;
}

void Individual :: init_hist(void)
{
  int i;

  hist_self.index = -1;
  hist_self.operation = OP_INIT;
  hist_self.how_created = CREATE_INIT;
  hist_self.num_mutate = 0;
  hist_self.num_recombine = 0;
  hist_self.delta = 0.0;
  hist_self.op1 = 0;
  hist_self.op2 = 0;
  hist_self.op3 = 0;
  hist_self.fitness_prev = 0.0;

  for (i=0; i<7; i++)
    hist_self.op_count[i] = 0;
}


int Individual :: create_mutate(void)
{
  if (hist_self.how_created == CREATE_MUTATE)
    return TRUE;

  return FALSE;
}

// Returns TRUE if individual has higher fitness than its first parent.
int Individual :: better(void)
{
  if (create_mutate()) {
    if (hist_self.fitness > hist_parent1.fitness)
      return 1;
    else if (hist_self.fitness == hist_parent1.fitness)
      return 0;
    else
      return -1;
  } else {
    if ((hist_self.fitness > hist_parent1.fitness) &&
	(hist_self.fitness > hist_parent2.fitness))
      return 1;
    else if ((hist_self.fitness == hist_parent1.fitness) ||
	     (hist_self.fitness == hist_parent2.fitness))
      return 0;
    else if ((hist_self.fitness > hist_parent1.fitness) ||
	     (hist_self.fitness > hist_parent2.fitness))
      return 0;
    else
      return -1;
  }
}

Real Individual :: get_fitness(void)
{
  return hist_self.fitness;
}

void Individual :: set_fitness(Real val)
{
  hist_self.fitness = val;
}

void Individual :: add_fitness(Real val)
{
  hist_self.fitness += val;
}

void Individual :: div_fitness(Real val)
{
  hist_self.fitness /= val;
}


void Individual :: set_creation(int type)
{
  hist_self.how_created = type;
}

int Individual :: get_creation(void)
{
  return hist_self.how_created;
}


void Individual :: duplicate_hist(history *hist_from, history *hist_to)
{
  int i;

  hist_to->index = hist_from->index;
  hist_to->operation = hist_from->operation;
  hist_to->how_created = hist_from->how_created;
  hist_to->num_mutate = hist_from->num_mutate;
  hist_to->num_recombine = hist_from->num_recombine;
  hist_to->delta = hist_from->delta;
  hist_to->fitness = hist_from->fitness;

  hist_to->op1 = hist_from->op1;
  hist_to->op2 = hist_from->op2;
  hist_to->op3 = hist_from->op3;
  hist_to->val1 = hist_from->val1;
  for (i=0; i<7; i++)
    hist_to->op_count[i] = hist_from->op_count[i];
}



void Individual :: print_fitness(void)
{
  printf("%6.1f", hist_self.fitness);
}



void Individual :: print_history2(history *hist)
{
  if (hist->how_created == CREATE_INIT)
    cout << "[-:";
  else if (hist->how_created == CREATE_MUTATE)
    cout << "[m:";
  else if (hist->how_created == CREATE_RECOMBINE)
    cout << "[r:";
  else 
    cout << "[?:";

  if (hist->operation == OP_WEIGHTS)
    cout << "w:";
  else if (hist->operation == OP_LINKS)
    cout << "l:";
  else if (hist->operation == OP_PROPERTIES)
    cout << "p:";
  else if (hist->operation == OP_INIT)
    cout << "-:";
  else
    cout << "?:";
  printf("%3d:%3d]", hist->num_mutate, hist->num_recombine);
}


void Individual :: print_history(history *hist)
{
  printf("%6.1f ", hist->fitness);
  print_history2(hist);
}

void Individual :: print_history_self(void)
{
  print_history(&hist_self);
}

void Individual :: print_history_self2(void)
{
  print_history2(&hist_self);
}

void Individual :: print_history(void)
{
  printf("[%3d:%3d:%3d:%3d:%3d:%3d]", hist_self.op_count[0],
	 hist_self.op_count[1], hist_self.op_count[2], 
	 hist_self.op_count[3], hist_self.op_count[4], 
	 hist_self.op_count[5]);
}

void Individual :: print_results(void)
{
  if (hist_self.how_created == CREATE_MUTATE) {
    if (hist_self.fitness > hist_parent1.fitness)
      cout << "  Higher!";
    else if (hist_self.fitness == hist_parent1.fitness)
      cout << "  Same.";
  } else if (hist_self.how_created == CREATE_RECOMBINE) {
    if ((hist_self.fitness > hist_parent1.fitness) &&
	(hist_self.fitness > hist_parent2.fitness))
      cout << "  Highest!";
    else if ((hist_self.fitness == hist_parent1.fitness) ||
	     (hist_self.fitness == hist_parent2.fitness))
      cout << "  Same.";
    else if ((hist_self.fitness > hist_parent1.fitness) ||
	     (hist_self.fitness > hist_parent2.fitness))
      cout << "  Middle.";
  }
}


void Individual :: print_history_full(void)
{
  Real diff, val;

  if (evaluated == 2) {
    printf("Evl                        ");
    val = get_fitness();
    diff = val - hist_self.fitness_prev;
    set_fitness(hist_self.fitness_prev);
    print_history(&hist_self);
    set_fitness(val);
    printf(" -> ");
    printf("%6.1f", val + diff);
    printf(" diff: %6.1f ", 2.0*diff);
    print_history();
    return;
  } else if (hist_self.how_created == CREATE_INIT) {
    print_history(&hist_self);
    return;
  } else if (hist_self.how_created == CREATE_MUTATE) {
    cout << "Mut ";
  } else if (hist_self.how_created == CREATE_RECOMBINE) {
    cout << "Rec ";
  }

  //  if (hist_self.operation == OP_INIT)
  //    return;

  if (hist_self.operation == OP_WEIGHTS) {
    printf("(weights %3d) ", hist_self.op1);
  } else if (hist_self.operation == OP_LINKS) {
    if (hist_self.how_created == CREATE_MUTATE) {
      if (hist_self.op1 > 0) {
	printf("(lnk %2d:%5.2lf)", hist_self.op1, hist_self.val1);
      } else {
	if (hist_self.op3) {
	  // deleted
	  printf("(lk %1d:%1d:%5.2lf)", -hist_self.op1, -hist_self.op2, hist_self.val1);
	  //	  printf("(lnk %2d:%5.2lf)", hist_self.op1+hist_self.op2, hist_self.val1);
	} else {
	  printf("(lk'%1d:%1d:%5.2lf)", -hist_self.op1, -hist_self.op2, hist_self.val1);
	  //	  printf("(lnk'%2d:%5.2lf)", hist_self.op1+hist_self.op2, hist_self.val1);
	}
      }
    } else {
      printf("(lnk %d:%2d:%2d) ", hist_self.op1,
	     hist_self.op2, hist_self.op3);
    }
  } else if (hist_self.operation == OP_PROPERTIES) {
    if (hist_self.how_created == CREATE_MUTATE)
      printf("(prop %2d:%2d)  ", hist_self.op1, hist_self.op2);
    else
      printf("(prop    %2d)  ", hist_self.op1);
  }

  if (hist_self.how_created == CREATE_MUTATE) {
    printf("              %6.3f   ", hist_self.delta);
    print_history(&hist_parent1);
  } else {
    print_history(&hist_parent1);
    cout << " + ";
    print_history(&hist_parent2);
  }

  cout << " -> ";
  print_history(&hist_self);
  print_history();
  print_results();
}


void Individual :: info(void)
{
  print_history();
}



void Individual :: make_random(void)
{
  if (!created) {
    cout << "Tree not created." << endl;
    init();
  }

  evaluated = FALSE;
  //  tree->make_random_depth(5, 1);
  tree->make_random_depth(random_int(5)+1, 0);
  if (!tree->verify()) {
    cout << "Make random:: error in verifying tree!" << endl;
    while(1) ;
    exit(-1);
  }
}


void Individual :: duplicate(Individual *ind)
{
  if (!created) {
    num_inputs = ind->num_inputs;
    num_outputs = ind->num_outputs;
    init();
  }

  evaluated = ind->evaluated;
  duplicate_hist(&(ind->hist_self), &hist_self);
  duplicate_hist(&(ind->hist_parent1), &hist_parent1);
  duplicate_hist(&(ind->hist_parent2), &hist_parent2);

  //  cout << "Ind tree::Duplicate() size: " << ind->tree->size1 << endl;
  tree->duplicate(ind->tree, NULL);
  if (!tree->verify()) {
    cout << "Duplicate:: error in verifying." << endl;
    while(1) ;
    exit(-1);
  }

  //  cout << "Ind tree::Duplicate() - finished" << endl;
}



int Individual :: mutate(Real alpha)
{
  int loc1;
  Tree *v_tmp1;

  evaluated = FALSE;

  if (TRUE) {
    if (tree->size1 == 0) {
      return FALSE;
    } else if (tree->size1 == 1) {
      loc1 = 1;
      v_tmp1 = tree;
    } else {
      loc1 = random_int(tree->size1) + 1;
      v_tmp1 = tree->get_tree1(loc1);
    }
  } else {
    if (tree->size2 == 0) {
      return FALSE;
    } else if (tree->size2 == 1) {
      loc1 = 1;
      v_tmp1 = tree;
    } else {
      loc1 = random_int(tree->size2) + 1;
      v_tmp1 = tree->get_tree2(loc1);
    }    

    hist_self.op_count[1]++;
  }
  

  if (random_int(2)) {
    hist_self.op_count[0]++;
    v_tmp1->mutate();
  } else {
    hist_self.op_count[1]++;
    v_tmp1->make_random_depth(random_int(4)+1, 
			      v_tmp1->return_type);
    v_tmp1->resize();
  }

  duplicate_hist(&hist_self, &hist_parent1);
  init_hist(&hist_parent2);
  hist_self.num_mutate++;
  hist_self.how_created = CREATE_MUTATE;

  tree->verify();

  return TRUE;
}


int Individual :: mutate(void)
{
  return mutate(1.0);
}


int Individual :: recombine(Individual *parent2)
{
  int loc1, loc2, i_tmp;
  Tree *v_tmp1, *v_tmp2, *v_tmp3;

  evaluated = FALSE;
  parent2->evaluated = FALSE;

  //  if (random_int(2)) {
  if (TRUE) {
    if (tree->size1 == 0) {
      return FALSE;
    } else if (tree->size1 == 1) {
      loc1 = 1;
      v_tmp1 = tree;
    } else {
      loc1 = random_int(tree->size1) + 1;
      v_tmp1 = tree->get_tree1(loc1);
    }

    if (parent2->tree->size1 == 0) {
      return FALSE;
    } else if (parent2->tree->size1 == 1) {
      loc2 = 1;
      v_tmp2 = parent2->tree;
    } else {
      loc2 = random_int(parent2->tree->size1) + 1;
      v_tmp2 = parent2->tree->get_tree1(loc2);
    }

    hist_self.op_count[3]++;
  } else {
    if (tree->size2 == 0) {
      return FALSE;
    } else if (tree->size2 == 1) {
      loc1 = 1;
      v_tmp1 = tree;
    } else {
      loc1 = random_int(tree->size2) + 1;
      v_tmp1 = tree->get_tree2(loc1);
    }    

    if (parent2->tree->size2 == 0) {
      return FALSE;
    } else if (parent2->tree->size2 == 1) {
      loc2 = 1;
      v_tmp2 = parent2->tree;
    } else {
      loc2 = random_int(parent2->tree->size2) + 1;
      v_tmp2 = parent2->tree->get_tree2(loc2);
    }

    hist_self.op_count[4]++;
  }

  // History for this individual.
  duplicate_hist(&hist_self, &hist_parent1);
  duplicate_hist(&(parent2->hist_self), &hist_parent2);
  hist_self.num_recombine++;
  hist_self.how_created = CREATE_RECOMBINE;

  // History for parent2.
  parent2->duplicate_hist(&(parent2->hist_self), &(parent2->hist_parent1));
  duplicate_hist(&hist_self, &(parent2->hist_parent2));
  parent2->hist_self.num_recombine++;
  parent2->hist_self.how_created = CREATE_RECOMBINE;



  v_tmp1->source = parent2;
  v_tmp2->source = this;
  v_tmp1->re_source();
  v_tmp2->re_source();

  if ((v_tmp1 == tree) && (v_tmp2 != parent2->tree)) {
    v_tmp1->parent = v_tmp2->parent;
    v_tmp1->child_id = v_tmp2->child_id;
    v_tmp1->parent->child[v_tmp1->child_id] = v_tmp1;
    v_tmp1->parent->resize();

    tree = v_tmp2;
    v_tmp2->parent = NULL;
    v_tmp2->child_id = -1;

  } else if ((v_tmp1 != tree) && (v_tmp2 == parent2->tree)) {
    v_tmp2->parent = v_tmp1->parent;
    v_tmp2->child_id = v_tmp1->child_id;
    v_tmp2->parent->child[v_tmp2->child_id] = v_tmp2;
    v_tmp2->parent->resize();

    parent2->tree = v_tmp1;
    v_tmp1->parent = NULL;
    v_tmp1->child_id = -1;

  } else if ((v_tmp1 == tree) && (v_tmp2 == parent2->tree)) {
    tree = v_tmp2;
    parent2->tree = v_tmp1;

  } else {
    v_tmp3 = v_tmp1->parent;
    i_tmp = v_tmp1->child_id;

    v_tmp1->parent = v_tmp2->parent;
    v_tmp1->child_id = v_tmp2->child_id;

    v_tmp2->parent = v_tmp3;
    v_tmp2->child_id = i_tmp;

    v_tmp1->parent->child[v_tmp1->child_id] = v_tmp1;
    v_tmp2->parent->child[v_tmp2->child_id] = v_tmp2;

    v_tmp1->parent->resize();
    v_tmp2->parent->resize();
  }

  if (!tree->verify()) {
    cout << "Recombine:: error in verifying tree!" << endl;
    while(1) ;
  }

  if (!parent2->tree->verify()) {
    cout << "Recombine:: error in verifying tree!" << endl;
    while(1) ;
  }

  return TRUE;
}


int Individual :: read_history(history *hist, FILE *file)
{
  int num;

  num = fscanf(file, "%lf\n", &(hist->fitness));
  if (num != 1) {
    cout << "Read history:: Error reading fitness" << endl;
    exit -1;
  }

  num = fscanf(file, "%d %d %d %d %d\n", &(hist->index),
	       &(hist->how_created), &(hist->operation),
	       &(hist->num_mutate), &(hist->num_recombine));
  if (num != 5) {
    cout << "Read history:: Error reading." << endl;
    exit -1;
  }    
}


int Individual :: read(FILE *file)
{
  int num, version;

  num = fscanf(file, "Version: %d\n", &version);
  if (num != 1) {
    cout << "Individual: error in reading version\n" << endl;
    return FALSE;
  } else if (version != GP_VERSION) {
    cout << "Individual: Using version #%d\n" << version << endl;
    return FALSE;
  }    

  if (!read_history(&hist_self, file)) {
    cout << "Individ: error in history.\n" << endl;
    return FALSE;
  }

  init();
  tree = tree_new();
  tree->source = this;
  tree->read(file);
  tree->re_source();

  return TRUE;
}


void Individual :: write_history(history *hist, FILE *file)
{
  fprintf(file, "%.7lf\n", hist->fitness);
  fprintf(file, "%d %d %d %d %d\n", hist->index,
	  hist->how_created, hist->operation,
	  hist->num_mutate, hist->num_recombine);
}


void Individual :: write(FILE *file)
{
  int i;

  fprintf(file, "Version: %d\n", GP_VERSION);
  write_history(&hist_self, file);
  if (tree != NULL)
    tree->write(file);
  else {
    cout << "Ind_tree::write()  Error - tree is empty." << endl;
    exit -1;
  }
}




int Individual :: set_input(Real *val)
{
  for (int i=0; i<num_inputs; i++)
    input[i] = val[i];
  return TRUE;
}

int Individual :: get_output(Real *val)
{

  return TRUE;
}


Real Individual :: calculate(void)
{
  return tree->calculate();
}


extern Real evaluate_individ(int init, Individual *net);


Real Individual :: evaluate(int init)
{
  Real val;

  if (!tree->verify()) {
    cout << "Evaluate:: error in verifying tree!" << endl;
    while(1) ;
  }

  val = evaluate_individ(FALSE, this);
  if (evaluated) {
    Real val_prev, val;

    val_prev = get_fitness();
    hist_self.fitness_prev = val_prev;
    set_fitness((val + val_prev)/2.0);
    evaluated = 2;
    return val;
  }    

  evaluated = 1;
  set_fitness(val);
  return val;
}

Real Individual :: evaluate(void) {
  return evaluate(FALSE);
}
