/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <iostream.h>
#include <stream.h>
#include "general.h"
#include "neuron.h"
#include "network.h"


void copy_linkindex(LinkIndex *to_li, LinkIndex *from_li)
{
  to_li->from_index = from_li->from_index;
  to_li->from_child = from_li->from_child;
  to_li->from_child_link_index = from_li->from_child_link_index;
  to_li->to_index = from_li->to_index;
}


Network :: Network(void)
{
  init();
}


int Network :: init(void)
{
  Made = FALSE;
  Num_neurons = 0;
  Num_inputs = 0;
  Num_outputs = 0;

  return TRUE;
}


void Network :: reset(void)
{
  int i;

  for (i=0; i<Num_neurons; i++)
    Neuron_list[i]->reset();

  for (i=0; i<Num_neurons; i++) {
    Neuron_value[i] = 0.0;
    Neuron_value_prev[i] = -2.0;
    Neuron_memory[i] = 1000.0;
    Neuron_max[i] = 0.0;
    Neuron_min[i] = 0.0;
  }
}


// Releases all neurons and initializes the network.
void Network :: clear(void)
{
  for (int i=0; i<Num_neurons; i++)
    release_neuron(Neuron_list[i]);
  init();
}


void Network :: next_iteration(void)
{
  int i, j;

  /*
  for (i=0; i<Num_neurons; i++)
    Neuron_list[i]->update();
  */

  Real sum, new_val=0.0;
  for (i=0; i<Num_neurons; i++) {
    Neuron_value_prev[i] = Neuron_value[i]; // Save previous value.
    if (Neuron_type[i] == NEURON_INPUT)
      continue;

    switch (Neuron_function[i])
      {
      case NEURON_OUTPUT:
      case NEURON_SIGMOID:
	sum = 0.0;
	for (j=0; j<Neuron_num_inputs[i]; j++)
	  sum += Neuron_input_weight[i][j] * Neuron_value[Neuron_input_index[i][j]];
	new_val = tanh(sum);
	break;
	
      case NEURON_SIN:
	sum = 0.0;
	for (j=0; j<Neuron_num_inputs[i]; j++)
	  sum += Neuron_input_weight[i][j] * Neuron_value[Neuron_input_index[i][j]];
	new_val = sin(sum);
	break;

      case NEURON_ASIN:
	sum = 0.0;
	for (j=0; j<Neuron_num_inputs[i]; j++)
	  sum += Neuron_input_weight[i][j] * Neuron_value[Neuron_input_index[i][j]];
	new_val = sum;
	if (new_val > 1.0)
	  new_val = 1.0;
	else if (new_val < -1.0)
	  new_val = -1.0;
	new_val = asin(new_val)/PI;
	break;

      case NEURON_SIGN:
	sum = 0.0;
	for (j=0; j<Neuron_num_inputs[i]; j++)
	  sum += Neuron_input_weight[i][j] * Neuron_value[Neuron_input_index[i][j]];
	new_val = sum;
	if (new_val < 0.0)
	  new_val = -1.0;
	else if (new_val > 0.0)
	  new_val = 1.0;
	break;

      case NEURON_SUM:
	sum = 0.0;
	for (j=0; j<Neuron_num_inputs[i]; j++)
	  sum += Neuron_input_weight[i][j] * Neuron_value[Neuron_input_index[i][j]];
	new_val = sum;
	break;

      case NEURON_PROD:
	new_val = 1.0;
	for (j=0; j<Neuron_num_inputs[i]; j++) {
	  new_val *= Neuron_input_weight[i][j] * Neuron_value[Neuron_input_index[i][j]];
	}
	break;

      case NEURON_SAW_WAVE:
	// Actually a triangle wave.
	Neuron_memory[i] += WAVE_STEP;
	sum = 0.0;
	for (j=0; j<Neuron_num_inputs[i]; j++)
	  sum += Neuron_input_weight[i][j] * Neuron_value[Neuron_input_index[i][j]];
	new_val = Neuron_memory[i] + sum;
	int tmp_int = int(new_val);
	tmp_int -= tmp_int % 4;
	new_val -= double(tmp_int);
	if (new_val > 2.0)
	  new_val = 4.0 - new_val;
	new_val -= 1.0;
      }

    if (new_val > MAX_VALUE)
      new_val = MAX_VALUE;
    else if (new_val < -MAX_VALUE)
      new_val = -MAX_VALUE;
    Neuron_value[i] = new_val;
  }

#ifndef OPTIMIZE
  if (Print[35]) {
    for (i=0; i<Num_neurons; i++) {
      printf("%f  ", Neuron_list[i]->Value-Neuron_value[i]);
      //      printf("%f:%f:%f  ", Neuron_list[i]->Value-Neuron_value[i],
      //	     Neuron_list[i]->Value, Neuron_value[i]);
    }
    printf("\n");
  }
#endif

}



void Network :: print_stats(void)
{
  int i;
  for (i=0; i<Num_neurons; i++)
    Neuron_list[i]->print_stats();
  print_state();
}

void Network :: print_state(void)
{
  int i;

  printf("In: %d, Out: %d, # %d\n", Num_inputs,
	 Num_outputs, Num_neurons);

  for (i=0; i<Num_neurons; i++) {
    printf("%d: ", i);
    Neuron_list[i]->print_state();
  }
}

void Network :: print_structure(void)
{
  int i;

  printf("In: %d, Out: %d, # %d\n", Num_inputs,
	 Num_outputs, Num_neurons);

  for (i=0; i<Num_neurons; i++) {
    Neuron_list[i]->print_structure();
  }
}


void Network :: print_neuron_state(int index)
{
  int i;

  printf("%2d: ", index);

  if (Neuron_type[index] == NEURON_INPUT)
    printf("I ");
  else if (Neuron_type[index] == NEURON_OUTPUT)
    printf("O ");
  else if (Neuron_type[index] == NEURON_HIDDEN)
    printf("H ");
  else if (Neuron_type[index] == NEURON_DELETED)
    printf("D ");
  else if (Neuron_type[index] == NEURON_UNUSED)
    printf("U ");
  else
    printf("Invalid neuron type.");


  if (Neuron_function[index] == NEURON_SIGMOID) {
    printf("sig ");
  } else if (Neuron_function[index] == NEURON_SIN) {
    printf("sin ");
  } else if (Neuron_function[index] == NEURON_SUM) {
    printf("sum ");
  } else if (Neuron_function[index] == NEURON_ASIN) {
    printf("asn ");
  } else if (Neuron_function[index] == NEURON_MIN) {
    printf("min ");
  } else if (Neuron_function[index] == NEURON_MAX) {
    printf("max ");
  } else if (Neuron_function[index] == NEURON_SIGN) {
    printf("sgn ");
  } else if (Neuron_function[index] == NEURON_PROD) {
    printf("prd ");
  } else if (Neuron_function[index] == NEURON_SAW_WAVE) {
    printf("saw ");
  } else {
    printf("nof ");
  }

  printf("in:%d : ", Neuron_num_inputs[index]);

  for (i=0; i<Neuron_num_inputs[index]; i++) {
    printf("%d:%.2f  ", Neuron_input_index[index][i], Neuron_input_weight[index][i]);
  }
  printf("\n");
}



void Network :: print_structure2(void)
{
  int i;

  for (i=0; i<Num_neurons; i++) {
    print_neuron_state(i);
  }
}


// Prints the network to a file in DOT format for
// creating images with "dot".
// See: http://www.research.att.com/sw/tools/graphviz/refs.html
void Network :: print_dot_file(FILE *file)
{
  int i, j;

  fprintf(file, "digraph G {\n");

  // Put inputs on top rank.
  fprintf(file, "  { rank=source");
  for (i=0; i<Num_inputs; i++) {
    fprintf(file, "; %d", Neuron_input_list[i]);
  }
  fprintf(file, "};\n");

  // Put outputs on bottom rank.
  fprintf(file, "  { rank=sink");
  for (i=0; i<Num_outputs; i++) {
    fprintf(file, "; %d", Neuron_output_list[i]);
  }
  fprintf(file, "};\n");


  // List rest of nodes.
  for (i=0; i<Num_neurons; i++) {
    if (Neuron_type[i] == NEURON_INPUT) {
      fprintf(file, "  %d [label=\"%c:Input\", style=bold];\n", i, i+'a');
    } else if (Neuron_type[i] == NEURON_OUTPUT) {
      fprintf(file, "  %d [label=\"%c:Output\", style=bold];\n", i, i+'a');
    } else if (Neuron_type[i] == NEURON_HIDDEN) {
      if (Neuron_function[i] == NEURON_SIGMOID) {
	fprintf(file, "  %d [label=\"%c:Sigmoid\"];\n", i, i+'a');
      } else if (Neuron_function[i] == NEURON_SUM) {
	fprintf(file, "  %d [label=\"%c:Sum\"];\n", i, i+'a');
      } else if (Neuron_function[i] == NEURON_SAW_WAVE) {
	fprintf(file, "  %d [label=\"%c:Oscillator\"];\n", i, i+'a');
      } else {
	fprintf(file, "  %d [label=%c];\n", i, i+'a');
      }
    }

    for (j=0; j<Neuron_num_inputs[i]; j++) {
      fprintf(file, "      %d -> %d [label=\"%.2f\"];\n", Neuron_input_index[i][j], i,
	      Neuron_input_weight[i][j]);
    }
  }

  fprintf(file, "}\n");
}



// Returns TRUE if at least one output has changed since previous update.
// Otherwise return FALSE.
int Network :: output_changed(void)
{
  int i, index;
  for (i=0; i<Num_outputs; i++) {
    index = Neuron_output_list[i];
    if (absv(Neuron_value[i] - Neuron_value_prev[i]) > 0.00001)
      return TRUE; // At least one value has changed.
  }

  return FALSE; // No values changed.
}


void Network :: get_output(double *val)
{
  int i;
  Real value;

  /*
  for (i=0; i<Num_outputs; i++) {
    val[i] = Output_neuron[i]->Value;
    if (val[i] > 1.0)
      val[i] = 1.0;
    else if (val[i] < -1.0)
      val[i] = -1.0;
  }
  */

  for (i=0; i<Num_outputs; i++) {
    value = Neuron_value[Neuron_output_list[i]];
    if (value > 1.0)
      value = 1.0;
    else if (value < -1.0)
      value = -1.0;
    val[i] = value;

    if (value < Neuron_min[i])
      Neuron_min[i] = value;
    else if (value > Neuron_max[i])
      Neuron_max[i] = value;
  }
}


void Network :: set_input(double *val)
{
  int i;

  /*
  for (i=0; i<Num_inputs; i++) {
    Input_neuron[i]->Value = val[i];
  }
  */

  for (i=0; i<Num_inputs; i++) {
    Neuron_value[Neuron_input_list[i]] = val[i];
  }
}


void Network :: set_first_value(double val)
{
  Neuron_value[0] = val;
}


void Network :: empty(void)
{
  int i;

  for (i=0; i<Num_neurons; i++)
    Neuron_list[i]->remove();

  for (i=0; i<Num_neurons; i++)
    release_neuron(Neuron_list[i]);

  init();
}


/////////////////////////////////////////
// Network build functions.
/////////////////////////////////////////

int Network :: valid_network(void)
{
  int i;

  for (i=0; i<Num_neurons; i++) {
    if (!Neuron_list[i]->valid_neuron()) {
      return FALSE;
    }
  }
  return TRUE;
}


int Network :: print_link_index(LinkIndex *current_link)
{
  printf("f:%d (c:%d, l#%d) t:%d",
	 current_link->from_index,
	 current_link->from_child,
	 current_link->from_child_link_index,
	 current_link->to_index);
  if ((current_link->to_index >= 0) &&
      (current_link->to_index < Num_neurons)) {
    printf(" (#inputs: %d)", Neuron_list[current_link->to_index]->Num_inputs);
  }
  printf("\n");
  return TRUE;
}


int Network :: valid_link_index(LinkIndex *current_link)
{
  if (current_link == NULL)
    return FALSE;

  if ((current_link->to_index < 0) ||
      (current_link->to_index >= Num_neurons)) {
    printf("Network :: valid_link_index() - False, invalid to_index: %d (%d).\n",
	   current_link->to_index, Num_neurons);
    return FALSE;
  }

  if ((current_link->from_child_link_index < 0) ||
      (current_link->from_child_link_index >= Neuron_list[current_link->from_child]->Num_inputs)) {
    printf("Network :: valid_link_index() - False, invalid from_child_link_index: %d (%d).\n",
	   current_link->from_child_link_index, Neuron_list[current_link->from_child]->Num_inputs);
    return FALSE;
  }

  return TRUE;
}


// This traverses the paths of merged neurons to correctly locate the link.
// Returns TRUE if a change was made, else FALSE.
int Network :: update_link(LinkIndex *current_link)
{
  int link_changed = FALSE;
  Neuron *neuron_from, *neuron_to;

  neuron_from = Neuron_list[current_link->from_index];
  while (1) {
    if (neuron_from->Type == NEURON_DELETED) {
      neuron_from = Neuron_list[neuron_from->Merged_to];
      link_changed = TRUE;
    } else {
      break;
    }
  }

  neuron_to = Neuron_list[current_link->to_index];
  while (1) {
    if (neuron_to->Type == NEURON_DELETED) {
      neuron_to = Neuron_list[neuron_to->Merged_to];
      link_changed = TRUE;
    } else {
      break;
    }
  }

  if (link_changed) {
    current_link->from_index = neuron_from->Id;
    current_link->to_index = neuron_to->Id;
    current_link->from_child_link_index = 0; // Set to something.
    find_existing_link(current_link);
    return TRUE;
  }

  return FALSE;
}


// This tries to find an existing link connecting from_index to to_index
// when the current_link does not already have one.
int Network :: find_existing_link(LinkIndex *current_link)
{
  if (current_link->from_child == current_link->to_index)
    // Already is an existing link.
    return TRUE;

  int link_index = Neuron_list[current_link->to_index]->is_input(Neuron_list[current_link->from_index]);
  if (link_index == -1) {
    // did not find an existing link.
    return FALSE;
  }

  current_link->from_child = current_link->to_index;
  current_link->from_child_link_index = link_index;
  return TRUE;
}


int Network :: first_neuron(Real weight, LinkIndex *current_link)
{
  int result;
  Neuron *new_neuron;

  if (Num_neurons != 0) {
    printf("Network :: first_neuron() - error, already have %d neurons.\n",
	   Num_neurons);
    return FALSE;
  } else if (Made) {
    clear();
  }
  Made = TRUE;


  // Initialize neuron.
  new_neuron = next_free_neuron();
  Neuron_list[0] = new_neuron;
  if (new_neuron == NULL) {
    printf("Network :: first_neuron() - error, invalid new neuron.\n");
    return FALSE;
  }

  new_neuron->init();
  new_neuron->Type = NEURON_HIDDEN;
  new_neuron->Id = Num_neurons;

  result = new_neuron->add_input(weight, new_neuron);

  current_link->from_index = 0;
  current_link->from_child = 0;
  current_link->from_child_link_index = result;
  current_link->to_index = 0;
  Num_neurons = 1;

#ifndef OPTIMIZE
  if (!valid_link_index(current_link)) {
    printf("Error in first_neuron().\n");
    print_link_index(current_link);
  }
  if (!valid_network()) {
    printf("Error, invalid network in first_neuron().\n");
    while (1);
  }
#endif

  return TRUE;
}


int Network :: create_inputs(int num)
{
  int i;
  Neuron *new_neuron;

  if (Num_inputs + num >= MAX_NETWORK_INPUTS) {
    if (Print[9]) {
      printf("network :: create_inputs() - fail because too many inputs: %d+%d\n",
	     num, Num_inputs);
    }
    return FALSE;
  }

  if (Num_neurons+num >= MAX_NEURONS) {
    if (Print[9]) {
      printf("network :: create_inputs() - fail because too many neurons: %d+%d\n",
	     num, Num_neurons);
    }
    return FALSE;
  }

  for (i=0; i<num; i++) {
    // Get neuron and put into network data structure.
    new_neuron = next_free_neuron();
    if (new_neuron == 0) {
      if (Print[9]) {
	printf("network :: create_inputs() - fail because unable to malloc new neuron.\n");
      }
      return FALSE;
    }
    Input_neuron[Num_inputs] = new_neuron;
    Neuron_list[Num_neurons] = new_neuron;

    // Initialize neuron.
    new_neuron->init();
    new_neuron->Type = NEURON_INPUT;
    new_neuron->Id = Num_neurons;

    Num_inputs++;
    Num_neurons++;
  }

  return TRUE;
}


int Network :: add_input(Real weight, LinkIndex *current_link)
{
  int result;
  Neuron *new_neuron;

  if (Num_inputs >= MAX_NETWORK_INPUTS) {
    if (Print[9]) {
      printf("network :: add_input() - fail because too many inputs: %d\n", Num_inputs);
    }
    return FALSE;
  }

  if (Num_neurons >= MAX_NEURONS) {
    if (Print[9]) {
      printf("network :: add_input() - fail because too many neurons: %d\n", Num_neurons);
    }
    return FALSE;
  }

  if (current_link == 0) {
    printf("Error :: current_link == 0\n");
    while (1);
  }

  // Update link.
  update_link(current_link);

  int current_neuron = current_link->to_index;

  // Get neuron and put into network data structure.
  new_neuron = next_free_neuron();
  Input_neuron[Num_inputs] = new_neuron;
  Neuron_list[Num_neurons] = new_neuron;

  // Initialize neuron.
  new_neuron->init();
  new_neuron->Type = NEURON_INPUT;
  new_neuron->Id = Num_neurons;

  // Add connection to current neuron.
  result = Neuron_list[current_neuron]->add_input(weight, new_neuron);
  if (result == -1) {
    release_neuron(new_neuron);
    if (Print[9]) {
      printf("network :: add_input() - fail because error in connect to current neuron: %d\n", result);
    }
    return FALSE;
  }

  current_link->from_index = Num_neurons;
  current_link->from_child = current_neuron;
  current_link->from_child_link_index = result;
  current_link->to_index = current_neuron;

  Num_inputs++;
  Num_neurons++;


#ifndef OPTIMIZE
  if (!valid_link_index(current_link)) {
    printf("Error in add_input().\n");
    print_link_index(current_link);
  }
  if (!valid_network()) {
    printf("Error, invalid network in add_input().\n");
    while (1);
  }
#endif

  return TRUE;
}


int Network :: connect_to_input(Real weight, int index, LinkIndex *current_link)
{
  int result;

  if ((index >= Num_inputs) || (index < 0)) {
    printf("network :: connect_to_input() - fail, invalid index: %d>=%d\n",
	   index, Num_inputs);
    while (1) ;
    return FALSE;
  }

  if (current_link == 0) {
    printf("Error :: current_link == 0\n");
    while (1);
  }


  // Update link.
  update_link(current_link);

  
  int current_neuron = current_link->to_index;
  Neuron *input_neuron = Input_neuron[index];
  if (input_neuron == 0) {
    printf("network :: connect_to_input() - error, invalid input_neuron.\n");
    while (1) ;
  }

  // Add connection to current neuron.
  result = Neuron_list[current_neuron]->add_input(weight, input_neuron);
  if (result == -1) {
    if (Print[9]) {
      printf("network :: connect_to_input() - fail because unable to add input to neuron %d (in:%d)\n",
	     current_link->to_index, Neuron_list[current_neuron]->Num_inputs);
    }
    return FALSE;
  }

  /*
  current_link->from_index = Num_neurons;
  current_link->from_child = current_neuron;
  current_link->from_child_link_index = result;
  current_link->to_index = current_neuron;
  */

#ifndef OPTIMIZE
  if (!valid_link_index(current_link)) {
    printf("Error in connect_to_input().\n");
    print_link_index(current_link);
  }
  if (!valid_network()) {
    printf("Error, invalid network in connect_to_input().\n");
    while (1);
  }
#endif

  return TRUE;
}


int Network :: add_output(Real weight, LinkIndex *current_link)
{
  int result;
  Neuron *new_neuron;

  if (Num_outputs >= MAX_NETWORK_OUTPUTS) {
    if (Print[9]) {
      printf("network :: add_output() - error, exceeded max outputs: %d.\n",
	     MAX_NETWORK_OUTPUTS);
    }
    return FALSE;
  }
  if (Num_neurons >= MAX_NEURONS) {
    if (Print[9]) {
      printf("network :: add_output() - error, exceeded max neurons: %d.\n",
	     MAX_NEURONS);
    }
    return FALSE;
  }

  // Update link.
  update_link(current_link);


  // Get neuron and put into network data structure.
  new_neuron = next_free_neuron();
  Output_neuron[Num_outputs] = new_neuron;
  Neuron_list[Num_neurons] = new_neuron;

  // Initialize neuron.
  new_neuron->init();
  new_neuron->Type = NEURON_OUTPUT;
  new_neuron->Id = Num_neurons;
  new_neuron->Function = NEURON_SUM;

  // Add connection to current neuron.
  int current_neuron = current_link->to_index;
  result = new_neuron->add_input(weight, Neuron_list[current_neuron]);
  if (result == -1) {
    release_neuron(new_neuron);
    return FALSE;
  }


  /*
    // don't change current edge.
  current_link->from_index = 0;
  current_link->from_child = 0;
  current_link->from_child_link_index = result;
  current_link->to_index = Num_neurons;
  */

  Num_outputs++;
  Num_neurons++;


#ifndef OPTIMIZE
  if (!valid_link_index(current_link)) {
    printf("Error in add_output().\n");
    print_link_index(current_link);
  }
  if (!valid_network()) {
    printf("Error, invalid network in add_output().\n");
    while (1);
  }
#endif

  return TRUE;
}
	       

int Network :: cut_link(LinkIndex *current_link)
{
  // Update link.
  update_link(current_link);

  if (current_link->from_child != current_link->to_index)
    return TRUE; // No link to cut.

  int neuron_index = current_link->to_index;

  if ((neuron_index < 0) || (neuron_index >= Num_neurons))
    return FALSE;

  Neuron_list[current_link->to_index]->Input[current_link->from_child_link_index].weight = 0.0;
  find_existing_link(current_link);

  return TRUE;
}


int Network :: duplicate_link(LinkIndex *current_link)
{
  // Update link.
  update_link(current_link);

  int result;
  int neuron_index = current_link->to_index;

  if ((neuron_index < 0) || (neuron_index >= Num_neurons))
    return FALSE;

  Real weight = Neuron_list[current_link->from_child]->Input[current_link->from_child_link_index].weight;
  result = Neuron_list[neuron_index]->add_input(weight,
						Neuron_list[current_link->from_index]);
  if (result == -1)
    return FALSE;
  
  current_link->from_child = current_link->to_index;
  current_link->from_child_link_index = result;


#ifndef OPTIMIZE
  if (!valid_link_index(current_link)) {
    printf("Error in duplicate_link(1).\n");
    print_link_index(current_link);
  }
  if (!valid_network()) {
    printf("Error, invalid network in duplicate_link(1).\n");
    while (1);
  }
#endif

  return TRUE;
}

int Network :: duplicate_link(Real weight, LinkIndex *current_link)
{
  // Update link.
  update_link(current_link);

  int result;
  int neuron_index = current_link->to_index;

  if ((neuron_index < 0) || (neuron_index >= Num_neurons))
    return FALSE;

  result = Neuron_list[neuron_index]->add_input(weight,
						Neuron_list[current_link->from_index]);
  if (result == -1)
    return FALSE;
  
  current_link->from_child = current_link->to_index;
  current_link->from_child_link_index = result;


#ifndef OPTIMIZE
  if (!valid_link_index(current_link)) {
    printf("Error in duplicate_link(2).\n");
    print_link_index(current_link);
  }
  if (!valid_network()) {
    printf("Error, invalid network in duplicate_link(2).\n");
    while (1);
  }
#endif

  return TRUE;
}


int Network :: adjust_weight(Real value, LinkIndex *current_link)
{
  // Update link.
  update_link(current_link);

  if (current_link->from_child == current_link->to_index) {
    // Adjust weight.
    Neuron_list[current_link->to_index]->Input[current_link->from_child_link_index].weight += value;
    return TRUE;
  }

  // Create link.
  int result = Neuron_list[current_link->to_index]->add_input(value,
							      Neuron_list[current_link->from_index]);
  if (result == -1)
    return FALSE;

  current_link->from_child = current_link->to_index;
  current_link->from_child_link_index = result;

  return TRUE;
}


int Network :: set_input_function(int function, Real value, LinkIndex *current_link)
{
  // Update link.
  update_link(current_link);

  Neuron *other_neuron = Neuron_list[current_link->from_index];

  if (other_neuron->Type == NEURON_INPUT) {
    return FALSE;
  }
  
  other_neuron->Function = function;
  if (function == NEURON_SAW_WAVE) {
    other_neuron->Speed = value;
  }

  return TRUE;
}


// Makes link negative.  If link does not exist,
// makes a link of negative weight.
int Network :: negate_weight(Real weight, LinkIndex *current_link)
{
  // Update link.
  update_link(current_link);

  int result;

  if (current_link->from_child != current_link->to_index) {
    // No link -> make negative.

    result = Neuron_list[current_link->to_index]->add_input(-weight,
						  Neuron_list[current_link->from_index]);
    if (result == -1)
      return FALSE;
  
    current_link->from_child = current_link->to_index;
    current_link->from_child_link_index = result;

  } else {
    Neuron_list[current_link->to_index]->Input[current_link->from_child_link_index].weight =
      -Neuron_list[current_link->to_index]->Input[current_link->from_child_link_index].weight;
  }

  return TRUE;
}

int Network :: set_output_function(int num, LinkIndex *current_link)
{
  // Update link.
  update_link(current_link);

  Neuron *current_neuron = Neuron_list[current_link->to_index];

  if ((num < 0) || (num >= NEURON_MAX_OUTPUT_FUNC))
    return FALSE;

  if (current_neuron->set_function(num))
    return TRUE;

  return FALSE;
}



// Split the current link: from -w1-> to
// to: from -w_new-> new -w1-> to
// current link becomes: from --> new
int Network :: split_link2(Real weight, LinkIndex *current_link)
{
  // Update link.
  update_link(current_link);

  int result, link_index;
  Neuron *new_neuron, *current_neuron;

  // Test that link exists.
  if (current_link->to_index != current_link->from_child)
    return FALSE;

  if (Num_neurons >= MAX_NEURONS)
    return FALSE;

  new_neuron = next_free_neuron();
  Neuron_list[Num_neurons] = new_neuron;

  // Initialize neuron.
  new_neuron->init();
  new_neuron->Type = NEURON_HIDDEN;
  new_neuron->Id = Num_neurons;


  link_index = current_link->from_child_link_index;
  current_neuron = Neuron_list[current_link->to_index];
  result = new_neuron->add_input(weight, current_neuron->Input[link_index].input);
  if (result == -1) {
    printf("Network :: split_link() - error in result: %d\n", result);
    return FALSE;
  }

  current_neuron->Input[link_index].input = new_neuron;

  current_link->from_child = Num_neurons;
  current_link->from_child_link_index = result;
  current_link->to_index = Num_neurons;

  Num_neurons++;

#ifndef OPTIMIZE
  if (!valid_link_index(current_link)) {
    printf("Error in split_link().\n");
    print_link_index(current_link);
  }
  if (!valid_network()) {
    printf("Error, invalid network in split_link().\n");
    while (1);
  }
#endif

  return TRUE;
}



// Split the current link: from -w1-> to
// to: from -w_new-> new -w1-> to
// current link becomes: new --> to
int Network :: split_link(Real weight, LinkIndex *current_link)
{
  // Update link.
  update_link(current_link);

  int result, link_index;
  Neuron *new_neuron, *current_neuron;

  // Test that link exists.
  if (current_link->to_index != current_link->from_child)
    return FALSE;

  if (Num_neurons >= MAX_NEURONS)
    return FALSE;

  new_neuron = next_free_neuron();
  Neuron_list[Num_neurons] = new_neuron;

  // Initialize neuron.
  new_neuron->init();
  new_neuron->Type = NEURON_HIDDEN;
  new_neuron->Id = Num_neurons;


  link_index = current_link->from_child_link_index;
  current_neuron = Neuron_list[current_link->to_index];
  result = new_neuron->add_input(weight, current_neuron->Input[link_index].input);
  if (result == -1) {
    printf("Network :: split_link() - error in result: %d\n", result);
    return FALSE;
  }

  current_neuron->Input[link_index].input = new_neuron;
  current_link->from_index = Num_neurons;

  Num_neurons++;

#ifndef OPTIMIZE
  if (!valid_link_index(current_link)) {
    printf("Error in split_link().\n");
    print_link_index(current_link);
  }
  if (!valid_network()) {
    printf("Error, invalid network in split_link().\n");
    while (1);
  }
#endif

  return TRUE;
}


// Makes a loop from the current to_index to itself.
int Network :: make_loop(Real weight, LinkIndex *current_link)
{
  // Update link.
  update_link(current_link);

  int result;
  Neuron *current_neuron;

  current_neuron = Neuron_list[current_link->to_index];
  result = current_neuron->add_input(weight, current_neuron);
  if (result == -1) {
    return FALSE;
  }
  
  current_link->from_index = current_link->to_index;
  current_link->from_child = current_link->to_index;
  current_link->from_child_link_index = result;

#ifndef OPTIMIZE
  if (!valid_link_index(current_link)) {
    printf("Error in make_loop().\n");
    print_link_index(current_link);
  }
  if (!valid_network()) {
    printf("Error, invalid network in make_loop().\n");
    while (1);
  }
#endif

  return TRUE;
}

// num is how many sibling links to count.
// Should not count dead links.
int Network :: next_link(int num, LinkIndex *current_link)
{
  // Update link.
  update_link(current_link);

  int i, index;
  Neuron *child = Neuron_list[current_link->from_child];

  //  if (current_neuron->Num_inputs < 2)
  //    return TRUE;

  // Test that there are valid inputs.
  if (!child->has_live_inputs())
    return FALSE;

  //  current_link->to_index = current_link->from_child;

  index = current_link->from_child_link_index;
  for (i=0; i<num; i++) {
    index++;
    if (index == child->Num_inputs)
      index = 0;
    while (child->dead_link(index)) {
      index++;
      if (index == child->Num_inputs)
	index = 0;
    }
  }

  current_link->from_index = child->Input[index].input->Id;
  current_link->from_child_link_index = index;

#ifndef OPTIMIZE
  if (!valid_link_index(current_link)) {
    printf("Error in next_link().\n");
    print_link_index(current_link);
  }
  if (!valid_network()) {
    printf("Error, invalid network in next_link().\n");
    while (1);
  }
#endif

  return TRUE;
}


// num is which input into /from_index/ to be current.
// If from_index is an input, this changes to a different child of /from_child/.
//  this also occurs if from_index has no inputs.
int Network :: parent_link(int num, LinkIndex *current_link)
{
  // Update link.
  update_link(current_link);

  int i, index;
  Neuron *parent;

  parent = Neuron_list[current_link->from_index];
  if ((parent->Type == NEURON_INPUT) || (!parent->has_live_inputs())) {
    return next_link(num, current_link);
  }

  index = 0;
  while (parent->dead_link(index)) {
    index++;
    if (index >= parent->Num_inputs) {
      printf("Network :: parent_link() - error finding living link.\n");
      while (1) ;
    }
  }

  for (i=0; i<num; i++) {
    index++;
    if (index == parent->Num_inputs)
      index = 0;
    while (parent->dead_link(index)) {
      index++;
      if (index == parent->Num_inputs)
	index = 0;
    }
  }

  current_link->from_child = current_link->from_index;
  current_link->from_index = parent->Input[index].input->Id;
  current_link->from_child_link_index = index;
  find_existing_link(current_link);

#ifndef OPTIMIZE
  if (!valid_link_index(current_link)) {
    printf("Error in parent_link().\n");
    print_link_index(current_link);
    while (1);
  }
  if (!valid_network()) {
    printf("Error, invalid network in parent_link().\n");
    while (1);
  }
#endif

  return TRUE;
}


// Delete current link and re-make connectin to_index to from_index.
int Network :: reverse_link(LinkIndex *current_link)
{
  if (Neuron_list[current_link->from_index]->Type == NEURON_INPUT) {
    // Fail because cannot reverse link into an input neuron.
    return FALSE;
  }

  // Update link.
  update_link(current_link);

  int result;
  Real weight = Neuron_list[current_link->from_child]->Input[current_link->from_child_link_index].weight;
  Neuron *new_from = Neuron_list[current_link->to_index];
  Neuron *new_to = Neuron_list[current_link->from_index];

  // cut link.
  cut_link(current_link);

  // Add reverse.
  result = new_to->add_input(weight, new_from);
  if (result == -1)
    return FALSE;

  current_link->from_index = new_from->Id;
  current_link->from_child = new_to->Id;
  current_link->from_child_link_index = result;
  current_link->to_index = new_to->Id;

  
#ifndef OPTIMIZE
  if (!valid_link_index(current_link)) {
    printf("Error in reverse_link().\n");
    print_link_index(current_link);
  }
  if (!valid_network()) {
    printf("Error, invalid network in reverse_link().\n");
    while (1);
  }
#endif

  return TRUE;
}



// Merges the input and output neurons by adding all the
// inputs fron neuronA to neuronB, then deleting neuronA.
// At least one neuron must be hidden, or 
// neuronA can be deleted and neuronB not deleted.
int Network :: merge_neurons(LinkIndex *current_link)
{
  // Update link.
  update_link(current_link);

  int i;
  int child_link_index;

  if (current_link->from_index == current_link->to_index) {
    // Trying to merge same neuron to itself.
    return cut_link(current_link);
  }

  Neuron *neuronA = Neuron_list[current_link->from_index];
  Neuron *neuronB = Neuron_list[current_link->to_index];

  if (neuronB->Type == NEURON_DELETED)
    // Cannot merge into a deleted neuron.
    return FALSE;

  else if (neuronA->Type == NEURON_INPUT)
    // Cannot merge an input into anything.
    return FALSE;

  else if (neuronB->Type != NEURON_HIDDEN) {
    if ((neuronA->Type == NEURON_INPUT) || (neuronA->Type == NEURON_OUTPUT))
      // Cannot merge input or output into a non-deleted, non-hidden neuron.
      return FALSE;
  }

  //  printf("Merge Neurons:\n");
  //  print_link_index(current_link);


  // Copy inputs from neuronA to neuronB.
  child_link_index = neuronB->Num_inputs;
  for (i=0; i<neuronA->Num_inputs; i++) {
    if (!neuronA->dead_link(i)) {
      if (!neuronB->add_input(neuronA->Input[i].weight,
			      neuronA->Input[i].input)) {
	// Add input failed.
	return FALSE;
      }
    }
  }

  if (neuronA != neuronB) {
    // Replace all neurons who input from neuronA to input from neuronB.
    for (i=0; i<Num_neurons; i++) {
      if (i == current_link->from_index) {
	continue; // Don't change these links, need for history purposes.
      } else if (i == current_link->to_index) {
	Neuron_list[i]->delete_inputs_from(neuronA);
      } else {
	Neuron_list[i]->replace_input_neuron(neuronA, neuronB);
      }
    }

    // Delete neuronA.
    if ((neuronA->Type == NEURON_INPUT) ||
	(neuronA->Type == NEURON_OUTPUT)) {
      neuronB->Type = neuronA->Type;
      if (neuronA->Same_as == -1)
	//      if (neuronA->Id == -1)
	neuronB->Same_as = neuronA->Id;
      else
	neuronB->Same_as = neuronA->Same_as;
    }

    neuronA->Type = NEURON_DELETED;
    neuronA->Merged_to = neuronB->Id;
  }

  
  // Set current edge to be link from first neuron into neuronA to neuronB.
  if (neuronA->Num_inputs > 0) {
    // Use neuronA's first input as from neuron.
    current_link->from_index = neuronA->Input[0].input->Id;
    current_link->from_child = neuronB->Id;

    if (child_link_index == neuronB->Num_inputs)
      current_link->from_child_link_index = child_link_index-1;
    else
      current_link->from_child_link_index = child_link_index;

  } else if (neuronB->Num_inputs > 0) {
    // Use neuronB's first input as a from neuron.
    current_link->from_index = neuronB->Input[0].input->Id;
    current_link->from_child = neuronB->Id;
    current_link->from_child_link_index = 0;

  } else {
    // Use first neuron as from neuron.
    current_link->from_index = 0;
    current_link->from_child = neuronB->Id;
    current_link->from_child_link_index = 0;
  }
  find_existing_link(current_link);


#ifndef OPTIMIZE
  if (!valid_link_index(current_link)) {
    printf("Error in merge_neurons().\n");
    print_link_index(current_link);
  }
  if (!valid_network()) {
    printf("Error, invalid network in merge_neurons().\n");
    while (1);
  }
#endif  
  return TRUE;
}


// Set the specified output neuron as deleted.
int Network :: delete_output(int index)
{
  if ((index < 0) || (index >= Num_outputs)) {
    // No such output neuron.
    return FALSE;
  }

  Output_neuron[index]->Type = NEURON_DELETED;
  return TRUE;
}


// Remakes network from objects and points to a faster data structure.
void Network :: rebuild_network(void)
{
  int i, j, k, num_inputs;
  Real sum;
  Real input_bk[MAX_INPUTS];
  Neuron *neuron, *input;

  // Set input list.
  for (i=0; i<Num_inputs; i++) {
    Neuron_input_list[i] = Input_neuron[i]->Id;
  }

  // Set output list.
  for (i=0; i<Num_outputs; i++) {
    Neuron_output_list[i] = Output_neuron[i]->Id;
  }


  // Build fast data structure.
  for (i=0; i<Num_neurons; i++) {
    neuron = Neuron_list[i];

    Neuron_function[i] = neuron->Function;
    Neuron_type[i] = neuron->Type;
    num_inputs = 0;

    // Save the input weights.
    for (j=0; j<neuron->Num_inputs; j++) {
      input_bk[j] = neuron->Input[j].weight;
    }

    for (j=0; j<neuron->Num_inputs; j++) {
      if (absv(neuron->Input[j].weight) <  0.0001)
	continue;

      // Find total sum of weights from this input.
      sum = neuron->Input[j].weight;
      input = neuron->Input[j].input;
      for (k=j+1; k<neuron->Num_inputs; k++) {
	if (neuron->Input[k].input == input) {
	  sum += neuron->Input[k].weight;
	  neuron->Input[k].weight = 0.0;
	}
      }

      if (sum != 0.0) {
	// Add input.
	Neuron_input_index[i][num_inputs] = input->Id;
	Neuron_input_weight[i][num_inputs] = sum;
	num_inputs++;
      }
    }
    Neuron_num_inputs[i] = num_inputs;

    // Restore the input weights.
    for (j=0; j<neuron->Num_inputs; j++) {
      neuron->Input[j].weight = input_bk[j];
    }
  }

  for (i=0; i<Num_neurons; i++) {
    Neuron_value[i] = 0.0;
    Neuron_memory[i] = 1000.0;
  }
}


// Remove dead links and neurons.
void Network :: finish_build(void)
{
  int i, j, index;
  Neuron *tmp;

  // Remove dead links.
  for (i=0; i<Num_neurons; i++) {
    Neuron_list[i]->remove_dead_links();
  }

  // Swap neurons with their /same_as/.
  for (i=0; i<Num_neurons; i++) {
    if (Neuron_list[i]->Same_as >= 0) {
      index = Neuron_list[i]->Same_as;

      // Test that this is a unique swap.
      if (Neuron_list[index]->Same_as != -1)
	printf("Same_as error!\n");
      Neuron_list[index]->Same_as = -2; // Show it has already swapped.

      // Swap!
      tmp = Neuron_list[i];
      Neuron_list[i] = Neuron_list[index];
      Neuron_list[index] = tmp;
    }
  }

  // Remove dead neurons from output list.
  i=0;
  while (i<Num_outputs) {
    if (Output_neuron[i]->Type == NEURON_DELETED) {
      for (j=i; j<Num_outputs-1; j++) {
	Output_neuron[j] = Output_neuron[j+1];
      }
      Num_outputs--;
    } else {
      i++;
    }
  }

  // Remove dead neurons from neuron list.
  i = 0;
  while (i<Num_neurons) {
    if (Neuron_list[i]->Type == NEURON_DELETED) {
      release_neuron(Neuron_list[i]);
      Num_neurons--;
      for (j=i; j<Num_neurons; j++)
	Neuron_list[j] = Neuron_list[j+1];
    } else {
      i++;
    }
  }

  // Relabel neuron indexes.
  for (i=0; i<Num_neurons; i++)
    Neuron_list[i]->Id = i;

  // Reset values.
  for (i=0; i<Num_neurons; i++)
    Neuron_list[i]->Value = 0.0;

  rebuild_network();
}


// Returns the average number of inputs each hidden unit has.
Real Network :: average_num_inputs(void)
{
  Real num_hidden = 0.0;
  Real sum_inputs = 0.0;

  for (int i=0; i<Num_neurons; i++) {
    if (Neuron_type[i] == NEURON_HIDDEN) {
      num_hidden += 1.0;
      sum_inputs += Neuron_num_inputs[i];
      //      printf("%d: %d (%d:%d)\n", i, Neuron_num_inputs[i], 
      //	     int(num_hidden), int(sum_inputs));
    }
  }
  if (num_hidden == 0.0)
    return 1.0;
  return sum_inputs / num_hidden;
}


// Returns the number of input units in the network.
int Network :: num_used_inputs(void)
{
  int i, j, index, num = 0;
  int input_used[Num_neurons];

  for (i=0; i<Num_inputs; i++) {
    input_used[Neuron_input_list[i]] = 0;
  }

  for (i=0; i<Num_neurons; i++) {
    if (Neuron_type[i] == NEURON_HIDDEN) {
      for (j=0; j<Neuron_num_inputs[i]; j++) {
	index = Neuron_input_index[i][j];
	if (Neuron_type[index] == NEURON_INPUT) {
	  input_used[index]++;
	}
      }
    }
  }

  for (i=0; i<Num_inputs; i++) {
    if (input_used[Neuron_input_list[i]] > 0) {
      num++;
    }
  }

  return num;
}

Real Network :: average_output_range(void)
{
  int i;
  Real range = 0.0;

  for (i=0; i<Num_outputs; i++) {
    range += Neuron_max[i] - Neuron_min[i];
  }

  return range / Real(Num_outputs);
}


// Returns number of different hidden nodes used as inputs for the outputs,
// only for those hidden units whose output values has changed.
int Network :: num_active_outputs(void)
{
  int i, j, index;
  int num_sources = 0;
  int sources[MAX_NETWORK_INPUTS];

  for (i=0; i<Num_outputs; i++) {    
    if (Neuron_max[i] > Neuron_min[i]) {
      index = Neuron_output_list[i];
      for (j=0; j<num_sources; j++) {
	if (Neuron_input_index[index][0] == sources[j])
	  break;
      }
      if (j == num_sources) {
	// New source!
	sources[num_sources] = Neuron_input_index[index][0];
	num_sources++;
      }
    }
  }

  return num_sources;
}


// Returns number of different hidden nodes used as inputs for the outputs.
int Network :: num_unique_output_sources(void)
{
  int i, j;
  int num_sources = 0;
  int sources[MAX_NETWORK_INPUTS];

  for (i=0; i<Num_neurons; i++) {
    if (Neuron_type[i] == NEURON_OUTPUT) {
      for (j=0; j<num_sources; j++) {
	if (Neuron_input_index[i][0] == sources[j])
	  break;
      }
      if (j == num_sources) {
	// New source!
	sources[num_sources] = Neuron_input_index[i][0];
	num_sources++;
      }
    }
  }

  return num_sources;
}


// Returns total number of links in the network.
int Network :: num_links(void)
{
  int sum = 0;
  for (int i=0; i<Num_neurons; i++) {
    sum += Neuron_num_inputs[i];
  }
  return sum;
}


/***** Not needed functions *****/
void sort(int *sort, int max)
{
  int i, j, tmp_I;

  for (i=max-1; i>0; i--)
    for (j=0; j<i; j++)
      {
	if (sort[j] > sort[j+1])
	  {
	    tmp_I = sort[j];
	    sort[j] = sort[j+1];
	    sort[j+1] = tmp_I;
	  }
      }
}


void Network :: duplicate(Network *net)
{
}




/***** I/O *****/


void Network :: read(ifstream *infile)
{
}


int Network :: read(FILE *file)
{
  return TRUE;
}


void Network :: write(ofstream *outfile)
{
}


void Network :: write(FILE *file)
{
}
