/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <iostream.h>
#include <stream.h>
#include "general.h"
#include "network.h"
#include "neuron.h"

#define MAX_LINKS 20
#define OFFSET_DELTA 0.05


/***** Neuron List Maintenance Functions *****/

#define NUM_MAKE 260

int neurons_avail = 0;
Neuron *first_free_neuron = NULL;

void release_neuron(Neuron *n)
{
  n->Type = NEURON_UNUSED;
  n->next_neuron = first_free_neuron;
  first_free_neuron = n;
  neurons_avail++;
}


void add_neurons(void)
{
  int i;
  Neuron *neuron_tmp;

  if (neurons_avail != 0)
    return;

  for (i=0; i<NUM_MAKE; i++) {
    neuron_tmp = new Neuron();
    release_neuron(neuron_tmp);
  }
}


Neuron *next_free_neuron(void)
{
  Neuron *next_neuron;

  if (neurons_avail == 0)
    add_neurons();

  if (first_free_neuron == NULL) {
    printf("next_free_neuron :: error - could not allocate.\n");
    return 0;
  }

  next_neuron = first_free_neuron;

  first_free_neuron = next_neuron->next_neuron;
  neurons_avail--;
  return next_neuron;
}



//////////////////////////////////////////////////////////////////
//           Neuron Class
//////////////////////////////////////////////////////////////////


Neuron :: Neuron(void)
{
  init();
  network = 0;
  Speed = 1.0;
  Memory = 1000.0; // Must start large enough so that adding sum_inputs() does not bring below 0.
}


Neuron :: Neuron(Network *n)
{
  init();
  network = n;
}


void Neuron :: init(void)
{
  Same_as = -1;
  Type = NEURON_HIDDEN;
  Function = NEURON_SIGMOID;
  Num_inputs = 0;
  Value = 0.0;
  Scale = 0.0;
  Offset = 0.0;
  Time_const = 0.0;
  prev_neuron = NULL;
  next_neuron = NULL;
}

void Neuron :: reset(void)
{
  Value = 0.0;
}


void Neuron :: print_stats(void)
{
  cout << "in:"<<Num_inputs<<"\n";
}

void Neuron :: print_stats(char * string)
{
  cout << string;
  print_stats();
}

void Neuron :: print_state(void)
{
  printf("%2d(%2d): ", Id, Same_as);

  if (Type == NEURON_INPUT)
    printf("I ");
  else if (Type == NEURON_OUTPUT)
    printf("O ");
  else if (Type == NEURON_HIDDEN)
    printf("H ");
  else if (Type == NEURON_DELETED)
    printf("D ");
  else if (Type == NEURON_UNUSED)
    printf("U ");
  else
    printf("Inv");

  if (Function == NEURON_SIGMOID) {
    printf("sig ");
  } else if (Function == NEURON_SIN) {
    printf("sin ");
  } else if (Function == NEURON_SUM) {
    printf("sum ");
  } else if (Function == NEURON_ASIN) {
    printf("asn ");
  } else if (Function == NEURON_MIN) {
    printf("min ");
  } else if (Function == NEURON_MAX) {
    printf("max ");
  } else if (Function == NEURON_SIGN) {
    printf("sgn ");
  } else if (Function == NEURON_PROD) {
    printf("prd ");
  } else if (Function == NEURON_SAW_WAVE) {
    printf("saw ");
  } else {
    printf("nof ");
  }

  printf("in:%d : ", Num_inputs);
}


void Neuron :: print_structure(void)
{
  int i;

  print_state();

  for (i=0; i<Num_inputs; i++) {
    printf("%d:%.2f  ", Input[i].input->Id, Input[i].weight);
  }
  printf("\n");
}


int Neuron :: valid_neuron(void)
{
  int i;

  if ((Type != NEURON_HIDDEN) && (Type != NEURON_INPUT) &&
      (Type != NEURON_OUTPUT) && (Type != NEURON_DELETED)) {
    printf("Neuron :: valid_neuron() - error.\n");
    while (1) ;
    return FALSE;
  }

  for (i=0; i<Num_inputs; i++) {
    if (Input[i].input == 0) {
      printf("Neuron %d, link #%d :: valid_neuron() - null input neuron.\n", Id, i);
      print_state();
      return FALSE;
    }
  }
  return TRUE;
}


// If in_neuron is an input to this neuron, returns the last link index of it,
// otherwise returns -1.
int Neuron :: is_input(Neuron *in_neuron)
{
  for (int i=Num_inputs-1; i>=0; i--) {
    if (Input[i].input == in_neuron)
      return i;
  }
  return -1;
}


int Neuron :: add_input(Real weight, Neuron *in_neuron)
{
  if (Num_inputs >= MAX_INPUTS) {
    if (Print[9]) {
      printf("neuron :: add_input() - fail because %d has too many inputs: %d.\n", Id, Num_inputs);
    }
    return -1;
  }

  if (Type == NEURON_INPUT) {
    if (Print[9]) {
      printf("neuron :: add_input() - fail because %d type is input: %d.\n", Id, Type);
    }
    return -1;
  }


  if (in_neuron == 0) {
    printf("neuron :: add_input() - error, in_neuron = NULL\n");
    while (1) ;
  }

  int index = Num_inputs;
  Input[Num_inputs].weight = weight;
  Input[Num_inputs].input = in_neuron;
  Num_inputs++;
  return index;
}

int Neuron :: duplicate_input(int index)
{
  return add_input(Input[index].weight,
		   Input[index].input);
}


void Neuron :: copy_link(int index_to, int index_from)
{
  Input[index_to].weight = Input[index_from].weight;
  Input[index_to].input = Input[index_from].input;
}


void Neuron :: cut_input(int index)
{
  if ((index < 0) || (index >= Num_inputs))
    return;
  
  Input[index].weight = 0.0;
}
  

void Neuron :: remove_input(int index)
{
  int i;
  Num_inputs--;
  for (i=index; i<Num_inputs; i++)
    copy_link(i, i+1);
}
  

int Neuron :: dead_link(int index)
{
  if ((index < 0) || (index >= Num_inputs)) {
    printf("neuron :: dead_link() - error, invalid index %d\n", index);
    while (1) ;
    return FALSE;
  }    

  if (Input[index].weight == 0.0)
    return TRUE;
  else if (Input[index].input == NULL)
    return TRUE;
  else if (Input[index].input->Type == NEURON_DELETED)
    return TRUE;
  return FALSE;
}


void Neuron :: remove_dead_links(void)
{
  int l=0;

  while (l<Num_inputs) {
    if (dead_link(l)) {
      remove_input(l);
    } else {
      l++;
    }
  }
}


// Returns TRUE if this neuron has at least one live input.
int Neuron :: has_live_inputs(void)
{
  int i;

  if (Num_inputs == 0)
    return FALSE;

  for (i=0; i<Num_inputs; i++) {
    if (!dead_link(i))
      return TRUE;
  }
  return FALSE;
}


int Neuron :: set_function(int function)
{
  if ((function < 0) || (function >= NEURON_MAX_OUTPUT_FUNC))
    return FALSE;

  Function = function;
  if (function == NEURON_SAW_WAVE)
    Speed = 1.0;

  return TRUE;
}



void Neuron :: remove()
{
}


double Neuron :: sum_inputs(void)
{
  if (Type == NEURON_INPUT)
    return Value;

  //  double sum = Offset;
  double sum = 0.0;
  for (int i=0; i<Num_inputs; i++) {
    sum += Input[i].input->Value * Input[i].weight;
  }
  return sum;
}


void Neuron :: update(void)
{
  if (Type == NEURON_INPUT)
    return;

  int i;
  double new_val=0.0, tmp;

  /*
  cout << "Type: " << Function << "  ";
  cout << "Offset: " << offset << "  ";
  cout << "Scale: " << scale << "  ";
  cout << "Sum inputs: " << sum_inputs() << "  ";
  */

  switch (Function)
    {
    case NEURON_OUTPUT:
    case NEURON_SIGMOID:
      //      new_val = 1.0/(1.0 + exp(-sum_inputs()));
      new_val = tanh(sum_inputs());
      break;
      
    case NEURON_SIN:
      new_val = sin(sum_inputs());
      break;

    case NEURON_ASIN:
      new_val = sum_inputs();
      if (new_val > 1.0)
	new_val = 1.0;
      else if (new_val < -1.0)
	new_val = -1.0;
      new_val = asin(new_val)/PI;
      break;

    case NEURON_MIN:
      new_val = Offset/Scale;
      for (i=0; i<Num_inputs; i++) {
	tmp = Input[i].input->Value * Input[i].weight;
	if (tmp < new_val)
	  new_val = tmp;
      }
      break;

    case NEURON_MAX:
      new_val = Offset;
      for (i=0; i<Num_inputs; i++) {
	tmp = Input[i].input->Value * Input[i].weight;
	if (tmp > new_val)
	  new_val = tmp;
      }
      break;

    case NEURON_SIGN:
      new_val = sum_inputs();
      if (new_val < 0.0)
	new_val = -1.0;
      else if (new_val > 0.0)
	new_val = 1.0;
      break;

    case NEURON_SUM:
      new_val = sum_inputs();
      break;

    case NEURON_PROD:
      for (i=0; i<Num_inputs; i++) {
	new_val *= Input[i].input->Value * Input[i].weight;
      }
      new_val += Offset;
      break;

    case NEURON_SAW_WAVE:
      Memory += WAVE_STEP * Speed; // * speed.
      new_val = Memory + sum_inputs();
      {
      int tmp_int = int(new_val);
      tmp_int -= tmp_int % 4;
      new_val -= double(tmp_int);
      }
      if (new_val > 2.0)
	new_val = 4.0 - new_val;
      new_val -= 1.0;
      break;
    }

  //  cout << "New val: " << new_val << "  ";
  //  cout << "T_const: " << Time_const << "  ";

  //  Value = Time_const * new_val + (1.0 - Time_const) * Value;
  //  cout << "New val: " << value << "  ";
  Value = new_val;

  if (Value > MAX_VALUE)
    Value = MAX_VALUE;
  else if (Value < -MAX_VALUE)
    Value = -MAX_VALUE;

  //  cout << "Value: " << Value << endl;
}


int Neuron :: duplicate(Neuron *original)
{
  int i;

  Type = original->Type;
  Function = original->Function;
  Scale = original->Scale;
  Offset = original->Offset;
  Time_const = original->Time_const;
  Value = original->Value;

  Num_inputs = 0;
  for (i=0; i<original->Num_inputs; i++) {
    if (!add_input(original->Input[i].weight, original->Input[i].input)) {
      return FALSE;
    }
  }
  return TRUE;
}



int Neuron :: reverse_link(int index)
{
  if (index >= Num_inputs)
    return -1;
  else if (index < 0)
    return -1;

  int result;
  result = Input[index].input->add_input(Input[index].weight, this);
  if (result == -1)
    return -1;

  cut_input(index);
  return result;
}


// Replaces all occurencs of input neuron /original/ with
// the neuron /replacement/.
void Neuron :: replace_input_neuron(Neuron *original, Neuron *replacement)
{
  int i;

  for (i=0; i<Num_inputs; i++) {
    if (Input[i].input == original) {
      Input[i].input = replacement;
      //      printf("%d : link# %d, Replacing: %d with %d.\n", 
      //	     Id, i, original->Id, replacement->Id);
    }
  }
}


// Deletes all input links from source neuron.
void Neuron :: delete_inputs_from(Neuron *source)
{
  int i;

  for (i=0; i<Num_inputs; i++) {
    if (Input[i].input == source)
      Input[i].weight = 0.0;
  }
}



/******** I/O *******/

void Neuron :: read(ifstream *infile, Neuron **neurons)
{
  *infile >> Function;
  *infile >> Offset;
  *infile >> Time_const;
  *infile >> Id;
}


int Neuron :: read(FILE *file)
{
  int num;

  num = fscanf(file, "%d %d %lf %lf %lf %d\n", &Function,
	       &Num_inputs, &Scale, &Offset,
	       &Time_const, &Id);
  if (num != 7)
    {
      printf("Neuron: error reading\n");
      return FALSE;
    }

  Scale = reduce(Scale);
  Offset = reduce(Offset);
  Time_const = reduce(Time_const);

  Num_inputs = 0;

  //  printf("%d %d %lf %lf %d\n", Function, num_inputs,
  //	  offset, time_const, id);

  return TRUE;
}


void Neuron :: write(ofstream *outfile)
{
  *outfile << Function;
  *outfile << ' ';
  *outfile << Num_inputs;
  *outfile << ' ';
  *outfile << Offset;
  *outfile << ' ';
  *outfile << Time_const;
  *outfile << ' ';
  *outfile << Id;
  *outfile << '\n';
}


void Neuron :: write(FILE *file)
{
  fprintf(file, "%d %d %.18lf %.18lf %.18lf %d\n", Function, Num_inputs,
	  Scale, Offset, Time_const, Id);
}
