/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stream.h>
#include <fstream.h>
#include "general.h"
#include "pop.h"
#include <limits.h>
#include "deme.h"
#include "individual.h"
#include "stats.h"

/***** Initialization functions *****/

Pop :: Pop(void)
{
  init("backup");
}

Pop :: Pop(char *n)
{
  init(n);
}


void Pop :: init(char *n)
{
  created = FALSE;
  stats_avg = new Stats();
  stats_best = new Stats();
  stats_best_vec = new Stats();
  stats_avg_vec = new Stats();
  stats_trials_avg = new Stats();
  stats_trials_best = new Stats();

  do_coevolve = FALSE;
  structure = STRUCT_CANONICAL;
  max_age = MAX_AGE;
  nghbr_size = 10;
  nghbr_type = NGHBR_NONE;

  pressure = 0.05;
  elitism = 3;
  best_ever = -DBL_MAX;
  num_demes = 1;
  num_individs = 50;

  max_generations = 100;
  hof_size = HOF_SIZE;

  sample_individ = NULL;

  log_file[0] = 0;
  set_name(n);
}


void Pop :: set_name(char *n)
{
  int i;

  for (i=0; i<20; i++) {
    pop_name[i] = n[i];
    if (pop_name[i] == 0)
      break;
  }
  pop_name[i] = 0;
}

void Pop :: set_logfile(char *n)
{
  int i;

  if (n == 0)
    return;

  for (i=0; i<20; i++) {
    log_file[i] = n[i];
    if (log_file[i] == 0)
      break;
  }
  log_file[i++] = '.';
  log_file[i++] = 'l';
  log_file[i++] = 'o';
  log_file[i++] = 'g';
  log_file[i] = 0;
}


void Pop :: set_sample_individ(Individual *sample)
{
  sample_individ = sample;
}


void Pop :: uncreate(void)
{
  /* not implemented */
  cout << "uncreate: not-implemented\n" << endl;
}


void Pop :: set_tournament_size(int size)
{
  int i;

  if (size < 3)
    size = 3;

  tournament_size = size;
  if (created) {
    for (i=0; i<num_demes; i++)
      deme[i]->tournament_size = size;
  }
}


void Pop :: set_replace_size(int size)
{
  if (size < 1)
    replace_size = 1;
  else if (size > 2)
    replace_size = 2;
  else
    replace_size = size;
}


void Pop :: create(void)
{
  int i;

  if (created)
    uncreate();
  created = TRUE;

  deme = (Deme**)malloc(num_demes*sizeof(Deme*));
  for (i=0; i<num_demes; i++) {
    deme[i] = new Deme(num_individs);
    deme[i]->adaptive_mut = adaptive_mut;
    deme[i]->elitism = elitism;
    deme[i]->pressure = pressure;
    deme[i]->replace_size = replace_size;
    deme[i]->tournament_size = tournament_size;
    deme[i]->structure = structure;
    deme[i]->max_age = max_age;
    deme[i]->nghbr_size = nghbr_size;
    deme[i]->nghbr_type = nghbr_type;
    deme[i]->recomb_rate = recomb_rate;
    deme[i]->pop = this;
    deme[i]->sample_individ = sample_individ;

    if (num_demes > 1)
      deme[i]->deme_id = i;
    deme[i]->create();
  }
}




/***** Update functions *****/

extern Real play_game(int init, Real *fitnessA, Real *fitnessB,
		      Individual *net1, Individual *net2);

void Pop :: evaluate_init(void)
{
  int i, j, k;
  int init_flag = TRUE;
  int indexs[num_demes];
  int new_eval[num_demes];
  Real score;
  Real fitnessA, fitnessB;
  Individual *inds[num_demes];

  cout << "Evaluate init." << num_demes << endl;

  while (init_flag) {
    // Play a Round-Robin.
    for (j=0; j<num_demes; j++) {
      indexs[j] = deme[j]->next_individual_init();
      if (indexs[j] == -1) {
	new_eval[j] = FALSE;
	indexs[j] = random_int(num_individs);
      } else {
	new_eval[j] = TRUE;
	deme[j]->individ[indexs[j]]->set_fitness(0.0);
      }
      inds[j] = deme[j]->individ[indexs[j]];
    }

    init_flag = FALSE;
    for (j=0; j<num_demes; j++) {
      if (new_eval[j])
	init_flag = TRUE;
    }
    if (!init_flag)
      break;

    for (j=0; j<num_demes; j++) {
      for (k=j+1; k<num_demes; k++) {
	score = play_game(TRUE, &fitnessA, &fitnessB, 
			  inds[j], inds[k]);
	inds[j]->add_fitness(fitnessA);
	inds[k]->add_fitness(fitnessB);
      }

      if (new_eval[j]) {
	inds[j]->div_fitness(4.0);
	if (Print[8]) {
	  inds[j]->print_history_self();
	  cout << endl;
	}
      }
    }
  }

  for (i=0; i<num_demes; i++)
    deme[i]->cycle_index = 0;
}

void Pop :: evaluate_steady_state(void)
{
#ifdef old
  int i, j, k;
  int indexs[num_demes];
  Real score;
  Real fitnessA, fitnessB;
  Real fitness_old[num_demes];
  Individual *inds[num_demes];

  cout << "Evaluating." << endl;

  for (i=0; i<num_individs; i++) {
    // Play a Round-Robin.
    for (j=0; j<num_demes; j++) {
      indexs[j] = deme[j]->next_individual();
      if (Print[8])
	cout << endl;
      
      inds[j] = deme[j]->individ[indexs[j]];
      fitness_old[j] = inds[j]->fitness;
      inds[j]->fitness = 0.0;
    }

    for (j=0; j<num_demes; j++) {
      for (k=j+1; k<num_demes; k++) {
	play_game(FALSE, &fitnessA, &fitnessB, 
			  inds[j], inds[k]);
	inds[j]->fitness += fitnessA;
	inds[k]->fitness += fitnessB;
      }
      inds[j]->fitness /= 4.0;
      if (Print[8]) {
	inds[j]->print_history_self();
	cout << endl;

      // Test for a re-evaluation.
      if (deme[j]->age[indexs[j]] > 0)
	inds[j]->set_fitness((inds[j]->fitness + fitness_old[j]) / 2.0;
      }
    }
  }

#endif
}


void Pop :: coevaluate_init(void)
{
  int i, j, index;
  
  // Reset fitness scores.
  for (i=0; i<num_demes; i++)
    for (j=0; j<num_individs; j++)
      deme[i]->individ[j]->set_fitness(0.0);

  // Solo evaluation.
  for (i=0; i<num_demes; i++) {
    cout << "Initial evaluation, deme #" << i << endl;
    while ((index = deme[i]->next_individual_init()) != -1)
      deme[i]->individ[index]->evaluate();
    //      deme[i]->individ[index]->evaluate(TRUE);
  }
}


void Pop :: coevaluate_hof(void)
{
  int i, j, k, l;
  Real fitnessA, fitnessB;

  // Reset fitness scores.
  for (i=0; i<num_demes; i++)
    for (j=0; j<num_individs; j++)
      deme[i]->individ[j]->set_fitness(0.0);

  // Special case, when opponent offset = 0.
  for (i=0; i<num_individs; i++)
    for (j=0; j<num_demes; j++)
      for (k=0; k<num_demes; k++) {
	if (j==k)
	  continue;  // don't play against own HOF.

	for (l=0; l<HOF_SIZE; l++) {
	  play_game(FALSE, &fitnessA, &fitnessB,
		    deme[j]->individ[i],
		    deme[k]->hof[l]);
	  deme[j]->individ[i]->add_fitness(fitnessA);
	}
      }

  // Divide fitness scores by number of games.
  for (i=0; i<num_demes; i++)
    for (j=0; j<num_individs; j++)
      deme[i]->individ[j]->div_fitness((Real)(num_demes*HOF_SIZE));


  // Solo evaluation.
  for (i=0; i<num_demes; i++)
    for (j=0; j<num_individs; j++)
      deme[i]->individ[j]->evaluate(FALSE);
}


void Pop :: coevaluate(void)
{
  int i, j, k, l;
  Real fitnessA, fitnessB;

  // Reset fitness scores.
  for (i=0; i<num_demes; i++)
    for (j=0; j<num_individs; j++)
      deme[i]->individ[j]->add_fitness(0.0);

  // Special case, when opponent offset = 0.
  for (i=0; i<num_individs; i++)
    for (k=0; k<num_demes; k++)
      for (l=k+1; l<num_demes; l++) {
	play_game(FALSE, &fitnessA, &fitnessB,
		  deme[k]->individ[i],
		  deme[l]->individ[i]);
	deme[k]->individ[i]->add_fitness(fitnessA);
	deme[l]->individ[i]->add_fitness(fitnessB);
      }

  for (i=0; i<num_individs; i++)
    for (j=1; j<num_opponents; j++)
      for (k=0; k<num_demes; k++)
	for (l=0; l<num_demes; l++) {
	  if (k == l)
	    continue;
	  play_game(FALSE, &fitnessA, &fitnessB,
		    deme[k]->individ[i],
		    deme[l]->individ[(i+j)%num_individs]);
	  deme[k]->individ[i]->add_fitness(fitnessA);
	  deme[l]->individ[(i+j)&num_individs]->add_fitness(fitnessB);
	}


  // Divide fitness scores by number of games.
  for (i=0; i<num_demes; i++)
    for (j=0; j<num_individs; j++)
      deme[i]->individ[j]->div_fitness((Real)(num_demes*num_opponents+1));

  // Solo evaluation.
  for (i=0; i<num_demes; i++)
    for (j=0; j<num_individs; j++)
      deme[i]->individ[j]->evaluate(FALSE);
}

void Pop :: post_evaluate(void)
{
  int i;

  for (i=0; i<num_demes; i++)
    deme[i]->post_evaluate();
}



void Pop :: make_next_generation(void)
{
  int i;

  for (i=0; i<num_demes; i++) {
    //    cout << "Pop :: Making next generation" << endl;
    deme[i]->make_next_generation();
  }
  //  cout << endl;
}


void Pop :: update_final_stats(void)
{
  //  printf("update last stats.\n");

  if (!stats_trials_best->built()) {
    stats_trials_best->num_single = 0;
    stats_trials_best->num_multi = max_generations + 1;
    stats_trials_best->multi_len = Num_fit_dims;
    stats_trials_best->build();
  }
  //  stats_trials_best->multi_start();
  //  stats_trials_best->m_trials += 1;
  stats_best_vec->send_stats(stats_trials_best);
  //  stats_best_vec->print_formatted();

  if (!stats_trials_avg->built()) {
    stats_trials_avg->num_single = 0;
    stats_trials_avg->num_multi = max_generations + 1;
    stats_trials_avg->multi_len = Num_fit_dims;
    stats_trials_avg->build();
  }
  //  stats_trials_avg->multi_start();
  //  stats_trials_avg->m_trials += 1;
  stats_avg_vec->send_stats(stats_trials_avg);
}


void Pop :: update_stats(void)
{
  int i;
  Real best_vec[Num_fit_dims];

  log = 0;
  if (log_file[0] != 0) {
    switch (generation) {
    case 0:
    case 1:
    case 5:
    case 10:
    case 20:
    case 30:
    case 50:
    case 75:
    case 100:
    case 125:
    case 150:
    case 175:
    case 200:
    case 250:
    case 300:
      log = fopen(log_file, "a");
      break;

    default:
      if (!(generation % 100))
	log = fopen(log_file, "a");
      break;
    }
  }

  deme[0]->update_stats((Config.num_trials == 1) || Print[36]);

  avg_fitness = deme[0]->avg_fitness;
  max_fitness = deme[0]->max_fitness;
  deme[0]->update_best_vec(best_vec);

  for (i=1; i<num_demes; i++) {
    deme[i]->update_stats((Config.num_trials == 1) || Print[36]);
    avg_fitness += deme[i]->avg_fitness;
    if (deme[i]->max_fitness > max_fitness) {
      max_fitness = deme[i]->max_fitness;
      deme[i]->update_best_vec(best_vec);
    }
  }
  avg_fitness = avg_fitness / (double)num_demes;
  stats_best->single_add(max_fitness);
  stats_avg->single_add(avg_fitness);

  stats_best_vec->multi_add(best_vec);
  stats_best_vec->multi_next();

  if (Num_fit_dims > 1) {
    for (i=0; i<num_demes; i++) {
      deme[i]->update_stats_vector(stats_avg_vec);
    }
    stats_avg_vec->multi_next();
  }

  //  cout << "Best: "<<max_fitness<< "   " << endl;
  //  cout << "Avg: "<<avg_fitness<< "\n" << endl;

  if (log) {
    fclose(log);
  }

  // Operator stats.
  for (i=0; i<Max_operator; i++) {
    Operation_hist[i] += Operation[i];
    Op_better_hist[i] += Op_better[i];
  }

  if (Print[30] && (Max_operator > 0)) {
    printf("  tot#: ");
    for (i=0; i<Max_operator; i++)
      printf("%5d ", Operation_hist[i]);
    printf("\n");

    printf("  betr: ");
    for (i=0; i<Max_operator; i++)
      printf("%5d ", Op_better_hist[i]);
    printf("\n");

    /*
    for (i=0; i<Max_operator; i++)
      printf("%5d ", Operation[i]);
    printf("\n");

    for (i=0; i<Max_operator; i++)
      printf("%5d ", Op_better[i]);
    printf("\n");
    */
  }

  for (i=0; i<Max_operator; i++) {
    Operation[i] = 0;
    Op_better[i] = 0;
  }
}


void Pop :: initialize(void)
{
  int i;

  if (do_coevolve) {
    for (i=0; i<num_demes; i++) {
      cout << "Creating deme #" << i << endl;
      deme[i]->initialize_simple();
    }
    cout << "Initial evaluation" << endl;
    coevaluate_init();
  } else {
    for (i=0; i<num_demes; i++) {
      deme[i]->initialize();
    }
  }
}

void Pop :: initialize(Individual *ind_seed)
{
  int i;

  if (do_coevolve) {
    for (i=0; i<num_demes; i++) {
      cout << "Creating deme #" << i << endl;
      deme[i]->initialize_simple(ind_seed);
    }
    cout << "Initial evaluation" << endl;
    coevaluate_init();
  } else {
    cout << "Pop :: Initialize with seed." << endl;
    for (i=0; i<num_demes; i++) {
      deme[i]->initialize(ind_seed);
    }
  }
}



void Pop :: evolve(void)
{
  int i;

  if (stats_best->built()) {
    stats_best->init();
  } else {
    stats_best->num_single = max_generations + 1;
    stats_best->build();
  }
  stats_best->single_start();

  if (stats_avg->built()) {
    stats_best->init();
  } else {
    stats_avg->num_single = max_generations + 1;
    stats_avg->build();
  }
  stats_avg->single_start();

  if (stats_best_vec->built()) {
    stats_best_vec->init();
  } else {
    stats_best_vec->num_single = 0;
    stats_best_vec->num_multi = max_generations + 1;
    stats_best_vec->multi_len = Num_fit_dims;
    stats_best_vec->build();
  }
  stats_best_vec->multi_start();
  stats_best_vec->m_trials += 1;

  if (stats_avg_vec->built()) {
    stats_avg_vec->init();
  } else {
    stats_avg_vec->num_single = 0;
    stats_avg_vec->num_multi = max_generations + 1;
    stats_avg_vec->multi_len = Num_fit_dims;
    stats_avg_vec->build();
  }
  stats_avg_vec->multi_start();
  stats_avg_vec->m_trials += num_individs * num_demes;

  for (i=0; i<Max_operator; i++) {
    Operation[i] = 0;
    Op_better[i] = 0;
    Operation_hist[i] = 0;
    Op_better_hist[i] = 0;
  } 

  generation = 0;
  update_stats();

  if (structure != STRUCT_BLIND_WATCHMAKER) {
    for (i=0; i<max_generations; i++) {
      generation++;
      if (!Print[15] && (Print[8] || Print[7]))
	cout << "Generation: "<< generation << endl;
      make_next_generation();
      if (structure != STRUCT_DETERMINISTIC_CROWDING) {
	post_evaluate();
	if (Print[39]) {
	  print_gen_results();
	}
      }
      update_stats();
      if ((Config.num_trials == 1) && (!(i%10))) {
	write_backup();
      }
      if (!Print[15] && (Print[8] || Print[7]))
	cout << endl;
    }

    if (Config.num_trials > 1) {
      update_final_stats();
    }

  } else {
    Run_BlindWatchmaker(deme[0]);
  }
}


void Pop :: evolve(int gen)
{
  max_generations = gen;

  evolve();
}


void Pop :: store_hof(void)
{
  int i;

  for (i=0; i<num_demes; i++)
    deme[i]->hof[generation%hof_size]->duplicate(deme[i]->individ[deme[i]->max_index]);
}

void Pop :: coevolve(int iterations)
{
  int i;

  stats_best->num_single = iterations+1;
  stats_best->build();
  stats_best->single_start();

  stats_avg->num_single = iterations+1;
  stats_avg->build();
  stats_avg->single_start();

  generation = 0;
  coevaluate();
  print_all2();
  update_stats();
  store_hof();

  for (i=0; i<iterations; i++) {
    generation++;

    if (!Print[15])
      cout << "Generation: "<< generation <<"\n" << endl;
    make_next_generation();

    if (generation < HOF_SIZE)
      coevaluate();
    else
      coevaluate_hof();

    print_all2();
    update_stats();
    store_hof();
    write_backup();
  }

  /*
  for (i=0; i<iterations; i++) {
    cout << "Generation: "<< i+1 <<"\n" << endl;
    evaluate_steady_state();
    update_stats();
    write_backup();
  }
  */
}


void Pop :: mutate_all(void)
{
  int i;

  for (i=0; i<num_demes; i++)
    deme[i]->mutate_all();
}


/******* Print functions *****/

void Pop :: print_all2(void)
{
  int i, j;

  for (i=0; i<num_individs; i++) {
    for (j=0; j<num_demes; j++) {
      //      deme[j]->individ[i]->print_history_self();
      deme[j]->individ[i]->print_fitness();
      deme[j]->individ[i]->print_history();
      cout << "   ";
    }
    cout << endl;
  }  
}


void Pop :: print_gen_results(void)
{
  int i;

  for (i=0; i<num_demes; i++)
    deme[i]->print_gen_results();
}

  

void Pop :: print_all(void)
{
  int i;

  for (i=0; i<num_demes; i++)
    deme[i]->print_all();
}


void Pop :: info(void)
{
  int i;

  for (i=0; i<num_demes; i++)
    deme[i]->info();
}


void Pop :: print_last_stats(void)
{
  stats_best_vec->print_last();
  stats_avg_vec->print_last();
}

void Pop :: print_stats(void)
{
  //  stats_best->print();
  //  stats_avg->print();

  //  cout << "# Average" << endl;
  //  stats_avg->print_formatted();

  if (structure != STRUCT_BLIND_WATCHMAKER) {
    //    cout << endl << "# Best" << endl;
    //    stats_best->print_formatted();
    printf("Best\n");
    stats_best_vec->print_short_formatted();
    /*
    printf("Average\n");
    stats_avg_vec->print_short_formatted();
    */
  }
}

void Pop :: print_final_stats(void)
{
  if (structure != STRUCT_BLIND_WATCHMAKER) {
    printf("Best\n");
    stats_trials_best->print_short_formatted();
    printf("Average\n");
    stats_trials_avg->print_short_formatted();
  }
}

/***** File I/O *****/

void Pop :: write_best(FILE *file)
{
  deme[0]->write_best(file);
}

void Pop :: write_best_new(FILE *file)
{
  deme[0]->write_best_new(file);
}


int Pop :: read_individ(FILE *file)
{
  return deme[0]->read_individ(file);
}


int Pop :: read(FILE *file)
{
  int i, num;

  num = fscanf(file, "%d %d %d\n", &num_demes, &structure, 
	       &num_individs);
  if (num != 3) {
    printf("pop: error in reading start: %d\n", num);
  }

  create();
  for (i=0; i<num_demes; i++) {
    if (deme[i]->read(file) == FALSE)
      return FALSE;
  }

  return TRUE;
}
      
void Pop :: write(FILE *file)
{
  int i;

  fprintf(file, "%d %d %d\n", num_demes, structure,
	  num_individs);

  for (i=0; i<num_demes; i++)
    deme[i]->write(file);
}


void Pop :: write_best_individ(char *filename)
{
  FILE *file;
  file = fopen(filename, "w");
  write_best(file);
  fclose(file);
}

void Pop :: write_backup(void)
{
  int i;
  char fname[50];
  FILE *file;

  sprintf(&fname[0], "%s_bk.pop", pop_name);
  file = fopen(fname, "w");
  write(file);
  fclose(file);

  sprintf(&fname[0], "%s_gen", pop_name);
  for (i=0; i<num_demes; i++)
    deme[i]->write_best(fname);
}
