/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#ifndef PROD_BODY_HEADER_FILE
#define PROD_BODY_HEADER_FILE

#include <iostream.h>
#include <fstream.h>
#include "general.h"
#include "ind_lsys.h"


//#define MAX_PROD_LENGTH 10
//#define PROD_LENGTH 25   // space allocated, allowing room for overflow

#define MAX_PROD_LENGTH 15
#define PROD_LENGTH 50   // space allocated, allowing room for overflow
#define MAX_STRING_LENGTH 10000 // 10000

extern int Pvalue_type;
extern int Max_prod_length;
extern int Prod_length;
extern int Max_string_length;
extern int Num_pbc_ops;

// Symbol types.
#define PBC_TYPE_NONE 0
#define PBC_TYPE_TERMINAL 1
#define PBC_TYPE_PRODUCTION 2
#define PBC_TYPE_BLOCK 3
#define PBC_TYPE_VALUE 4
#define PBC_TYPE_CONSTANT 5
#define PBC_TYPE_VARIABLE 6
#define PBC_TYPE_FUNCTION 7


// Operators.
#define NUM_PBC_OPS 5
#define PBC_OP_NONE 0
#define PBC_OP_ADD 1
#define PBC_OP_SUB 2
#define PBC_OP_EQUAL 3
#define PBC_OP_DIV 4
#define PBC_OP_MULT 5
#define PBC_OP_POW 6

#define OP_ADD '+'
#define OP_SUB '-'
#define OP_MULT '*'
#define OP_DIV '/'
#define OP_POW '^'
#define OP_EQUAL '='



// Comparators.
#define PBC_COMP_TRUE 0
#define PBC_COMP_LESS 1

#define COMP_LESS '<'
#define COMP_GREATER '>'


// Blocks.
#define SYMBOL_PUSH '['
#define SYMBOL_POP ']'
#define SYMBOL_BLK_START '{'
#define SYMBOL_BLK_END '}'
#define SYMBOL_PUSH_EDGE '('
#define SYMBOL_POP_EDGE ')'


typedef struct Production_Value
{
  char type;
  char symbol;
  Real value;
} Prod_Value;  


typedef struct Production_Condition
{
  char comparator;
  Prod_Value arg1, arg2;
} Prod_Cond;


typedef struct Production_Parameter
{
  char op;
  Prod_Value arg1, arg2;
} Prod_Param;


typedef struct Production_Character
{
  char type;
  char symbol;
  int match;

  Prod_Param *param;
} Prod_Char;


extern Real Prob_mult_seq;
extern Real Prob_branch_edge;
extern Real Prob_branch;

void set_probabilities(void);

int is_op(char c);

void copy_pvalue(Prod_Value *pv_to, Prod_Value *pv_from);
void copy_param(Prod_Param *p_from, Prod_Param *p_to);
void copy_pchar(Prod_Char *pc_from, Prod_Char *pc_to);

char random_terminal(void);
char random_func(void);
char random_production(void);
void random_pvalue(Prod_Value *pvalue, int deflt_flag);

Real calc_prod_value(Real *arg, Constant *constant,
		     Prod_Value *pval);
Real calc_prod_param(Real *arg, Constant *constant,
		     Prod_Param *param);

int read_integer(char *line);
int read_pvalue(char *line, Prod_Value *pvalue);
int read_param(char *line, Prod_Param *param);

void fprint_pvalue(FILE *file, Prod_Value *pvalue);
void fprint_param(FILE *file, Prod_Param *param);


class Prod_Body
{
public:
  /*** Variables ***/
  int tested;
  int used;

  int empty;
  Prod_Cond cond;

  int length;
  //  Prod_Char string[PROD_LENGTH];
  Prod_Char *string;

  Real max[MAX_PROD_VARS];
  Real min[MAX_PROD_VARS];
  Prod_Head phead;
  Constant *constant; // pointer to constants in parent.
  Individual *parent;

  /*** Methods ***/

  Prod_Body(void);
  void initialize_pvalue(Prod_Value *pvalue);
  void initialize_param(Prod_Param *param);
  void initialize_pchar(int index);
  void initialize(void);

  bool is_block(int i);
  bool is_block_start(int i);
  bool is_mult_start(int i);
  bool is_mult_end(int i);
  bool is_block_end(int i);
  bool is_production(int i);
  bool is_terminal(int i);
  int num_productions(void);

  bool valid_cond(void);
  bool valid(void);

  void copy_symbol(int s_from, int s_to);
  void move_symbol(int s_from, int s_to);
  int insert_space(int s_from, int len);
  int symbol_test(int index);

  int delete_single_symbol(int index);
  int delete_block(int index, int del_inside);
  int delete_symbol(int index, int del_inside);
  void delete_sequence(int l1, int l2);
  void delete_string(void);

  int cond_repair_pvalue(char v, Prod_Value *pval);
  int make_cond_smaller_old1(void);
  int make_cond_smaller(void);
  int make_cond_larger(void);
  int repair_cond(void);
  int repair(void);
  void make_random_production(int i);
  void random_single_param(int i, int p, int deflt_flag);
  void random_full_param(int i, int p);
  void make_random_terminal(int i);
  void make_random_sequence(int start, int length);
  void make_random_branch(int start, int length);
  void make_random_branch_edge(int start, int length);
  void make_random_mult(int start, int length);

  void duplicate_pchar(int to_loc, int from_loc, Prod_Body *from_body);
  void duplicate_string(Prod_Body *body);
  void duplicate_cond(Prod_Body *body);
  void duplicate(Prod_Body *body);
  int duplicate(int len, LString *string);
  void set_cond_true(void);
  void set_cond_greater_zero(void);
  int is_cond_true(void);
  int test_cond(Real *arg);
  int write_body(Real *arg, Tape *tape_dest);

  void mutate_pvalue(int index, Prod_Value *pvalue);
  void mutate_param_single(int index);
  int mutate_param_full(int index);
  int mutate_param(int index);
  void reset_single_param(int index);
  int mutate_symbol(int index);
  int mutate_type(int index);
  int mutate_character_orig(void);
  int mutate_character(void);
  int insert_prod_rule(int i);
  int add_symbol(int i);
  void make_random(void);
  void make_random_condition(void);

  int reorder(void);
  int replicate(void);
  int replicate_old1(void);
  int repair_arg2(void);
  int usable_cond(void);
  int always_true(void);
  int make_valid_cond_arg2(void);
  int mutate_cond_old1(void);
  int mutate_cond(void);
  int mutate_old1(void);
  int mutate(void);
  int mutate_pheno(void);

  int block_beginning(int i);
  int block_ending(int i);
  void pick_block(int *loc1, int *loc2);
  void random_block(int *l1, int *l2);
  int block_length(int l1, int l2);
  int get_loc(int start, int len);
  void random_subblock(int *l1, int *l2);

  int match_parens(char end_type, char end_symbol, int i);
  int match_parens(void);

  int recombine_blocks(Prod_Body *body);
  int recombine(Prod_Body *body);

  int fprint_char(FILE *file, int i);
  int fprint(FILE *file);

  int read(FILE *file);
  int read_old(FILE *file);
};

#endif
