/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/


#include <math.h>
#include "quat.h"
#include "algebra.h"
#include <stream.h>
//include "general.hh"


#define Q_S 0
#define Q_X 1
#define Q_Y 2
#define Q_Z 3

#define Q_SMALL (1e-11)


void qt_print(Quat q, char *string)
{
    int i;
  Real l = 0.0;

  if (string)
    cout << string;
  for (i=0; i<4; i++)
    cout << q[i]<<"  ";

  for (i=0; i<4; i++)
    l += sqr(q[i]);
  l = sqrt(l);
  cout << "("<<l<<")\n";
}


void qt_make(Quat q, Real x, Real y, Real z, Real ang)
{
  Real norm, sin_ang;

  norm = 1.0 / sqrt(sqr(x) + sqr(y) + sqr(z));
  x *= norm;
  y *= norm; 
  z *= norm;

  ang = ang/2.0;
  q[Q_S] = cos(ang);

  sin_ang = sin(ang);
  q[Q_X] = x * sin_ang;
  q[Q_Y] = y * sin_ang;
  q[Q_Z] = z * sin_ang;
}


Real qt_norm(Quat q)
{
  return sqrt(sqr(q[0])+sqr(q[1])+sqr(q[2])+sqr(q[3]));
}


void qt_invert(Quat q_inv, Quat q)
{
  q_inv[Q_X] = -q[Q_X];
  q_inv[Q_Y] = -q[Q_Y];
  q_inv[Q_Z] = -q[Q_Z];
  q_inv[Q_S] = q[Q_S];
}

void qt_normalize(Quat q)
{
  Real val;

  val = 1.0/sqrt(sqr(q[0])+sqr(q[1])+sqr(q[2])+sqr(q[3]));
  q[0] *= val;
  q[1] *= val;
  q[2] *= val;
  q[3] *= val;
}

void qt_mult(Quat q_d, Quat q1, Quat q2)
{
  Quat q_tmp;

  q_tmp[Q_X] = q1[Q_X]*q2[Q_S] + q1[Q_Y]*q2[Q_Z]
	   - q1[Q_Z]*q2[Q_Y] + q1[Q_S]*q2[Q_X];

  q_tmp[Q_Y] = -q1[Q_X]*q2[Q_Z] + q1[Q_Y]*q2[Q_S]
	   + q1[Q_Z]*q2[Q_X] + q1[Q_S]*q2[Q_Y];
					
  q_tmp[Q_Z] = q1[Q_X]*q2[Q_Y] - q1[Q_Y]*q2[Q_X]
	   + q1[Q_Z]*q2[Q_S] + q1[Q_S]*q2[Q_Z];

  q_tmp[Q_S] = -q1[Q_X]*q2[Q_X] - q1[Q_Y]*q2[Q_Y]
	   - q1[Q_Z]*q2[Q_Z] + q1[Q_S]*q2[Q_S];

  q_d[0] = q_tmp[0];
  q_d[1] = q_tmp[1];
  q_d[2] = q_tmp[2];
  q_d[3] = q_tmp[3];
}


/* quaternion -> rotation matrix */
void qt_to_col_matrix(Real mat[4][4], Quat q)
{
  int i, j;
  Real xx = q[Q_X]*q[Q_X];
  Real yy = q[Q_Y]*q[Q_Y];
  Real zz = q[Q_Z]*q[Q_Z];
  Real ss = q[Q_S]*q[Q_S];

  Real xy = q[Q_X]*q[Q_Y];
  Real xz = q[Q_X]*q[Q_Z];
  Real xs = q[Q_X]*q[Q_S];
  Real yz = q[Q_Y]*q[Q_Z];
  Real ys = q[Q_Y]*q[Q_S];
  Real zs = q[Q_Z]*q[Q_S];


  mat[0][0] = ss + xx - 0.5;
  mat[0][1] = xy - zs;
  mat[0][2] = xz + ys;
  mat[0][3] = 0.0;

  mat[1][0] = xy + zs;
  mat[1][1] = ss + yy - 0.5;
  mat[1][2] = yz - xs;
  mat[1][3] = 0.0;

  mat[2][0] = xz - ys;
  mat[2][1] = yz + xs;
  mat[2][2] = zz + ss - 0.5;
  mat[2][3] = 0.0;

  mat[3][0] = 0.0;
  mat[3][1] = 0.0;
  mat[3][2] = 0.0;
  mat[3][3] = 1.0;

  for (i=0; i<3; i++)
    for (j=0; j<3; j++)
      mat[i][j] *= 2.0;
}


void qt_to_row_matrix(Real mat[4][4], Quat q)
{
  Real xx = 2.0 * q[Q_X]*q[Q_X];
  Real yy = 2.0 * q[Q_Y]*q[Q_Y];
  Real zz = 2.0 * q[Q_Z]*q[Q_Z];
  Real ss = 2.0 * q[Q_S]*q[Q_S];

  Real xy = 2.0 * q[Q_X]*q[Q_Y];
  Real xz = 2.0 * q[Q_X]*q[Q_Z];
  Real xs = 2.0 * q[Q_X]*q[Q_S];
  Real yz = 2.0 * q[Q_Y]*q[Q_Z];
  Real ys = 2.0 * q[Q_Y]*q[Q_S];
  Real zs = 2.0 * q[Q_Z]*q[Q_S];


  mat[0][0] = ss + xx - 1.0;
  mat[1][0] = xy - zs;
  mat[2][0] = xz + ys;
  mat[3][0] = 0.0;

  mat[0][1] = xy + zs;
  mat[1][1] = ss + yy - 1.0;
  mat[2][1] = yz - xs;
  mat[3][1] = 0.0;

  mat[0][2] = xz - ys;
  mat[1][2] = yz + xs;
  mat[2][2] = zz + ss - 1.0;
  mat[3][2] = 0.0;

  mat[0][3] = 0.0;
  mat[1][3] = 0.0;
  mat[2][3] = 0.0;
  mat[3][3] = 1.0;
}


void qt_deriv(Quat dq, Real vang[3], Quat q)
{
  Real m_tmp[4][3];

  m_tmp[0][0] = -q[Q_X];
  m_tmp[0][1] = -q[Q_Y];
  m_tmp[0][2] = -q[Q_Z];

  m_tmp[1][0] = q[Q_S];
  m_tmp[1][1] = -q[Q_Z];
  m_tmp[1][2] = q[Q_Y];

  m_tmp[2][0] = q[Q_Z];
  m_tmp[2][1] = q[Q_S];
  m_tmp[2][2] = -q[Q_X];

  m_tmp[3][0] = -q[Q_Y];
  m_tmp[3][1] = q[Q_X];
  m_tmp[3][2] = q[Q_S];

  vmv443_mult(dq, m_tmp, vang);
  vs4_mult(dq, 0.5);
}
