/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "general.h"
#include "stats.h"

Stats :: Stats(void)
{
  num_single = 0;
  num_multi = 0;
  multi_len = 0;
  Built = FALSE;

  //  return this;
}

void Stats :: free(void)
{
  
}


void Stats :: init(void)
{
  int i, j;
  
  s_trials = 0;
  m_trials = 0;

  for (i=0; i<num_single; i++)
    {
      stat_savg(i) = 0.0;
      stat_sstd(i) = 0.0;
    }

  for (i=0; i<num_multi; i++)
    for (j=0; j<multi_len; j++)
      {
	stat_mavg(i, j) = 0.0;
	stat_mstd(i, j) = 0.0;
      }
}


int Stats :: built(void)
{
  return Built;
}


int Stats :: unbuild(void)
{
  printf("Stats::unbuild - not implemented\n");
  return FALSE;
}


int Stats :: build(void)
{
  int i, l = multi_len;

  if (Built)
    unbuild();

  Built = TRUE;

  multi_ptr = (Stat_multi*)malloc(num_multi*sizeof(Stat_multi));
  for (i=0; i<num_multi; i++)
    {
      stat_mavg_ptr(i) = (double*)malloc(l*sizeof(double));
      if (stat_mavg_ptr(i) == 0)
	{
	  printf("Error: could not alloc stat_mavg_ptr\n");
	  return 0;
	}

      stat_mstd_ptr(i) = (double*)malloc(l*sizeof(double));
      if (stat_mstd_ptr(i) == 0)
	{
	  printf("Error: could not alloc stat_mstd_ptr\n");
	  return 0;
	}
    }

  single_ptr = (Stat_single*)malloc(num_single*sizeof(Stat_single));
  if (single_ptr == 0)
    {
      printf("Error: could not alloc stat_sptr\n");
      return 0;
    }
  
  init();
  return TRUE;
}


void Stats :: single_start(void)
{
  
  s_ctr = 0;
  s_trials++;
}

void Stats :: multi_start(void)
{
   int i;

   for (i=0; i<num_multi; i++) {
      stat_mstart(i);
   }

   m_ctr = 0;
}


void Stats :: multi_add(double val)
{
   int i = m_ctr;

   stat_mavg(i, stat_mmctr(i)) += val;
   stat_mstd(i, stat_mmctr(i)) += val*val;
   stat_mmctr(i)++;
}

void Stats :: multi_add(double *val)
{
   int i = m_ctr, j;

   for (j=0; j<multi_len; j++) {
     stat_mavg(i, stat_mmctr(i)) += val[j];
     stat_mstd(i, stat_mmctr(i)) += val[j]*val[j];
     stat_mmctr(i)++;
   }
}

void Stats :: single_add(double val)
{
  if (s_ctr >= num_single)
    {
      printf("Stats: single overflow: %d >= %d\n", s_ctr, num_single);
      exit(-1);
    }

  stat_savg(s_ctr) += val;
  stat_sstd(s_ctr) += val*val;
  s_ctr++;
}


// Add another entry without incrementing m_ctr.
void Stats :: multi_again(void)
{
  stat_mmctr(m_ctr) = 0;
}


// Increments m_ctr so as to start adding in next set.
void Stats :: multi_next(void)
{
  m_ctr++;
  stat_mmctr(m_ctr) = 0;
}

void Stats :: print(void)
{
  int i, j;
  double t, avg, std;

  printf("stats - print\n");

  t = (double)s_trials;
  if (t != 0.0)
    {
      for (i=0; i<num_single; i++)
	{
	  avg = stat_savg(i)/t;
	  if (t > 1.0)
	    std = sqrt((stat_sstd(i) - 
			avg * stat_savg(i))/(t-1.0));
	  else
	    std = 0.0;
	  printf("%.2f:%.2f\n", avg, std);
	}
      if (num_single && num_multi)
	printf("\n");
    }

  t = (double)m_trials;
  if (num_multi && (t != 0.0))
    {
      for (i=0; i<multi_len; i++)
	{
	  for (j=0; j<num_multi; j++)
	    {
	      avg = stat_mavg(j, i)/t;
	      if (t > 1.0)
		std = sqrt((stat_mstd(j,i) -
			    avg*stat_mavg(j,i))/(t-1.0));
	      else
		std = 0.0;
	      printf("%.2f:%.2f  ", avg, std);
	    }
	  printf("\n");
	}
    }
}   


void Stats :: send_stats(Stats *stats)
{
  int i, j;
  double t;

  if ((num_single > 0) && (s_trials > 0)) {
    t = (double)s_trials;
    stats->single_start();
    for (i=0; i<num_single; i++) {
      stats->single_add(stat_savg(i)/t);
    }
  }

  if ((num_multi > 0) && (m_trials > 0)) {
    t = (double)m_trials;
    stats->multi_start();
    stats->m_trials += 1;;
    for (i=0; i<num_multi; i++) {
      for (j=0; j<multi_len; j++) {
	stats->multi_add(stat_mavg(i,j)/t);	
      }
      stats->multi_next();
    }
  }
}


void Stats :: print_last(void)
{
  int i, j;
  double t, avg, std;

  t = (double)s_trials;
  if (t != 0.0) {
    printf("Trials: %.2f\n", t);
    i = num_single-1;
    avg = stat_savg(i)/t;
    if (t <= 1.0) {
      printf("%d %.2f\n", i, avg);
      
    } else {
      //      std = sqrt(stat_sstd(i)/t - avg*avg);
      std = (stat_sstd(i) - avg * stat_savg(i))/(t-1.0);
      if (std > 0.0) {
	std = sqrt(std);
      } else {
	std = 0.0;
      }
      printf("%d %.2f:%.2f\n", i, avg, std);
    }
  }

  if (num_single && num_multi)
    printf("\n");

  t = (double)m_trials;
  if (num_multi && (t != 0.0)) {
    printf("Trials: %.2f\n", t);
    j = num_multi-1;
    printf("%d  ", j);
    for (i=0; i<multi_len; i++) {
      avg = stat_mavg(j, i)/t;
      if (t <= 1.0) {
	printf("%.2f  ", avg);
      } else {
	//	std = sqrt(stat_mstd(j,i)/(t) - avg*avg);
	std = (stat_mstd(j,i) - avg*stat_mavg(j,i))/(t-1.0);
	if (std > 0.0) {
	  std = sqrt(std);
	} else {
	  std = 0.0;
	}
	printf("%.2f:%.2f  ", avg, std);
      }
    }
    printf("\n");
  }
}


void Stats :: print_formatted(void)
{
  int i, j;
  double t, avg, std;

  t = (double)s_trials;
  if (t != 0.0) {
    printf("Trials: %.2f\n", t);
    for (i=0; i<num_single; i++) {
      avg = stat_savg(i)/t;
      if (t <= 1.0) {
	printf("%d %.2f\n", i, avg);

      } else {
	//	    std = sqrt((stat_sstd(i) - 
	//			avg * stat_savg(i))/(t-1.0));
	//	std = sqrt(stat_sstd(i)/t - avg*avg);
	std = (stat_sstd(i) - avg * stat_savg(i))/(t-1.0);
	if (std > 0.0) {
	  std = sqrt(std);
	} else {
	  std = 0.0;
	}

	printf("%d %.2f:%.2f\n", i, avg, std);
      }
    }
    if (num_single && num_multi)
      printf("\n");
  }

  t = (double)m_trials;
  if (num_multi && (t != 0.0)) {
    printf("Trials: %.2f\n", t);
    for (j=0; j<num_multi; j++) {
      printf("%d  ", j);
      for (i=0; i<multi_len; i++) {
	avg = stat_mavg(j, i)/t;
	if (t <= 1.0) {
	  printf("%.2f  ", avg);
	} else {
	  // Commented one seems correct to.
	  //		std = sqrt((stat_mstd(j,i) -
	  //			    avg*stat_mavg(j,i))/(t-1.0));
	  //	  std = sqrt(stat_mstd(j,i)/(t) - avg*avg);

	  std = (stat_mstd(j,i) - avg*stat_mavg(j,i))/(t-1.0);
	  if (std > 0.0) {
	    std = sqrt(std);
	  } else {
	    std = 0.0;
	  }

	  printf("%.2f:%.2f  ", avg, std);
	}
      }
      printf("\n");
    }
  }

  /* old
  t = (double)m_trials;
  if (num_multi && (t != 0.0))
    {
      for (i=0; i<multi_len; i++)
	{
	  for (j=0; j<num_multi; j++)
	    {
	      avg = stat_mavg(j, i)/t;
	      if (t > 1.0)
		std = sqrt((stat_mstd(j,i) -
			    avg*stat_mavg(j,i))/(t-1.0));
	      else
		std = 0.0;
	      printf("%.2f:%.2f  ", avg, std);
	    }
	  printf("\n");
	}
    }
  */
}   



void Stats :: print_short_formatted(void)
{
  int i, j;
  double t, avg, std;

  t = (double)s_trials;
  if (t != 0.0) {
    printf("Trials: %.2f\n", t);
    for (i=0; i<num_single; i+=1) {
      avg = stat_savg(i)/t;
      if (t <= 1.0) {
	printf("%d %.2f\n", i, avg);

      } else {
	//	    std = sqrt((stat_sstd(i) - 
	//			avg * stat_savg(i))/(t-1.0));
	//	std = sqrt(stat_sstd(i)/t - avg*avg);
	std = (stat_sstd(i) - avg * stat_savg(i))/(t-1.0);
	if (std > 0.0) {
	  std = sqrt(std);
	} else {
	  std = 0.0;
	}

	printf("%d %.2f:%.2f\n", i, avg, std);
      }
    }
    if (num_single && num_multi)
      printf("\n");
  }

  t = (double)m_trials;
  if (num_multi && (t != 0.0)) {
    printf("Trials: %.2f\n", t);
    for (j=0; j<num_multi; j+=1) {
      printf("%d  ", j);
      for (i=0; i<multi_len; i++) {
	avg = stat_mavg(j, i)/t;
	if (t <= 1.0) {
	  printf("%.2f  ", avg);
	} else {
	  // Commented one seems correct to.
	  //		std = sqrt((stat_mstd(j,i) -
	  //			    avg*stat_mavg(j,i))/(t-1.0));
	  //	  std = sqrt(stat_mstd(j,i)/(t) - avg*avg);

	  std = (stat_mstd(j,i) - avg*stat_mavg(j,i))/(t-1.0);
	  if (std > 0.0) {
	    std = sqrt(std);
	  } else {
	    std = 0.0;
	  }

	  printf("%.2f:%.2f  ", avg, std);
	}
      }
      printf("\n");
    }
  }
}   

