/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include <values.h>
#include <stream.h>
#include <fstream.h>
#include "general.h"
#include "graphic.h"
#include "lsystem.h"
#include "pop.h"
#include "bars3d.h"
#include "algebra.h"
#include "network.h"

#define ANGLE_TOLERANCE 0.0002
#define SAME_AS_TOLERANCE 0.1
#define DIST_TOLERANCE 0.0001
#define DIST_COLLISION 5.0

#define MAX_SETTLE_ANGLE 1.5
//#define MAX_SETTLE_ANGLE 0.15
//#define MAX_SETTLE_ANGLE 0.088

#define GROUND_TOLERANCE_MAX 1.0
#define GROUND_TOLERANCE_MIN 0.00001

#define MAX_CYCLE_DEPTH 19
#define OFFSET_STEP_3D 0.125
#define MAX_ARM_LENGTH 4
#define MAX_LINE_LENGTH 3

#define SENSOR_ANGLE PI/9.0;

#define DRAW_STEP 0.015
#define TIME_STEP 0.0002
#define JOINT_MOVE_STEP 0.001  // 0.025

#define WAIT_CHECK_FAR 0.004
#define WAIT_CHECK_NEAR 0.0004

#define PARAM_2_WEIGHT 1.0   // 0.25
#define PARAM_2_ADJUST 1.0 // 0.25

#define NET_TRIAL_LEN 8.0
#define REMOTE_TRIAL_LEN 4.0

Real Ground_tolerance3d;


Real tries = 0.0;
Real passes = 0.0;
int builds = 0;
int Rotating = FALSE;
Real Max_settle_angle = MAX_SETTLE_ANGLE;



int lines_intersect_3d(Real l1_a[3], Real l1_b[3],
		       Real l2_a[3], Real l2_b[3])
{
  Real l1l2[3];
  Real l1[3], l2[3];
  Real norm[3];
  Real len;

  vec_subt(l1, l1_b, l1_a);
  vec_subt(l2, l2_b, l2_a);
  vec_cross(norm, l1, l2);

  len = sqrt(sqr(norm[0]) + sqr(norm[1]) + sqr(norm[2]));
  if (len == 0) {
    // Lines are parallel.
#ifndef OPTIMIZE
    if (Print[23])
      printf("Lines are parallel.\n");
#endif
    return FALSE;
  }

  norm[0] /= len;
  norm[1] /= len;
  norm[2] /= len;

  vec_subt(l1l2, l1_a, l2_a);
  len = vec_dot(l1l2, norm);

#ifndef OPTIMIZE
  if (Print[23]) {
    printf("l1 - l2 : %.4f  ", len);

    vec_subt(l1l2, l1_a, l2_b);
    len = vec_dot(l1l2, norm);
    printf("%.4f  ", len);

    vec_subt(l1l2, l1_b, l2_a);
    len = vec_dot(l1l2, norm);
    printf("%.4f  ", len);

    vec_subt(l1l2, l1_b, l2_b);
    len = vec_dot(l1l2, norm);
    printf("%.4f\n", len);
  }
#endif  
  return TRUE;
}


// M = R1 * R2 * R3;
void make_identity_mat(Real mat[3][3])
{
  mat[0][0] = 1.0;
  mat[0][1] = 0.0;
  mat[0][2] = 0.0;

  mat[1][0] = 0.0;
  mat[1][1] = 1.0;
  mat[1][2] = 0.0;

  mat[2][0] = 0.0;
  mat[2][1] = 0.0;
  mat[2][2] = 1.0;
}


void make_rotX_mat(Real angle, Real mat[3][3])
{
  Real cos_ang = cos(angle);
  Real sin_ang = sin(angle);

  mat[0][0] = 1.0;
  mat[0][1] = 0.0;
  mat[0][2] = 0.0;

  mat[1][0] = 0.0;
  mat[1][1] = cos_ang;
  mat[1][2] = -sin_ang;

  mat[2][0] = 0.0;
  mat[2][1] = sin_ang;
  mat[2][2] = cos_ang;
}

void make_rotY_mat(Real angle, Real mat[3][3])
{
  Real cos_ang = cos(angle);
  Real sin_ang = sin(angle);

  mat[0][0] = cos_ang;
  mat[0][1] = 0.0;
  mat[0][2] = sin_ang;

  mat[1][0] = 0.0;
  mat[1][1] = 1.0;
  mat[1][2] = 0.0;

  mat[2][0] = -sin_ang;
  mat[2][1] = 0.0;
  mat[2][2] = cos_ang;

  /*
  mat[0][0] = cos_ang;
  mat[0][1] = 0.0;
  mat[0][2] = -sin_ang;

  mat[1][0] = 0.0;
  mat[1][1] = 1.0;
  mat[1][2] = 0.0;

  mat[2][0] = sin_ang;
  mat[2][1] = 0.0;
  mat[2][2] = cos_ang;
  */
}

void make_rotZ_mat(Real angle, Real mat[3][3])
{
  Real cos_ang = cos(angle);
  Real sin_ang = sin(angle);

  mat[0][0] = cos_ang;
  mat[0][1] = -sin_ang;
  mat[0][2] = 0.0;

  mat[1][0] = sin_ang;
  mat[1][1] = cos_ang;
  mat[1][2] = 0.0;

  mat[2][0] = 0.0;
  mat[2][1] = 0.0;
  mat[2][2] = 1.0;
}




BarStruct3D :: BarStruct3D(void)
{
  ground_pt = (int*)malloc(MAX_BARS3*sizeof(int));
  ground_pt_prev = (int*)malloc(MAX_BARS3*sizeof(int));
  network = 0;
  Is_ball = FALSE;
}


void BarStruct3D :: initialize_network(void)
{
  if (network == 0) {
    network = new Network();
  } else {
    network->clear();
  }
}


void BarStruct3D :: initialize(void)
{
  Bar_stack_ptr = 0;
  Num_bars = 1;
  Num_joints = 0;
  Num_sensors = 0;
  ref_pt = 0;

  Time_step = TIME_STEP;
  Time = 0.0;

  num_grnd_pts = 0;
  num_grnd_pts_prev = 0;

  Bar[0].type = BAR_TYPE_FIXED;
  Bar[0].moves = FALSE;
  Bar[0].angle_a = 0.0;
  Bar[0].length = 0;
  Bar[0].end_loc[0] = 0.0;
  Bar[0].end_loc[1] = 0.0;
  Bar[0].end_loc[2] = 0.0;
  Bar[0].abs_angle = 0.0;
  Bar[0].parent = 0;
  Bar[0].moves = FALSE;
  Bar[0].same_as = -1;
  Bar[0].joint_index = 0;
  Bar[0].num_children = 0;
  Bar[0].num_from = 0;
  Bar[0].weight = 1.0;
  Bar[0].sensor_value = 0.0;
  make_identity_mat(Bar[0].rotation_matrix);
  make_identity_mat(Bar[0].abs_heading);
}

void BarStruct3D :: set_translate(void)
{
  if (COM[0] - draw_loc[0] > DRAW_TOLERANCE) {
    draw_loc[0] = COM[0] - DRAW_TOLERANCE;
  } else if (draw_loc[0] - COM[0] > DRAW_TOLERANCE) {
    draw_loc[0] = COM[0] + DRAW_TOLERANCE;
  }

  if (COM[2] - draw_loc[2] > DRAW_TOLERANCE) {
    draw_loc[2] = COM[2] - DRAW_TOLERANCE;
  } else if (draw_loc[2] - COM[2] > DRAW_TOLERANCE) {
    draw_loc[2] = COM[2] + DRAW_TOLERANCE;
  }

  win1->translate(-draw_loc[0], 0.0, -draw_loc[2]);
}


void BarStruct3D :: draw_struct_bw(void)
{
  int i;

  // Draw creature.
  win1->SetLineWidth(2.0);
  win1->startLine();
  win1->SetColor(col_side3);
  for (i=1; i<Num_bars; i++) {
    if (Bar[i].num_children > 0) {
      win1->addVertex3v(Bar[Bar[i].parent].end_loc);
      win1->addVertex3v(Bar[i].end_loc);
    }
  }

  win1->SetColor(col_green2);
  for (i=1; i<Num_bars; i++) {
    if (Bar[i].num_children == 0) {
      win1->addVertex3v(Bar[Bar[i].parent].end_loc);
      win1->addVertex3v(Bar[i].end_loc);
    }

    win1->starLocation(Bar[i].end_loc[0], Bar[i].end_loc[1],
		       Bar[i].end_loc[2]);
  }

  win1->stopLine();
}



void BarStruct3D :: draw_struct_tree(void)
{
  int i, ctr;
  Real x, z;

  set_translate();

  if (!Option[10]) { // option[10 => blind watchmaker
    // Draw surface.
    /*
    if (Option[8] != 0)
      win1->SetColor(col_surf3);
    else
      win1->SetColor(col_surf2);
    */

    if (Option[8] != 10) {
      win1->SetColor(col_surf3);
      win1->drawPlaneY(21000.0, -1.0, 21000.0);
    
      if (Option[8] != 3)
	win1->SetLineWidth(1.0);
      else
	win1->SetLineWidth(2.0);    

      win1->startLine();
      // Stripes on road.
      ctr = 0;
      for (x=-10036.0; x<=10000.0; x+=100.0) {
	ctr++;
	if ((ctr%5) == 1)
	  win1->SetColor(col_stripe1);
	else if ((ctr%5) == 2)
	  win1->SetColor(col_stripe2);
	win1->addVertex(x, 0.0, -10000.0);
	win1->addVertex(x, 0.0, 10000.0);
      }
      ctr = 0;
      for (z=-10021.0; z<=10000.0; z+=100.0) {
	ctr++;
	if ((ctr%5) == 1)
	  win1->SetColor(col_stripe1);
	else if ((ctr%5) == 2)
	  win1->SetColor(col_stripe2);
	win1->addVertex(-10000, 0.0, z);
	win1->addVertex(10000, 0.0, z);
      }
      win1->stopLine();

      // Draw shadow.
      if (!Option[8])
	win1->SetColor(col_grey4);
      else
	win1->SetColor(col_grey6);

      if (Option[8] != 3)
	win1->SetLineWidth(4.0);
      else
	win1->SetLineWidth(4.0);
      win1->startLine();


      win1->starShadowLocation(Bar[0].end_loc[0], 0.01,
			       Bar[0].end_loc[2]);
      for (i=1; i<Num_bars; i++) {
	win1->starShadowLocation(Bar[i].end_loc[0], 0.01,
				 Bar[i].end_loc[2]);
	win1->addVertex(Bar[Bar[i].parent].end_loc[0], 0.01,
			Bar[Bar[i].parent].end_loc[2]);
	win1->addVertex(Bar[i].end_loc[0], 0.01, Bar[i].end_loc[2]);
      }
    }
  }    

  if (Option[10]) {
    win1->startLine();
  } else if (Option[8] == 10) {
    win1->SetLineWidth(3.0);
    //    printf("set line width.\n");
    win1->startLine();
  }

  // Draw creature.
  if (Option[8] != 10) {
    win1->SetColor(col_side3);
  } else {
    //    win1->SetColor(col_black);
    win1->SetColor(col_side3);
  }
  for (i=1; i<Num_bars; i++) {
    if (Bar[i].num_children > 0) {
      win1->addVertex3v(Bar[Bar[i].parent].end_loc);
      win1->addVertex3v(Bar[i].end_loc);
    }
  }

  if (Option[8] != 10) {
    win1->SetColor(col_green3);
  } else {
    //    win1->SetColor(col_black);
    win1->SetColor(col_green3);
  }
  for (i=1; i<Num_bars; i++) {
    if (Bar[i].num_children == 0) {
      win1->addVertex3v(Bar[Bar[i].parent].end_loc);
      win1->addVertex3v(Bar[i].end_loc);
    }

    win1->starLocation(Bar[i].end_loc[0], Bar[i].end_loc[1],
		       Bar[i].end_loc[2]);
  }

  win1->stopLine();
}



void BarStruct3D :: draw_struct(void)
{
  int i, index, ctr;
  Real x, z;

  set_translate();


  if (Is_ball) {
    // Draw ball.
    win1->push_orientation();
    win1->translate(Ball_loc[0], Ball_radius, Ball_loc[2]);
    win1->drawSphere(Ball_radius, col_white, col_red);
    win1->pop_orientation();
  }


  if (!Option[10]) {
    // Draw surface.
    /*
    if (Option[8] != 0)
      win1->SetColor(col_surf3);
    else
      win1->SetColor(col_surf2);
    */

    if (Option[8] != 10) {
      if (Option[8] != 11) {
	win1->SetColor(col_surf3);
      } else {
	win1->SetColor(col_surf2);
      }
      win1->drawPlaneY(21000.0, -1.0, 21000.0);

      if (Option[8] != 11) {
	if (Option[8] != 3)
	  win1->SetLineWidth(1.0);
	else
	  win1->SetLineWidth(2.0);
	
	win1->startLine();
	// Stripes on road.
	ctr = 0;
	for (x=-10036.0; x<=10000.0; x+=100.0) {
	  ctr++;
	  if ((ctr%5) == 1)
	    win1->SetColor(col_stripe1);
	  else if ((ctr%5) == 2)
	    win1->SetColor(col_stripe2);
	  win1->addVertex(x, -0.2, -10000.0);
	  win1->addVertex(x, -0.2, 10000.0);
	}
	ctr = 0;
	for (z=-10021.0; z<=10000.0; z+=100.0) {
	  ctr++;
	  if ((ctr%5) == 1)
	    win1->SetColor(col_stripe1);
	  else if ((ctr%5) == 2)
	    win1->SetColor(col_stripe2);
	  win1->addVertex(-10000, -0.02, z);
	  win1->addVertex(10000, -0.02, z);
	}

	win1->stopLine();
      }
    }

    if (Terminal_set == 28) {
      int i;
      int dir = -1;
      for (i=0; i<4; i++) {
	if (network->Neuron_value[network->Neuron_input_list[i]] == 1.0) {
	  dir = i;
	  break;
	}
      }

      win1->SetLineWidth(3.0);
      if (dir == -1) {
	/*
	printf("draw_struct() - Error: invalid direction.\n");
	printf("%f %f %f %f\n", 
	       network->Neuron_value[network->Neuron_input_list[0]],
	       network->Neuron_value[network->Neuron_input_list[1]],
	       network->Neuron_value[network->Neuron_input_list[2]],
	       network->Neuron_value[network->Neuron_input_list[3]]);
	*/
	;

      } else if (dir == 0) {
	win1->SetColor(col_red);
	win1->startPolygon();
	win1->addVertex(COM[0]+100.0, 2.0, COM[2]-10.0);
	win1->addVertex(COM[0]+130.0, 2.0, COM[2]);
	win1->addVertex(COM[0]+100.0, 2.0, COM[2]+10.0);
	win1->stopList();

      } else if (dir == 1) {
	win1->SetColor(col_red);
	win1->startPolygon();
	win1->addVertex(COM[0]-100.0, 2.0, COM[2]-10.0);
	win1->addVertex(COM[0]-130.0, 2.0, COM[2]);
	win1->addVertex(COM[0]-100.0, 2.0, COM[2]+10.0);
	win1->stopList();

      } else if (dir == 2) {
	win1->SetColor(col_red);
	win1->startPolygon();
	win1->addVertex(COM[0]-10.0, 2.0, COM[2]+100.0);
	win1->addVertex(COM[0]+10.0, 2.0, COM[2]+100.0);
	win1->addVertex(COM[0], 2.0, COM[2]+130.0);
	win1->stopList();

      } else {
	win1->SetColor(col_red);
	win1->startPolygon();
	win1->addVertex(COM[0]-10.0, 2.0, COM[2]-100.0);
	win1->addVertex(COM[0]+10.0, 2.0, COM[2]-100.0);
	win1->addVertex(COM[0], 2.0, COM[2]-130.0);
	win1->stopList();
      }
    }

    // Draw center of mass.
    /*
      win1->SetLineWidth(2.0);
      win1->startLine();
      win1->SetColor(col_red);
      win1->starLocation(COM[0], COM[1], COM[2]);
      win1->stopLine();
    */

    // Draw shadow.
    if (!Option[8])
      win1->SetColor(col_grey4);
    else if (Option[8] == 10) {
      win1->SetColor(col_grey10);
    } else
      win1->SetColor(col_grey6);

    if (Option[8] != 10) {
      win1->SetLineWidth(5.0);
    } else {
      win1->SetLineWidth(8.0);
    }

    win1->startLine();
    for (i=0; i<Num_bars; i++) {
      if (Bar[i].moves) {
	win1->starShadowLocationBig(Bar[i].end_loc[0], -0.01,
				    Bar[i].end_loc[2]);
      }
    }
    win1->stopLine();


    if (Option[8] == 10) {
      win1->SetLineWidth(6.0);
    } else if (Option[8] != 3) {
      win1->SetLineWidth(4.0);
    } else {
      win1->SetLineWidth(6.0);
    }
    win1->startLine();

    win1->starShadowLocation(Bar[0].end_loc[0], -0.01,
			     Bar[0].end_loc[2]);

    for (i=1; i<Num_bars; i++) {
      if (!Bar[i].moves) {
	win1->starShadowLocation(Bar[i].end_loc[0], -0.01,
				 Bar[i].end_loc[2]);
      }

      win1->addVertex(Bar[Bar[i].parent].end_loc[0], -0.01,
		      Bar[Bar[i].parent].end_loc[2]);
      win1->addVertex(Bar[i].end_loc[0], 0.01, Bar[i].end_loc[2]);
    }
  }

  if (Option[10])
    win1->startLine();

  // Draw creature.
  if ((Option[8] == 0) || (Option[8] == 10))
    win1->SetColor(col_black);
  else {
    if (Option[14] == 1) {
      win1->SetColor(col_blue3);
    } else {
      win1->SetColor(col_blue);
    }
  }

  for (i=0; i<Num_bars; i++) {
    if (Bar[i].same_as >= 0)
      continue;

    if (Bar[i].ground_pt) {
      win1->starLocation(Bar[i].end_loc[0], Bar[i].end_loc[1],
			 Bar[i].end_loc[2]);
    }
  }

  // Fixed bars:
  if (Option[8] == 10) {
    win1->SetColor(col_black);
  } else if (Option[14] == 1) {
    win1->SetColor(col_blue3);
  } else {
    win1->SetColor(col_blue);
  }
  if (!Bar[0].ground_pt) {
    win1->starLocation(Bar[0].end_loc[0], Bar[0].end_loc[1],
		       Bar[0].end_loc[2]);
  }

  for (i=0; i<Num_bars; i++) {
    if (!Bar[i].moves) {
      win1->addVertex3v(Bar[Bar[i].parent].end_loc);
      win1->addVertex3v(Bar[i].end_loc);
    }

    if (Bar[i].same_as >= 0)
      continue;

    if ((Bar[i].num_children==0) || 
	(Bar[i].type == BAR_TYPE_FIXED))
      if (!Bar[i].ground_pt) {
	win1->starLocation(Bar[i].end_loc[0], Bar[i].end_loc[1],
			   Bar[i].end_loc[2]);
      }
  }

  // Moving bars:
  if (Option[8] == 10) {
    win1->SetColor(col_black);
  } else  if (Option[8] != 4) {
    win1->SetColor(col_green6); // col_green4
  } else {
    if (Option[14] == 1) {
      win1->SetColor(col_blue3);
    } else {
      win1->SetColor(col_blue);
    }
  }
  for (i=0; i<Num_bars; i++) {
    if (Bar[i].moves) {
      win1->addVertex3v(Bar[Bar[i].parent].end_loc);
      win1->addVertex3v(Bar[i].end_loc);
    }
  }


  double col_sensor[3] = {0.35, 0.0, 0.35};
  for (i=0; i<Num_sensors; i++) {
    index = Sensor_index[i];
    col_sensor[0] = Bar[index].sensor_value;
    win1->SetColor(col_sensor);
    win1->starLocation(Bar[index].end_loc[0], Bar[index].end_loc[1],
		       Bar[index].end_loc[2]);
  }
  win1->stopLine();

  win1->SetLineWidth(6.0);
  if (Option[8] == 10) {
    win1->SetColor(col_black);
  } else  if (Option[8] != 4) {
    win1->SetColor(col_green5);
  } else {
    if (Option[14] == 1) {
      win1->SetColor(col_blue3);
    } else {
      win1->SetColor(col_blue);
    }
  }
  win1->startLine();
  for (i=0; i<Num_bars; i++) {
    if (Bar[i].same_as >= 0)
      continue;
    
    if (Bar[i].moves)
      if (!Bar[i].ground_pt) {
	win1->starLocationBig(Bar[i].end_loc[0], Bar[i].end_loc[1],
			      Bar[i].end_loc[2]);
      }
  }
  win1->stopLine();
}


void BarStruct3D :: draw_bars(int arg)
{
  if (win1 == 0) {
    printf("win1 not defined\n");
    return;
  }

  if (Draw_arg[4]) {
    printf("%.4f: %.2f %.2f %.2f\n", Time, COM[0], COM[1], COM[2]);
    Draw_arg[4] = FALSE;
  }


  if (Draw_arg[2]) {
    Time += Time_step;
    calculate_positions(Time);
    settle_structure();

  } else if (Draw_arg[1]) {
    Real prev_time = Time;
    while (Time < prev_time + DRAW_STEP) {
      Time += Time_step;

#ifndef OPTIMIZE
      if (Print[12]) {
	printf("Draw_bars :: update positions and settle...\n");
      }
#endif
      calculate_positions(Time);
      if (Is_ball) {
	update_ball_position();
      }
      if (Terminal_set == 28) {
	// Update network.
	Real time = Time;
	time = floor(Time / 5.0) * 5.0;
	if (Time - time < Time_step) {
	  randomize_remote();
	}
      }

      if (Print[34]) {
	printf("Calculated positions (ref: %d)...\n", ref_pt);
	print_struct();
	printf("Recalc positions: %.4f\n", Time);
	calculate_positions_new(Time);
      }
      if (!settle_structure()) {
	printf("draw_bars() - Bars failed to settle.\n");
      }
      if (Print[34]) {
	printf("Calculated positions (ref: %d)...\n", ref_pt);
	print_struct();
	printf("Recalc positions: %.4f\n", Time);
	calculate_positions_new(Time);
      }
    }
  }

  if (!Option[10]) {
    if (Build_type != 10)
      draw_struct();
    else
      draw_struct_tree();
  } else
    draw_struct_bw();
}


// Originally 0.15
#define JOINT_NOISE 0.2

// Does not add connection noise between i and j if the link i-j is
// part of a cycle.
void BarStruct3D :: add_connection_noise(void)
{
  int i, j, index;
  Real tmp_mat1[3][3], tmp_mat2[3][3], tmp_mat3[3][3];
  Real tmp_mat4[3][3], noise_rot[3][3];


  for (i=0; i<Num_bars; i++) {
    for (j=0; j<Bar[i].num_children; j++) {
      if (Bar[i].in_cycle[j]) {
	continue;
      }

      index = Bar[i].child[j];
      Bar[index].angle_a = random_dbl()*JOINT_NOISE - (JOINT_NOISE/2.0) + Bar[index].true_angle_a;
      if (Bar[index].moves)
	Bar[index].angle_b = random_dbl()*JOINT_NOISE - (JOINT_NOISE/2.0) + Bar[index].true_angle_b;

      // Add noise to the rotation.
      //      make_rotX_mat(random_dbl()*0.04-0.02, tmp_mat1);
      //      make_rotY_mat(random_dbl()*0.04-0.02, tmp_mat2);
      //      make_rotZ_mat(random_dbl()*0.04-0.02, tmp_mat3);      

      make_rotX_mat(random_dbl()*JOINT_NOISE - (JOINT_NOISE/2.0), tmp_mat1);
      make_rotY_mat(random_dbl()*JOINT_NOISE - (JOINT_NOISE/2.0), tmp_mat2);
      make_rotZ_mat(random_dbl()*JOINT_NOISE - (JOINT_NOISE/2.0), tmp_mat3);      

      if (random_int(2)) {
	mmm3_mult(tmp_mat4, tmp_mat1, tmp_mat2);
	mmm3_mult(noise_rot, tmp_mat4, tmp_mat3);
      } else {
	mmm3_mult(tmp_mat4, tmp_mat3, tmp_mat2);
	mmm3_mult(noise_rot, tmp_mat4, tmp_mat1);
      }

      mmm3_mult(Bar[i].delta_rot_mat, noise_rot,
		Bar[i].true_delta_rot);
      mm3_transpose(Bar[i].inverse_delta_rot, Bar[i].delta_rot_mat);
    }
  }
}


// Does not add connection noise between i and j if either is in a cycle.
void BarStruct3D :: add_connection_noise2(void)
{
  int i, j, index;
  Real tmp_mat1[3][3], tmp_mat2[3][3], tmp_mat3[3][3];
  Real tmp_mat4[3][3], noise_rot[3][3];


  for (i=0; i<Num_bars; i++) {
    if (Bar[i].in_cycle2)
      continue;

    for (j=0; j<Bar[i].num_children; j++) {
      index = Bar[i].child[j];
      //      if (Bar[index].in_cycle2)
      //	continue;

      Bar[index].angle_a = random_dbl()*JOINT_NOISE - (JOINT_NOISE/2.0) + Bar[index].true_angle_a;
      if (Bar[index].moves)
	Bar[index].angle_b = random_dbl()*JOINT_NOISE - (JOINT_NOISE/2.0) + Bar[index].true_angle_b;

      // Add noise to the rotation.
      //      make_rotX_mat(random_dbl()*0.04-0.02, tmp_mat1);
      //      make_rotY_mat(random_dbl()*0.04-0.02, tmp_mat2);
      //      make_rotZ_mat(random_dbl()*0.04-0.02, tmp_mat3);      

      make_rotX_mat(random_dbl()*JOINT_NOISE - (JOINT_NOISE/2.0), tmp_mat1);
      make_rotY_mat(random_dbl()*JOINT_NOISE - (JOINT_NOISE/2.0), tmp_mat2);
      make_rotZ_mat(random_dbl()*JOINT_NOISE - (JOINT_NOISE/2.0), tmp_mat3);      

      if (random_int(2)) {
	mmm3_mult(tmp_mat4, tmp_mat1, tmp_mat2);
	mmm3_mult(noise_rot, tmp_mat4, tmp_mat3);
      } else {
	mmm3_mult(tmp_mat4, tmp_mat3, tmp_mat2);
	mmm3_mult(noise_rot, tmp_mat4, tmp_mat1);
      }

      mmm3_mult(Bar[i].delta_rot_mat, noise_rot,
		Bar[i].true_delta_rot);
      mm3_transpose(Bar[i].inverse_delta_rot, Bar[i].delta_rot_mat);
    }
  }
}


int BarStruct3D :: max_line_length(void)
{
  int i, j, k, index_stack, index_new, stack_ptr, max_length;
  int stack[MAX_BARS3];
  // int stack[MAX_LINE_LENGTH+1];

  for (i=0; i<Num_bars; i++)
    Bar[i].line_length = 0;

  max_length = 0;
  for (i=1; i<Num_bars; i++) {
    if (Bar[i].in_cycle2) {
      Bar[i].line_length = 0;
      continue;
    }

    //    printf("Line length: %d\n", i);
    stack_ptr = 1;
    stack[0] = i;
    Bar[i].updated = 0;

    for (j=0; j<stack_ptr; j++) {
      index_stack = stack[j];
      for (k=0; k<Bar[index_stack].num_children; k++) {
	index_new = Bar[stack[j]].child[k];
	if (Bar[index_new].in_cycle2)
	  continue;
	
	Bar[index_new].updated = Bar[index_stack].updated+1;
	stack[stack_ptr++] = index_new;
	//	printf("Line: %d -> %d (length: %d) (%d:%d)\n", index_stack, index_new,
	//	       Bar[index_new].updated, stack_ptr, Bar[stack_ptr-1].updated);
      }
    }

    Bar[i].line_length = Bar[stack[stack_ptr-1]].updated;
  }

  // First bar is special.
  if (Bar[0].in_cycle2) {
    Bar[0].line_length = 0;

  } else if (Bar[0].num_children == 0) {
    Bar[0].line_length = 0;

  } else if (Bar[0].num_children == 1) {
    Bar[0].line_length = Bar[Bar[0].child[0]].line_length + 1;

  } else {
    int len1 = Bar[Bar[0].child[0]].line_length;
    int len2 = Bar[Bar[0].child[1]].line_length;
    int len3;
    if (len2 > len1) {
      int len3 = len1;
      len1 = len2;
      len2 = len3;
    }

    for (i=2; i<Bar[0].num_children; i++) {
      len3 = Bar[Bar[0].child[i]].line_length;
      if (len3 > len1) {
	len2 = len1;
	len1 = len3;
      } else if (len3 > len2) {
	len2 = len3;
      }
    }
    Bar[0].line_length = len1 + len2 + 2;
  }


  // Calculate maximum length of a line.
  max_length = Bar[0].line_length;
  for (i=1; i<Num_bars; i++) {
    if (Bar[i].line_length > max_length)
      max_length = Bar[i].line_length;
  }


  if (Print[26]) {
    printf("Line length:\n");
    for (i=0; i<Num_bars; i++) {
      printf("%d: %d\n", i, Bar[i].line_length);
    }
  }


  return max_length;
}




// Test an arm for length.
int BarStruct3D :: arm_length(int node, int depth)
{
  int length, max_length = depth;
  int i, child, parent;

  Bar[node].updated = depth;


  // Test same_as.
  if (Bar[node].same_as != -1) {
    child = Bar[node].same_as;
    if (Bar[child].updated == 0) {
      length = arm_length(Bar[node].same_as, depth);
      if (length > max_length)
	max_length = length;
    }
  }

  // Test parent.
  parent = Bar[node].parent;
  if (Bar[parent].updated == 0) {
    length = arm_length(parent, depth+1);
    if (length > max_length)
      max_length = length;
  }
  
  // Test children.
  for (i=0; i<Bar[node].num_children; i++) {
    child = Bar[node].child[i];
    if (Bar[child].updated == 0) {
      length = arm_length(child, depth+1);
      if (length > max_length)
	max_length = length;
    }
  }

  // Test from list.
  for (i=0; i<Bar[node].num_from; i++) {
    child = Bar[node].from_list[i];
    if (Bar[child].updated == 0) {
      length = arm_length(child, depth+1);
      if (length > max_length)
	max_length = length;
    }
  }

  Bar[node].updated = 0;
  return max_length;
}


// Returns the length of the longest arm.
int BarStruct3D :: max_arm_length(void)
{
  int i, j;
  int index_j;
  int prev_i, prev_j;
  int max_length = 1;
  int length1, length2;

  // Initialize data structures.
  for (i=0; i<Num_bars; i++) {
    Bar[i].updated = 0; // distance.
    for (j=0; j<Bar[i].num_children; j++) {
      if (Bar[i].in_cycle[j]) {
	Bar[i].updated = 1;
	Bar[Bar[i].child[j]].updated = 1;
      }
    }

    if (Bar[i].same_as != -1) {
      if (Bar[Bar[i].same_as].updated == 1)
	Bar[i].updated = 1;
      else if (Bar[i].updated == 1)
	Bar[Bar[i].same_as].updated = 1;
    }
  }

#ifndef OPTIMIZE
  if (Print[25]) {
    for (i=0; i<Num_bars; i++) {
      printf("%d(%d): ", i,Bar[i].updated);
      if (Bar[i].same_as == -1) {
	for (j=0; j<Bar[i].num_children; j++) {
	  printf("%d(%d)  ", Bar[i].child[j], Bar[i].in_cycle[j]);
	}
	if (Bar[i].num_from > 0) {
	  printf(":  ");
	  for (j=0; j<Bar[i].num_from; j++) {
	    printf("%d  ", Bar[i].from_list[j]);
	  }
	}
	
      } else {
	printf(" = %d", Bar[i].same_as);
      }
      printf("\n");
    }
  }
#endif


  // Test nodes.
  for (i=0; i<Num_bars; i++) {
    // Test all children.
    for (j=0; j<Bar[i].num_children; j++) {
      if (Bar[i].in_cycle[j]) {
	continue; // Don't need to test length of edge in cycle.
      }

      index_j = Bar[i].child[j];

      prev_i = Bar[i].updated;
      prev_j = Bar[index_j].updated;

      Bar[i].updated = 1;
      length1 = arm_length(index_j, 2);
      Bar[index_j].updated = 1;
      length2 = arm_length(i, 2);

      Bar[i].updated = prev_i;
      Bar[index_j].updated = prev_j;

#ifndef OPTIMIZE
      if (Print[25]) {
	printf("Arm length: %d - %d(%d):  %d, %d\n", i, index_j, j,
	       length1, length2);
      }
#endif

      if (length1 > length2) {
	if (length2 > max_length)
	  max_length = length2;
      } else {
	if (length1 > max_length)
	  max_length = length1;
      }
    }
  }


#ifndef OPTIMIZE
  if (Print[25]) {
    for (i=0; i<Num_bars; i++) {
      printf("%d(%d): ", i,Bar[i].updated);
      if (Bar[i].same_as == -1) {
	for (j=0; j<Bar[i].num_children; j++) {
	  printf("%d(%d)  ", Bar[i].child[j], Bar[i].in_cycle[j]);
	}
	if (Bar[i].num_from > 0) {
	  printf(":  ");
	  for (j=0; j<Bar[i].num_from; j++) {
	    printf("%d  ", Bar[i].from_list[j]);
	  }
	}
	
      } else {
	printf(" = %d", Bar[i].same_as);
      }
      printf("\n");
    }
  }
#endif

#ifndef OPTIMIZE
  if (Print[25]) {
    printf("Max Arm length: %d\n", max_length);
  }
#endif

  return max_length;
}


// Test to see if arms (non-cycle edges) are too long.
int BarStruct3D :: test_arm_lengths(void)
{
  int i, j;
  int index_j;
  int length1, length2;

  // Initialize data structures.
  for (i=0; i<Num_bars; i++) {
    Bar[i].updated = 0; // distance.
  }


  // Test nodes.
  for (i=0; i<Num_bars; i++) {
    // Test all children.
    for (j=0; j<Bar[i].num_children; j++) {
      if (Bar[i].in_cycle[j]) {
	continue; // Don't need to test length of edge in cycle.
      }

      index_j = Bar[i].child[j];

      Bar[i].updated = 1;
      length1 = arm_length(index_j, 2);
      Bar[index_j].updated = 1;
      length2 = arm_length(i, 2);
      Bar[index_j].updated = 0;

#ifndef OPTIMIZE
      if (Print[25]) {
	printf("Arm length: %d - %d(%d):  %d, %d\n", i, index_j, j,
	       length1, length2);
      }
#endif

      if ((length1 > MAX_ARM_LENGTH) && (length2 > MAX_ARM_LENGTH)) {
#ifndef OPTIMIZE
	if (Print[9]) {
	  printf("Arm too long: %d - %d(%d):  %d,%d\n", i, index_j, j,
		 length1, length2);
	}
#endif
	return FALSE;
      }
    }
  }

  return TRUE;
}



int BarStruct3D :: test_children(int node, int root, int depth)
{
  if (depth > MAX_CYCLE_DEPTH)
    return FALSE;

  int i, child, parent;

  if (Bar[node].same_as != -1) {
    if (Bar[node].same_as == root) {
      return TRUE;
    }

    if (Bar[Bar[node].same_as].updated == 0) {
      Bar[node].updated = depth + 1;
      Bar[Bar[node].same_as].updated = depth + 1;
      int ret_val = test_children(Bar[node].same_as, root, depth);
      Bar[node].updated = 0;
      Bar[Bar[node].same_as].updated = 0;
      return ret_val;
    }
  }

  // Test parent.
  parent = Bar[node].parent;
  if (depth > 2) {
    if (parent == root) {
      // Cycle!
      return TRUE;
    }
  }
  if (Bar[parent].updated == 0) {
    Bar[parent].updated = depth + 1;
    if (test_children(parent, root, depth+1)) {
      // Cycle!
      Bar[parent].updated = 0;
      return TRUE;
    }
    Bar[parent].updated = 0;
  }

  // Test children.
  for (i=0; i<Bar[node].num_children; i++) {
    child = Bar[node].child[i];
    if (child == root) {
      Bar[node].in_cycle[i] = TRUE;
      Bar[node].in_cycle2 = TRUE;
      Bar[child].in_cycle2 = TRUE;
      //      printf("In cycle: %d & %d\n", node, child);
      return TRUE;
    }

    if (Bar[child].updated != 0)
      continue;

    Bar[child].updated = depth + 1;
    if (test_children(child, root, depth+1)) {
      // cycle!
      Bar[child].updated = 0;
      Bar[node].in_cycle[i] = TRUE;
      Bar[node].in_cycle2 = TRUE;
      Bar[child].in_cycle2 = TRUE;
      //      printf("In cycle: %d & %d\n", node, child);
      return TRUE;
    }
    Bar[child].updated = 0;
  }

  // Test from list.
  for (i=0; i<Bar[node].num_from; i++) {
    child = Bar[node].from_list[i];
    if (depth > 2) {
      if (child == root) {
	return TRUE;
      }
    }

    if (Bar[child].updated != 0)
      continue;

    Bar[child].updated = depth + 1;
    if (test_children(child, root, depth+1)) {
      // cycle!
      Bar[child].updated = 0;
      return TRUE;
    }
    Bar[child].updated = 0;
  }
    
  return FALSE;
}



void BarStruct3D :: label_cycles(void)
{
  int i, j;
  int index_j;

  // Initialize data structures.
  for (i=0; i<Num_bars; i++) {
    Bar[i].updated = 0; // distance.
    Bar[i].in_cycle2 = FALSE;
    for (j=0; j<Bar[i].num_children; j++) {
      Bar[i].in_cycle[j] = FALSE;
    }
  }

  // Find cycles.
  for (i=0; i<Num_bars; i++) {
    Bar[i].updated = 1; // distance.

    // Test all children.
    for (j=0; j<Bar[i].num_children; j++) {
      if (Bar[i].in_cycle[j]) {
	// If already in cycle, don't need to check.
	continue;
      }

      index_j = Bar[i].child[j];
      Bar[index_j].updated = 2;
      if (test_children(index_j, i, 2)) {
	Bar[i].in_cycle[j] = TRUE;
	Bar[i].in_cycle2 = TRUE;
	Bar[index_j].in_cycle2 = TRUE;
	//	printf("In cycle: %d & %d\n", i, index_j);
      }
      Bar[index_j].updated = 0;
    }
    Bar[i].updated = 0;
  }

#ifndef OPTIMIZE
  if (Print[25]) {
    printf("Labelling cycles:\n");
    for (i=0; i<Num_bars; i++) {
      printf("%d (%d): ", i, Bar[i].in_cycle2);
      if (Bar[i].same_as == -1) {
	for (j=0; j<Bar[i].num_children; j++) {
	  printf("%d(%d)  ", Bar[i].child[j], Bar[i].in_cycle[j]);
	}
	if (Bar[i].num_from > 0) {
	  printf(":  ");
	  for (j=0; j<Bar[i].num_from; j++) {
	    printf("%d  ", Bar[i].from_list[j]);
	  }
	}
	
      } else {
	printf(" = %d", Bar[i].same_as);
      }
      printf("\n");
    }
  }
#endif
}


void BarStruct3D :: initialize_intersection(void)
{
  int i, j;
  int pi, pj, jt_i, jt_j;
  int near;

  last_check_near = 0.0;
  last_check_far = 0.0;

  for (i=1; i<Num_bars-1; i++) {
    pi = Bar[i].parent;
    jt_i = Bar[i].joint_index;

    near = 0;
    for (j=i+1; j<Num_bars; j++) {
      jt_j = Bar[j].joint_index;
      if (jt_i == jt_j) {
	// Part of same joint structure.
	bar_check[i][j] = 0;
	continue;
      }

      pj = Bar[j].parent;      
      if (pj == i) {
	// Bars are connected.
	bar_check[i][j] = 0;
	continue;
      }

      // Test for "same_as" connected.
      if ((Bar[j].same_as == i) ||
	  (Bar[j].same_as == pi) ||
	  ((Bar[j].same_as == Bar[i].same_as) &&
	   (Bar[j].same_as != -1))) {
	// Same as connection.
	bar_check[i][j] = 0;
	continue;
      }

      near++;
      bar_check[i][j] = 2;
    }
    num_near[i] = near;
  }

  /*
  printf("\nInitialize Intersections\n");
  for (i=1; i<Num_bars; i++) {
    printf("%2d (%2d): ", i, num_near[i]);
    for (j=i+1; j<Num_bars; j++)
      printf("%d", bar_check[i][j]);
    printf("\n");
  }
  */
}


int BarStruct3D :: is_intersection(void)
{
  int i, j;
  int pi, pj;
  int jt_i;
  int check_val, near;
  Real ci[3], cj[3];
  Real dist, length, dist_apart;

  if (Time - last_check_far >= WAIT_CHECK_FAR) {
    last_check_far = Time;
    last_check_near = Time;
    check_val = 1;

    // Test to to make sure "same_as" nodes have same locatino.
    for (i=1; i<Num_bars; i++) {
      if (Bar[i].same_as != -1) {
	j = Bar[i].same_as;
	if ((absv(Bar[i].end_loc[0] - Bar[j].end_loc[0]) > SAME_AS_TOLERANCE) ||
	    (absv(Bar[i].end_loc[1] - Bar[j].end_loc[1]) > SAME_AS_TOLERANCE) ||
	    (absv(Bar[i].end_loc[2] - Bar[j].end_loc[2]) > SAME_AS_TOLERANCE)) {
	  if (Print[9]) {
	    printf("Bars %d(%.4f:%.4f:%.4f) and %d(%.4f:%.4f:%.4f) (%lf:%lf:%lf) moved apart.\n",
		   i, Bar[i].end_loc[0], Bar[i].end_loc[1], Bar[i].end_loc[2],
		   j, Bar[j].end_loc[0], Bar[j].end_loc[1], Bar[i].end_loc[2],
		   absv(Bar[i].end_loc[0] - Bar[j].end_loc[0]),
		   absv(Bar[i].end_loc[1] - Bar[j].end_loc[1]),
		   absv(Bar[i].end_loc[2] - Bar[j].end_loc[2]));
	  }
	  return TRUE;
	}
      }
    }

  } else if (Time - last_check_near >= WAIT_CHECK_NEAR) {
    last_check_near = Time;
    check_val = 2;
  } else {
    // No need to test.
    //    printf("No test: %.5f, %.5f, %.5f\n", Time, last_check_near, last_check_far);
    return FALSE;
  }

  //  printf("Test %d: %.5f, %.5f, %.5f\n", check_val, Time, last_check_near, last_check_far);


  for (i=1; i<Num_bars; i++) {
    pi = Bar[i].parent;
    Bar[i].center[0] = (Bar[i].end_loc[0] + Bar[pi].end_loc[0])/2.0;
    Bar[i].center[1] = (Bar[i].end_loc[1] + Bar[pi].end_loc[1])/2.0;
    Bar[i].center[2] = (Bar[i].end_loc[2] + Bar[pi].end_loc[2])/2.0;
  }
    
  for (i=1; i<Num_bars; i++) {
    pi = Bar[i].parent;
    jt_i = Bar[i].joint_index;
    ci[0] = Bar[i].center[0];
    ci[1] = Bar[i].center[1];
    ci[2] = Bar[i].center[2];

    if (check_val == 2) {
      if (num_near[i] == 0)
	continue;
    }

    near = 0;
    for (j=i+1; j<Num_bars; j++) {
      if (bar_check[i][j] < check_val) {
	// Don't need to test yet.
	continue;
      }

      cj[0] = Bar[j].center[0];
      cj[1] = Bar[j].center[1];
      cj[2] = Bar[j].center[2];

      dist = sqr(ci[0]-cj[0])+sqr(ci[1]-cj[1])+sqr(ci[2]-cj[2]);
      length = sqr(Bar[i].len_short+Bar[j].len_short);

      /*
      if (Time < 4.0*TIME_STEP) {
	//	printf("%d - %d: %.2f:%.2f:%.2f : %.2f:%.2f:%.2f  ", i, j,
	//	       Bar[i].end_loc[0], Bar[i].end_loc[1], Bar[pi].end_loc[2],	       
	//	       Bar[j].end_loc[0], Bar[j].end_loc[1], Bar[pi].end_loc[2]);
	printf("%d - %d : dist: %.2f, length: %.2f  -> %d\n", i, j,
	       dist, length, bar_check[i][j]);
      }
      */
      
      if (dist > length) {
	// Too far apart.
	if (dist < 4.0*length) {
	  near++;
	  bar_check[i][j] = 2;
	} else
	  bar_check[i][j] = 1;
	continue;
      }

      pj = Bar[j].parent;
      dist_apart = lines_distance_3d(Bar[pi].end_loc,
				     Bar[i].end_loc,
				     Bar[pj].end_loc,
				     Bar[j].end_loc);
      if (dist_apart < DIST_COLLISION) {
#ifndef OPTIMIZE
	if (Print[10] || Print[9]) {
	  printf("Intersection between %d and %d\n", i, j);
	  printf("%.2f:%.2f:%.2f : %.2f:%.2f:%.2f  -  %.2f:%.2f:%.2f : %.2f:%.2f:%.2f\n",
		 Bar[pi].end_loc[0], Bar[pi].end_loc[1], Bar[pi].end_loc[2],
		 Bar[i].end_loc[0], Bar[i].end_loc[1], Bar[pi].end_loc[2],
		 Bar[pj].end_loc[0], Bar[pj].end_loc[1], Bar[pi].end_loc[2],
		 Bar[j].end_loc[0], Bar[j].end_loc[1], Bar[pi].end_loc[2]);
	}
#endif
	return TRUE;
      }
      if (dist_apart < 4.0*length) {
	near++;
	bar_check[i][j] = 2;
      } else
	bar_check[i][j] = 1;
    }
    num_near[i] = near;
  }


#ifndef OPTIMIZE
  if (Print[19]) {
    for (i=1; i<Num_bars; i++) {
      for (j=i+1; j<Num_bars; j++)
	printf("%d", bar_check[i][j]);
      printf("\n");
    }
  }
#endif

  return FALSE;
}


void BarStruct3D :: make_joint_rot_mat_network(int index, Real rot_mat[3][3])
{
  /*
  if (Bar[index].angle == Bar[index].desired_angle) {
    mm3_copy(rot_mat, Bar[index].rotation_matrix);
    return;
  }
  */

  Real angle;

  if (Bar[index].angle == Bar[index].desired_angle) {
    ;

  } else if (absv(Bar[index].angle - Bar[index].desired_angle) <= JOINT_MOVE_STEP) {
    Bar[index].angle = Bar[index].desired_angle;

  } else if (Bar[index].angle < Bar[index].desired_angle) {
    Bar[index].angle += JOINT_MOVE_STEP;

  } else {
    Bar[index].angle -= JOINT_MOVE_STEP;
  }

  angle = Bar[index].angle;
  if (Bar[index].rot_axis == ROT_X_AXIS) {
    make_rotX_mat(angle, rot_mat);
  } else if (Bar[index].rot_axis == ROT_Y_AXIS) {
    make_rotY_mat(angle, rot_mat);
  } else if (Bar[index].rot_axis == ROT_Z_AXIS) {
    make_rotZ_mat(angle, rot_mat);
  } else {
    printf("ERROR :: bars3d - Invalid rotation axis (%d:%d)!  ", index,
	   Bar[index].rot_axis);
    printf("type: %d, moves: %d, angle: %.4f - %.4f\n", Bar[index].type,
	   Bar[index].moves, Bar[index].angle_a, Bar[index].angle_b);
    make_identity_mat(rot_mat);
  }
}


void BarStruct3D :: make_joint_rot_mat(int index, Real time, Real rot_mat[3][3])
{
  Real angle, theta;

  if (Build_type == 5) {
    make_joint_rot_mat_network(index, rot_mat);
    return;
  }

  theta = time * Bar[index].move_spd + Bar[index].move_offset;
  theta -= floor(theta);
  theta *= 2.0;
  if (theta > 1.0)
    theta = 2.0 - theta;

  angle = Bar[index].angle_a*(1.0-theta) + Bar[index].angle_b*theta;
  if (Bar[index].rot_axis == ROT_X_AXIS) {
    make_rotX_mat(angle, rot_mat);
  } else if (Bar[index].rot_axis == ROT_Y_AXIS) {
    make_rotY_mat(angle, rot_mat);
  } else if (Bar[index].rot_axis == ROT_Z_AXIS) {
    make_rotZ_mat(angle, rot_mat);
  } else {
    printf("ERROR :: bars3d - Invalid rotation axis (%d:%d)!  ", index,
	   Bar[index].rot_axis);
    printf("type: %d, moves: %d, angle: %.4f - %.4f\n", Bar[index].type,
	   Bar[index].moves, Bar[index].angle_a, Bar[index].angle_b);
    make_identity_mat(rot_mat);
  }
}


// Calculate position & orientation of child's parent.
void BarStruct3D :: calculate_up(int child, Real time)
{
  int parent;
  Real len;
  Real loc_delta[3], rot_mat[3][3], inv_rot[3][3];

  if (Bar[child].moves) {
    make_joint_rot_mat(child, time, rot_mat);
    mm3_transpose(inv_rot, rot_mat);
    mmm3_mult(rot_mat, Bar[child].rotation_matrix, inv_rot);
  } else {
    mm3_copy(rot_mat, Bar[child].rotation_matrix);
  }

  // Calculate new location.
  len = Bar[child].length;
  loc_delta[0] = rot_mat[0][0]*len;
  loc_delta[1] = rot_mat[1][0]*len;
  loc_delta[2] = rot_mat[2][0]*len;

  parent = Bar[child].parent;
  Bar[parent].end_loc[0] = Bar[child].end_loc[0] - loc_delta[0];
  Bar[parent].end_loc[1] = Bar[child].end_loc[1] - loc_delta[1];
  Bar[parent].end_loc[2] = Bar[child].end_loc[2] - loc_delta[2];

  mmm3_mult(Bar[parent].rotation_matrix, rot_mat,
	    Bar[child].inverse_delta_rot);
  

#ifndef OPTIMIZE
  if (Print[12]) {
    Real angle = 0.0;
    if (!Bar[child].moves)
      angle = Bar[child].angle_a;

    printf("Calculating parent %d from child %d (mvs:%d).\n", parent, child, Bar[child].moves);
    printf("%d(%d): len:%.2f  axis: %c angle:%.5f\n", parent, child,
	     len, Bar[child].rot_axis+'w', angle);
    m3_print(Bar[child].rotation_matrix, "Child rotation matrix:\n");
    if (Bar[child].moves)
      m3_print(inv_rot, "Inverse joint rotation matrix:\n");
    printf("child (%d) inverse delta rotation matrix:\n", child);
    m3_print(Bar[child].inverse_delta_rot, 0);
    printf("parent (%d) rotation matrix:\n", parent);
    m3_print(Bar[parent].rotation_matrix, 0);
    printf("child (%d):  %.3f, %.3f, %.3f\n", child,
	   Bar[child].end_loc[0], Bar[child].end_loc[1],
	   Bar[child].end_loc[2]);
    printf("delta:  %.3f, %.3f, %.3f\n",
	   loc_delta[0], loc_delta[1], loc_delta[2]);
    printf("parent (%d): %.3f %.3f %.3f\n", parent,
	   Bar[parent].end_loc[0],
	   Bar[parent].end_loc[1], Bar[parent].end_loc[2]);
  }
#endif

  Bar[parent].updated = TRUE;
  if (parent != 0)
    calculate_up(parent, time);
}


// Calculate position of bar i from position of its parent.
// Child_orientation = Joint_rotation * Delta_rotation * Parent_orientation
void BarStruct3D :: calc_bar_position(int i, Real time)
{
  int parent;
  Real len;
  Real loc_delta[3];
  Real rot_mat[3][3], tmp_mat[3][3];

  // Calculate orientation from parent.
  parent = Bar[i].parent;
  mmm3_mult(Bar[i].rotation_matrix, 
	    Bar[parent].rotation_matrix,
	    Bar[i].delta_rot_mat);

  // Calculate new location.
  len = Bar[i].length;
  loc_delta[0] = Bar[i].rotation_matrix[0][0]*len;
  loc_delta[1] = Bar[i].rotation_matrix[1][0]*len;
  loc_delta[2] = Bar[i].rotation_matrix[2][0]*len;

  Bar[i].end_loc[0] = Bar[parent].end_loc[0] + loc_delta[0];
  Bar[i].end_loc[1] = Bar[parent].end_loc[1] + loc_delta[1];
  Bar[i].end_loc[2] = Bar[parent].end_loc[2] + loc_delta[2];


  // If bar moves, change the resulting orientation.
  if (Bar[i].moves) {
    make_joint_rot_mat(i, time, rot_mat);
    mmm3_mult(tmp_mat, Bar[i].rotation_matrix, rot_mat);
    mm3_copy(Bar[i].rotation_matrix, tmp_mat);
  }


#ifndef OPTIMIZE
  if (Print[12]) {
    Real angle = 0.0;
    if (!Bar[i].moves)
      angle = Bar[i].angle_a;

    printf("%d(%d): len:%.2f  %c angle:%.5f:\n", i, parent,
	   len, Bar[i].rot_axis+'w', angle);
    if (Print[15] || Print[34]) {
      m3_print(Bar[parent].rotation_matrix, "Parent rotation matrix:\n");
      m3_print(Bar[i].delta_rot_mat, "Child delta rotation matrix:\n");
      if (Bar[i].moves)
	m3_print(rot_mat, "Child joint rotation matrix.\n");
      m3_print(Bar[i].rotation_matrix, "Child rotation matrix:\n");
    }
    printf("parent (%d): %.3f %.3f %.3f\n", parent,
	   Bar[parent].end_loc[0],
	   Bar[parent].end_loc[1], Bar[parent].end_loc[2]);
    printf("delta:  %.3f, %.3f, %.3f\n",
	   loc_delta[0], loc_delta[1], loc_delta[2]);
    printf("child (%d):  %.3f, %.3f, %.3f\n", i,
	   Bar[i].end_loc[0], Bar[i].end_loc[1],
	   Bar[i].end_loc[2]);
  }

#endif
}


void BarStruct3D :: calc_positions(Real time)
{
  int i;

#ifndef OPTMIZE
  if (Print[12])
    printf("Calc_positions - ref: %d, time: %.6f\n", 
	   ref_pt, time);
#endif

  for (i=1; i<Num_bars; i++) {
    if (Bar[i].updated) {
      continue;
    }
    calc_bar_position(i, time);
    Bar[i].updated = TRUE;
  }
}


void BarStruct3D :: calculate_positions(Real time)
{
  int i;
  Real tmp_mat[3][3];

  if (Num_bars == 0)
    return;

  if (Build_type == 5) {
    if ((!network->output_changed()) && (time > 0.0)) {
      return; // No need to update positions of no new joint positions.
    }
  }

#ifndef OPTIMIZE
  if (Print[12]) {
    printf("Calculate positions - ref: %d, time: %.4f\n",
	   ref_pt, time);
  }
#endif
  Time = time;

  for (i=0; i<Num_bars; i++) {
    Bar[i].updated = FALSE;
    Bar[i].prev_loc[0] = Bar[i].end_loc[0];
    Bar[i].prev_loc[1] = Bar[i].end_loc[1];
    Bar[i].prev_loc[2] = Bar[i].end_loc[2];
  }

  // Network controller stuff.
  if (Build_type == 5) {
    int update_network_flag = TRUE;
    for (i=0; i<Num_bars; i++) {
      if (Bar[i].moves) {
	if (Bar[i].angle != Bar[i].desired_angle) {
	  update_network_flag = FALSE;
	  break;
	}
      }
    }

    if (update_network_flag) {
      update_network();
    }
  }


#ifndef OPTIMIZE
  if (Print[12] && Print[15]) {
    m3_print(Bar[ref_pt].rotation_matrix, 
	     "Reference point rotation:\n");
    printf("Reference point: %d.\n", ref_pt);
    m3_print(rotation_matrix, 
	     "Global rotation matrix:\n");
  }
#endif

  // Update orientation of the reference point.
  mmm3_mult(tmp_mat, rotation_matrix, 
      	    Bar[ref_pt].rotation_matrix);
  //  mmm3_mult(tmp_mat, Bar[ref_pt].rotation_matrix, 
  //    	    rotation_matrix);
  mm3_copy(Bar[ref_pt].rotation_matrix, tmp_mat);
  Bar[ref_pt].updated = TRUE;

#ifndef OPTIMIZE
  if (Print[12] && Print[15]) {
    m3_print(Bar[ref_pt].rotation_matrix, 
	     "Reference point rotation:\n");
  }
#endif

  if (ref_pt != 0)
    calculate_up(ref_pt, time);
  make_identity_mat(rotation_matrix);

  calc_positions(time);

  calculate_com();

  if (Print[12]) {
    print_struct();
  }
}


void BarStruct3D :: calculate_positions_new(Real time)
{
  int i;
  int saved_ref_pt;
  Real tmp_mat[3][3];
  Real old_rot_mat[3][3];
  Real saved_com[3];

  if (Num_bars == 0)
    return;

  for (i=0; i<Num_bars; i++) {
    Bar[i].updated = FALSE;
    Bar[i].saved_loc[0] = Bar[i].end_loc[0];
    Bar[i].saved_loc[1] = Bar[i].end_loc[1];
    Bar[i].saved_loc[2] = Bar[i].end_loc[2];
  }
  saved_com[0] = COM[0];
  saved_com[1] = COM[1];
  saved_com[2] = COM[2];

  saved_ref_pt = ref_pt;
  ref_pt = 3;

  
#ifndef OPTIMIZE
  if (Print[12] && Print[15]) {
    m3_print(rotation_matrix, 
	     "Global rotation:\n");
    printf("Reference point: %d.\n", ref_pt);
    m3_print(Bar[ref_pt].rotation_matrix, 
	     "Original Reference point rotation:\n");
  }
#endif

  mm3_copy(old_rot_mat, Bar[ref_pt].rotation_matrix);
  mmm3_mult(tmp_mat, rotation_matrix, 
      	    Bar[ref_pt].rotation_matrix);
  //  mmm3_mult(tmp_mat, Bar[ref_pt].rotation_matrix, 
  //    	    rotation_matrix);
  mm3_copy(Bar[ref_pt].rotation_matrix, tmp_mat);

#ifndef OPTIMIZE
  if (Print[12] && Print[15]) {
    m3_print(Bar[ref_pt].rotation_matrix, 
	     "Rotated Reference point rotation:\n");
  }
#endif

  if (ref_pt != 0) {
    calculate_up(ref_pt, time);
  }
  ref_pt = saved_ref_pt;

  calc_positions(time);
  mm3_copy(Bar[ref_pt].rotation_matrix, old_rot_mat);

  calculate_com();

  if (Print[12] || Print[34]) {
    print_struct();
    printf("recalc over.\n\n");
  }

  for (i=0; i<Num_bars; i++) {
    Bar[i].end_loc[0] = Bar[i].saved_loc[0];
    Bar[i].end_loc[1] = Bar[i].saved_loc[1];
    Bar[i].end_loc[2] = Bar[i].saved_loc[2];
  }
  COM[0] = saved_com[0];
  COM[1] = saved_com[1];
  COM[2] = saved_com[2];
}


int BarStruct3D :: cycle_for_intersections(void)
{
  Real phase;

  for (phase = 1.0; phase >= 0.0; phase -= 0.025) {
    calculate_positions(phase);
    if (is_intersection())
      return TRUE;
  }

  return FALSE;
}


// Calculates the center-of-mass of the structure.
void BarStruct3D :: calculate_com(void)
{
  int i;

  COM[0] = 0.0;
  COM[1] = 0.0;
  COM[2] = 0.0;

  for (i=0; i<Num_bars; i++) {
    COM[0] += Bar[i].end_loc[0];
    COM[1] += Bar[i].end_loc[1];
    COM[2] += Bar[i].end_loc[2];
  }

  COM[0] /= Real(Num_bars);
  COM[1] /= Real(Num_bars);
  COM[2] /= Real(Num_bars);
}


void BarStruct3D :: print_struct(void)
{
  printf("Structure at time: %.6f\n", Time);
  printf("CoM: %.6f, %.6f %.6f\n", COM[0], COM[1], COM[2]);
  for (int i=0; i<Num_bars; i++) {
    if ((Bar[i].type == BAR_TYPE_JOINT) && (Bar[i].num_children > 0)) {
      printf("%3d: %8.3f : %10.5f : %8.3f, %2d (t:%d; s:%.1f; o:%.3f, a:%.2f; b:%.2f; w:%.3f)\n",
	     i, Bar[i].end_loc[0], Bar[i].end_loc[1],
	     Bar[i].end_loc[2], Bar[i].same_as,
	     Bar[i].type, Bar[i].move_spd, Bar[i].move_offset,
	     Bar[i].angle_a, Bar[i].angle_b, Bar[i].weight);

    } else {
      printf("%3d: %8.3f : %10.5f : %8.3f, %2d (t:%d, a:%.2f; b:%.2f; w:%.3f)\n",
	     i, Bar[i].end_loc[0], Bar[i].end_loc[1],
	     Bar[i].end_loc[2], Bar[i].same_as,
	     Bar[i].type,
	     Bar[i].angle_a, Bar[i].angle_b, Bar[i].weight);
    }
  }
}


// Moves the entire structure by x:y.
void BarStruct3D :: translate_struct(Real x, Real y,
				     Real z)
{
  int i;

#ifndef OPTIMIZE
  if (Print[34] || Print[12]) {
    printf("Translate: %.4f, %.4f %.4f\n", x, y, z);
    printf("  CoM: %.6f, %.6f %.6f\n", COM[0], COM[1], COM[2]);
  }
#endif

  COM[0] += x;
  COM[1] += y;
  COM[2] += z;

  for (i=0; i<Num_bars; i++) {
    Bar[i].end_loc[0] += x;
    Bar[i].end_loc[1] += y;
    Bar[i].end_loc[2] += z;
  }
}



// Rotates the structure by angle.
void BarStruct3D :: rotate_struct_Xaxis(Real angle)
{
  int i;
  Real y, z;
  Real cos_ang, sin_ang;
  Real rot_mat[3][3], tmp_mat[3][3];

#ifndef OPTIMIZE
  if (Print[34] || Print[12]) {
    printf("Rotate X axis: %.4f\n", angle);
    printf("  CoM: %.6f, %.6f %.6f\n", COM[0], COM[1], COM[2]);
  }
#endif

  
  make_rotX_mat(angle, rot_mat);
  mmm3_mult(tmp_mat, rot_mat, rotation_matrix);
  mm3_copy(rotation_matrix, tmp_mat);

  cos_ang = cos(angle);
  sin_ang = sin(angle);

  y = cos_ang*COM[1] - sin_ang * COM[2];
  z = sin_ang*COM[1] + cos_ang * COM[2];
  COM[1] = y;
  COM[2] = z;

  for (i=0; i<Num_bars; i++) {
    y = cos_ang*Bar[i].end_loc[1] - sin_ang * Bar[i].end_loc[2];
    z = sin_ang*Bar[i].end_loc[1] + cos_ang * Bar[i].end_loc[2];
    Bar[i].end_loc[1] = y;
    Bar[i].end_loc[2] = z;
  }

#ifndef OPTIMIZE
  if (Print[12]) {
    print_struct();
    if (Print[15]) {
      printf("Recalc positions: %.3f\n", Time);
      calculate_positions_new(Time);
    }
  }
#endif
}



// Rotates the structure by angle.
void BarStruct3D :: rotate_struct_Yaxis(Real angle)
{
  int i;
  Real x, z;
  Real cos_ang, sin_ang;
  Real rot_mat[3][3], tmp_mat[3][3];

#ifndef OPTIMIZE
  if (Print[34] || Print[12]) {
    printf("Rotate Y-axis: %.4f\n", angle);
    printf("  CoM: %.6f, %.6f %.6f\n", COM[0], COM[1], COM[2]);
  }
#endif

  
  make_rotY_mat(angle, rot_mat);
  mmm3_mult(tmp_mat, rot_mat, rotation_matrix);
  mm3_copy(rotation_matrix, tmp_mat);

  cos_ang = cos(angle);
  sin_ang = sin(angle);

  x = cos_ang*COM[0] + sin_ang * COM[2];
  z = -sin_ang*COM[0] + cos_ang * COM[2];

  if (Print[12] && Print[15]) {
    m3_print(rot_mat, "Rotation-Y matrix:\n");
    printf("angle: %f,  cos: %f   sin: %f\n",
	   angle, cos_ang, sin_ang);
    printf("COM %.6f(%.3f:%.3f) : %.6f : %.6f(%.3f:%.3f)\n",
	   x, cos_ang*COM[0], -sin_ang*COM[2],
	   COM[1], z, sin_ang*COM[0], cos_ang*COM[2]);
  }
  COM[0] = x;
  COM[2] = z;


  for (i=0; i<Num_bars; i++) {
    x = cos_ang*Bar[i].end_loc[0] + sin_ang * Bar[i].end_loc[2];
    z = -sin_ang*Bar[i].end_loc[0] + cos_ang * Bar[i].end_loc[2];

    if (Print[12] && Print[15]) {
      printf("%d: %.3f(%.3f:%.3f) : %.3f(%.3f:%.3f)\n", i,
	     x, cos_ang*Bar[i].end_loc[0], -sin_ang*Bar[i].end_loc[2],
	     z, sin_ang*Bar[i].end_loc[0], cos_ang*Bar[i].end_loc[2]);
    }

    Bar[i].end_loc[0] = x;
    Bar[i].end_loc[2] = z;
  }
#ifndef OPTIMIZE
  if (Print[12]) {
    print_struct();
    if (Print[15]) {
      printf("Recalc positions: %.3f\n", Time);
      calculate_positions_new(Time);
    }
  }
#endif
}



// Rotates the structure by angle.
void BarStruct3D :: rotate_struct_Zaxis(Real angle)
{
  int i;
  Real x, y;
  Real cos_ang, sin_ang;
  Real rot_mat[3][3], tmp_mat[3][3];

#ifndef OPTIMIZE
  if (Print[34] || Print[12]) {
    printf("Rotate Z-axis: %.4f\n", angle);
    printf("  CoM: %.6f, %.6f %.6f\n", COM[0], COM[1], COM[2]);
  }
#endif

  
  make_rotZ_mat(angle, rot_mat);
  mmm3_mult(tmp_mat, rot_mat, rotation_matrix);
  mm3_copy(rotation_matrix, tmp_mat);

  cos_ang = cos(angle);
  sin_ang = sin(angle);

  x = cos_ang*COM[0] - sin_ang * COM[1];
  y = sin_ang*COM[0] + cos_ang * COM[1];

  if (Print[12] && Print[15]) {
    m3_print(rot_mat, "Rotation-Z matrix:\n");
    printf("angle: %f,  cos: %f   sin: %f\n",
	   angle, cos_ang, sin_ang);
    printf("COM %.6f(%.3f:%.3f) : %.6f(%.3f:%.3f) : %.6f\n",
	   x, cos_ang*COM[0], -sin_ang*COM[1],
	   y, sin_ang*COM[0], cos_ang*COM[1],
	   COM[2]);
  }
  COM[0] = x;
  COM[1] = y;

  for (i=0; i<Num_bars; i++) {
    x = cos_ang*Bar[i].end_loc[0] - sin_ang * Bar[i].end_loc[1];
    y = sin_ang*Bar[i].end_loc[0] + cos_ang * Bar[i].end_loc[1];

    if (Print[12] && Print[15]) {
      printf("%d: %.3f(%.3f:%.3f) : %.3f(%.3f:%.3f)\n", i,
	     x, cos_ang*Bar[i].end_loc[0], -sin_ang*Bar[i].end_loc[1],
	     y, sin_ang*Bar[i].end_loc[0], cos_ang*Bar[i].end_loc[1]);
    }

    Bar[i].end_loc[0] = x;
    Bar[i].end_loc[1] = y;
  }

#ifndef OPTIMIZE
  if (Print[12]) {
    print_struct();
    if (Print[15]) {
      printf("Recalc positions: %.3f\n", Time);
      calculate_positions_new(Time);
    }
  }
#endif
}


// Rotates the structure by angle.
void BarStruct3D :: rotate_struct(int axis, Real angle)
{
  if (axis == ROT_X_AXIS)
    rotate_struct_Xaxis(angle);
  else if (axis == ROT_Y_AXIS)
    rotate_struct_Yaxis(angle);
  else if (axis == ROT_Z_AXIS)
    rotate_struct_Zaxis(angle);
  else {
    printf("BarStruct3D::rotate_struct - invalid axis %d\n", axis);
  }
}

/*********************************************************************/


int BarStruct3D :: same_point(int index1, int index2)
{
  if ((absv(Bar[index1].end_loc[0]-
	    Bar[index2].end_loc[0])<DIST_TOLERANCE) &&
      (absv(Bar[index1].end_loc[2]-
	    Bar[index2].end_loc[2])<DIST_TOLERANCE)) {
    return TRUE;
  }
  return FALSE;
}


int BarStruct3D :: settle_right(int right_min_index, int rotate_full)
{
  int i, bal_index = 0;
  Real slope, slope2;
  Real angle;
  Real old_x, old_y;

  slope = 9999999.0;

#ifndef OPTIMIZE
  if (Print[12])
    printf("Settle right:\n");
#endif

  // Find left balance point.
  for (i=0; i<Num_bars; i++) {
    if (i == right_min_index) {
      continue;
    } else if (absv(Bar[right_min_index].end_loc[0]-
		    Bar[i].end_loc[0]) < DIST_TOLERANCE) {
      // Points are too close.
      continue;

    } else if (absv(Bar[right_min_index].end_loc[1]-
		    Bar[i].end_loc[1]) < DIST_TOLERANCE) {
      // i is a ground point too!
      continue;
    }

    if (Bar[i].end_loc[0] > Bar[right_min_index].end_loc[0]) {
      slope2 = (Bar[i].end_loc[1]-Bar[right_min_index].end_loc[1]) /
	(Bar[i].end_loc[0]-Bar[right_min_index].end_loc[0]);
      if (slope2 < slope) {
	bal_index = i;
	slope = slope2;
      }
    }
  }

#ifndef OPTIMIZE
  if (Print[12]) {
    printf("  pivot: %2d  %.4f %.4f %.4f\n", right_min_index,
	   Bar[right_min_index].end_loc[0],
	   Bar[right_min_index].end_loc[1],
	   Bar[right_min_index].end_loc[2]);
    printf("  other: %2d  %.4f %.4f %.4f  (%f)\n", bal_index,
	   Bar[bal_index].end_loc[0],
	   Bar[bal_index].end_loc[1],
	   Bar[bal_index].end_loc[2],
	   Bar[bal_index].end_loc[0]-Bar[right_min_index].end_loc[0]);
  }
#endif

  // Rotate clockwise and re-settle().
  angle = atan(slope);
  if (!rotate_full) {
    if (angle > MAX_SETTLE_ANGLE) {
      Rotating = TRUE;
      angle = MAX_SETTLE_ANGLE;
    } else if (angle < -MAX_SETTLE_ANGLE) {
      Rotating = FALSE;
      angle = -MAX_SETTLE_ANGLE;
    }
  }

  /*
  if (angle == 0.0) {
    printf("Settle right.  Angle warning = 0.0\n");
    Print[12] = TRUE;

  } else if (absv(angle) < 0.0000001) {
    printf("Settle right.  Angle warning: %lf\n", angle);
    Print[12] = TRUE;
  }
  */

  old_x = Bar[right_min_index].end_loc[0];
  old_y = Bar[right_min_index].end_loc[1];
  rotate_struct(ROT_Z_AXIS, -angle);
  translate_struct(old_x-Bar[right_min_index].end_loc[0],
		   old_y-Bar[right_min_index].end_loc[1], 0.0);

#ifndef OPTIMIZE
  if (Print[12]) {
    printf("Settled orientation...\n");
    print_struct();
  }
#endif

  return bal_index;
}


int BarStruct3D :: settle_left(int left_min_index, int rotate_full)
{
  int i, bal_index = 0;
  Real slope, slope2;
  Real angle;
  Real old_x, old_y;

  slope = -9999999.0;

#ifndef OPTIMIZE
  if (Print[12])
    printf("Settle left:\n");
#endif

  // Find left balance point.
  for (i=0; i<Num_bars; i++) {
    if (i == left_min_index) {
      continue;
    } else if (absv(Bar[left_min_index].end_loc[0]-
		    Bar[i].end_loc[0]) < DIST_TOLERANCE) {
      // Points are too close.
      continue;

    } else if (absv(Bar[left_min_index].end_loc[1]-
		    Bar[i].end_loc[1]) < DIST_TOLERANCE) {
      // i is a ground point too!
      continue;
    }

    if (Bar[i].end_loc[0] < Bar[left_min_index].end_loc[0]) {
      slope2 = (Bar[i].end_loc[1]-Bar[left_min_index].end_loc[1]) /
	(Bar[i].end_loc[0]-Bar[left_min_index].end_loc[0]);
      if (slope2 > slope) {
	bal_index = i;
	slope = slope2;
      }
    }
  }

#ifndef OPTIMIZE
  if (Print[12]) {
    printf("  pivot: %2d  %.4f %.4f %.4f\n", left_min_index,
	   Bar[left_min_index].end_loc[0],
	   Bar[left_min_index].end_loc[1],
	   Bar[left_min_index].end_loc[2]);
    printf("  other: %2d  %.4f %.4f %.4f  (%f)\n", bal_index,
	   Bar[bal_index].end_loc[0],
	   Bar[bal_index].end_loc[1],
	   Bar[bal_index].end_loc[2],
	   Bar[bal_index].end_loc[0]-Bar[left_min_index].end_loc[0]);
  }
#endif

  // Rotate clockwise, re-position and re-settle().
  angle = atan(slope);
  if (!rotate_full) {
    if (angle > MAX_SETTLE_ANGLE) {
      Rotating = TRUE;
      angle = MAX_SETTLE_ANGLE;
    } else if (angle < -MAX_SETTLE_ANGLE) {
      Rotating = FALSE;
      angle = -MAX_SETTLE_ANGLE;
    }
  }

  /*
  if (angle == 0.0) {
    printf("Settle left.  Angle warning = 0.0\n");
    Print[12] = TRUE;

  } else if (absv(angle) < 0.0000001) {
    printf("Settle left.  Angle warning: %lf\n", angle);
    Print[12] = TRUE;
  }
  */

  old_x = Bar[left_min_index].end_loc[0];
  old_y = Bar[left_min_index].end_loc[1];  
  rotate_struct(ROT_Z_AXIS, -angle);
  translate_struct(old_x-Bar[left_min_index].end_loc[0],
		   old_y-Bar[left_min_index].end_loc[1], 0.0);
  return bal_index;
}


Bar_struct_3d *OBar;

int pt_compare(const void *elt1, const void *elt2)
{
  int index1, index2;

  index1 = *(int*)elt1;
  index2 = *(int*)elt2;

  if (OBar[index1].end_loc[0] < OBar[index2].end_loc[0])
    return -1;
  else if (OBar[index1].end_loc[0] == OBar[index2].end_loc[0]) {
    if (OBar[index1].end_loc[2] < OBar[index2].end_loc[2])
      return -1;
    else if (OBar[index1].end_loc[2] > OBar[index2].end_loc[2])
      return 1;
    return 0;
  }
  return 1;
}


inline Real calc_inv_slope(Real *p1, Real *p2)
{
  if (absv(p2[2]-p1[2]) < DIST_TOLERANCE)
    if (p2[0]<p1[0]) {
      if (p2[2]<p1[2])
	return DBL_MAX;
      else
	return -DBL_MAX;
    } else {
      if (p2[0]<p1[0])
	return -DBL_MAX;
      else
	return DBL_MAX;
    }

  return (p2[0]-p1[0]) / (p2[2]-p1[2]);
}


inline int slopes_equal(Real slope1, Real slope2)
{
  //  printf("Slopes equal: %f  %f.\n", slope1, slope2);

  if (absv(slope2) > 1.0) {
    //    printf(">1: %d  %f\n", (absv(slope1/slope2-1.0) < ANGLE_TOLERANCE),
    //	   absv(slope1/slope2) - 1.0);
    if (absv(slope1/slope2-1.0) < ANGLE_TOLERANCE)
      return TRUE;
    else
      return FALSE;
  }
  //  printf("<1: %d  %f\n", (absv(slope1-slope2) < ANGLE_TOLERANCE),
  //	 absv(slope1-slope2));

  if (absv(slope1-slope2) < ANGLE_TOLERANCE)
    return TRUE;
  else
    return FALSE;
}

inline Real calculate_slope(Real p1[3], Real p2[3])
{
  Real div;
  Real dist_2 = p2[2] - p1[2];

#ifndef OPTIMIZE
  if (Print[12]) {
    printf("Calculate slope:\n");
    v_print(p1, "#1: ");
    v_print(p2, "#2: ");
  }
#endif

  if (absv(dist_2) < 1.0)
    div = 1.0;
  else
    div = absv(dist_2);

  if (absv(p2[0]-p1[0])/div<DIST_TOLERANCE) {
    if (p2[2] < p1[2]) {
      if (p2[0] <= p1[0])
	return DBL_MAX;
      else
	return -DBL_MAX;
    } else {
      if (p2[0] < p1[0])
	return -DBL_MAX;
      else
	return DBL_MAX;
    }
  }

  div = (p2[2] - p1[2])/(p2[0] - p1[0]);
#ifndef OPTIMIZE
  if (Print[12]) {
    printf("%f / %f = %f\n", (p2[2] - p1[2]), (p2[0] - p1[0]), div);
  }
#endif

  if (absv(div) < ANGLE_TOLERANCE)
    div = 0.0;
  return div;
}


Real BarStruct3D :: calc_slope(int p1, int p2)
{
  return calculate_slope(Bar[p1].end_loc, Bar[p2].end_loc);
}

Real BarStruct3D :: calc_slope_COM(int p1)
{
  return calculate_slope(Bar[p1].end_loc, COM);
}


// Returns r, where:
// r = (AC dot AB) / |AB|^2
// A = p1; B = p2, C = COM.
inline Real BarStruct3D :: calc_COM_proj(int p1, int p2)
{
  Real val, num, denom;

  num = (COM[0]-Bar[p1].end_loc[0])*(Bar[p2].end_loc[0]-Bar[p1].end_loc[0]);  
  num += (COM[2]-Bar[p1].end_loc[2])*(Bar[p2].end_loc[2]-Bar[p1].end_loc[2]);

  denom = sqr(Bar[p2].end_loc[0]-Bar[p1].end_loc[0]) +
    sqr(Bar[p2].end_loc[2]-Bar[p1].end_loc[2]);

  /*
  printf("calc_COM_proj: %f / %f\n", num, denom);
  printf("(%f - %f) * (%f - %f) = %f * %f = %f\n",
	 COM[0], Bar[p1].end_loc[0], Bar[p2].end_loc[0], Bar[p1].end_loc[0],
	 (COM[0]-Bar[p1].end_loc[0]), (Bar[p2].end_loc[0]-Bar[p1].end_loc[0]),
	 (COM[0]-Bar[p1].end_loc[0])*(Bar[p2].end_loc[0]-Bar[p1].end_loc[0]));
   
  printf("(%f - %f) * (%f - %f) = %f * %f = %f\n",
	 COM[2], Bar[p1].end_loc[2], Bar[p2].end_loc[2], Bar[p1].end_loc[2],
	 (COM[2]-Bar[p1].end_loc[2]), (Bar[p2].end_loc[2]-Bar[p1].end_loc[2]),
	 (COM[2]-Bar[p1].end_loc[2])*(Bar[p2].end_loc[2]-Bar[p1].end_loc[2]));
  */

  val = num / denom;
  if (absv(val) < DIST_TOLERANCE)
    val = 0.0;
  return val;
}


// Returns TRUE if COM is above the line from p1 to p2.
// Assumes p1.x <= p2.x  (sortof)
int BarStruct3D :: COM_above_line(int p1, int p2)
{
  Real z;

#ifndef OPTIMIZE
  if (Print[12])
    printf("COM above line:  %d  %d\n", p1, p2);
#endif

  // First, handle special case where ...
  //  if (Bar[p1].end_loc[0] == Bar[p2].end_loc[0]) {
  if (absv(Bar[p1].end_loc[0]-Bar[p2].end_loc[0]) < DIST_TOLERANCE) {
    // ... Slope is infinite.
    if (Bar[p1].end_loc[2] < Bar[p2].end_loc[2]) {
      if (COM[0]+DIST_TOLERANCE < Bar[p1].end_loc[0]) // originally: <=
	return TRUE;
      else
	return FALSE;
    } else {
      if (COM[0]-DIST_TOLERANCE <= Bar[p1].end_loc[0])
	return FALSE;
      else
	return TRUE;
    }
  }

  Real slope = calc_slope(p1, p2);
  z = slope*(COM[0] - Bar[p1].end_loc[0]) +
    Bar[p1].end_loc[2];
  
#ifndef OPTIMIZE
  if (Print[12]) {
    printf("slope: %f  z: %f  COM: %f  : %d\n", slope,
	   z, COM[2], Bar[p1].end_loc[0] <= Bar[p2].end_loc[0]);
  }
#endif

  if (Bar[p1].end_loc[0] <= Bar[p2].end_loc[0]) {
    return (z+DIST_TOLERANCE < COM[2]);
  }
  return (z-DIST_TOLERANCE > COM[2]);


#ifdef OLD
  if (z+DIST_TOLERANCE <= COM[2]) {
    //    return TRUE;
    return (Bar[p1].end_loc[0] <= Bar[p2].end_loc[0]);
  }

  //  return FALSE;
  return (Bar[p1].end_loc[0] > Bar[p2].end_loc[0]);
#endif
}


int BarStruct3D :: build_upper_hull(int num_grnd_pts,
				    int *grnd_pt,
				    HullPt *hull_pt)
{
  int i;
  int num_hull_pts;
  Real slope=0.0;

#ifndef OPTIMIZE
  if (Print[12])
    printf("Building upper hull:\n");
#endif

  // Create upper hull.
  num_hull_pts = 1;
  hull_pt[0].id = grnd_pt[0];
  hull_pt[1].id = grnd_pt[1];
  hull_pt[0].slope = calc_slope(hull_pt[0].id,
				hull_pt[1].id);

  for (i=2; i<num_grnd_pts; i++) {
    for (; num_hull_pts>0; num_hull_pts--) {
      slope = calc_slope(hull_pt[num_hull_pts].id,
			 grnd_pt[i]);
#ifndef OPTIMIZE
      if (Print[12]) {
	printf("(%d:%d) %d-%d: %.4f vs %d-%d: %.4f\n",
	       i, num_hull_pts,
	       hull_pt[num_hull_pts-1].id, hull_pt[num_hull_pts].id,
	       hull_pt[num_hull_pts-1].slope,
	       hull_pt[num_hull_pts].id, grnd_pt[i], slope);
      }
#endif
      if (slope < hull_pt[num_hull_pts-1].slope)
	break;
#ifndef OPTIMIZE
      if (Print[12])
	printf("replace!\n");
#endif
    }
    if (num_hull_pts == 0) {
      slope = calc_slope(hull_pt[0].id, grnd_pt[i]);
    }
    hull_pt[num_hull_pts].slope = slope;
    num_hull_pts++;
    hull_pt[num_hull_pts].id = grnd_pt[i];
  }
  num_hull_pts++;

#ifndef OPTIMIZE
  if (Print[12]) {
    printf("Hull points: %d\n", num_hull_pts);
    for (i=0; i<num_hull_pts; i++) {
      printf("%d: %.4f %.4f %.4f  (%.4f)\n",
	     hull_pt[i].id,
	     Bar[hull_pt[i].id].end_loc[0],
	     Bar[hull_pt[i].id].end_loc[1],
	     Bar[hull_pt[i].id].end_loc[2],
	     hull_pt[i].slope);
    }
  }
#endif

  return num_hull_pts;
}


int BarStruct3D :: build_lower_hull(int num_grnd_pts,
				    int *grnd_pt,
				    HullPt *hull_pt)
{
  int i;
  int num_hull_pts;
  Real slope=0.0;

#ifndef OPTIMIZE
  if (Print[12])
    printf("Building lower hull:\n");
#endif

  // Create upper hull.
  num_hull_pts = 1;
  hull_pt[1].id = grnd_pt[num_grnd_pts-2];
  hull_pt[0].slope = calc_slope(hull_pt[0].id,
				hull_pt[1].id);

  for (i=num_grnd_pts-3; i>=0; i--) {
    for (; num_hull_pts>0; num_hull_pts--) {
      slope = calc_slope(hull_pt[num_hull_pts].id,
			 grnd_pt[i]);
#ifndef OPTIMIZE
      if (Print[12]) {
	printf("(%d:%d) %d-%d: %.4f vs %d-%d: %.4f\n",
	       i, num_hull_pts,
	       hull_pt[num_hull_pts-1].id, hull_pt[num_hull_pts].id,
	       hull_pt[num_hull_pts-1].slope,
	       hull_pt[num_hull_pts].id, grnd_pt[i], slope);
      }
#endif
      if (slope < hull_pt[num_hull_pts-1].slope)
	break;
#ifndef OPTIMIZE
      if (Print[12])
	printf("replace!\n");
#endif
    }
    if (num_hull_pts == 0) {
      slope = calc_slope(hull_pt[0].id, grnd_pt[i]);
    }
    hull_pt[num_hull_pts].slope = slope;
    num_hull_pts++;
    hull_pt[num_hull_pts].id = grnd_pt[i];
  }

#ifndef OPTIMIZE
  if (Print[12]) {
    printf("Hull points: %d\n", num_hull_pts);
    for (i=0; i<num_hull_pts; i++) {
      printf("%d: %.4f %.4f %.4f  (%.4f)\n",
	     hull_pt[i].id,
	     Bar[hull_pt[i].id].end_loc[0],
	     Bar[hull_pt[i].id].end_loc[1],
	     Bar[hull_pt[i].id].end_loc[2],
	     hull_pt[i].slope);
    }
  }
#endif

  return num_hull_pts;
}


int BarStruct3D :: is_balanced_1pt(int num_grnd_pts,
				   int *grnd_pt)
{
#ifndef OPTIMIZE
  if (Print[12]) {
    printf("One ground point.\n");
  }
#endif
  if ((COM[0] == Bar[grnd_pt[0]].end_loc[0]) &&
      (COM[2] == Bar[grnd_pt[0]].end_loc[2])) {
    // COM is over the ground point.
    return TRUE;
  }
    
  grnd_pt[1] = grnd_pt[0];
  return FALSE;
}


int BarStruct3D :: is_balanced_2pt(int num_grnd_pts,
				   int *grnd_pt)
{
  Real r = calc_COM_proj(grnd_pt[0], grnd_pt[1]);

#ifndef OPTIMIZE
  if (Print[12]) {
    printf("Two ground points.\n");
    printf("%d: r=%.4f\n", grnd_pt[0], r);
  }
#endif

  if ((r < 0.0) || (r > 1.0)) {
    // If two ground points are not at equal heights, rotate on lower.
    if (Bar[grnd_pt[0]].end_loc[1] < Bar[grnd_pt[1]].end_loc[1]) {
      // Must rotate on point 0.
      grnd_pt[1] = grnd_pt[0];
      return FALSE;

    } else if (Bar[grnd_pt[0]].end_loc[1] > Bar[grnd_pt[1]].end_loc[1]) {
      // Must rotate on point 1.
      grnd_pt[0] = grnd_pt[1];
      return FALSE;
    }

    // Groudn points are at equal heights, rotate on one closer to COM.
    if (r < 0.0) {
      // Must rotate on point 0.
      grnd_pt[1] = grnd_pt[0];
      return FALSE;
      
    } else if (r > 1.0) {
      // Must rotate on point 1.
      grnd_pt[0] = grnd_pt[1];
      return FALSE;
    }
  }


#ifdef OLD
  if (r < 0.0) {
    // Must rotate on point 0.
    grnd_pt[1] = grnd_pt[0];
    return FALSE;
      
  } else if (r > 1.0) {
    // Must rotate on point 1.
    grnd_pt[0] = grnd_pt[1];
    return FALSE;
  }
#endif

  Real slope_hull = calc_slope(grnd_pt[0], grnd_pt[1]);
  Real slope_COM = calc_slope_COM(grnd_pt[0]);

#ifndef OPTIMIZE
  if (Print[12]) {
    printf("slope_hull: %.4f  slope_COM: %.4f: %f %d\n",
	   slope_hull, slope_COM, slope_hull-slope_COM,
	   slope_hull==slope_COM);
  }
#endif
 
  if (slopes_equal(slope_COM, slope_hull)) {
    // COM is over the edge between the two ground points -> balanced.
    return TRUE;
  }

  // Not balanced, rotate on edge.
  return FALSE;
}


// 3 points -> 2 unique ground points.
// Possibly a special case where the entire model is 2D.
int BarStruct3D :: is_balanced_3pt(int *grnd_pt, HullPt *hull_pt)
{
#ifndef OPTIMIZE
  if (Print[12]) {
    printf("Points on a line:\n");
    printf("%d - %d\n", hull_pt[0].id, hull_pt[1].id);
  }
#endif

  hull_pt[0].slope_COM = calc_slope_COM(hull_pt[0].id);
  if (slopes_equal(hull_pt[0].slope_COM, hull_pt[0].slope)) {
    Real r = calc_COM_proj(hull_pt[0].id, hull_pt[1].id);

#ifndef OPTIMIZE
    if (Print[12]) {
      printf("Slopes are equal. r= %.4f\n", r);
    }
#endif

    if (r < 0.0) {
      // Must rotate on point 0.
      grnd_pt[0] = hull_pt[0].id;
      grnd_pt[1] = grnd_pt[0];
      return FALSE;

    } else if (r > 1.0) {
      // Must rotate on point 0.
      grnd_pt[1] = hull_pt[1].id;
      grnd_pt[0] = grnd_pt[1];
      return FALSE;
    }

    return TRUE;
  }
    

  hull_pt[0].slope_COM = calc_slope_COM(hull_pt[0].id);
#ifndef OPTIMIZE
  if (Print[12]) {
    printf("slope hull: %f   slope com: %f (%f).\n",
	   hull_pt[0].slope, hull_pt[0].slope_COM,
	   (hull_pt[0].slope/hull_pt[0].slope_COM));
  }
#endif

  grnd_pt[0] = hull_pt[0].id;
  grnd_pt[1] = hull_pt[1].id;
  return FALSE;
}



// Returns TRUE if object is balanced,
// returns FALSE if object needs to be rotated to settle.
// If FALSE, the first two points in grnd_pt are the edge
// to rotate about - if these are the same, then rotate
// about that point.
// Assumes ground points are sorted via increasing x coords.
int BarStruct3D :: is_balanced(int num_grnd_pts,
			       int *grnd_pt)
{
  int i;
  int num_pts, inside, inside_line;
  int min_index;
  Real min_r=0.0, r_val;
  HullPt hull_pt[num_grnd_pts+10];

  // Handle special cases of 1 and 2 ground points specially.
  if (num_grnd_pts == 1) {
    return is_balanced_1pt(num_grnd_pts, grnd_pt);
    
  } else if (num_grnd_pts == 2) {
    // Linear case.
    return is_balanced_2pt(num_grnd_pts, grnd_pt);
  }


  // At least 3 points.
  // Test special case of COM above a ground point.
  for (i=0; i<num_grnd_pts; i++) {
    if ((absv(COM[0]-Bar[grnd_pt[i]].end_loc[0]) < DIST_TOLERANCE) &&
	(absv(COM[2]-Bar[grnd_pt[i]].end_loc[2]) < DIST_TOLERANCE)) {
      // COM is over a ground point -> structure is balanced.
#ifndef OPTIMIZE
      if (Print[12]) {
	printf("COM is over ground point #%d -> Balanced.\n", i);
      }
#endif
      return TRUE;
    }
  }


  num_pts = build_upper_hull(num_grnd_pts, grnd_pt,
			     hull_pt);
  num_pts += build_lower_hull(num_grnd_pts, grnd_pt,
			      &(hull_pt[num_pts-1]));
#ifndef OPTIMIZE
  if (Print[12]) {
    printf("Number of Hull points: %d\n", num_pts);
    for (i=0; i<num_pts; i++) {
      printf("%d: %.4f %.4f %.4f  (%.4f)\n",
	     hull_pt[i].id,
	     Bar[hull_pt[i].id].end_loc[0],
	     Bar[hull_pt[i].id].end_loc[1],
	     Bar[hull_pt[i].id].end_loc[2],
	     hull_pt[i].slope);
    }
  }
#endif


  if (num_pts == 3) {
    return is_balanced_3pt(grnd_pt, hull_pt);
  }


  /********* Standard Case **********/

  inside = TRUE;
  min_index = -1;

  // Calculate slopes for initial point.
  hull_pt[0].r = calc_COM_proj(hull_pt[0].id, hull_pt[1].id);
  hull_pt[0].slope_COM = calc_slope_COM(hull_pt[0].id);
  if (!COM_above_line(hull_pt[0].id, hull_pt[1].id)) {
    inside_line = TRUE;

  } else if (slopes_equal(hull_pt[0].slope_COM, hull_pt[0].slope)) {
    inside_line = TRUE;
#ifndef OPTIMIZE
    if (Print[12]) {
      printf("First edge: On line between hullpoints.\n");
    }
#endif

  } else {
    inside_line = FALSE;
  }

  if (inside_line == FALSE) {
#ifndef OPTIMIZE
    if (Print[12]) {
      printf("First: %d(%d)-%d: inside = FALSE.  r=%f\n",
	     hull_pt[0].id, 0, hull_pt[1].id, hull_pt[0].r);
    }
#endif

    if ((hull_pt[0].r>=0.0) && (hull_pt[0].r<=1.0)) {
#ifndef OPTIMIZE
      if (Print[12]) {
	printf("First edge (%d:%d): r: %.4f,  slope hull: %f,  slope com: %f  (%f).\n",
	       hull_pt[0].id, hull_pt[1].id,
	       hull_pt[0].r, hull_pt[0].slope, hull_pt[0].slope_COM,
	       hull_pt[0].slope_COM/hull_pt[0].slope);
      }
#endif
      grnd_pt[0] = hull_pt[0].id;
      grnd_pt[1] = hull_pt[1].id;
      return FALSE;
    }

    inside = FALSE;
    hull_pt[0].inside = FALSE;
    min_index = 0;
    if (hull_pt[0].r < 0.0) {
      r_val = -hull_pt[0].r;
    } else {
      r_val = hull_pt[0].r - 1.0;
    }
    min_r = r_val;

  } else {
    hull_pt[0].inside = TRUE;
#ifndef OPTIMIZE
      if (Print[12]) {
	printf("First: %d(%d)-%d: inside = TRUE.\n", hull_pt[0].id, 0,
	       hull_pt[1].id);
      }
#endif
  }


  // Copy info from first point to last point.
  hull_pt[num_pts-1].slope = hull_pt[0].slope;

  // Add first edge to end.
  hull_pt[num_pts].id = hull_pt[1].id;


  for (i=1; i<num_pts; i++) {
    if (!COM_above_line(hull_pt[i].id, hull_pt[i+1].id)) {
      inside_line = TRUE;

    } else {
      hull_pt[i].slope_COM = calc_slope_COM(hull_pt[i].id);
      if (slopes_equal(hull_pt[i].slope_COM, hull_pt[i].slope)) {
	inside_line = TRUE;
#ifndef OPTIMIZE
	if (Print[12]) {
	  printf("On line between hullpoints: %d:%d\n", hull_pt[i].id,
		 hull_pt[i+1].id);
	}
#endif
      } else {
	inside_line = FALSE;
      }
    }

    if (inside_line == FALSE) {
      // COM above upper hull -> outside footprint.
      inside = FALSE;
      hull_pt[i].inside = FALSE;
      hull_pt[i].r = calc_COM_proj(hull_pt[i].id, hull_pt[i+1].id);

#ifndef OPTIMIZE
      if (Print[12]) {
	printf("%d(%d)-%d: above line.\n", (int)hull_pt[i].id, i,
	       (int)hull_pt[i+1].id);
	printf("%d: r: %.4f  hull_slope: %.4f\n",
	       (int)hull_pt[i].id, (float)hull_pt[i].r,
	       hull_pt[i].slope);
      }
#endif
	
      if ((hull_pt[i].r >= 0) && (hull_pt[i].r <= 1.0)) {
#ifndef OPTIMIZE
	if (Print[12]) {
	  printf("%d: r: %.4f  hull_slope: %.4f  COM_slope: %.4f\n",
		 hull_pt[i].id, hull_pt[i].r,
		 hull_pt[i].slope, hull_pt[i].slope_COM);
	  printf("Rotate on edge: %d - %d.\n",
		 hull_pt[i].id, hull_pt[i+1].id);
	}
#endif
	// i - i+1 is the edge to rotate on.
	grnd_pt[0] = hull_pt[i].id;
	grnd_pt[1] = hull_pt[i+1].id;
	return FALSE;

      }

      // Test if this is best edge to rotate on so far.
      // r_val is closeness of r to range 0-1.0.
      if (hull_pt[i].r < 0.0) {
	r_val = -hull_pt[i].r;
      } else {
	r_val = hull_pt[i].r - 1.0;
      }

      if ((r_val < min_r) || (min_index == -1)) {
	min_r = r_val;
	min_index = i;
      }

    } else {
      hull_pt[i].inside = TRUE;
#ifndef OPTIMIZE
      if (Print[12]) {
	printf("%d - %d: inside TRUE. (prev: %d)\n",
	       hull_pt[i].id, hull_pt[i+1].id,
	       hull_pt[i-1].inside);
      }
#endif
    }
  }

  if (inside == FALSE) {
#ifndef OPTIMIZE
    if (Print[12]) {
      printf("Not Balanced.  Minimum is : %d (%.4f) - %d\n",
	     hull_pt[min_index].id, min_r, hull_pt[min_index+1].id);
    }
#endif
    grnd_pt[0] = hull_pt[min_index].id;
    grnd_pt[1] = hull_pt[min_index+1].id;
    return FALSE;
  }


#ifndef OPTIMIZE
  if (Print[12])
    printf("Balanced2.\n");
#endif

  if (inside == TRUE)
    return TRUE;

  printf("is_balanced() - error, did not find rotation point/edge.\n");
  grnd_pt[0] = -1;

  return FALSE;  
}




// Assumes pivot[0].x < pivot[1].x
int BarStruct3D :: rotate_to_balance(int *pivot)
{
  int p0 = pivot[0];
  int p1 = pivot[1];
  //  int rotate_full = (p0 != p1);
  int rotate_full = FALSE;
  Real angle = 0.0;

#ifndef OPTIMIZE
  if (Print[12])
    printf("Rotate to balance: %d %d\n", p0, p1);
#endif

  if (p0 != p1) {
    // Need to rotate to align edge with z-axis.
    if (absv(Bar[p0].end_loc[0]-
	     Bar[p1].end_loc[0])>DIST_TOLERANCE) {
      // Rotate to align.
      angle = -atan((Bar[p0].end_loc[0]-Bar[p1].end_loc[0])/
		    (Bar[p0].end_loc[2]-Bar[p1].end_loc[2]));
#ifndef OPTIMIZE
      if (Print[12]) {
	printf("%.4f / %.4f = %.4f(%.4f)\n",
	       Bar[p0].end_loc[0]-Bar[p1].end_loc[0],
	       Bar[p0].end_loc[2]-Bar[p1].end_loc[2],
	       (Bar[p0].end_loc[0]-Bar[p1].end_loc[0])/
	       (Bar[p0].end_loc[2]-Bar[p1].end_loc[2]),
	       angle);
      }
#endif

      rotate_struct_Yaxis(angle);
    }

    if (Bar[p0].end_loc[0] < Bar[p1].end_loc[0]) {
      if (COM[0] > Bar[p1].end_loc[0])
	p0 = p1;
    } else if (COM[0] < Bar[p1].end_loc[0]) {
      p0 = p1;
    }

  } else {
#ifndef OPTIMIZE
    if (Print[12])
      printf("p0 == p1\n");
#endif
    // Align pivot and COM along x-axis.
    if (absv(Bar[p0].end_loc[2]-COM[2])>DIST_TOLERANCE) {
      angle = atan((Bar[p0].end_loc[2]-COM[2])/
		   (Bar[p0].end_loc[0]-COM[0]));
#ifndef OPTIMIZE
      if (Print[12]) {
	printf("Rotating to align COM with pivot on X-axis (%d):\n  %.4f/%.4f=%.4f -> %.4f\n",
	       p0,
	       (Bar[p0].end_loc[2]-COM[2]),
	       (Bar[p0].end_loc[0]-COM[0]),
	       ((Bar[p0].end_loc[2]-COM[2])/
		    (Bar[p0].end_loc[0]-COM[0])),
	       angle);
      }
#endif

      rotate_struct_Yaxis(angle);
    }
#ifndef OPTIMIZE
    else if (Print[12]) {
      printf("%.4f-%.4f < TOLERANCE.\n",
	     Bar[p0].end_loc[2], COM[2]);
    }
#endif
  }

  // Rotate about p0.
  if (COM[0] < Bar[p0].end_loc[0])
    p1 = settle_left(p0, rotate_full);
  else
    p1 = settle_right(p0, rotate_full);

  if (angle != 0.0) {
    rotate_struct_Yaxis(-angle);
  }

#ifndef OPTIMIZE
  if (Print[12]) {
    printf("%d(%f:%f:%f)\n%d(%f:%f:%f)\n", p0,
	   Bar[p0].end_loc[0], Bar[p0].end_loc[1],
	   Bar[p0].end_loc[2], p1,
	   Bar[p1].end_loc[0], Bar[p1].end_loc[1],
	   Bar[p1].end_loc[2]);
  }
#endif
  return TRUE;
}
  

int BarStruct3D :: find_ground_pts(int *grnd_pt)
{
  int i, j, k, num_pts;
  Real min_y;

  // Find ground points.
  min_y = Bar[0].end_loc[1];
  for (i=1; i<Num_bars; i++) {
    if (Bar[i].end_loc[1] < min_y)
      min_y = Bar[i].end_loc[1];
  }

  num_pts = 0;
  for (i=0; i<Num_bars; i++) {
    if (absv(Bar[i].end_loc[1]-min_y) < DIST_TOLERANCE) {
      // Another ground point.
      grnd_pt[num_pts++] = i;
    }
#ifndef OPTIMIZE
      else if (Print[12]) {
	if (grnd_pt[num_pts-1] == i) {
	  printf("Not included: %d(%f:%f:%f) min:%f(%f)\n",
		 i, Bar[i].end_loc[0], Bar[i].end_loc[1],
		 Bar[i].end_loc[2], min_y,
		 (Bar[i].end_loc[1]-min_y));
	}
      }
#endif
  }

  // Sort by smallest x (smallest z for same x).
  OBar = &Bar[0];
  qsort((void*)grnd_pt, num_pts, sizeof(int),
	pt_compare);

#ifndef OPTIMIZE
  if (Print[12]) {
    printf("Ground points: %d  (find_grnd_pts 1)\n", num_pts);
    for (i=0; i<num_pts; i++) {
      printf("%d(%d): ", grnd_pt[i], i);
      v_print(Bar[grnd_pt[i]].end_loc, 0);
    }
    printf("CoM: %.6f, %.6f %.6f\n", COM[0], COM[1], COM[2]);
  }
#endif

  // Need to delete redundant points (those with same x and z).
  for (i=0; i<num_pts-1; i++) {
    for (j=i+1; j<num_pts; j++) {
      if (sqrt(sqr(Bar[grnd_pt[j]].end_loc[0]-Bar[grnd_pt[i]].end_loc[0]) +
	       sqr(Bar[grnd_pt[j]].end_loc[2]-Bar[grnd_pt[i]].end_loc[2]))
	  < DIST_TOLERANCE) {

	// A redundant point.
	if (Print[12]) {
	  printf("Redundant %d(%.3f:%.3f) w/ %d(%.3f:%.3f)\n",
		 grnd_pt[i], Bar[grnd_pt[i]].end_loc[0],
		 Bar[grnd_pt[i]].end_loc[2],
		 grnd_pt[j], Bar[grnd_pt[j]].end_loc[0],
		 Bar[grnd_pt[j]].end_loc[2]);
	}

	num_pts--;
	for (k=j; k<num_pts; k++)
	  grnd_pt[k] = grnd_pt[k+1];
	j--;
      }
    }
  }
      
#ifndef OPTIMIZE
  if (Print[12]) {
    printf("Ground points: %d (find_grnd_pts 2)\n", num_pts);
    for (i=0; i<num_pts; i++) {
      printf("%d(%d): ", grnd_pt[i], i);
      v_print(Bar[grnd_pt[i]].end_loc, 0);
    }
    printf("CoM: %.6f, %.6f %.6f\n", COM[0], COM[1], COM[2]);
  }
#endif

  return num_pts;
}


// Distributes mass of each vertex to the nearest ground point.
void BarStruct3D :: distribute_mass(void)
{
  int i;
  int ctr, index, index2, stack_ptr, dist;
  int node_stack[Num_bars];
  int num_visited[Num_bars];
  int from_list[Num_bars][3];
  Real support_weight[Num_bars];

  if (num_grnd_pts == 0)
    return;

  // Test to see if re-distribution is necessary.
  if (num_grnd_pts == num_grnd_pts_prev) {
    int recalc = FALSE;

    for (i=0; i<num_grnd_pts; i++) {
      if (ground_pt[i] != ground_pt_prev[i]) {
	recalc = TRUE;
	//	printf("Ground point %d different: %d - %d.\n", i,
	//	       ground_pt[i], ground_pt_prev[i]);
	break;
      }
    }
    if (!recalc) {
      // Don't need to re-distribute mass.
      return;
    }
  }
  /*
 else {
    printf("Num ground points different: %d (%d)\n",
	   num_grnd_pts, num_grnd_pts_prev);
  }
  */

  // Put ground points on node list.
  stack_ptr = 0;
  for (i=0; i<Num_bars; i++) {
    if ((Bar[i].ground_pt) && (Bar[i].ground_pt_prev)) {
      Bar[i].updated = 1;
      node_stack[stack_ptr] = i;
      stack_ptr++;

    } else {
      Bar[i].updated = 0;
    }
  }

  /*
  printf("Ground points: %d (%d).\n", stack_ptr, num_grnd_pts);
  for (i=0; i<stack_ptr; i++) {
    printf("%d: %.5f:%.7f:%.5f (%d)\n", ground_pt[i],
	   Bar[node_stack[i]].end_loc[0],
	   Bar[node_stack[i]].end_loc[1],
	   Bar[node_stack[i]].end_loc[2],
	   Bar[node_stack[i]].ground_pt);
  }
  */

  // Initialize variables.
  for (i=0; i<Num_bars; i++) {
    num_visited[i] = 0;
    support_weight[i] = Bar[i].weight;
  }


  // Do a breadth first search to find distance from ground points.
  ctr = 0;
  while (ctr < Num_bars) {
    index = node_stack[ctr];
    ctr++;
    dist = Bar[index].updated + 1;

    // Test parent.
    if (index != 0) {
      index2 = Bar[index].parent;
      if (Bar[index2].updated == 0) {
	Bar[index2].updated = dist;

	from_list[index2][0] = index;
	num_visited[index2] = 1;

	node_stack[stack_ptr] = index2;
	stack_ptr++;

      } else if (Bar[index2].updated == dist) {
	if (num_visited[index2] < 3) {
	  from_list[index2][num_visited[index2]] = index;
	  num_visited[index2]++;
	}
      }
    }

    // Test children.
    for (i=0; i<Bar[index].num_children; i++) {
      index2 = Bar[index].child[i];
      if (Bar[index2].updated == 0) {
	Bar[index2].updated = dist;

	from_list[index2][0] = index;
	num_visited[index2] = 1;

	node_stack[stack_ptr] = index2;
	stack_ptr++;

      } else if (Bar[index2].updated == dist) {
	if (num_visited[index2] < 3) {
	  from_list[index2][num_visited[index2]] = index;
	  num_visited[index2]++;
	}
      }
    }
  }

  // Go in reverse to distribute the weight.
  ctr = Num_bars-1;
  while (ctr >= num_grnd_pts) {
    index = node_stack[ctr];
    ctr--;

    // Distribute weight among its "children".
    if (num_visited[index] == 1) {
      support_weight[from_list[index][0]] += support_weight[index];

    } else if (num_visited[index] == 2) {
      support_weight[from_list[index][0]] += support_weight[index]*0.5;
      support_weight[from_list[index][1]] += support_weight[index]*0.5;

    } else if (num_visited[index] == 3) {
      support_weight[from_list[index][0]] += support_weight[index]*0.333333;
      support_weight[from_list[index][1]] += support_weight[index]*0.333333;
      support_weight[from_list[index][2]] += support_weight[index]*0.333333;
    }
  }

  for (i=0; i<num_grnd_pts; i++) {
    Bar[ground_pt[i]].weight_supported = support_weight[ground_pt[i]];
  //    Bar[node_stack[i]].weight_supported = support_weight[node_stack[i]] / total_weight;
  }

  // Print tree.
  /*
  for (i=0; i<Num_bars; i++) {
    printf("%d: w:%.4f:%.4f; d:%d,  #:%d: ", i, Bar[i].weight,
	   support_weight[i], Bar[i].updated, num_visited[i]);
    for (int j=0; j<num_visited[i]; j++) {
      printf("%d ", from_list[i][j]);
    }
    printf("\n");
  }

  for (i=0; i<num_grnd_pts; i++) {
    printf("%d: %.4f (%.4f).\n", node_stack[i],
    	   Bar[node_stack[i]].weight_supported, total_weight);
  }
  printf("\n");
  */
}



// Translates and rotates the structure until it
// is in a stable state, resting on the surface.
int BarStruct3D :: settle_struct(void)
{
  int i, index;
  int settle_attempts = 0;
  int max_tries = Num_bars;
  Real min_y, move_x, move_z;
  Real val, val_sum;

#ifndef OPTIMIZE
  if (Print[12]) {
    printf("\n\n -- Settle Struct (ref: %d) -- \n", ref_pt);
    printf("\n  Attempt #0.\n");
    print_struct();
  }
#endif

  // First, find ground points.
  // Convert to convex hull.
  // Determine if need to adjust and find that edge.
  // Rotate, settle and un-rotate.
  Rotating = FALSE;
  if (max_tries < 10)
    max_tries = 10;
  num_grnd_pts = find_ground_pts(ground_pt);
  while (!is_balanced(num_grnd_pts, ground_pt)) {
    if (settle_attempts++ > max_tries) {
      printf("Settle struct failed:  Too many attempts: %d.\n", settle_attempts);
      if (Print[12] != FALSE) {
	//      while(1);
	return FALSE;
      }
      // Turn printing on so as to catch the error.
      print_struct();
      Print[12] = TRUE;
      settle_attempts = 0;
      num_grnd_pts = find_ground_pts(ground_pt);
      continue;

    } else if (ground_pt[0] < 0) {
      printf("Settle struct failed.  Invalid ground point: %d.\n", ground_pt[0]);
      return FALSE;
    }
      
#ifndef OPTIMIZE
    if (Print[12]) {
      printf("Pivots: %d %d\n", ground_pt[0], ground_pt[1]);
    }
#endif
    rotate_to_balance(ground_pt);

#ifndef OPTIMIZE
    if (Print[12]) {
      printf("\n  Attempt #%d.\n", settle_attempts);
    }
#endif
    num_grnd_pts = find_ground_pts(ground_pt);
  } 

#ifndef OPTIMIZE
  if (Print[14])
    printf("Total number of settle attempts: %d.\n", settle_attempts);
#endif

  // Settled!
  // Find minimum y.
  min_y = Bar[0].end_loc[1];
  for (i=1; i<Num_bars; i++) {
    if (Bar[i].end_loc[1] < min_y)
      min_y = Bar[i].end_loc[1];
  }

  // Determine the ground points.
  for (i=0; i<Num_bars; i++) {
    if (Bar[i].end_loc[1] - min_y < DIST_TOLERANCE) {
      Bar[i].ground_pt = TRUE;
    } else {
      Bar[i].ground_pt = FALSE;
    }
  }

  // Determine sequential ground points.
  num_grnd_pts = 0;
  for (i=0; i<Num_bars; i++) {
    if (Bar[i].ground_pt && Bar[i].ground_pt_prev) {
      ground_pt[num_grnd_pts] = i;
      num_grnd_pts++;
    }
  }


#ifdef NEW
  // Find average rotational displacement among sequential ground points.
  if (num_grnd_pts > 1) {
    //    printf("\n %f  %f\n", move_x, move_z);
    Real avg_angle = 0.0;
    for (i=0; i<num_grnd_pts; i++) {
      index = ground_pt[i];
      avg_angle += angle_btw_pts(COM[0], COM[2],
				 Bar[index].prev_loc[0],
				 Bar[index].prev_loc[2],
				 Bar[index].end_loc[0],
				 Bar[index].end_loc[2])
	* Bar[index].weight_supported;
      //      printf("%d: angle %.4f (%.4f)\n", index, avg_angle,
      //	     Bar[index].weight_supported);
    }

    if (avg_angle > ANGLE_TOLERANCE) {
      //      printf("avg_angle: %.4f\n", avg_angle);

      // Translate COM to 0,0,0
      Real tmp_x = COM[0], tmp_z = COM[2];
      translate_struct(-tmp_x, 0.0, -tmp_z);
      
      // Rotate
      rotate_struct_Yaxis(avg_angle);
    
      // Restore COM location.
      translate_struct(tmp_x, 0.0, tmp_z);
    }
  }
#endif



  // Find average linear displacement of sequential ground points.`
  move_x = 0.0;
  move_z = 0.0;
  if (num_grnd_pts > 0) {
    // Distribute mass on ground points.
    distribute_mass();

    for (i=0; i<num_grnd_pts; i++) {
      int index = ground_pt[i];
      //      move_x += Bar[index].prev_loc[0] - Bar[index].end_loc[0];
      //      move_z += Bar[index].prev_loc[2] - Bar[index].end_loc[2];
      move_x += (Bar[index].prev_loc[0] - Bar[index].end_loc[0])
	* Bar[index].weight_supported;
      move_z += (Bar[index].prev_loc[2] - Bar[index].end_loc[2])
	* Bar[index].weight_supported;
    }

    // Set ref_pt as one with largest weight on it.
    int max_index = ground_pt[0];
    Real max_weight = Bar[max_index].weight_supported;
    for (i=0; i<num_grnd_pts; i++) {
      if (Bar[ground_pt[i]].weight_supported > max_weight) {
	max_index = ground_pt[i];
	max_weight = Bar[max_index].weight_supported;
      }
    }
    ref_pt = max_index;
  }
  translate_struct(move_x, -min_y, move_z);


  // Swap ground points.
  int *tmp = ground_pt;
  ground_pt = ground_pt_prev;
  ground_pt_prev = tmp;
  num_grnd_pts_prev = num_grnd_pts;


  val_sum = 0.0;
  for (i=0; i<num_grnd_pts; i++) {
      index = ground_pt[i];
      val = (Bar[index].prev_loc[0] - Bar[index].end_loc[0])
	* Bar[index].weight_supported;
      val += (Bar[index].prev_loc[2] - Bar[index].end_loc[2])
	* Bar[index].weight_supported;
      drag_distance += val;
      drag_distance_abs2 += absv(val);
      val_sum += val;
  }
  drag_distance_abs1 += absv(val_sum);

  for (i=0; i<Num_bars; i++) {
    Bar[i].ground_pt_prev = Bar[i].ground_pt;
  }


#ifndef OPTIMIZE
  if (Print[12]) {
    printf("Balanced, translating to ground (pts: %d).\n", num_grnd_pts);
  }
#endif

#ifndef OPTIMIZE
  if (Print[12]) {
    print_struct();

    if (Print[15]) {
      printf("Recalc positions: %.3f\n", Time);
      calculate_positions_new(Time);
      print_struct();
    }
  }
#endif

#ifndef OPTIMIZE
  if (Print[12]) {
    printf("-- Settled --\n  %.1f:%.1f:%.1f  (%.3f:%.3f:%.3f)\n  CoM: %.3f, %.3f, %.3f.\n\n",
	   Bar[1].end_loc[0], Bar[1].end_loc[1],
	   Bar[1].end_loc[2], move_x, min_y, move_z,
	   COM[0], COM[1], COM[2]);
  }
#endif

#ifndef OPTIMIZE
  if (Print[17])
    print_struct();
  else if (Print[18])
    printf("CoM: %.6f, %.6f %.6f  (Time: %.4f)\n", COM[0], COM[1], COM[2], Time);
#endif
  return TRUE;
}


int BarStruct3D :: settle_structure(void)
{
  return settle_struct();
}


/*********************************************************************/

Real Angle_sum = 0.0;

void BarStruct3D :: update_ball_sensors(void)
{
  int i, index, parent;
  Real angle;
  Real vec_to_ball[3];
  Real vec_sensor[3];

  for (i=0; i<Num_sensors; i++) {
    index = Sensor_index[i];
    vec_subt(vec_to_ball, Ball_loc, Bar[index].end_loc);

    parent = Bar[index].parent;
    vec_subt(vec_sensor, Bar[index].end_loc, Bar[parent].end_loc);

    angle = angle_btw_vec3(vec_to_ball, vec_sensor) / PI;

    /*
    v_print(vec_to_ball, "to ball: ");
    v_print(vec_sensor, "sensor: ");
    printf("angle: %.4f\n\n", angle);
    */

    if (angle < 0.4) {
      Bar[index].sensor_value = 1.0 - angle;
      Angle_sum += Bar[index].sensor_value - 0.5;
      //      printf("Angle_sum: %.4f  (%.4f)\n", Angle_sum, Bar[index].sensor_value);
    } else {
      Bar[index].sensor_value = 0.0;
    }
    //    printf("angle: %.4f\n", angle);
  }
}



void BarStruct3D :: update_ball_position(void)
{
  int i;
  Real dist;

  for (i=0; i<Num_bars; i++) {
    dist = sqr(Bar[i].end_loc[0] - Ball_loc[0]) +
      sqr(Bar[i].end_loc[1] - Ball_loc[1]) +
      sqr(Bar[i].end_loc[2] - Ball_loc[2]);

    /*
    if (dist < sqr(Ball_radius + 2.0)) {
      printf("Collision w/ ball (%d:%.2f:%.2f:%.2f)\n", i,
	     Bar[i].end_loc[0], Bar[i].end_loc[1], Bar[i].end_loc[2]);
    }
    */
  }

  if (Num_sensors > 0) {
    update_ball_sensors();
  }
}


Real BarStruct3D :: animate_struct(Real time_length)
{
  Real prev_time=Time;
  Real time, start_time;

  start_time = Time;
  time = Time + time_length;
  for (; Time <= time; ) {
    if (Build_type == 5) {
      if (!network->output_changed()) {
	// Reached a stable state.
	Time = time + TIME_STEP;
	continue;
      }
    }  

    calculate_positions(Time);
    if (Print[34]) {
      printf("Calculated positions (ref: %d)...\n", ref_pt);
      print_struct();
      printf("Recalc positions: %.4f\n", Time);
      calculate_positions_new(Time);
    }

    if (is_intersection()) {
      // Bars collide.
      if (Print[9]) {
	printf("Bars collided at time: %.4f.\n", Time);
      }
      return Time-start_time;
    }

    if (!settle_structure()) {
      printf("Animate_struct() - structure failed to settle at time: %.4f.\n", Time);
      return -1.0;
    }

    if (Is_ball) {
      update_ball_position();
    }

    if (Print[34]) {
      printf("Settled positions...\n");
      print_struct();
      printf("Recalc positions: %.4f\n", Time);
      calculate_positions_new(Time);
      printf("\n\n");
    }


    if (COM[0] < com_min[0])
      com_min[0] = COM[0];
    else if (COM[0] > com_max[0])
      com_max[0] = COM[0];

    if (COM[1] < com_min[1])
      com_min[1] = COM[1];
    else if (COM[1] > com_max[1])
      com_max[1] = COM[1];

    if (COM[2] < com_min[2])
      com_min[2] = COM[2];
    else if (COM[2] > com_max[2])
      com_max[2] = COM[2];

    if (win1 != 0) {
      if (time - prev_time > DRAW_STEP) {
	draw_struct();
	prev_time = time;
      }
    }

    if (!Rotating)
      Time += TIME_STEP;
  }

  return time_length;
}


void BarStruct3D :: make_delta_rot_mat(int index, Real angle)
{
  if (Bar[index].rot_axis == ROT_Y_AXIS) {
    make_rotY_mat(angle, Bar[index].delta_rot_mat);

  } else if (Bar[index].rot_axis == ROT_Z_AXIS) {
    make_rotZ_mat(angle, Bar[index].delta_rot_mat);

  } else {
    make_identity_mat(Bar[index].delta_rot_mat);
  }
}


void calc_rel_heading(Real relative[3][3], Real parent[3][3], Real child[3][3])
{
  Real parent_inverse[3][3];

  mm3_transpose(parent_inverse, parent); // get inverse/transpose
  //  mmm3_mult(relative, child, parent_inverse);
  mmm3_mult(relative, parent_inverse, child);

#ifndef OPTIMIZE
  /*
  if (Print[12]) {
    m3_print(parent, "Parent:\n");
    m3_print(child, "Child:\n");
    m3_print(relative, "Relative rotation matrix:\n");
  }
  */
#endif
}

int headings_equal(Real heading1[3][3], Real heading2[3][3])
{
  if (heading1[0][0] != heading2[0][0])
    return FALSE;

  if (heading1[1][0] != heading2[1][0])
    return FALSE;

  if (heading1[2][0] != heading2[2][0])
    return FALSE;
  return TRUE;
}

#ifdef OLD
  // Compare headings with other children for redundancy.
  for (i=0; i<Bar[parent].num_children; i++) {
    child = Bar[parent].child[i];
    if (headings_equal(Bar[child].delta_rot_mat, relative_heading)) {
      *p_index = child;
#ifndef OPTIMIZE
      if (Print[12])
	printf("Same heading as: %d.\n", child);
#endif
      return TRUE;
    }
  }
#endif

int headings_equal_all(Real heading1[3][3], Real heading2[3][3])
{
  if (heading1[0][0] != heading2[0][0])
    return FALSE;
  if (heading1[0][1] != heading2[0][1])
    return FALSE;
  if (heading1[0][2] != heading2[0][2])
    return FALSE;

  if (heading1[1][0] != heading2[1][0])
    return FALSE;
  if (heading1[1][1] != heading2[1][1])
    return FALSE;
  if (heading1[1][2] != heading2[1][2])
    return FALSE;

  if (heading1[2][0] != heading2[2][0])
    return FALSE;
  if (heading1[2][1] != heading2[2][1])
    return FALSE;
  if (heading1[2][2] != heading2[2][2])
    return FALSE;
  return TRUE;
}


// len : length of bar.
// rot_mat : rotation matrix to transform from world coords to child.
// type : joint type.
// rot_axis : rotation axis for moving joints.
// rot_angle : amount to rotate about rot_axis for joints.
// spd : speed to rotate.
// offset : phase offset.
// p_index : index of parent.

// Returns:
// FALSE : when error and this error should cause error in build.
// -1 : when fails to add bar and do not change parent.
// child index :  when add bar succeeds.
int BarStruct3D :: add_bar(Real len, Real heading[3][3], int type, int rot_axis,
			   Real rot_angle_a, Real rot_angle_b, Real spd,
			   Real offset, int *p_index)
{
  int i;
  int child, child2;
  int parent = *p_index;
  int same_as;
  Real relative_heading[3][3];


#ifndef OPTIMIZE
  if (Print[12]) {
    printf("Add bar (%d): type:%d, axis:%c, len:%.1f, ang:%.3f:%.3f, spd:%.2f, ofst:%.3f\n",
	   parent, type, rot_axis+'w', len, rot_angle_a, rot_angle_b, spd, offset);
  }
#endif

  if (Num_bars >= MAX_BARS3) {
#ifndef OPTIMIZE
    if (Print[9] || Print[12])
      printf("Maximum number of bars exceeded.\n");
#endif
    return FALSE;
  }

  if (Bar[parent].type == BAR_TYPE_SENSOR) {
    printf("Error: Parent Bar #%d is sensor.\n", parent);
    *p_index = Bar[parent].parent;
    return -1; // Will cause fatal error
  }


  // Adjust length
  if (len == 0.0)
    len = 1.0;
  else if (len > 5.0)
    len = 5.0;
  len *= 10.0;

  // Adjust angle.
  rot_angle_a = simplify_angle_180(rot_angle_a);
  rot_angle_b = simplify_angle_180(rot_angle_b);
  if (absv(rot_angle_a - rot_angle_b) < ANGLE_TOLERANCE)
    if (type == BAR_TYPE_JOINT)
      type = BAR_TYPE_FIXED;

  if (type != BAR_TYPE_JOINT) {
    spd = 0.0;
    offset = 0.0;

  } else {
    // Adjust speed.
    if (spd > 2.0)
      spd = 2.0;
    else if (spd < -2.0)
      spd = -2.0;

    // Adjust angle.
    if (rot_angle_b > PI+ANGLE_TOLERANCE) {
      rot_angle_b = -(2.0*PI - rot_angle_b);
    }
  }

  if (Bar[parent].num_children >= MAX_CHILDREN3D) {
#ifndef OPTIMIZE
    if (Print[9] || Print[12]) {
      printf("Too many children for bar #%d : %d.\n", parent, Bar[parent].num_children);
    }
#endif
    return FALSE;
  }


  // Add bar (without commitment).
  child = Num_bars;
  Bar[child].parent = parent;
  Bar[child].moves = FALSE;
  Bar[child].rot_axis = rot_axis;
  Bar[child].move_spd = spd;
  Bar[child].move_offset = offset;
  Bar[child].num_children = 0;
  Bar[child].num_from = 0;
  Bar[child].length = len;
  Bar[child].len_short = len/2.0;
  Bar[child].type = type;
  Bar[child].angle_a = rot_angle_a;
  Bar[child].angle_b = rot_angle_b;
  //  Bar[child].weight = 1.0;
  calc_rel_heading(relative_heading, Bar[parent].abs_heading, heading);  // Calculate relative heading.
  mm3_copy(Bar[child].delta_rot_mat, relative_heading);
  mm3_transpose(Bar[child].inverse_delta_rot, relative_heading);
  mm3_copy(Bar[child].abs_heading, heading);
  if (Bar[parent].type == BAR_TYPE_JOINT) {
    Bar[child].joint_index = parent;
  } else {
    Bar[child].joint_index = Bar[parent].joint_index;
  }

  // Test for conflicts when parent is a moving joint.
  if (Bar[parent].type == BAR_TYPE_JOINT) {
    // Joint bends.
    if (Bar[parent].rot_axis == ROT_Z_AXIS) {
      // Joint bends about Z axis.
      if (relative_heading[1][0] == 1.0) {
	// New bar goes on Y axis.
	if ((Bar[parent].angle_a == PI/2.0) ||
	    (Bar[parent].angle_b == PI/2.0)) {
	  // Will be a collision.
#ifndef OPTIMIZE
	  if (Print[12]) {
	    printf("Parent bends to same location (+).\n");
	    m3_print(Bar[child].delta_rot_mat, "Child delta rotation matrix:\n");
	  }
#endif
	  return -1; // TRUE;
	}

      } else if (relative_heading[1][0] == -1.0) {
	if ((Bar[parent].angle_a == -PI/2.0) ||
	    (Bar[parent].angle_b == -PI/2.0)) {
	  // Will be a collision.
#ifndef OPTIMIZE
	  if (Print[12]) {
	    printf("Parent bends to same location (-).\n");
	    m3_print(Bar[child].delta_rot_mat, "Child delta rotation matrix:\n");
	  }
#endif
	  return -1; // TRUE;
	}
      }
    }
  }


  // Calculate the bars location at time 0.
  calc_bar_position(child, 0.0);


  // Test for collision/heading conflicts with other bars.
  //  if (Bar[parent].parent >= 0) {
    if (Bar[child].delta_rot_mat[0][0] == -1.0) {
      // Collides with parent.
      if (Bar[parent].same_as == -1) {
	*p_index = parent;
      } else {
	*p_index = Bar[parent].same_as;
      }
      
#ifndef OPTIMIZE
      if (Print[12])
	printf("Same heading as parent: %d (%d).\n", parent,
	       Bar[parent].same_as);
#endif
      return -1; // TRUE;
    }
    //  }

  // Test for redundancy with existing parent's children.
  for (i=0; i<Bar[parent].num_children; i++) {
    child2 = Bar[parent].child[i];
    if ((absv(Bar[child2].delta_rot_mat[0][0] - Bar[child].delta_rot_mat[0][0]) < DIST_TOLERANCE) &&
	(absv(Bar[child2].delta_rot_mat[1][0] - Bar[child].delta_rot_mat[1][0]) < DIST_TOLERANCE) &&
	(absv(Bar[child2].delta_rot_mat[2][0] - Bar[child].delta_rot_mat[2][0]) < DIST_TOLERANCE)) {
      if (Bar[child2].type == BAR_TYPE_SENSOR) {
	; // Don't change parent.
      } else if (Bar[child2].same_as == -1) {
	*p_index = child2;
      } else {
	*p_index = Bar[child2].same_as;
      }

#ifndef OPTIMIZE
      if (Print[12])
	printf("Same heading as child: %d (%d:%d).  New parent: %d\n", child2,
	       Bar[child2].same_as, Bar[child2].type, *p_index);
#endif
      return -1; // TRUE;
    }
  }



  // Test for redunancy with parent.
  if (Bar[parent].parent >= 0) {
    child2 = Bar[parent].parent;
    if ((absv(Bar[child2].end_loc[0] - Bar[child].end_loc[0]) < DIST_TOLERANCE) &&
	(absv(Bar[child2].end_loc[1] - Bar[child].end_loc[1]) < DIST_TOLERANCE) &&
	(absv(Bar[child2].end_loc[2] - Bar[child].end_loc[2]) < DIST_TOLERANCE)) {
      if (Bar[child2].same_as == -1) {
	*p_index = child2;
      } else {
	*p_index = Bar[child2].same_as;
      }

#ifndef OPTIMIZE
      if (Print[12])
	printf("Same location as parent: %d (%d).\n", child2,
	       Bar[child2].same_as);
#endif
      return -1; // TRUE;
    }
  }

  // Test for redundancy with existing parent's children.
  for (i=0; i<Bar[parent].num_children; i++) {
    child2 = Bar[parent].child[i];
    if ((absv(Bar[child2].end_loc[0] - Bar[child].end_loc[0]) < DIST_TOLERANCE) &&
	(absv(Bar[child2].end_loc[1] - Bar[child].end_loc[1]) < DIST_TOLERANCE) &&
	(absv(Bar[child2].end_loc[2] - Bar[child].end_loc[2]) < DIST_TOLERANCE)) {
      if (Bar[child2].same_as == -1) {
	*p_index = child2;
      } else {
	*p_index = Bar[child2].same_as;
      }

#ifndef OPTIMIZE
      if (Print[12]) {
	printf("Same location as %d child, %d (%d).\n", i, child2,
	       Bar[child2].same_as);
	v_print(Bar[child].end_loc, "New child: ");
	m3_print(Bar[child].delta_rot_mat, "New Child delta rotation matrix:\n");
	
	v_print(Bar[child2].end_loc, "Old child: ");
	m3_print(Bar[child2].delta_rot_mat, "Old Child delta rotation matrix:\n");
      }      
#endif
      return -1; // TRUE;
    }
  }

  // Not same as existing child -> add bar to parent (add now commited).
  Bar[parent].child[Bar[parent].num_children] = child;
  Bar[parent].num_children++;

  // Test for connections to other bars.
  same_as = -1;
  for (i=0; i<Num_bars; i++) {
    if (i == child)
      continue;

    if ((absv(Bar[i].end_loc[0] - Bar[child].end_loc[0]) < DIST_TOLERANCE) &&
	(absv(Bar[i].end_loc[1] - Bar[child].end_loc[1]) < DIST_TOLERANCE) &&
	(absv(Bar[i].end_loc[2] - Bar[child].end_loc[2]) < DIST_TOLERANCE)) {
      same_as = i;
      break;
    }
  }
  Bar[child].same_as = same_as;

  if (same_as != -1) {
    // Connect same_as to parent.
    int index = Bar[same_as].num_from;
    Bar[same_as].from_list[index] = parent;
    Bar[same_as].num_from += 1;
    if (Bar[same_as].num_from >= MAX_CHILDREN3D) {
      if (Print[9] || Print[12]) {
	printf("Too many from nodes for %d.\n", same_as);
      }
      return FALSE;
    }
  }
  Num_bars++;
  if (same_as != -1) {
    *p_index = Bar[child].same_as;
  } else if (type != BAR_TYPE_SENSOR) {
    *p_index = child;
  }


#ifndef OPTIMIZE
  if (Print[34] || Print[12]) {
    printf("P: %d: t:%d, mvs:%d, spd:%.2f, angle(%.2f:%.2f)\n",
	   parent, Bar[parent].type, Bar[parent].moves,
	   Bar[parent].move_spd,
	   Bar[parent].angle_a, Bar[parent].angle_b);
    m3_print(Bar[child].delta_rot_mat, "Child delta rotation matrix:\n");    
    m3_print(Bar[child].abs_heading, "Child, absolute heading\n");    
    printf("C: %d(%d): t:%d, mvs:%d, spd:%.2f angle(%.2f:%.2f)\n",
	   child, Bar[child].same_as, Bar[child].type, Bar[child].moves,
	   Bar[child].move_spd,
	   Bar[child].angle_a, Bar[child].angle_b);
    printf("Calculated position: %7.3f %7.3f %7.3f\n", Bar[child].end_loc[0],
	   Bar[child].end_loc[1], Bar[child].end_loc[2]);
    printf("\n");
  }
#endif

  if (Print[9]) {
    if (child == 0)
      printf(" -  Add successful: %d.\n", child);
  }
  return child;
}






// len : length of bar.
// rot_mat : rotation matrix to transform from world coords to child.
int BarStruct3D :: add_branch(Real len, Real rotation[3][3], int *p_index)
{
  int i;
  int child, child2;
  int parent = *p_index;
  int same_as;


#ifndef OPTIMIZE
  if (Print[12]) {
    printf("Add branch (%d): len:%.1f\n", parent, len);
  }
#endif

  if (Num_bars >= MAX_BARS3) {
#ifndef OPTIMIZE
    if (Print[9] || Print[12])
      printf("Maximum number of branches exceeded.\n");
#endif
    return FALSE;
  }

  // Adjust length
  if (len == 0.0)
    len = 1.0;
  else if (len > 5.0)
    len = 5.0;
  len *= 10.0;

  if (Bar[parent].num_children >= MAX_CHILDREN3D) {
#ifndef OPTIMIZE
    if (Print[9] || Print[12]) {
      printf("Too many children for branch #%d : %d.\n", parent, Bar[parent].num_children);
    }
#endif
    return FALSE;
  }


  // Add bar (without commitment).
  child = Num_bars;
  Bar[child].parent = parent;
  Bar[child].moves = FALSE;
  Bar[child].rot_axis = -1;
  Bar[child].move_spd = 0.0;
  Bar[child].move_offset = 0.0;
  Bar[child].num_children = 0;
  Bar[child].num_from = 0;
  Bar[child].length = len;
  Bar[child].len_short = len/2.0;
  Bar[child].type = BAR_TYPE_FIXED;
  
  mm3_copy(Bar[child].delta_rot_mat, rotation);
  mm3_transpose(Bar[child].inverse_delta_rot, rotation);

  // Calculate the bars location at time 0.
  calc_bar_position(child, 0.0);

  // Test for redundancy with existing parent's children.
  for (i=0; i<Bar[parent].num_children; i++) {
    child2 = Bar[parent].child[i];
    if ((absv(Bar[child2].end_loc[0] - Bar[child].end_loc[0]) < DIST_TOLERANCE) &&
	(absv(Bar[child2].end_loc[1] - Bar[child].end_loc[1]) < DIST_TOLERANCE) &&
	(absv(Bar[child2].end_loc[2] - Bar[child].end_loc[2]) < DIST_TOLERANCE)) {
      if (Bar[child2].same_as == -1) {
	*p_index = child2;
      } else {
	*p_index = Bar[child2].same_as;
      }

#ifndef OPTIMIZE
      if (Print[12]) {
	printf("Same location as %d child, %d (%d).\n", i, child2,
	       Bar[child2].same_as);
	v_print(Bar[child].end_loc, "New child: ");
	m3_print(Bar[child].delta_rot_mat, "New Child delta rotation matrix:\n");
	
	v_print(Bar[child2].end_loc, "Old child: ");
	m3_print(Bar[child2].delta_rot_mat, "Old Child delta rotation matrix:\n");
      }      
#endif
      return TRUE;
    }
  }

  // Not same as existing child -> add bar to parent (add now commited).
  Bar[parent].child[Bar[parent].num_children] = child;
  Bar[parent].num_children++;

  // Test for connections to other bars.
  same_as = -1;
  for (i=0; i<Num_bars; i++) {
    if (i == child)
      continue;

    if ((absv(Bar[i].end_loc[0] - Bar[child].end_loc[0]) < DIST_TOLERANCE) &&
	(absv(Bar[i].end_loc[1] - Bar[child].end_loc[1]) < DIST_TOLERANCE) &&
	(absv(Bar[i].end_loc[2] - Bar[child].end_loc[2]) < DIST_TOLERANCE)) {
      same_as = i;
      break;
    }
  }
  Bar[child].same_as = same_as;

  if (same_as != -1) {
    // Connect same_as to parent.
    int index = Bar[same_as].num_from;
    Bar[same_as].from_list[index] = parent;
    Bar[same_as].num_from += 1;
    if (Bar[same_as].num_from >= MAX_CHILDREN3D) {
      if (Print[9] || Print[12]) {
	printf("Too many from nodes for %d.\n", same_as);
      }
      return FALSE;
    }
  }
  Num_bars++;
  if (same_as == -1) {
    *p_index = child;
  } else {
    *p_index = Bar[child].same_as;
  }


#ifndef OPTIMIZE
  if (Print[34] || Print[12]) {
    printf("P: %d: t:%d, mvs:%d, spd:%.2f, angle(%.2f:%.2f)\n",
	   parent, Bar[parent].type, Bar[parent].moves,
	   Bar[parent].move_spd,
	   Bar[parent].angle_a, Bar[parent].angle_b);
    m3_print(Bar[child].delta_rot_mat, "Child delta rotation matrix:\n");    
    printf("C: %d(%d): t:%d, mvs:%d, spd:%.2f angle(%.2f:%.2f)\n",
	   child, Bar[child].same_as, Bar[child].type, Bar[child].moves,
	   Bar[child].move_spd,
	   Bar[child].angle_a, Bar[child].angle_b);
    printf("Calculated position: %7.3f %7.3f %7.3f\n", Bar[child].end_loc[0],
	   Bar[child].end_loc[1], Bar[child].end_loc[2]);
    printf("\n");
  }
#endif
  return child;
}





void init_heading_matrix(Real mat[3][3])
{
  mat[0][0] = 1;
  mat[0][1] = 0;
  mat[0][2] = 0;

  mat[1][0] = 0;
  mat[1][1] = 1;
  mat[1][2] = 0;

  mat[2][0] = 0;
  mat[2][1] = 0;
  mat[2][2] = 1;
}


// Twist clockwise.
/* Multiply by:
  1  0  0
  0  0 -1
  0  1  0  */
void rot_xaxis_pos(Real mat[3][3])
{
  Real tmp[3][3];

  tmp[0][0] = mat[0][0];
  tmp[0][1] = mat[0][2];
  tmp[0][2] = -mat[0][1];

  tmp[1][0] = mat[1][0];
  tmp[1][1] = mat[1][2];
  tmp[1][2] = -mat[1][1];

  tmp[2][0] = mat[2][0];
  tmp[2][1] = mat[2][2];
  tmp[2][2] = -mat[2][1];

#ifdef OLD
  tmp[0][0] = mat[0][0];
  tmp[0][1] = mat[0][1];
  tmp[0][2] = mat[0][2];

  tmp[1][0] = -mat[2][0];
  tmp[1][1] = -mat[2][1];
  tmp[1][2] = -mat[2][2];

  tmp[2][0] = mat[1][0];
  tmp[2][1] = mat[1][1];
  tmp[2][2] = mat[1][2];
#endif
  mm3_copy(mat, tmp);
}

// Twist counter-clockwise.
/* Multiply by:
  1  0  0
  0  0  1
  0 -1  0  */
void rot_xaxis_neg(Real mat[3][3])
{
  Real tmp[3][3];

  tmp[0][0] = mat[0][0];
  tmp[0][1] = -mat[0][2];
  tmp[0][2] = mat[0][1];

  tmp[1][0] = mat[1][0];
  tmp[1][1] = -mat[1][2];
  tmp[1][2] = mat[1][1];

  tmp[2][0] = mat[2][0];
  tmp[2][1] = -mat[2][2];
  tmp[2][2] = mat[2][1];

#ifdef OLD
  tmp[0][0] = mat[0][0];
  tmp[0][1] = mat[0][1];
  tmp[0][2] = mat[0][2];

  tmp[1][0] = mat[2][0];
  tmp[1][1] = mat[2][1];
  tmp[1][2] = mat[2][2];

  tmp[2][0] = -mat[1][0];
  tmp[2][1] = -mat[1][1];
  tmp[2][2] = -mat[1][2];
#endif
  mm3_copy(mat, tmp);
}


// Turn left.
/* Multiply by:
  0  0  1
  0  1  0
 -1  0  0  */
void rot_yaxis_pos(Real mat[3][3])
{
  Real tmp[3][3];

  tmp[0][0] = -mat[0][2];
  tmp[0][1] = mat[0][1];
  tmp[0][2] = mat[0][0];

  tmp[1][0] = -mat[1][2];
  tmp[1][1] = mat[1][1];
  tmp[1][2] = mat[1][0];

  tmp[2][0] = -mat[2][2];
  tmp[2][1] = mat[2][1];
  tmp[2][2] = mat[2][0];

#ifdef OLD
  tmp[0][0] = mat[2][0];
  tmp[0][1] = mat[2][1];
  tmp[0][2] = mat[2][2];

  tmp[1][0] = mat[1][0];
  tmp[1][1] = mat[1][1];
  tmp[1][2] = mat[1][2];

  tmp[2][0] = -mat[0][0];
  tmp[2][1] = -mat[0][1];
  tmp[2][2] = -mat[0][2];
#endif
  mm3_copy(mat, tmp);
}

// Turn right.
/* Multiply by:
  0  0 -1
  0  1  0
  1  0  0  */
void rot_yaxis_neg(Real mat[3][3])
{
  Real tmp[3][3];

  tmp[0][0] = mat[0][2];
  tmp[0][1] = mat[0][1];
  tmp[0][2] = -mat[0][0];

  tmp[1][0] = mat[1][2];
  tmp[1][1] = mat[1][1];
  tmp[1][2] = -mat[1][0];

  tmp[2][0] = mat[2][2];
  tmp[2][1] = mat[2][1];
  tmp[2][2] = -mat[2][0];

#ifdef OLD
  tmp[0][0] = -mat[2][0];
  tmp[0][1] = -mat[2][1];
  tmp[0][2] = -mat[2][2];

  tmp[1][0] = mat[1][0];
  tmp[1][1] = mat[1][1];
  tmp[1][2] = mat[1][2];

  tmp[2][0] = mat[0][0];
  tmp[2][1] = mat[0][1];
  tmp[2][2] = mat[0][2];
#endif
  mm3_copy(mat, tmp);
}


// Turn up. M'= MR
/* Multiply by:
  0 -1  0
  1  0  0
  0  0  1  */
void rot_zaxis_pos(Real mat[3][3])
{
  Real tmp[3][3];

  tmp[0][0] = mat[0][1];
  tmp[0][1] = -mat[0][0];
  tmp[0][2] = mat[0][2];

  tmp[1][0] = mat[1][1];
  tmp[1][1] = -mat[1][0];
  tmp[1][2] = mat[1][2];

  tmp[2][0] = mat[2][1];
  tmp[2][1] = -mat[2][0];
  tmp[2][2] = mat[2][2];

#ifdef OLD
  tmp[0][0] = -mat[1][0];
  tmp[0][1] = -mat[1][1];
  tmp[0][2] = -mat[1][2];

  tmp[1][0] = mat[0][0];
  tmp[1][1] = mat[0][1];
  tmp[1][2] = mat[0][2];

  tmp[2][0] = mat[2][0];
  tmp[2][1] = mat[2][1];
  tmp[2][2] = mat[2][2];
#endif
  mm3_copy(mat, tmp);
}

// Turn down.
/* Multiply by:
  0  1  0
 -1  0  0
  0  0  1  */
void rot_zaxis_neg(Real mat[3][3])
{
  Real tmp[3][3];

  tmp[0][0] = -mat[0][1];
  tmp[0][1] = mat[0][0];
  tmp[0][2] = mat[0][2];

  tmp[1][0] = -mat[1][1];
  tmp[1][1] = mat[1][0];
  tmp[1][2] = mat[1][2];

  tmp[2][0] = -mat[2][1];
  tmp[2][1] = mat[2][0];
  tmp[2][2] = mat[2][2];

#ifdef OLD
  tmp[0][0] = mat[1][0];
  tmp[0][1] = mat[1][1];
  tmp[0][2] = mat[1][2];

  tmp[1][0] = -mat[0][0];
  tmp[1][1] = -mat[0][1];
  tmp[1][2] = -mat[0][2];

  tmp[2][0] = mat[2][0];
  tmp[2][1] = mat[2][1];
  tmp[2][2] = mat[2][2];
#endif
  mm3_copy(mat, tmp);
}


// Build structure using parameterized absolute coordinates.
// This builds the kinematic model for bending pieces.
int BarStruct3D :: build_structure_right(ModelStruct *model)
{
  int i, j, x, y;
  int reps, bar_index, num_bars_prev;
  int ret_val;
  Real heading[3][3];
  Real len, spd, offset;
  int length = model->lstring_length;
  LString *command = model->lstring;


  int axis = ROT_Z_AXIS;
  waste_length = 0;

  initialize();
  init_heading_matrix(heading);

  offset = 0.0;
  bar_index = 0;
  make_identity_mat(rotation_matrix);


  for (i=0; i<length; i++) {
    if (Print[13]) {
      printf("%d: %c(%d)\n", i, command[i].symbol,
	     int(command[i].arg));

    } else if (Print[37]) {
      if ((command[i].symbol == TERM_PUSH) ||
	  (command[i].symbol == TERM_POP)) {
	printf("%d:  %c\n", i, command[i].symbol);
      } else {	
	//	printf("%c(%d) ", command[i].symbol, int(command[i].arg));
	printf("%d:  %c(%.2f)\n", i, command[i].symbol, command[i].arg);
      }

    } else if (Print[38]) {
      if (command[i].symbol == TERM_PUSH) {
	printf("[");

      } else if (command[i].symbol == TERM_POP) {
	printf("] ");

      } else {	
	printf("%c(%.4f) ", command[i].symbol, command[i].arg);
      }
    }      

    if (command[i].symbol == 0) {
      if (Print[0])
	printf("Build:: unexpected early termination of command sequence @:%d(!%d)\n", i, length);
      return FALSE;

    } else if (command[i].symbol == TERM_PUSH) {
      if (Bar_stack_ptr == BAR_STACK_SIZE) {
	printf("Bar stack exceeded.\n");
	return FALSE;
      }
      Bar_stack[Bar_stack_ptr].bar_index = bar_index;
      Bar_stack[Bar_stack_ptr].axis = axis;
      Bar_stack[Bar_stack_ptr].offset = offset;
      for (x=0; x<3; x++)
	for (y=0; y<3; y++)
	  Bar_stack[Bar_stack_ptr].heading[x][y] = heading[x][y];
      Bar_stack_ptr++;

    } else if (command[i].symbol == TERM_POP) {
      if (Bar_stack_ptr == 0) {
	printf("Bar stack over-popped.\n");
	return FALSE;
      }
      Bar_stack_ptr--;
      bar_index = Bar_stack[Bar_stack_ptr].bar_index;
      axis = Bar_stack[Bar_stack_ptr].axis;
      offset = Bar_stack[Bar_stack_ptr].offset;
      for (x=0; x<3; x++)
	for (y=0; y<3; y++)
	  heading[x][y] = Bar_stack[Bar_stack_ptr].heading[x][y];

    } else if (command[i].symbol == TERM_BACKWARD) {
      // Go up parent.
      if (bar_index != 0)
	bar_index = Bar[bar_index].parent;
      else
	waste_length++;

    } else if (command[i].symbol == TERM_FORWARD) {
      len = command[i].arg;
      if (len == 0.0)
	continue;
      len = 1.0;

      num_bars_prev = Num_bars;
      ret_val = add_bar(1.0, heading, BAR_TYPE_FIXED, -1, 0.0, 0.0,
			0.0, 0.0, &bar_index);
      if (ret_val == FALSE) {
	if (Print[9])
	  printf("bars3d :: build_structure() 'forward' - error adding bar.\n");
	return FALSE;
      }
      if (num_bars_prev == Num_bars)
	waste_length++;

    } else if (command[i].symbol == TERM_COLOR_SENSOR) {
      // Arg should probably have some sort of weight associated with it.
      num_bars_prev = Num_bars;
      ret_val = add_bar(0.4, heading, BAR_TYPE_SENSOR, -1, 0.0, 0.0,
			0.0, 0.0, &bar_index);
      /*      if (ret_val == -1) {
	printf("Bars3d :: build_structure_right() - error adding sensor.\n");
	model->error = TRUE;
	return FALSE;
	} */
      if (ret_val == FALSE) {
	if (Print[9])
	  printf("bars3d :: build_structure() 'color sensor' - error adding bar.\n");
	return FALSE;
      }
      if (num_bars_prev == Num_bars)
	waste_length++;

    } else if (command[i].symbol == TERM_JOINT_REVOLUTE) {
      spd = command[i].arg / 2.0;
      if (spd == 0.0)
	spd = 1.0;

      len = command[i].arg;
      if (len == 0.0)
	continue;
      len = 1.0;

      num_bars_prev = Num_bars;
      ret_val = add_bar(1.0, heading, BAR_TYPE_JOINT, ROT_Z_AXIS,
			0.0, PI/2.0, spd, offset, &bar_index);
      if (ret_val == FALSE) {
	if (Print[9])
	  printf("bars3d :: build_structure() 'revolute-1' - error adding bar: %d.\n", ret_val);
	return FALSE;
      }
      if (num_bars_prev == Num_bars)
	waste_length++;

    } else if (command[i].symbol == TERM_JOINT_REVOLUTE2) {
      spd = command[i].arg / 2.0;
      if (spd == 0.0)
	spd = 1.0;

      len = command[i].arg;
      if (len == 0.0)
	continue;
      len = 1.0;

      num_bars_prev = Num_bars;
      ret_val = add_bar(1.0, heading, BAR_TYPE_JOINT, ROT_Z_AXIS,
			-PI/4.0, PI/4.0, spd, offset, &bar_index);
      if (ret_val == FALSE) {
	if (Print[9])
	  printf("bars3d :: build_structure() 'revolute-2' - error adding bar: %d.\n", ret_val);
	return FALSE;
      }
      if (num_bars_prev == Num_bars)
	waste_length++;

    } else if (command[i].symbol == TERM_JOINT_TWIST) {
      spd = command[i].arg / 2.0;
      if (spd == 0.0)
	spd = 1.0;

      len = command[i].arg;
      if (len == 0.0)
	continue;
      len = 1.0;

      num_bars_prev = Num_bars;
      ret_val = add_bar(1.0, heading, BAR_TYPE_JOINT, ROT_X_AXIS,
			0.0, PI/2.0, spd, offset, &bar_index);
      if (ret_val == FALSE) {
	if (Print[9])
	  printf("bars3d :: build_structure() 'twist' - error adding bar: %d.\n", ret_val);
	return FALSE;
      }
      if (num_bars_prev == Num_bars)
	waste_length++;

    } else if (command[i].symbol == TERM_JOINT_TWIST180) {
      spd = command[i].arg / 2.0;
      if (spd == 0.0)
	spd = 1.0;

      len = command[i].arg;
      if (len == 0.0)
	continue;
      len = 1.0;

      num_bars_prev = Num_bars;
      ret_val = add_bar(1.0, heading, BAR_TYPE_JOINT, ROT_X_AXIS,
			-PI/2.0, PI/2.0, spd/2.0, offset, &bar_index);
      if (ret_val == FALSE) {
	if (Print[9])
	  printf("bars3d :: build_structure() 'twist-180' - error adding bar: %d.\n", ret_val);
	return FALSE;
      }
      if (num_bars_prev == Num_bars)
	waste_length++;

    } else if (command[i].symbol == TERM_TURN_UP) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	rot_zaxis_pos(heading);

    } else if (command[i].symbol == TERM_TURN_DOWN) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	rot_zaxis_neg(heading);

    } else if (command[i].symbol == TERM_TURN_LEFT) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	rot_yaxis_pos(heading);

    } else if (command[i].symbol == TERM_TURN_RIGHT) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	rot_yaxis_neg(heading);

    } else if (command[i].symbol == TERM_TURN_CLOCKWISE) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	rot_xaxis_pos(heading);

    } else if (command[i].symbol == TERM_TURN_CNTR_CLOCKWISE) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	rot_xaxis_neg(heading);

    } else if (command[i].symbol == TERM_OFFSET_INCREASE) {      
      offset += OFFSET_STEP_3D * floor(command[i].arg);
      if (offset > 1.0) {
	offset -= floor(offset);
      }

    } else if (command[i].symbol == TERM_OFFSET_DECREASE) {
      offset -= OFFSET_STEP_3D * floor(command[i].arg);
      if (offset < 0.0) {
	offset -= floor(offset);
      }

    } else {
#ifndef OPTIMIZE
	if (Print[0])
	  printf("Build:: unknown command: (%c:%d)\n", 
		 command[i].symbol, int(command[i].symbol));
#endif
    }     
  }

  //  return TRUE;

  if (Num_bars <= 1) {
    if (Print[9])
      printf("Num bars: %d <= 1.\n", Num_bars);
    return FALSE;
  }

  for (i=0; i<Num_bars; i++) {
    if (Bar[i].type == BAR_TYPE_JOINT) {
      if (Bar[i].num_children == 0) {
	Bar[i].type = BAR_TYPE_FIXED;
      } else {
	Num_joints++;
	Bar[i].moves = TRUE;
      }

    } else if (Bar[i].type == BAR_TYPE_SENSOR) {
      Sensor_index[Num_sensors++] = i;
    }
  }
  if (Num_joints > Max_joints) {
    if (Print[9]) {
      printf("Build_structure() - error, too many joints.\n");
    }
    return FALSE;
  }

  total_weight = Real(Num_bars);
  for (i=0; i<Num_bars; i++)
    Bar[i].weight /= total_weight;

  calculate_positions(0.0);
  if (Print[34]) {
    printf("Recalc positions: %.4f\n", Time);
    calculate_positions_new(Time);
  }


#ifndef OPTIMIZE
  if (Print[27])
    print_struct();
#endif

  /*
  if (!settle_structure()) {
    printf("Settle structure failed in build_right.\n");
    model->error = TRUE;
    return FALSE;
  }    
  */

  //  translate_struct(-COM[0], 0.0, -COM[2]);
  for (i=0; i<Num_bars; i++) {
    Bar[i].prev_loc[0] = Bar[i].end_loc[0];
    Bar[i].prev_loc[1] = Bar[i].prev_loc[1];
  }
  initialize_intersection();

  return TRUE;
}



// Build structure using parameterized absolute coordinates.
// This builds the kinematic model for bending pieces.
#define TREE_ROTATE (PI/7.0)
int BarStruct3D :: build_structure_tree(ModelStruct *model)
{
  int i, x, y;
  int reps, bar_index, num_bars_prev;
  int ret_val;
  Real heading[3][3], tmp_mat[3][3], rot_mat[3][3];
  Real len, offset;
  int length = model->lstring_length;
  LString *command = model->lstring;


  int axis = ROT_Z_AXIS;
  waste_length = 0;

  initialize();
  init_heading_matrix(heading);
  rot_zaxis_pos(heading);

  offset = 0.0;
  bar_index = 0;
  make_identity_mat(rotation_matrix);

  for (i=0; i<length; i++) {
    if (Print[13]) {
      printf("%d: %c(%d)\n", i, command[i].symbol,
	     int(command[i].arg));
    }

    if (command[i].symbol == 0) {
      if (Print[0])
	printf("Build:: unexpected early termination of command sequence @:%d(!%d)\n", i, length);
      return FALSE;

    } else if (command[i].symbol == TERM_PUSH) {
      if (Bar_stack_ptr == BAR_STACK_SIZE) {
	printf("Bar stack exceeded.\n");
	return FALSE;
      }
      Bar_stack[Bar_stack_ptr].bar_index = bar_index;
      Bar_stack[Bar_stack_ptr].axis = axis;
      Bar_stack[Bar_stack_ptr].offset = offset;
      for (x=0; x<3; x++)
	for (y=0; y<3; y++)
	  Bar_stack[Bar_stack_ptr].heading[x][y] = heading[x][y];
      Bar_stack_ptr++;

    } else if (command[i].symbol == TERM_POP) {
      if (Bar_stack_ptr == 0) {
	printf("Bar stack over-popped.\n");
	return FALSE;
      }
      Bar_stack_ptr--;
      bar_index = Bar_stack[Bar_stack_ptr].bar_index;
      axis = Bar_stack[Bar_stack_ptr].axis;
      offset = Bar_stack[Bar_stack_ptr].offset;
      for (x=0; x<3; x++)
	for (y=0; y<3; y++)
	  heading[x][y] = Bar_stack[Bar_stack_ptr].heading[x][y];

    } else if (command[i].symbol == TERM_BACKWARD) {
      // Go up parent.
      if (bar_index != 0)
	bar_index = Bar[bar_index].parent;
      else
	waste_length++;

    } else if (command[i].symbol == TERM_FORWARD) {
      len = command[i].arg;
      if (len == 0.0)
	continue;
      len = 1.0;

      num_bars_prev = Num_bars;
      // The following line is original:
      //      ret_val = add_branch(1.0, heading, &bar_index);
      ret_val = add_bar(1.0, heading, BAR_TYPE_FIXED, -1, 0.0, 0.0,
			0.0, 0.0, &bar_index);
      if (ret_val == FALSE)
	return FALSE;
      if (num_bars_prev == Num_bars)
	waste_length++;
      make_identity_mat(rot_mat);

    } else if (command[i].symbol == TERM_TURN_UP) {
      reps = int(command[i].arg) % 12;
      // M' = MR
      make_rotZ_mat(reps*TREE_ROTATE, rot_mat);
      mmm3_mult(tmp_mat, heading, rot_mat);
      mm3_copy(heading, tmp_mat);

    } else if (command[i].symbol == TERM_TURN_DOWN) {
      reps = -int(command[i].arg) % 12;
      // M' = MR
      make_rotZ_mat(reps*TREE_ROTATE, rot_mat);
      mmm3_mult(tmp_mat, heading, rot_mat);
      mm3_copy(heading, tmp_mat);

    } else if (command[i].symbol == TERM_TURN_LEFT) {
      reps = int(command[i].arg) % 12;
      // M' = MR
      make_rotY_mat(reps*TREE_ROTATE, rot_mat);
      mmm3_mult(tmp_mat, heading, rot_mat);
      mm3_copy(heading, tmp_mat);

    } else if (command[i].symbol == TERM_TURN_RIGHT) {
      reps = -int(command[i].arg) % 12;
      // M' = MR
      make_rotY_mat(reps*TREE_ROTATE, rot_mat);
      mmm3_mult(tmp_mat, heading, rot_mat);
      mm3_copy(heading, tmp_mat);

    } else if (command[i].symbol == TERM_TURN_CLOCKWISE) {
      reps = int(command[i].arg) % 12;
      // M' = MR
      make_rotX_mat(reps*TREE_ROTATE, rot_mat);
      mmm3_mult(tmp_mat, heading, rot_mat);
      mm3_copy(heading, tmp_mat);

    } else if (command[i].symbol == TERM_TURN_CNTR_CLOCKWISE) {
      reps = -int(command[i].arg) % 12;
      // M' = MR
      make_rotX_mat(reps*TREE_ROTATE, rot_mat);
      mmm3_mult(tmp_mat, heading, rot_mat);
      mm3_copy(heading, tmp_mat);

    } else {
#ifndef OPTIMIZE
	if (Print[0])
	  printf("Build:: unknown command: (%c:%d)\n", 
		 command[i].symbol, int(command[i].symbol));
#endif
    }     
  }

  if (Num_bars <= 1) {
    if (Print[9])
      printf("Num bars: %d <= 1.\n", Num_bars);
    return FALSE;
  }

  return TRUE;
}




// Build structure using parameterized absolute coordinates.
// This builds the kinematic model for bending pieces.
int BarStruct3D :: build_network(ModelStruct *model)
{
  int i, j, x, y;
  int axis = ROT_Z_AXIS;
  int ret_val, num;
  int reps, bar_index, num_bars_prev;
  Real len, spd, offset, weight;
  Real heading[3][3];
  int length = model->lstring_length;

  waste_length = 0;

  LString *command = model->lstring;
  LinkIndex current_link;

  //  builds++; // Total number of times this function is called.

  // Initialize morphology variables.
  initialize();
  init_heading_matrix(heading);
  offset = 0.0;
  bar_index = 0;
  make_identity_mat(rotation_matrix);

  // Initialize network variables.
  Link_stack_ptr = 0;
  initialize_network();
  network->first_neuron(0.25, &current_link);
  //  network->first_neuron(1.0, &current_link);
  //  network->first_neuron(PARAM_2_WEIGHT, &current_link);
  //  //  network->first_neuron(0.0, &current_link);

  //  if ((Terminal_set < 40) || (Terminal_set == 50)) {
  if (Terminal_set < 40) {
    network->set_input_function(NEURON_SAW_WAVE, 1.0, &current_link);
  } else {
    //    network->set_input_function(NEURON_SAW_WAVE, 1.0, &current_link);
    network->set_input_function(NEURON_SIGMOID, 1.0, &current_link);
  }

  if (Terminal_set == 28) {
    // Create 4 input neurons.
    if (!network->create_inputs(4)) {
      printf("bars3d :: error creating input neurons.\n");
      return FALSE;
    }
  }


  // Start processing build string.
  for (i=0; i<length; i++) {
    if (Print[13]) {
      network->print_structure();
      printf("%d: %c(%d)\n", i, command[i].symbol,
	     int(command[i].arg));
    }

    /*
    if (Terminal_set == 50) {
      if ((command[i].symbol == TERM_SPLIT_LINK) ||
	  (command[i].symbol == TERM_DUPLICATE_LINK) ||
	  (command[i].symbol == TERM_MAKE_LOOP)) {
	// Convert weights to +/-1.
	if (((int)command[i].arg)%2) {
	  command[i].arg = 4.0;
	} else {
	  command[i].arg = -4.0;
	}
      }
    }
    */


    if (command[i].symbol == 0) {
      if (Print[0])
	printf("Build:: unexpected early termination of command sequence @:%d(!%d)\n", i, length);
      return FALSE;

    } else if (command[i].symbol == TERM_PUSH) {
      if (Bar_stack_ptr == BAR_STACK_SIZE) {
	printf("Bar stack exceeded.\n");
	return FALSE;
      }
      copy_linkindex(&(Bar_stack[Bar_stack_ptr].current_link),
      		     &current_link);

      Bar_stack[Bar_stack_ptr].bar_index = bar_index;
      Bar_stack[Bar_stack_ptr].axis = axis;
      Bar_stack[Bar_stack_ptr].offset = offset;
      for (x=0; x<3; x++)
	for (y=0; y<3; y++)
	  Bar_stack[Bar_stack_ptr].heading[x][y] = heading[x][y];
      Bar_stack_ptr++;

    } else if (command[i].symbol == TERM_POP) {
      if (Bar_stack_ptr == 0) {
	printf("Bar stack over-popped.\n");
	return FALSE;
      }
      Bar_stack_ptr--;
      copy_linkindex(&current_link,
		     &(Bar_stack[Bar_stack_ptr].current_link));
      bar_index = Bar_stack[Bar_stack_ptr].bar_index;
      axis = Bar_stack[Bar_stack_ptr].axis;
      offset = Bar_stack[Bar_stack_ptr].offset;
      for (x=0; x<3; x++)
	for (y=0; y<3; y++)
	  heading[x][y] = Bar_stack[Bar_stack_ptr].heading[x][y];

    } else if (command[i].symbol == TERM_PUSH_EDGE) {
      if (Link_stack_ptr == EDGE_STACK_SIZE) {
	printf("link stack exceeded: %d.\n", Link_stack_ptr);
	return FALSE;
      }
      copy_linkindex(&(Link_stack[Link_stack_ptr]),
		     &current_link);
      Link_stack_ptr++;

    } else if (command[i].symbol == TERM_POP_EDGE) {
      if (Link_stack_ptr == 0) {
	printf("Link stack over-popped.\n");
	return FALSE;
      }
      Link_stack_ptr--;
      copy_linkindex(&current_link,
		     &(Link_stack[Link_stack_ptr]));

    } else if (command[i].symbol == TERM_CONNECT_TO_INPUT) {
      //      weight = (command[i].arg) * PARAM_2_WEIGHT;
      num = int(command[i].arg);
      if (num < 0) {
	num = -num;
      }
      num = num % 4;
      //      printf("connect to input: %d.\n", num);
      if (!network->connect_to_input(1.0, num, &current_link)) {
	waste_length++;
      }

    } else if (command[i].symbol == TERM_ADD_INPUT) {
      weight = (command[i].arg) * PARAM_2_WEIGHT;
      if (!network->add_input(weight, &current_link)) {
	waste_length++;
      }

    } else if (command[i].symbol == TERM_ADD_OUTPUT) {
      weight = (command[i].arg) * PARAM_2_WEIGHT;
      if (!network->add_output(weight, &current_link)) {
	waste_length++;
      }

    } else if (command[i].symbol == TERM_SPLIT_LINK) {
      weight = (command[i].arg) * PARAM_2_WEIGHT;
      if (!network->split_link(weight, &current_link)) {
	waste_length++;
      }
      
    } else if (command[i].symbol == TERM_MAKE_LOOP) {
      weight = (command[i].arg) * PARAM_2_WEIGHT;
      if (!network->make_loop(weight, &current_link)) {
	waste_length++;
      }
      
    } else if (command[i].symbol == TERM_DUPLICATE_LINK) {
      weight = (command[i].arg) * PARAM_2_WEIGHT;
      if (!network->duplicate_link(weight, &current_link)) {
	waste_length++;
      }

    } else if (command[i].symbol == TERM_SPLIT_LINK_NEG) {
      weight = -(command[i].arg) * PARAM_2_WEIGHT;
      if (!network->split_link(weight, &current_link)) {
	waste_length++;
      }
      
    } else if (command[i].symbol == TERM_DUPLICATE_LINK_NEG) {
      weight = -(command[i].arg) * PARAM_2_WEIGHT;
      if (!network->duplicate_link(weight, &current_link)) {
	waste_length++;
      }

    } else if (command[i].symbol == TERM_ADJUST_WEIGHT) {
      weight = (command[i].arg) * PARAM_2_ADJUST;
      if (!network->adjust_weight(weight, &current_link)) {
	waste_length++;
      }

    } else if (command[i].symbol == TERM_INCREASE_WEIGHT) {
      weight = (command[i].arg) * PARAM_2_ADJUST;
      if (!network->adjust_weight(weight, &current_link)) {
	waste_length++;
      }

    } else if (command[i].symbol == TERM_DECREASE_WEIGHT) {
      weight = -(command[i].arg) * PARAM_2_ADJUST;
      if (!network->adjust_weight(weight, &current_link)) {
	waste_length++;
      }

    } else if (command[i].symbol == TERM_NEGATE_WEIGHT) {
      weight = -(command[i].arg) * PARAM_2_WEIGHT;
      if (!network->negate_weight(weight, &current_link)) {
	waste_length++;
      }

    } else if (command[i].symbol == TERM_DELETE_LINK) {
      // Deletes link.  Since this link is dead, all other
      // operations on it are killed, so must go to next TERM_POP.
      if (!network->cut_link(&current_link)) {
	waste_length++;
      }

    } else if (command[i].symbol == TERM_REVERSE_LINK) {
      if (!network->reverse_link(&current_link)) {
	waste_length++;
      }

    } else if (command[i].symbol == TERM_NEXT_LINK) {
      num = int(command[i].arg);
      if (num != 0) {
	network->next_link(num, &current_link);
      } else {
	waste_length++;
      }

    } else if (command[i].symbol == TERM_SET_OUTPUT_FUNCTION) {
      num = int(command[i].arg);
      if (!network->set_output_function(num, &current_link)) {
	waste_length++;
      }

    } else if (command[i].symbol == TERM_PARENT_NEURON) {
      num = int(command[i].arg);
      if (num != 0) {
	network->parent_link(num, &current_link);
      } else {
	waste_length++;
      }

    } else if (command[i].symbol == TERM_MERGE_NEURONS) {
      if (!network->merge_neurons(&current_link)) {
	waste_length++;
      }

    } else if (command[i].symbol == TERM_BACKWARD) {
      // Go up parent.
      if (bar_index != 0)
	bar_index = Bar[bar_index].parent;
      else
	waste_length++;

    } else if (command[i].symbol == TERM_FORWARD) {
      len = command[i].arg;
      if (len == 0.0)
	continue;
      len = 1.0;

      num_bars_prev = Num_bars;
      ret_val = add_bar(1.0, heading, BAR_TYPE_FIXED, -1, 0.0, 0.0,
			0.0, 0.0, &bar_index);
      if (ret_val == FALSE) {
	if (Print[9])
	  printf("bars3d :: build_structure() 'forward' - error adding bar.\n");
	return FALSE;
      }
      if (num_bars_prev == Num_bars)
	waste_length++;

    } else if (command[i].symbol == TERM_COLOR_SENSOR) {
      // Arg should probably have some sort of weight associated with it.
      num_bars_prev = Num_bars;
      ret_val = add_bar(0.4, heading, BAR_TYPE_SENSOR, -1, 0.0, 0.0,
			0.0, 0.0, &bar_index);
      if (ret_val == FALSE) {
	if (Print[9])
	  printf("bars3d :: build_structure() 'color sensor' - error adding bar: %d.\n", ret_val);
	return FALSE;
      }
      if (num_bars_prev == Num_bars) {
	waste_length++;
      } else {
	// Add input node.
	weight = (command[i].arg) * PARAM_2_WEIGHT;
	//	weight = 1.0;
	if (!network->add_input(weight, &current_link)) {
	  waste_length++;
	  if (Print[9])
	    printf("bars3d :: build_network() 'color sensor' - error adding input: %d.\n", ret_val);
	  return FALSE;
	}
      }      

    } else if (command[i].symbol == TERM_JOINT_REVOLUTE) {
      num_bars_prev = Num_bars;
      ret_val = add_bar(1.0, heading, BAR_TYPE_JOINT, ROT_Z_AXIS,
			0.0, PI/2.0, 1.0, offset, &bar_index);
      if (ret_val == FALSE) {
	if (Print[9])
	  printf("bars3d :: build_network() 'revolute-1' - error adding bar: %d.\n", ret_val);
	return FALSE;
      }

      if (num_bars_prev == Num_bars) {
	waste_length++;
      } else {
	// Add output node.
	//	weight = (command[i].arg) * PARAM_2_WEIGHT;
	weight = 1.0;
	if (!network->add_output(weight, &current_link)) {
	  waste_length++;
	  if (Print[9])
	    printf("bars3d :: build_network() 'revolute-1' - error adding output: %d.\n", ret_val);
	  return FALSE;
	}
      }      

    } else if (command[i].symbol == TERM_JOINT_REVOLUTE2) {
      spd = command[i].arg / 2.0;
      if (spd == 0.0)
	spd = 1.0;

      len = command[i].arg;
      if (len == 0.0)
	continue;
      len = 1.0;

      num_bars_prev = Num_bars;
      ret_val = add_bar(1.0, heading, BAR_TYPE_JOINT, ROT_Z_AXIS,
			-PI/4.0, PI/4.0, spd, offset, &bar_index);
      if (ret_val == FALSE) {
	if (Print[9])
	  printf("bars3d :: build_network() 'revolute-2' - error adding bar: %d.\n", ret_val);
	return FALSE;
      }

      if (num_bars_prev == Num_bars) {
	waste_length++;
      } else {
	// Add output node.
	// weight = (command[i].arg) * PARAM_2_WEIGHT;
	weight = 1.0;
	if (!network->add_output(weight, &current_link)) {
	  waste_length++;
	  if (Print[9])
	    printf("bars3d :: build_network() 'revolute-2' - error adding output: %d.\n", ret_val);
	  return FALSE;
	}
      }      

    } else if (command[i].symbol == TERM_JOINT_TWIST) {
      spd = command[i].arg / 2.0;
      if (spd == 0.0)
	spd = 1.0;

      len = command[i].arg;
      if (len == 0.0)
	continue;
      len = 1.0;

      num_bars_prev = Num_bars;
      ret_val = add_bar(1.0, heading, BAR_TYPE_JOINT, ROT_X_AXIS,
			0.0, PI/2.0, spd, offset, &bar_index);
      if (ret_val == FALSE) {
	if (Print[9])
	  printf("bars3d :: build_network() 'twist' - error adding bar: %d.\n", ret_val);
	return FALSE;
      }

      if (num_bars_prev == Num_bars) {
	waste_length++;
      } else {
	// Add output node.
	//	weight = (command[i].arg) * PARAM_2_WEIGHT;
	weight = 1.0;
	if (!network->add_output(weight, &current_link)) {
	  waste_length++;
	  if (Print[9])
	    printf("bars3d :: build_network() 'twist' - error adding output.\n");
	  return FALSE;
	}
      }      

    } else if (command[i].symbol == TERM_JOINT_TWIST180) {
      spd = command[i].arg / 2.0;
      if (spd == 0.0)
	spd = 1.0;

      len = command[i].arg;
      if (len == 0.0)
	continue;
      len = 1.0;

      num_bars_prev = Num_bars;
      ret_val = add_bar(1.0, heading, BAR_TYPE_JOINT, ROT_X_AXIS,
			-PI/2.0, PI/2.0, spd/2.0, offset, &bar_index);
      if (ret_val == FALSE) {
	if (Print[9])
	  printf("bars3d :: build_network() 'twist-180' - error adding bar: %d.\n", ret_val);
	return FALSE;
      }

      if (num_bars_prev == Num_bars) {
	waste_length++;
      } else {
	// Add output node.
	// weight = (command[i].arg) * PARAM_2_WEIGHT;
	weight = 1.0;
	if (!network->add_output(weight, &current_link)) {
	  waste_length++;
	  if (Print[9])
	    printf("bars3d :: build_network() 'twist-180' - error adding output: %d.\n", ret_val);
	  return FALSE;
	}
      }      

    } else if (command[i].symbol == TERM_TURN_UP) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	rot_zaxis_pos(heading);

    } else if (command[i].symbol == TERM_TURN_DOWN) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	rot_zaxis_neg(heading);

    } else if (command[i].symbol == TERM_TURN_LEFT) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	rot_yaxis_pos(heading);

    } else if (command[i].symbol == TERM_TURN_RIGHT) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	rot_yaxis_neg(heading);

    } else if (command[i].symbol == TERM_TURN_CLOCKWISE) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	rot_xaxis_pos(heading);

    } else if (command[i].symbol == TERM_TURN_CNTR_CLOCKWISE) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	rot_xaxis_neg(heading);

    } else {
#ifndef OPTIMIZE
	if (Print[0])
	  printf("Build:: unknown command: (%c:%d)\n", 
		 command[i].symbol, int(command[i].symbol));
#endif
    }     
  }

  /*
  if (network->Num_inputs == 0) {
    return FALSE;
  }
  */

  /*
  printf("> %f  (%d:%d:%d) (%d:%d)\n", passes/tries, int(passes),
	 int(tries), builds,
	 network->Num_outputs, Num_bars);
  */

#ifdef OLD
  if (network->Num_outputs == 0) {
    if (Print[9])
      printf("Build network :: Not enough outputs: %d.\n", network->Num_outputs);
    return FALSE;

  } else if ((Terminal_set < 30) && (Num_bars < 3)) {
    if (Print[9])
      printf("Fitness distance :: Not enough bars: %d.\n", Num_bars);
    return FALSE;
  }
#endif

  // Finish morphology.
  if (Num_bars > 1) {
    int ctr = 0;
    for (i=0; i<Num_bars; i++) {
      if (Bar[i].type == BAR_TYPE_JOINT) {
	if (Bar[i].num_children == 0) {
	  Bar[i].type = BAR_TYPE_FIXED;
	  Bar[i].moves = FALSE;
	  network->delete_output(ctr);

	} else {
	  Bar[i].moves = TRUE;
	  Bar[i].desired_angle = 0.0;
	  Bar[i].angle = -JOINT_MOVE_STEP/2.0;

	  // Link output node with actuated joint.
	  Joint_index[Num_joints++] = i;
	}
	ctr++; // Count number of added joints.

      } else if (Bar[i].type == BAR_TYPE_SENSOR) {
	Sensor_index[Num_sensors++] = i;
      }
    }
    if (Num_joints > Max_joints) {
      if (Print[9]) {
	printf("Build_structure() - error, too many joints.\n");
      }
      return FALSE;
    }

    total_weight = Real(Num_bars);
    for (i=0; i<Num_bars; i++)
      Bar[i].weight /= total_weight;

    calculate_positions(0.0);
    if (Print[34]) {
      printf("Recalc positions: %.4f\n", Time);
      calculate_positions_new(Time);
    }

    if (!settle_structure()) {
      print_struct();
      printf("Settle structure failed in build_network.\n");
      model->error = TRUE;
      return FALSE;
    }    
    //    translate_struct(-COM[0], 0.0, -COM[2]);
    translate_struct(-Bar[0].end_loc[0], 0.0, -Bar[0].end_loc[2]);
    for (i=0; i<Num_bars; i++) {
      Bar[i].prev_loc[0] = Bar[i].end_loc[0];
      Bar[i].prev_loc[1] = Bar[i].prev_loc[1];
    }
    initialize_intersection();
  }

  /*
  if (Print[32]) {
    network->print_structure();
    printf("\n");
  }
  */

  network->finish_build();

  if (Print[32]) {
    //    network->print_structure();
    //    printf("\n");
    network->print_structure2();
    printf("Avg in: %.4f,  # unique sources: %d.\n",
	   network->average_num_inputs(),
	   network->num_unique_output_sources());
  }

  if (Print[40]) {
    network->print_dot_file(stdout);
  }

#ifdef OLD
  //  tries += 1.0;
  if ((network->average_num_inputs() < 1.0) ||
      (network->num_unique_output_sources() < 2)) {
    if (Print[9]) {
      //      printf("> %f  (%d:%d)\n", passes/tries, int(passes), int(tries));
      printf("Build_network() - not enough connectivity: %.4f : %d.\n",
	     network->average_num_inputs(),
	     network->num_unique_output_sources());
    }
    //    return FALSE;
  }
#endif

  /*
  printf("Connectivity: %.4f : %d.\n", network->average_num_inputs(),
	 network->num_unique_output_sources());
  passes += 1.0;
  printf("> %f  (%d:%d:%d)\n", passes/tries, int(passes), int(tries), builds);
  */

  if (Print[6]) {
    printf("Network: I:%d O:%d T:%d, l:%d, avg_in:%.4f,  num_src:%d\n", 
	   network->Num_inputs, network->Num_outputs, network->Num_neurons,
	   network->num_links(),
	   network->average_num_inputs(),
	   network->num_unique_output_sources());
  }

  return TRUE;
}



int BarStruct3D :: build_structure(ModelStruct *model)
{
  int ret_val;

  if (Build_type == 1) {
    ret_val = build_structure_right(model);

  } else if (Build_type == 5) {
    ret_val = build_network(model);

  } else if (Build_type == 10) {
    ret_val = build_structure_tree(model);

  } else {
    printf("Bars3d :: invalid build type.\n");
    ret_val = -1;
  }

  return ret_val;	     
}


void BarStruct3D :: initialize_evaluation(void)
{
  int i;

  // Initialize network.
  if (network) {
    network->reset();
    if ((Terminal_set >= 40) && (Terminal_set < 50)) {
      // In this setting the network is made of just sigmoids so the first
      // Neuron has to be set to a non-zero value to start the network
      // (because the transformation function of the sigmoid is 0->0).
      network->set_first_value(1.0);
    }
  }

  // Initialize ball position.
  if (Is_ball) {
    Angle_sum = 0.0;
    Ball_radius = 20.0;

    Real heading = random_dbl() * 2.0 * PI;
    Real distance = random_dbl()*100.0 + 100.0;

    Ball_loc[0] = cos(heading)*distance;
    Ball_loc[1] = random_dbl()*100.0;
    Ball_loc[2] = sin(heading)*distance;

    for (i=0; i<Num_bars; i++) {
      Bar[i].sensor_value = 0.0;
    }
  }

  drag_distance = 0.0;
  drag_distance_abs1 = 0.0;
  drag_distance_abs2 = 0.0;

  // Give random weights;
  for (i=0; i<Num_bars; i++) {
    if (Bar[i].same_as == -1) {
      Bar[i].weight = 0.75 + 0.5*random_dbl();
    } else
      Bar[i].weight = 0.0;

    if ((Bar[i].num_children > 0) &&
	(Bar[i].type != BAR_TYPE_FIXED)) {
      Bar[i].weight += 0.5;
    }
  }

  total_weight = 0.0;
  for (i=0; i<Num_bars; i++) {
    total_weight += Bar[i].weight;
  }    
  for (i=0; i<Num_bars; i++)
    Bar[i].weight /= total_weight;
}

int BarStruct3D :: fitness_com_height(ModelStruct *model)
{
  int fail = FALSE;
  Real distance=0.0, val;

  if (Num_bars < 3)
    fail = TRUE;
  else if (Num_joints == 0)
    fail = TRUE;

  /*
  if (win1)
    animate_struct(-1);
    */

  if (!fail) {
    initialize_evaluation();
    Time = 0.0;
    com_max[0] = COM[0];
    com_max[1] = COM[1];
    com_max[2] = COM[2];
    com_min[0] = COM[0];
    com_min[1] = COM[1];
    com_min[2] = COM[2];

    val = animate_struct(1.0);
    if (val == -1.0) {
      model->error = TRUE;
      printf("Animate structure error #3\n");
      return FALSE;
    } else if (val < 1.0) {
      return FALSE;
    }

    if ((com_max[0] != com_min[0]) ||
	(com_max[1] != com_min[1]) ||
	(com_max[2] != com_min[2])) {
      com_max[0] = COM[0];
      com_max[1] = COM[1];
      com_max[2] = COM[2];
      com_min[0] = COM[0];
      com_min[1] = COM[1];
      com_min[2] = COM[2];

      val = animate_struct(10.0);
      if (val == -1.0) {
	model->error = TRUE;
	printf("Animate structure error #4\n");
	return FALSE;
      } else if (val < 10.0) {
	return FALSE;
      }

      distance = com_max[1]-com_min[1];
    } else {
      distance = 0.0;
    }
  }

  if (fail) {
    model->fitness[0] = 0.0;
    if (model->fv_length > 1) { 
      model->fitness[1] = Num_bars;
    }

  } else {
    model->fitness[0] = 1.0 + distance;
    if (model->fv_length > 1) { 
      model->fitness[1] = Num_bars;
    }
  }


  if (Print[6]) {
    printf("COM Max: %.2f:%.2f:%.2f\n", com_max[0], com_max[1], com_max[2]);
    printf("COM Min: %.2f:%.2f:%.2f\n", com_min[0], com_min[1], com_min[2]);
    printf("#bars: %d, #joints: %d.\n", Num_bars, Num_joints);
    printf("Fitness: %.1f\n", model->fitness[0]);
  }

  if (model->fitness[0] > 100000.0) {
    printf("Bars3d :: fitness_com_height() - error, fitness too high.\n");
    model->error = TRUE;
  } else {
    model->error = FALSE;
  }
  translate_struct(-COM[0], 0.0, -COM[2]);

  return TRUE;
}



Real BarStruct3D :: run_distance(ModelStruct *model)
{
  Real dist_x, dist_z;
  Real distance, val;
  Real com_init[3];

  initialize_evaluation();
  Time = 0.0;
  com_max[0] = COM[0];
  com_max[1] = COM[1];
  com_max[2] = COM[2];
  com_min[0] = COM[0];
  com_min[1] = COM[1];
  com_min[2] = COM[2];

  //  val = animate_struct(1.0);
  val = animate_struct(TIME_STEP);
  if (val == -1.0) {
    model->error = TRUE;
    printf("Animate structure error #1\n");
    return 0.0;

  } else if (val < TIME_STEP) {
#ifndef OPTIMIZE
    if (Print[9])
      printf("Animate structure(TIME_STEP) ended prematurely\n");
#endif
    return 0.0;
  }

  if ((com_max[0] != com_min[0]) ||
      (com_max[1] != com_min[1]) ||
      (com_max[2] != com_min[2]) ||
      (network)) {
    com_max[0] = COM[0];
    com_max[1] = COM[1];
    com_max[2] = COM[2];
    com_min[0] = COM[0];
    com_min[1] = COM[1];
    com_min[2] = COM[2];
    com_init[0] = COM[0];
    com_init[1] = COM[1];
    com_init[2] = COM[2];

    if (network) {
      val = animate_struct(2.0);
      if (val == -1.0) {
	model->error = TRUE;
	printf("Animate structure error #2\n");
	return 0.0;
	
      } else if (val < 2.0) {
#ifndef OPTIMIZE
	if (Print[9])
	  printf("Animate structure(2.0) ended prematurely\n");
#endif
	return 0.0;
      } else if (network->num_active_outputs() < 2) {
	return 0.0;
      }

      val = animate_struct(NET_TRIAL_LEN);
      if (val == -1.0) {
	model->error = TRUE;
	printf("Animate structure error #2\n");
	return 0.0;
	
      } else if (val < NET_TRIAL_LEN) {
	//      printf("Animate structure ended prematurely\n");
	return 0.0;
      }
      
    } else {
      val = animate_struct(10.0);
      if (val == -1.0) {
	model->error = TRUE;
	printf("Animate structure error #2\n");
	return 0.0;
	
      } else if (val < 10.0) {
	//      printf("Animate structure ended prematurely\n");
	return 0.0;
      }
    }

    // Old distance function: distance from max to min.
    //    distance = sqrt(sqr(com_max[0]-com_min[0]) +
    //		    sqr(com_max[2]-com_min[2]));

    // Calc x_dist travelled.
    if ((com_max[0] - com_init[0]) > (com_init[0] - com_min[0])) {
      dist_x = com_max[0] - com_init[0];
    } else {
      dist_x = com_init[0] - com_min[0];
    }

    // Calc z_dist travelled.
    if ((com_max[2] - com_init[2]) > (com_init[2] - com_min[2])) {
      dist_z = com_max[2] - com_init[2];
    } else {
      dist_z = com_init[2] - com_min[2];
    }    
    distance = sqrt(sqr(dist_x) + sqr(dist_z));

  } else {
    distance = 0.0;
  }

  return distance;
}



void BarStruct3D :: randomize_remote(void)
{
  // Switch directions...
  int dir;
  Real inputs[4];
  for (int i=0; i<4; i++) {
    inputs[i] = 0.0;
  }

  dir = random_int(4);
  inputs[dir] = 1.0;
  network->set_input(inputs);
  printf("New direction: %d\n", dir);
}


#ifdef OLD
    if (direction[i] == 0) {
      // Want to go +X.
      dist = com_max[0] - com_init[0] - (com_init[0] - com_min[0]);
      if ((com_max[2] - com_init[2]) > (com_init[2] - com_min[2])) {
	dist -= com_max[2] - com_init[2];
      } else {
	dist -= com_init[2] - com_min[2];
      }

    } else if (direction[i] == 1) {
      // Want to go -X.
      dist = com_init[0] - com_min[0] - (com_max[0] - com_init[0]);
      if ((com_max[2] - com_init[2]) > (com_init[2] - com_min[2])) {
	dist -= com_max[2] - com_init[2];
      } else {
	dist -= com_init[2] - com_min[2];
      }

    } else if (direction[i] == 2) {
      // Want to go +Z.
      dist = com_max[2] - com_init[2] - (com_init[2] - com_min[2]);
      if ((com_max[0] - com_init[0]) > (com_init[0] - com_min[0])) {
	dist -= com_max[0] - com_init[0];
      } else {
	dist -= com_init[0] - com_min[0];
      }

    } else {
      // Want to go -Z.
      dist = com_init[2] - com_min[2] - (com_max[2] - com_init[2]);
      if ((com_max[0] - com_init[0]) > (com_init[0] - com_min[0])) {
	dist -= com_max[0] - com_init[0];
      } else {
	dist -= com_init[0] - com_min[0];
      }
    }
    if (dist > 0.0) {
      distance += dist;
      //      distance *= dist;
    }
#endif

#define DISTANCE_ERROR -999999.0

// Returns -99999.0 on error.
// Otherwise returns distance.
Real BarStruct3D :: run_distance_remote(ModelStruct *model)
{
  int i, j, dir, direction[8];
  Real dist, distance, val;
  Real com_init[3];
  Real inputs[4];

  initialize_evaluation();
  Time = 0.0;

  // Initialize direction vector to a random permutation
  // of the number 0..4.
  for (i=0; i<4; i++) {
    direction[i] = i;
  }
  for (i=0; i<4; i++) {
    j = random_int(3);
    if (j==i) {
      j = 3;
    }
    dir = direction[i];
    direction[i] = direction[j];
    direction[j] = dir;
  }

  for (i=0; i<4; i++) {
    direction[i+4] = i;
  }
  for (i=0; i<4; i++) {
    j = random_int(3);
    if (j==i) {
      j = 3;
    }
    dir = direction[i+4];
    direction[i+4] = direction[j+4];
    direction[j+4] = dir;
  }


  // Take one time-step to see that if there are any errors.
  for (j=0; j<4; j++) {
    inputs[j] = 0.0;
  }
  network->set_input(inputs);

  val = animate_struct(TIME_STEP);
  if (val == -1.0) {
    model->error = TRUE;
    printf("bars3d :: run_distance_remote() - Animate structure error #1\n");
    return DISTANCE_ERROR;

  } else if (val < TIME_STEP) {
#ifndef OPTIMIZE
    if (Print[9])
      printf("bars3d :: run_distance_remote() - Animate structure(TIME_STEP) ended prematurely\n");
#endif
    return DISTANCE_ERROR;
  }

  distance = 0.0;
  for (i=0; i<8; i++) {
    // Select random direction, and set inputs.
    for (j=0; j<4; j++) {
      inputs[j] = 0.0;
    }
    inputs[direction[i]] = 1.0;
    network->set_input(inputs);
    //    printf("setting direction to: %d\n", direction[i]);

    // Run for 1 second to give Genobot a chance to
    // realize it should switch directions and do so.
    val = animate_struct(1.0);
    if (val == -1.0) {
      model->error = TRUE;
      printf("Animate structure error #2\n");
      return DISTANCE_ERROR;
	
    } else if (val < 1.0) {
#ifndef OPTIMIZE
      if (Print[9])
	printf("Animate structure(1.0) ended prematurely\n");
#endif
      return DISTANCE_ERROR;
    }

    // Initialize Center Of Mass.
    com_max[0] = COM[0];
    com_max[1] = COM[1];
    com_max[2] = COM[2];
    com_min[0] = COM[0];
    com_min[1] = COM[1];
    com_min[2] = COM[2];
    com_init[0] = COM[0];
    com_init[1] = COM[1];
    com_init[2] = COM[2];

    // Run Genobot and measure how far it goes in desired
    // direction.
    val = animate_struct(REMOTE_TRIAL_LEN);
    if (val == -1.0) {
      model->error = TRUE;
      printf("Animate structure error #2\n");
      return DISTANCE_ERROR;
	
    } else if (val < REMOTE_TRIAL_LEN) {
#ifndef OPTIMIZE
      if (Print[9])
	printf("Animate structure(%.1f) ended prematurely\n", REMOTE_TRIAL_LEN);
#endif
      return DISTANCE_ERROR;
    }

    if (direction[i] == 0) {
      // Want to go +X.
      dist = com_max[0] - com_init[0] - (com_init[0] - com_min[0])/2.0;
      if ((com_max[2] - com_init[2]) > (com_init[2] - com_min[2])) {
	dist -= (com_max[2] - com_init[2])/2.0;
      } else {
	dist -= (com_init[2] - com_min[2])/2.0;
      }

    } else if (direction[i] == 1) {
      // Want to go -X.
      dist = com_init[0] - com_min[0] - (com_max[0] - com_init[0])/2.0;
      if ((com_max[2] - com_init[2]) > (com_init[2] - com_min[2])) {
	dist -= (com_max[2] - com_init[2])/2.0;
      } else {
	dist -= (com_init[2] - com_min[2])/2.0;
      }

    } else if (direction[i] == 2) {
      // Want to go +Z.
      dist = com_max[2] - com_init[2] - (com_init[2] - com_min[2])/2.0;
      if ((com_max[0] - com_init[0]) > (com_init[0] - com_min[0])) {
	dist -= (com_max[0] - com_init[0])/2.0;
      } else {
	dist -= (com_init[0] - com_min[0])/2.0;
      }

    } else {
      // Want to go -Z.
      dist = com_init[2] - com_min[2] - (com_max[2] - com_init[2])/2.0;
      if ((com_max[0] - com_init[0]) > (com_init[0] - com_min[0])) {
	dist -= (com_max[0] - com_init[0])/2.0;
      } else {
	dist -= (com_init[0] - com_min[0])/2.0;
      }
    }
    distance += dist;
  }

  if (distance < -1.0) {
    distance = -1.0 / distance;
  } else if (distance < 1.0) {
    distance = (distance + 1.0) / 2.0;
  } else {
    distance += 1.0;
  }

  //  printf("distance: %f  (tries: %d)\n", distance, i);
  return distance;
}




Real BarStruct3D :: fitness_distance_value4(Real distance)
{
  Real fitness;

  if (distance-drag_distance_abs1 <= 0) {
    fitness = 0.0;

  } else {
    fitness = distance - drag_distance_abs1;
    drag_distance_abs2 = sqrt(drag_distance_abs2);
    
    if (fitness/2.0 < drag_distance_abs2)
      fitness /= 2.0;
    else
      fitness -= drag_distance_abs2;

    /*
    if (max_arm_len > 1.0)
      fitness /= max_arm_len;

    len = waste_length;
    if (len > fitness / 2.0)
      fitness /= 2.0;
    else
      fitness -= Real(len);
    */

    // If a network used, factor in network performance.
    /*
    if (network) {
      if (network->num_active_outputs() < 2) {
	fitness = 0.0;
      } else {
	fitness += fitness_network_value();
      }
    }
    */

    if (fitness >= 2.0)
      fitness = floor(fitness/2.0)*2.0;
    else if (distance >= 1.0)
      fitness = 1.0;
    else
      fitness = 0.0;

    if (fitness < 1.0)
      fitness = 0.0;

  }

  if (Print[6]) {
    printf("dist:%6.1f (%.2f, %.2f, %.2f): fitness: %.2f\n", distance,
	   drag_distance, drag_distance_abs1, drag_distance_abs2, fitness);
  }

  return fitness;
}


int BarStruct3D :: fitness_distance4(ModelStruct *model)
{
  int i, fail = FALSE;
  int max_line_len=0;
  Real distance, fitness;
  Real min_fitness = 0.0;


  if (Num_bars < 4) {
    fail = TRUE;
    if (Print[9])
      printf("Fitness distance :: Not enough bars: %d.\n", Num_bars);

  } else if (Num_joints == 0) {
    fail = TRUE;
    if (Print[9])
      printf("Fitness distance :: Not enough joints: %d.\n", Num_joints);

  } else {
    label_cycles();
    max_line_len = max_line_length();
    if (max_line_len > MAX_LINE_LENGTH) {
      fail = TRUE;
      if (Print[9])
	printf("Fitness distance :: Max line length too long: %d (>%d)\n",
	       max_line_len, MAX_LINE_LENGTH);
    }
  }

  if (!fail) {
    for (i=0; i<Num_bars; i++) {
      Bar[i].true_angle_a = Bar[i].angle_a;
      Bar[i].true_angle_b = Bar[i].angle_b;
      mm3_copy(Bar[i].true_delta_rot, Bar[i].delta_rot_mat);
    }

    min_fitness = 9999999.0;
    for (i=0; i<Number_trials; i++) {
      //      if ((i != (Number_trials-1)) || (i == 0)) {
      if (i != 0) {
	add_connection_noise2();
	//	calculate_positions(0.0);
	//	print_struct();
      }
      distance = run_distance(model);
      if (distance == 0.0) {
	fail = TRUE;
	min_fitness = 0.0;
	break;

      } else {
	fitness = fitness_distance_value4(distance);
	if (fitness < min_fitness) {
	  min_fitness = fitness;
	  if (min_fitness == 0.0)
	    break;
	}
      }
    }
  }

  fitness = min_fitness;
  model->fitness[0] = fitness;
  if (model->fv_length > 1) { 
    model->fitness[1] = Num_bars;
  }

  if (Print[6]) {
    max_arm_len = Real(max_arm_length());
    printf("COM %.2f:%.2f ; %.2f:%.2f\n", com_min[0], com_max[0],
	   com_min[1], com_max[1]);
    printf("#bars:%d  #joints:%d  #sensors:%d  max-arm:%.0f  max-line:%d  len:%d  waste:%d\n",
	   Num_bars, Num_joints, Num_sensors, max_arm_len, max_line_len, model->lstring_length,
	   waste_length);
    printf("Fitness #4: %.1f\n", model->fitness[0]);
    if (network) {
      printf("Network: I:%d O:%d T:%d l:%d, avg_in:%.4f:%.4f,  num_src:%d:%d\n", 
	     network->Num_inputs, network->Num_outputs, network->Num_neurons,
	     network->num_links(),
	     network->average_num_inputs(), network->average_output_range(),
	     network->num_unique_output_sources(), network->num_active_outputs());
    }
  }

  if (model->fitness[0] > 100000.0) {
    printf("Fitness distance :: Error - fitness too high: %.3f\n", model->fitness[0]);
    model->error = TRUE;
  }

  //  translate_struct(-COM[0], 0.0, -COM[2]);
  translate_struct(-Bar[0].end_loc[0], 0.0, -Bar[0].end_loc[2]);
  ref_pt = 0;
  //  ref_rotation = 0.0;
  //  make_identity_mat(Bar[0].rotation_matrix);
  //  make_identity_mat(rotation_matrix);
  
  return TRUE;
}





Real BarStruct3D :: fitness_distance_value3(Real distance)
{
  Real fitness;

  if (distance-drag_distance_abs1 <= 0) {
    fitness = 0.0;

  } else {
    fitness = distance - drag_distance_abs1;
    drag_distance_abs2 = sqrt(drag_distance_abs2);
    
    if (fitness/2.0 < drag_distance_abs2)
      fitness /= 2.0;
    else
      fitness -= drag_distance_abs2;

    /*
    if (max_arm_len > 1.0)
      fitness /= max_arm_len;

    len = waste_length;
    if (len > fitness / 2.0)
      fitness /= 2.0;
    else
      fitness -= Real(len);
    */

    // If a network used, factor in network performance.
    if (network) {
      if (network->num_active_outputs() < 2) {
	fitness = 0.0;
      } else {
	fitness += fitness_network_value();
      }
    }

    if (fitness >= 2.0)
      fitness = floor(fitness/2.0)*2.0;
    else if (distance >= 1.0)
      fitness = 1.0;
    else
      fitness = 0.0;

    if (fitness < 1.0)
      fitness = 0.0;

  }

  if (Print[6]) {
    printf("dist:%6.1f (%.2f, %.2f, %.2f): fitness: %.2f\n", distance,
	   drag_distance, drag_distance_abs1, drag_distance_abs2, fitness);
  }

  return fitness;
}


int BarStruct3D :: fitness_distance3(ModelStruct *model)
{
  int i, fail = FALSE;
  Real distance, fitness;
  Real min_fitness = 0.0;


  if (Num_bars < 3) {
    fail = TRUE;
    if (Print[9])
      printf("Fitness distance :: Not enough bars: %d.\n", Num_bars);

  } else if (Num_joints == 0) {
    fail = TRUE;
    if (Print[9])
      printf("Fitness distance :: Not enough joints: %d.\n", Num_joints);

  } else {

    label_cycles();
    //      max_arm_len = Real(max_arm_length());
    for (i=0; i<Num_bars; i++) {
      Bar[i].true_angle_a = Bar[i].angle_a;
      Bar[i].true_angle_b = Bar[i].angle_b;
      mm3_copy(Bar[i].true_delta_rot, Bar[i].delta_rot_mat);
    }
  }


  if (!fail) {
    min_fitness = 9999999.0;
    for (i=0; i<Number_trials; i++) {
      if ((i != (Number_trials-1)) || (i == 0)) {
	add_connection_noise();
	//	calculate_positions(0.0);
	//	print_struct();
      }
      distance = run_distance(model);
      if (distance == 0.0) {
	fail = TRUE;
	min_fitness = 0.0;
	break;

      } else {
	fitness = fitness_distance_value3(distance);
	if (fitness < min_fitness) {
	  min_fitness = fitness;
	  if (min_fitness == 0.0)
	    break;
	}
      }
    }
  }

  fitness = min_fitness;
  model->fitness[0] = fitness;
  if (model->fv_length > 1) { 
    model->fitness[1] = Num_bars;
  }

  if (Print[6]) {
    max_arm_len = Real(max_arm_length());
    printf("COM %.2f:%.2f ; %.2f:%.2f\n", com_min[0], com_max[0],
	   com_min[1], com_max[1]);
    printf("#bars:%d  #joints:%d  #sensors:%d  max-arm:%.0f  len:%d  waste:%d\n",
	   Num_bars, Num_joints, Num_sensors, max_arm_len, model->lstring_length, waste_length);
    printf("Fitness #3: %.1f\n", model->fitness[0]);
    if (network) {
      printf("Network: I:%d O:%d T:%d l:%d, avg_in:%.4f:%.4f,  num_src:%d:%d\n", 
	     network->Num_inputs, network->Num_outputs, network->Num_neurons,
	     network->num_links(),
	     network->average_num_inputs(), network->average_output_range(),
	     network->num_unique_output_sources(), network->num_active_outputs());
    }
  }

  if (model->fitness[0] > 100000.0) {
    printf("Fitness distance :: Error - fitness too high: %.3f\n", model->fitness[0]);
    model->error = TRUE;
  }

  //  translate_struct(-COM[0], 0.0, -COM[2]);
  translate_struct(-Bar[0].end_loc[0], 0.0, -Bar[0].end_loc[2]);
  ref_pt = 0;
  //  ref_rotation = 0.0;
  //  make_identity_mat(Bar[0].rotation_matrix);
  //  make_identity_mat(rotation_matrix);
  
  return TRUE;
}





Real BarStruct3D :: fitness_distance_value2(Real distance)
{
  Real fitness;

  if (distance-drag_distance_abs1 <= 0) {
    fitness = 0.0;

  } else {
    Real len;
    
    fitness = distance - drag_distance_abs1;
    drag_distance_abs2 = sqrt(drag_distance_abs2);
    
    if (fitness/2.0 < drag_distance_abs2)
      fitness /= 2.0;
    else
      fitness -= drag_distance_abs2;

    if (max_arm_len > 1.0)
      fitness /= max_arm_len;
    
    len = waste_length;
    if (len > fitness / 2.0)
      fitness /= 2.0;
    else
      fitness -= Real(len);
    
    // If a network used, factor in network performance.
    if (network) {
      if (network->num_active_outputs() < 2) {
	fitness = 0.0;
      } else {
	fitness += fitness_network_value();
	if (!network->output_changed()) {
	  if (fitness > 1.0)
	    fitness = sqrt(fitness);
	  else
	    fitness *= fitness;
	}
      }
    }

    // Factor in body ratio to encourage big bodies.
    if ((fitness > 1.0) || (distance/max_arm_len > 1.0)) {
      if (body_ratio/Real(Num_joints) > fitness)
	fitness *= 2.0;
      else
	fitness += body_ratio/Real(Num_joints);
    }


    if (fitness >= 2.0)
      fitness = floor(fitness/2.0)*2.0;
    else if (distance/max_arm_len >= 1.0)
      fitness = 1.0;
    else
      fitness = 0.0;

    if (fitness < 1.0)
      fitness = 0.0;
  }

  if (Print[6]) {
    printf("dist:%6.1f (%.2f, %.2f, %.2f): fitness: %.2f\n", distance,
	   drag_distance, drag_distance_abs1, drag_distance_abs2, fitness);
  }

  return fitness;
}


int BarStruct3D :: fitness_distance2(ModelStruct *model)
{
  int i, j, fail = FALSE;
  Real distance, fitness;
  Real min_fitness = 0.0;

  if (Num_bars < 4) {
    fail = TRUE;
    if (Print[9])
      printf("Fitness distance :: Not enough bars: %d.\n", Num_bars);

  } else if (Num_joints == 0) {
    fail = TRUE;
    if (Print[9])
      printf("Fitness distance :: Not enough joints: %d.\n", Num_joints);

  } else {
    fail = TRUE;
    for (i=1; i<Num_bars; i++) {
      //      if (Bar[i].num_children > 1) {
      if ((Bar[i].num_children > 1) ||
          (Bar[i].same_as > -1)) {
	fail = FALSE;
	break;
      }
    }
    if (fail) {
      if (Print[9]) {
	printf("Fitness distance :: single line - fail (a).\n");
	max_arm_len = -1.0;
      }

    } else {
      label_cycles();
      max_arm_len = Real(max_arm_length());
      for (i=0; i<Num_bars; i++) {
	Bar[i].true_angle_a = Bar[i].angle_a;
	Bar[i].true_angle_b = Bar[i].angle_b;
	mm3_copy(Bar[i].true_delta_rot, Bar[i].delta_rot_mat);
      }

      body_ratio = 0.0;
      for (i=0; i<Num_bars; i++) {
	for (j=0; j<Bar[i].num_children; j++) {
	  if (Bar[i].in_cycle[j])
	    body_ratio += 1.0;
	}
      }
      if (body_ratio == 0.0) {
	fail = TRUE;
	if (Print[9]) {
	  printf("Fail because :  body ratio = 0.0\n");
	}
      }
    }
  }

  if (!fail) {
    min_fitness = 9999999.0;
    for (i=0; i<Number_trials; i++) {
      if (i != 0) {
	add_connection_noise();
      }
      distance = run_distance(model);
      if (distance == 0.0) {
	fail = TRUE;
	min_fitness = 0.0;
	if (Print[9])
	  printf(" Fail because :  Distance = 0.0\n");
	break;

      } else {
	fitness = fitness_distance_value2(distance);
	if (fitness < min_fitness) {
	  min_fitness = fitness;
	  if (min_fitness == 0.0)
	    break;
	}
      }
    }
  }

  fitness = min_fitness;
  model->fitness[0] = fitness;
  if (model->fv_length > 1) { 
    model->fitness[1] = Num_bars;
  }

  if (Print[6] && (!fail)) {
    printf("COM %.2f:%.2f ; %.2f:%.2f\n", com_min[0], com_max[0],
	   com_min[1], com_max[1]);
    printf("#bars:%d  #joints:%d  #sensors:%d  max-arm:%.0f  body-ratio:%.2f:%.2f  len:%d  waste:%d\n",
	   Num_bars, Num_joints, Num_sensors, max_arm_len, body_ratio,
	   (body_ratio/Real(Num_joints)), model->lstring_length, waste_length);
    printf("Fitness #2: %.1f\n", model->fitness[0]);
    if (network) {
      printf("Network: I:%d O:%d T:%d l:%d, avg_in:%.4f:%.4f,  num_src:%d:%d\n", 
	     network->Num_inputs, network->Num_outputs, network->Num_neurons,
	     network->num_links(),
	     network->average_num_inputs(), network->average_output_range(),
	     network->num_unique_output_sources(), network->num_active_outputs());
    }
  }

  if (model->fitness[0] > 100000.0) {
    printf("Fitness distance :: Error - fitness too high: %.3f\n", model->fitness[0]);
    model->error = TRUE;
  }

  //  translate_struct(-COM[0], 0.0, -COM[2]);
  translate_struct(-Bar[0].end_loc[0], 0.0, -Bar[0].end_loc[2]);

  ref_pt = 0;
  //  ref_rotation = 0.0;
  //  make_identity_mat(Bar[0].rotation_matrix);
  //  make_identity_mat(rotation_matrix);
  
  return TRUE;
}


#ifdef OLD
Real BarStruct3D :: fitness_distance_value(Real distance)
{
  Real fitness;

  if (distance-drag_distance_abs1 <= 0) {
    fitness = 0.0;

  } else {
    Real len;
    
    fitness = distance - drag_distance_abs1;
    drag_distance_abs2 = sqrt(drag_distance_abs2);
    
    if (fitness/2.0 < drag_distance_abs1)
      fitness /= 2.0;
    else
      fitness -= drag_distance_abs1;
    
    len = waste_length;
    if (len > fitness / 2.0)
      fitness /= 2.0;
    else
      fitness -= Real(len);

    if (max_arm_len > 1.0)
      fitness /= max_arm_len;
    
    if (fitness >= 2.0)
      fitness = floor(fitness/2.0)*2.0;
    else if (distance >= 1.0)
      fitness = 1.0;
    else
      fitness = 0.0;

    if (fitness < 1.0)
      fitness = 0.0;

  }

  if (Print[6]) {
    printf("dist:%6.1f (%.2f, %.2f, %.2f): fitness: %.2f\n", distance,
	   drag_distance, drag_distance_abs1, drag_distance_abs2, fitness);
  }

  return fitness;
}
#endif


Real BarStruct3D :: fitness_distance_value(Real distance)
{
  Real fitness;

  if (distance-drag_distance_abs1 <= 0) {
    fitness = 0.0;

  } else {
    Real len;
    
    fitness = distance - drag_distance_abs1;
    drag_distance_abs2 = sqrt(drag_distance_abs2);
    
    if (fitness/2.0 < drag_distance_abs2)
      fitness /= 2.0;
    else
      fitness -= drag_distance_abs2;

    if (max_arm_len > 1.0)
      fitness /= max_arm_len;

    len = waste_length;
    if (len > fitness / 2.0)
      fitness /= 2.0;
    else
      fitness -= Real(len);

    // If a network used, factor in network performance.
    if (network && (Terminal_set != 28)) {
      if (network->num_active_outputs() < 2) {
	fitness = 0.0;
      } else {
	fitness += fitness_network_value();
      }
    }

    if (fitness >= 2.0)
      fitness = floor(fitness/2.0)*2.0;
    else if (distance >= 1.0)
      fitness = 1.0;
    else
      fitness = 0.0;

    if (fitness < 1.0)
      fitness = 0.0;

  }

  if (Print[6]) {
    printf("dist:%6.1f (%.2f, %.2f, %.2f): fitness: %.2f\n", distance,
	   drag_distance, drag_distance_abs1, drag_distance_abs2, fitness);
  }

  return fitness;
}


int BarStruct3D :: fitness_distance(ModelStruct *model)
{
  int i, fail = FALSE;
  Real distance, fitness;
  Real min_fitness = 0.0;

  if (Num_bars < 3) {
    fail = TRUE;
    if (Print[9])
      printf("Fitness distance :: Not enough bars: %d.\n", Num_bars);

  } else if (Num_joints == 0) {
    fail = TRUE;
    if (Print[9])
      printf("Fitness distance :: Not enough joints: %d.\n", Num_joints);

  } else {
    fail = TRUE;
    for (i=1; i<Num_bars; i++) {
      //      if (Bar[i].num_children > 1) {
      if ((Bar[i].num_children > 1) ||
          (Bar[i].same_as > -1)) {
	fail = FALSE;
	break;
      }
    }
    if (fail) {
      if (Print[9]) {
	printf("Fitness distance :: single line - fail (b).\n");
	max_arm_len = -1.0;
      }

    } else {
      label_cycles();
      max_arm_len = Real(max_arm_length());
      for (i=0; i<Num_bars; i++) {
	Bar[i].true_angle_a = Bar[i].angle_a;
	Bar[i].true_angle_b = Bar[i].angle_b;
	mm3_copy(Bar[i].true_delta_rot, Bar[i].delta_rot_mat);
      }
    }
  }


  if (!fail) {
    min_fitness = 9999999.0;
    for (i=0; i<Number_trials; i++) {
      if (i != 0)
	add_connection_noise();
      distance = run_distance(model);
      if (distance == 0.0) {
	fail = TRUE;
	min_fitness = 0.0;
	break;

      } else {
	fitness = fitness_distance_value(distance);
	if (fitness < min_fitness) {
	  min_fitness = fitness;
	  if (min_fitness == 0.0)
	    break;
	}
      }
    }
  }

  fitness = min_fitness;
  model->fitness[0] = fitness;
  if (model->fv_length > 1) { 
    model->fitness[1] = Num_bars;
  }

  if (Print[6]) {
    printf("COM %.2f:%.2f ; %.2f:%.2f\n", com_min[0], com_max[0],
	   com_min[1], com_max[1]);
    printf("#bars:%d  #joints:%d  #sensors:%d  max-arm:%.0f  len:%d  waste:%d\n",
	   Num_bars, Num_joints, Num_sensors, max_arm_len, model->lstring_length, waste_length);
    printf("Fitness #1: %.1f\n", model->fitness[0]);
    if (network) {
      printf("Network: I:%d O:%d T:%d l:%d, avg_in:%.4f:%.4f,  num_src:%d:%d\n", 
	     network->Num_inputs, network->Num_outputs, network->Num_neurons,
	     network->num_links(),
	     network->average_num_inputs(), network->average_output_range(),
	     network->num_unique_output_sources(), network->num_active_outputs());
    }
  }

  if (model->fitness[0] > 100000.0) {
    printf("Fitness distance :: Error - fitness too high: %.3f\n", model->fitness[0]);
    model->error = TRUE;
  }

  //  translate_struct(-COM[0], 0.0, -COM[2]);
  translate_struct(-Bar[0].end_loc[0], 0.0, -Bar[0].end_loc[2]);
  ref_pt = 0;
  //  ref_rotation = 0.0;
  //  make_identity_mat(Bar[0].rotation_matrix);
  //  make_identity_mat(rotation_matrix);
  
  return TRUE;
}



Real BarStruct3D :: fitness_ball_value(Real distance)
{
  Real fitness;

  if (distance-drag_distance_abs1 <= 0) {
    fitness = 0.0;

  } else if (Angle_sum == 0.0) {
    fitness = 0.0;
    //    printf("Angle_sum == 0.0\n");

  } else {
    Real len;

    fitness = distance - drag_distance_abs1;
    drag_distance_abs2 = sqrt(drag_distance_abs2);

    //    fitness += Angle_sum / Real(Num_sensors);

    if (fitness/2.0 < drag_distance_abs2)
      fitness /= 2.0;
    else
      fitness -= drag_distance_abs2;

    if (max_arm_len > 1.0)
      fitness /= max_arm_len;

    len = waste_length;
    if (len > fitness / 2.0)
      fitness /= 2.0;
    else
      fitness -= Real(len);

    if (fitness >= 2.0)
      fitness = floor(fitness/2.0)*2.0;
    else if (distance >= 1.0)
      fitness = 1.0;
    else
      fitness = 0.0;

    if (fitness < 1.0)
      fitness = 0.0;

  }

  if (Print[6]) {
    printf("dist:%6.1f (%.2f, %.2f, %.2f): ang-sum:%.2f:%d  fitness: %.2f\n",
	   distance, drag_distance, drag_distance_abs1, drag_distance_abs2,
	   Angle_sum, Num_sensors, fitness);
  }

  return fitness;
}


int BarStruct3D :: fitness_ball(ModelStruct *model)
{
  int i, fail = FALSE;
  Real distance, fitness;
  Real min_fitness = 0.0;

  if (Num_bars < 3) {
    fail = TRUE;
    if (Print[9])
      printf("Fitness ball :: Not enough bars: %d.\n", Num_bars);

  } else if (Num_joints == 0) {
    fail = TRUE;
    if (Print[9])
      printf("Fitness ball :: Not enough joints: %d.\n", Num_joints);

  } else if (Num_sensors == 0) {
    fail = TRUE;
    if (Print[9])
      printf("Fitness ball :: Not enough sensors: %d.\n", Num_joints);

  } else {
    fail = TRUE;
    for (i=1; i<Num_bars; i++) {
      if (Bar[i].num_children > 1) {
	fail = FALSE;
	break;
      }
    }
    if (fail) {
      if (Print[9]) {
	printf("Fitness distance :: single line - fail.\n");
      }

    } else {
      label_cycles();
      max_arm_len = Real(max_arm_length());
      for (i=0; i<Num_bars; i++) {
	Bar[i].true_angle_a = Bar[i].angle_a;
	Bar[i].true_angle_b = Bar[i].angle_b;
	mm3_copy(Bar[i].true_delta_rot, Bar[i].delta_rot_mat);
      }
    }
  }

  if (!fail) {
    Is_ball = TRUE;
    min_fitness = 9999999.0;
    for (i=0; i<Number_trials; i++) {
      if (i != 0)
	add_connection_noise();

      // Run distance requires the creature to move after
      // one timestep - with a network, may want to wait a
      // few seconds to let the network startup.

      distance = run_distance(model);
      if (distance == 0.0) {
	fail = TRUE;
	min_fitness = 0.0;
	break;

      } else {
	fitness = fitness_ball_value(distance);
	if (fitness < min_fitness) {
	  min_fitness = fitness;
	  if (min_fitness == 0.0)
	    break;
	}
      }
    }
    
    /*
    if (!fail)
      min_fitness += 1.0;
    */
  }
  
  fitness = min_fitness;
  model->fitness[0] = min_fitness;
  if (model->fv_length > 1) { 
    model->fitness[1] = Num_bars;
  }

  if (Print[6] && (!fail)) {
    printf("COM %.2f:%.2f ; %.2f:%.2f  Ang: %.4f\n", com_min[0], com_max[0],
	   com_min[1], com_max[1], Angle_sum);
    printf("#bars:%d  #joints:%d  #sensors:%d  max-arm:%.0f  len:%d  waste:%d\n",
	   Num_bars, Num_joints, Num_sensors, max_arm_len, model->lstring_length, waste_length);
    printf("Fitness #5: %.1f\n", model->fitness[0]);
  }

  if (model->fitness[0] > 100000.0) {
    printf("Fitness distance :: Error - fitness too high: %.3f\n", model->fitness[0]);
    model->error = TRUE;
    //  } else {
    //    model->error = FALSE;
  }

  //  translate_struct(-COM[0], 0.0, -COM[2]);
  translate_struct(-Bar[0].end_loc[0], 0.0, -Bar[0].end_loc[2]);
  ref_pt = 0;
  //  ref_rotation = 0.0;
  //  make_identity_mat(Bar[0].rotation_matrix);
  //  make_identity_mat(rotation_matrix);
  
  return TRUE;
}


Real BarStruct3D :: fitness_remote_value(Real distance)
{
  Real fitness;

  if (distance-drag_distance_abs1 <= 0) {
    fitness = 0.0;

  } else {
    //    fitness = distance - drag_distance_abs1;
    fitness = distance;
    drag_distance_abs2 = sqrt(drag_distance_abs2);
    
    if (fitness/2.0 < drag_distance_abs2)
      fitness /= 2.0;
    else
      fitness -= drag_distance_abs2;

    // If a network used, factor in network performance.
    /*
    if (network) {
      if (network->num_active_outputs() < 2) {
	fitness = 0.0;
      } else {
	fitness += fitness_network_value();
      }
    }
    */

    if (fitness >= 20.0) {
      fitness = floor(fitness);
    }
  }

  if (Print[6]) {
    printf("dist:%6.1f (%.2f, %.2f, %.2f): fitness: %.2f\n", distance,
	   drag_distance, drag_distance_abs1, drag_distance_abs2, fitness);
  }

  return fitness;
}


int BarStruct3D :: fitness_remote(ModelStruct *model)
{
  int i, fail = FALSE;
  int max_line_len;
  Real distance, fitness = 0.0, val;
  Real min_fitness = 0.0;


  if (Num_bars < 5) {
    fail = TRUE;
    if (Print[9])
      printf("Fitness remote :: Not enough bars: %d.\n", Num_bars);

  } else if (Num_joints < 2) {
    fail = TRUE;
    if (Print[9])
      printf("Fitness remote :: Not enough joints: %d.\n", Num_joints);

  } else if (network->num_used_inputs() < 4) {
    fail = TRUE;
    if (Print[9]) {
      printf("Fitness remote :: Not enough used inputs: %d.\n",
	     network->num_used_inputs());
    }

  } else {
    label_cycles();
    max_line_len = max_line_length();
    if (max_line_len > MAX_LINE_LENGTH) {
      fail = TRUE;
      if (Print[9])
	printf("Fitness distance :: Max line length too long: %d (>%d)\n",
	       max_line_len, MAX_LINE_LENGTH);
    }
  }

  //  network->print_structure();
  //  network->print_structure2();
  //  printf("Net num used: %d\n", network->num_used_inputs());


  if (!fail) {
    for (i=0; i<Num_bars; i++) {
      Bar[i].true_angle_a = Bar[i].angle_a;
      Bar[i].true_angle_b = Bar[i].angle_b;
      mm3_copy(Bar[i].true_delta_rot, Bar[i].delta_rot_mat);
    }

    min_fitness = 9999999.0;
    for (i=0; i<Number_trials; i++) {
      if (i != 0) {
	add_connection_noise2();
      }

      distance = run_distance_remote(model);
      if (distance == DISTANCE_ERROR) {
	fail = TRUE;
	min_fitness = 0.0;
	break;

      } else {
	val = fitness_remote_value(distance);
	fitness += val;
	//	printf("fitness: %f:%f\n", val, fitness);
	if (val < min_fitness) {
	  min_fitness = val;
	  //	  if (min_fitness == 0.0)
	  //	    break;
	}
      }      
    }
  }
  
  //  fitness = min_fitness;
  //  model->fitness[0] = min_fitness;
  model->fitness[0] = fitness / (Real)Number_trials;
  if (model->fv_length > 1) { 
    model->fitness[1] = Num_bars;
  }

  if (Print[6] && (!fail)) {
    printf("COM %.2f:%.2f ; %.2f:%.2f  Ang: %.4f\n", com_min[0], com_max[0],
	   com_min[1], com_max[1], Angle_sum);
    printf("#bars:%d  #joints:%d  #sensors:%d  max-arm:%.0f  len:%d  waste:%d\n",
	   Num_bars, Num_joints, Num_sensors, max_arm_len, model->lstring_length, waste_length);
    printf("#inputs:%d(%d)  #output:%d  total:%d\n",
	   network->Num_inputs, network->num_used_inputs(), network->Num_outputs, network->Num_neurons);
    printf("Fitness #6: %.1f\n", model->fitness[0]);
  }

  if (model->fitness[0] > 100000.0) {
    printf("Fitness distance :: Error - fitness too high: %.3f\n", model->fitness[0]);
    model->error = TRUE;
    //  } else {
    //    model->error = FALSE;
  }

  //  translate_struct(-COM[0], 0.0, -COM[2]);
  translate_struct(-Bar[0].end_loc[0], 0.0, -Bar[0].end_loc[2]);
  ref_pt = 0;
  //  ref_rotation = 0.0;
  //  make_identity_mat(Bar[0].rotation_matrix);
  //  make_identity_mat(rotation_matrix);
  
  return TRUE;
}



int BarStruct3D :: fitness_tree(ModelStruct *model)
{
  if (Num_bars < 3) {
    model->fitness[0] = 0.0;
    if (Print[9])
      printf("Fitness tree :: Not enough branches: %d.\n", Num_bars);

  } else {
    model->fitness[0] = 1.0;
  }
  if (model->fv_length > 1) { 
    model->fitness[1] = Num_bars;
  }

  if (Print[6]) {
    printf("COM %.2f:%.2f ; %.2f:%.2f\n", com_min[0], com_max[0],
	   com_min[1], com_max[1]);
    printf("#bars:%d  #joints:%d  #sensors:%d  waste:%d\n",
	   Num_bars, Num_joints, Num_sensors, waste_length);
    printf("Fitness Tree: %.1f\n", model->fitness[0]);
  }

  return TRUE;
}



void BarStruct3D :: update_network(void)
{
  int i, index;
  Real theta;
  Real inputs[Num_sensors];
  Real outputs[network->Num_outputs];

  // If sensors, set inputs first.
  if (Num_sensors > 0) {
    for (i=0; i<Num_sensors; i++) {
      index = Sensor_index[i];
      inputs[i] = Bar[index].sensor_value;
    }
    network->set_input(inputs);
  }

  network->next_iteration();
  network->get_output(outputs);

#ifndef OPTIMIZE
  if (Print[33]) {
    if (!network->output_changed()) {
      printf("Output unchanged.\n");
    } else {
      for (i=0; i<Num_joints; i++) {
	printf("%.4f  ", outputs[i]);
      }
      printf("\n");
    }
  }
#endif



  for (i=0; i<Num_joints; i++) {
    theta = (outputs[i] + 1.0) / 2.0; // theta in range [0, 1].

    index = Joint_index[i];
    Bar[index].desired_angle = Bar[index].angle_a*(1.0-theta) + Bar[index].angle_b*theta;
  }
}



Real BarStruct3D :: fitness_network_value(void)
{
  if (!network)
    return 0.0;

  // Fitness to encourage connected networks.
  Real val_num, val_avg, value;
  
  val_num = Real(network->num_active_outputs());
  val_avg = network->average_output_range();
  value = (val_num+val_avg) + val_num * val_avg;
  value /= 2.0;

  return value;
}


int BarStruct3D :: fitness_network(ModelStruct *model)
{
  /*
  if (network->Num_inputs == 0)
    return FALSE;
  */
  if (network->Num_outputs == 0) {
    if (Print[9]) {
      printf("bars3d :: fitness_network() - Fail because number of outputs = 0.\n");
    }
    return FALSE;
  }

  int i, high;
  Real max, min, fitness;
  Real inputs[network->Num_inputs];
  Real outputs[network->Num_outputs];

  high = TRUE;
  max = -1.0;
  min = 1.0;
  fitness = 0.0;

  network->set_first_value(1.0);
  for (i=0; i<network->Num_inputs; i++)
    inputs[i] = 1.0;

  network->set_input(inputs);

  for (i=0; i<1000; i++) {
    network->next_iteration();
    network->get_output(outputs);

#ifndef OPTIMIZE
    if (Print[33]) {
      printf("%.4f\n", outputs[0]);
    }
#endif

    if (outputs[0] < min)
      min = outputs[0];
    else if (outputs[0] > max)
      max = outputs[0];

    if (high) {
      if (outputs[0] > 0.9) {
	fitness += max - min;
	high = FALSE;
	min = max;
      }
    } else {
      if (outputs[0] < -0.9) {
	fitness += max - min;
	high = TRUE;
	max = min;
      }
    }
  }

  fitness += max - min;
  /*
  if (fitness > 0.0) {
    if (high) {
      fitness += (1.0 - outputs[0])/2.0;
    } else {
      fitness += (outputs[0] + 1.0)/2.0;
    }
  }
  */

  if (Print[9]) {
    printf("Waste length: %d   fit: %.4f\n", waste_length, fitness);
    network->print_structure();
    printf("Avg in: %.4f,  # unique sources: %d.\n",
	   network->average_num_inputs(),
	   network->num_unique_output_sources());
  }


  fitness -= Real(waste_length)/1000.0;
  if (fitness < 0.0)
    fitness = 0.0;

  //  network->print_state();

  //  printf("Output: %f\n", outputs[0]);

  model->fitness[0] = fitness;
  if (model->fv_length > 1) { 
    model->fitness[1] = Num_bars;
  }

  if (Print[6]) {
    printf("COM %.2f:%.2f ; %.2f:%.2f\n", com_min[0], com_max[0],
	   com_min[1], com_max[1]);
    printf("#inputs:%d  #output:%d  total:%d  waste:%d\n",
	   network->Num_inputs, network->Num_outputs, network->Num_neurons,
	   waste_length);
    printf("Fitness Network: %.1f\n", model->fitness[0]);

    if (Print[14]) {
      network->print_structure();
    }
  }

  return TRUE;
}

/*
  for (i=0; i<1000; i++) {
    network->next_iteration();
    network->get_output(outputs);
    if (outputs[0] < min)
      min = outputs[0];
    else if (outputs[0] > max)
      max = outputs[0];

    if ((high) && (min < -0.9)) {
      if (outputs[0] > 0.9) {
	fitness += 1.0;
	high = FALSE;
      }
    } else {
      if (outputs[0] < -0.9) {
	fitness += 1.0;
	high = TRUE;
      }
    }
  }

  fitness += max - min;
  if (fitness > 1.0) {
    if (high) {
      fitness += (1.0 - outputs[0])/2.0;
    } else {
      fitness += (outputs[0] + 1.0)/2.0;
    }
  }

*/

#define PARITY_THRESHOLD 0.9

int BarStruct3D :: calculate_parity(int num, Real *inputs)
{
  int i;
  int correct = 0;
  int parity = 0;
  Real outputs[network->Num_outputs];

  for (i=0; i<num; i++) {
    if (inputs[i] > 0.0) {
      parity++;
    }
  }
  parity %= 2;


  network->reset();
  network->set_input(inputs);

  network->next_iteration();
  network->next_iteration();
  network->next_iteration();
  /*
  network->next_iteration();
  network->next_iteration();
  */

  for (i=0; i<num; i++) {
    network->next_iteration();
    network->get_output(outputs);

    if (outputs[0] > PARITY_THRESHOLD) {
      if (parity) {
	correct = 2;
      } else {
	correct = -1;
	;
      }
      break;

    } else if (outputs[0] < -PARITY_THRESHOLD) {
      if (!parity) {
	correct = 2;
      } else {
	correct = -1;
	;
      }
      break;
    }
  }

  if (correct == 0) {
    if (parity && (outputs[0] > 0.0)) {
      correct = 1;
    } else if ((!parity) && (outputs[0] < 0.0)) {
      correct = 1;
    }
  }

  /*
  printf("%.0f %.0f %0.f %.0f  -> %.2f(%d)  %d\n",
	 (float)inputs[0],
	 (float)inputs[1],
	 (float)inputs[2],
	 (float)inputs[3],
	 (float)outputs[0], parity, correct);
  */

  return correct;
}

int BarStruct3D :: fitness_net_parity(ModelStruct *model)
{
  if (network->Num_inputs == 0) {
    if (Print[9]) {
      printf("bars3d :: fitness_net_parity() - Fail because number of inputs = 0.\n");
    }
    return FALSE;
  }

  if (network->Num_outputs == 0) {
    if (Print[9]) {
      printf("bars3d :: fitness_net_parity() - Fail because number of outputs = 0.\n");
    }
    return FALSE;
  }

  int i, length;
  Real fitness;
  Real inputs[MAX_NETWORK_INPUTS+1]; // Extra 1 for overflow.
  
  length = 7;

  fitness = 0.0;

  if (network->Num_inputs != length) {
    if (Print[9]) {
      printf("bars3d :: fitness_net_parity() - Fail because number of inputs < %d, length.\n", length);
    }
    return FALSE;
  }

  if (network->Num_outputs != 1) {
    if (Print[9]) {
      printf("bars3d :: fitness_net_parity() - Fail because number of outputs != 1.\n");
    }
    return FALSE;
  }




  if (network->Num_inputs < length) {
    if (Print[9]) {
      printf("bars3d :: fitness_net_parity() - Fail because number of inputs < %d, length.\n", length);
    }
    return FALSE;
  }



  // Test all parity situations by starting with LOW, and
  // incremently flipping to HIGH by "adding 1".

  // Set all bits LOW.
  for (i=0; i<length; i++) {
    inputs[i] = -1.0;
  }

  // Clear extra bits explicitly.
  for (; i<MAX_NETWORK_INPUTS; i++) {
    inputs[i] = 0.0;
  }

  while (1) {
    fitness += (Real)calculate_parity(length, inputs);
    //    printf("fitness = %.0f\n", (float)fitness);

    // "Add 1".
    for (i=0; i<length; i++) {
      if (inputs[i] == -1.0) {
	inputs[i] = 1.0;
	break;
      }
      inputs[i] = -1.0;
    }

    if (i == length) {
      // Overflow -> done!
      break;
    }
  }

  //  fitness -= pow(2.0, (float)length);
  //  printf("penalty: %.0f\n", pow(2.0, (float)length));


  if (Print[9]) {
    printf("Waste length: %d   fit: %.4f\n", waste_length, fitness);
    network->print_structure();
    printf("Avg in: %.4f,  # unique sources: %d.\n",
	   network->average_num_inputs(),
	   network->num_unique_output_sources());
  }


  model->fitness[0] = fitness;
  if (model->fv_length > 2) { 
    model->fitness[1] = network->num_links() + network->Num_neurons;
    model->fitness[2] = network->num_links();
    model->fitness[3] = network->Num_neurons;
  }

  if (Print[6]) {
    printf("#inputs:%d  #output:%d  total:%d  links:%d  waste:%d\n",
	   network->Num_inputs, network->Num_outputs, network->Num_neurons,
	   network->num_links(), waste_length);
    printf("Fitness Net Parity: %.1f:%.1f\n", fitness, model->fitness[0]);

    if (Print[14]) {
      network->print_structure();
    }
  }

  return TRUE;
}


int BarStruct3D :: calculate_parity2(int num, Real *inputs)
{
  int i;
  int correct = 0;
  int parity = 0;
  Real outputs[network->Num_outputs];

  for (i=0; i<num; i++) {
    if (inputs[i] == 1.0) {
      parity++;
    }
  }
  parity %= 2;


  network->reset();
  network->set_input(inputs);

  /*
  for (i=0; i<num; i+=2) {
    network->next_iteration();
  }
  */

  network->next_iteration();


  for (i=0; i<num; i++) {
    network->next_iteration();
    network->get_output(outputs);
    if (outputs[0] > 0.9) {
      if (parity) {
	correct = 2;
      } else {
	correct = -1;
      }
      break;

    } else if (outputs[0] < -0.9) {
      if (!parity) {
	correct = 2;
      } else {
	correct = -1;
      }
      break;
    }
  }

  if (correct == 0) {
    if (parity && (outputs[0] > 0.0)) {
      correct = 1;
    } else if ((!parity) && (outputs[0] < -0.0)) {
      correct = 1;
    }
  }

  /*
  printf("%2d: ", num);
  for (int j=0; j<num; j++) {
    printf("%.0f ", (float)inputs[j]);
  }
  printf(" ->  %.2f(%d:%d)  %d\n\n",(float)outputs[0], parity, i, correct);
  */

  return correct;
}


int BarStruct3D :: fitness_net_parity2(ModelStruct *model)
{
  if (network->Num_inputs == 0) {
    if (Print[9]) {
      printf("bars3d :: fitness_net_parity() - Fail because number of inputs = 0.\n");
    }
    return FALSE;
  }

  if (network->Num_outputs == 0) {
    if (Print[9]) {
      printf("bars3d :: fitness_net_parity() - Fail because number of outputs = 0.\n");
    }
    return FALSE;
  }

  int i, length;
  Real fitness;
  Real inputs[MAX_NETWORK_INPUTS+1]; // Extra 1 for overflow.
  
  length = (int)model->build_args[0];

  fitness = 0.0;

  if (network->Num_inputs != length) {
    if (Print[9]) {
      printf("bars3d :: fitness_net_parity() - Fail because number of inputs < %d, length.\n", length);
    }
    return FALSE;
  }

  if (network->Num_outputs != 1) {
    if (Print[9]) {
      printf("bars3d :: fitness_net_parity() - Fail because number of outputs != 1.\n");
    }
    return FALSE;
  }


  // Test all parity situations by starting with LOW, and
  // incremently flipping to HIGH by "adding 1".

  // Clear all bits explicitly.
  for (i=0; i<MAX_NETWORK_INPUTS; i++) {
    inputs[i] = 0.0;
  }

  // Set all bits LOW.
  for (i=0; i<length; i++) {
    inputs[i] = -1.0;
  }

  while (1) {
    /*
    printf("send ", length);
    for (int j=0; j<length; j++) {
      printf("%.0f ", (float)inputs[j]);
    }
    printf("\n");
    */

    fitness += (Real)calculate_parity2(length, inputs);

    // "Add 1".
    for (i=0; i<length; i++) {
      if (inputs[i] == -1.0) {
	inputs[i] = 1.0;
	break;
      }
      inputs[i] = -1.0;
    }

    if (i == length) {
      // Overflow -> done!
      break;
    }
  }

  //  fitness -= pow(2.0, (float)length);
  //  printf("penalty: %.0f\n", pow(2.0, (float)length));


  if (Print[9]) {
    printf("Waste length: %d   fit: %.4f\n", waste_length, fitness);
    network->print_structure();
    printf("Avg in: %.4f,  # unique sources: %d.\n",
	   network->average_num_inputs(),
	   network->num_unique_output_sources());
  }


  model->fitness[0] = fitness;
  if (model->fv_length > 2) { 
    model->fitness[1] = network->num_links() + network->Num_neurons;
    model->fitness[2] = network->num_links();
    model->fitness[3] = network->Num_neurons;
  }

  if (Print[6]) {
    /*
    printf("   #inputs:%d  #output:%d  total:%d  links:%d  waste:%d\n",
	   network->Num_inputs, network->Num_outputs, network->Num_neurons,
	   network->num_links(), waste_length);
    */
    printf("Fitness Net Parity (%d): %.1f:%.1f\n", length,
	   fitness, model->fitness[0]);

    if (Print[14]) {
      network->print_structure();
    }
  }

  return TRUE;
}




int BarStruct3D :: fitness_struct(ModelStruct *model)
{
  if (Build_type != 10) {
    if (Fitness_func == 1)
      return fitness_distance(model);
    else if (Fitness_func == 2)
      return fitness_distance2(model);
    else if (Fitness_func == 3)
      return fitness_distance3(model);
    else if (Fitness_func == 4)
      return fitness_distance4(model);
    else if (Fitness_func == 5)
      return fitness_ball(model);
    else if (Fitness_func == 6)
      return fitness_remote(model);
    else if (Build_type == 5) {
      if (Fitness_func == 11)
	return fitness_net_parity(model);
      else if (Fitness_func == 12)
	return fitness_net_parity2(model);
      else if (Fitness_func == 13)
	return fitness_net_parity2(model);
      else {
	return fitness_network(model);
      }

    } else {
      printf("ERROR - No fitness function.\n");
    }

  } else if (Build_type == 10) {
    return fitness_tree(model);
  } else {
    printf("ERROR - No fitness function.\n");
  }

  return FALSE;
}


int BarStruct3D :: fitness_blind_watchmaker(int fv_len, 
					    Real *fitness_vec)
{
  int fail = FALSE;

  if (Num_bars < 3)
    fail = TRUE;
  else if (Build_type != 10) {
    if (Num_joints == 0)
    fail = TRUE;
  } else {
    Real num_children = 0.0;
    Real num_parents = 0.0;
    for (int i=0; i<Num_bars; i++) {
      if (Bar[i].num_children != 0) {
	num_parents += 1.0;
	num_children += Bar[i].num_children;
      }
    }
    if (num_children < 1.5 * num_parents)
      fail = TRUE;
    else if (num_parents < 3.0)
      fail = TRUE;
  }
  /*
  else if (cycle_for_intersections()) {
    fail = TRUE;
  }
  */

  if (fail) {
    fitness_vec[0] = 0.0;
  } else {
    fitness_vec[0] = 1.0;
  }

  if (Print[6]) {
    printf("#bars: %d, #joints: %d.\n", Num_bars, Num_joints);
    printf("Fitness: %.1f\n", fitness_vec[0]);
  }

  return (!fail);
}


